#include "windows.h"
#include "acop.h"
#include <stdio.h>
#include <io.h>
#include <time.h>
#include <sys/timeb.h>
#include "errors.h"

#define _TINE_     0x00
//#define _MKI3_     0x01
//#define _CA_       0x02
#define _SIMULATE_ 0x03
#define _HW_       0x04
#define MAKELONGLINKID(lid,tg) (((tg)<<16) | ((lid)&0xffff) )
#define GETSHORTLINKID(lid) ((lid)&0xffff)

#ifdef _TINE_
#include "rpcdefs.h"  // TINE rpc definitions
#include "tine.h"
#ifndef DEFAULT_PROTOCOL
#define DEFAULT_PROTOCOL "TINE"
#endif
#define SYSTEM_DELAY 500
GBLK gblk[MAXGLB];
int nglobals = 0;
#else
#define RPCERR(ec,txt) sprintf(txt,"ERR: %d",ec)
#endif // _TINE_

#ifdef _MKI3_
#ifndef DEFAULT_PROTOCOL
#define DEFAULT_PROTOCOL "MKI3"
#endif
#include "tinexref.h"  // include some TINE flavoring definitions
#include "mki3.h"
char computerName[48];
char fromID[16];
#endif // _MKI3_

#ifdef _CA_
#ifndef DEFAULT_PROTOCOL
#define DEFAULT_PROTOCOL "CA"
#endif
#include "tinexref.h"  // include some TINE flavoring definitions
#include "caccess.h"
#endif // _CA_

#ifdef _SIMULATE_
#include "math.h"
#include "tinexref.h"  // include some TINE flavoring definitions
#ifndef SYSTEM_DELAY
#define SYSTEM_DELAY 50
#endif
#ifndef DEFAULT_PROTOCOL
#define DEFAULT_PROTOCOL "SIMULATE"
#endif
int simulatedPollingRate = 0;
void simulate_bkg(void);
void simulate(void *data,int arraysize,int format);
#endif // _SIMULATE_

#ifdef _HW_
#ifndef _TINE_
#define _TINE_     0x00
#endif
#include "tinexref.h"  // include some TINE flavoring definitions
#include "hwAcop.h"
int hwInitialized = 0;
HWND hWndHw = 0;
#endif // _HW_

LPSTR AccessModes[]={"READ","WRITE","POLL","REFRESH","RECEIVE",
                     "POLL.NETWORK","REFRESH.NETWORK",
                     "READ.CONNECT","WRITE.CONNECT",
                     "POLL.CONNECT","REFRESH.CONNECT",
                     "POLL.WRITE","REFRESH.WRITE",
                     "READ.DISCARD","WRITE.DISCARD"};

LPSTR AccessProtocols[] = {
#ifdef _TINE_
	"TINE",
#endif // _TINE_
#ifdef _MKI3_
	"MKI3",
#endif // _MKI3_
#ifdef _CA_
	"CA",
#endif // _CA_
#ifdef _SIMULATE_
	"SIMULATE"
#endif // _SIMULATE_
};

#define CompletionRemote 0
#define CompletionLocal 1

#define getDTYPEformat(d,tag) (\
  (d).arraySize == 0 ? -1 :\
  (dt=LOBYTE((d).type)) == VT_EMPTY ? getFormatFromString((d).custFormat,tag,NULL) :\
  dt == VT_UI1  ? CF_BYTE :\
  dt == VT_I2   ? CF_SHORT :\
  dt == VT_I4   ? CF_LONG :\
  dt == VT_R4   ? CF_FLOAT :\
  dt == VT_R8   ? CF_DOUBLE :\
  dt == VT_BSTR ? CF_TEXT :\
  CF_NULL )

long numlinks = 0;
LRESULT CALLBACK ACOPCallback(HWND hwnd,UINT uMsg,WPARAM wParam,LPARAM lParam);
#ifdef _MKI3_
LRESULT CALLBACK MKICallback(HWND hWndMKI,UINT uMsg,WPARAM wParam,LPARAM lParam);
#endif
#ifdef _HW_
LRESULT CALLBACK hwCallback(HWND hwnd,UINT uMsg,WPARAM wParam,LPARAM lParam);
#endif
long SetCompletion(DevStruct *ds,short errorcode);

char dataTagIn[9], dataTagOut[9];
HWND hWnd, hWndMKI;
void (*callBack)(long hLink);

double gDataTimeStamp = 0;
double gDataTimeStampOffset = 0;
double putDataTimeStamp(double toffset, time_t tsec, int tmsec)
{
  double ts;
  ts = (double)tsec;
  ts += (double)tmsec/1000.0;
  ts += toffset;
  return ts;
}
double makeDataTimeStamp(void)
{
  struct timeb tb;
  double ts;
  ftime(&tb);
  ts = (double)tb.time;
  ts += (double)(tb.millitm)/1000.0;
  ts += gDataTimeStampOffset;
  return ts;
}
#ifdef _TINE_
short findGlbEntry(short id)
{
  int i;

  for (i=0; i<nglobals; i++) if (gblk[i].id == id) return i;
  return -1;
}
short findGlbID(char *name)
{
  int i;

  for (i=0; i<nglobals; i++) if (!strncmp(name,gblk[i].name,16)) return gblk[i].id;
  return -1;
}
char *getGlobalDataTag(char *accessProtocol)
{
  char *c = NULL;
  if ((c=strpbrk(accessProtocol,DELIMITER)) != NULL) *c++ = 0;
  return c;
}
float getGlobalsTolerance(void *data,short fmt)
{
  switch (fmt)
  {
    case CF_BYTE:   return (float)(*(BYTE *)data); 
    case CF_SHORT:  return (float)(*(short *)data);
    case CF_LONG:   return (float)(*(long *)data);
    case CF_FLOAT:  return (*(float *)data);
    case CF_DOUBLE: return (float)(*(double *)data);
    default: return 0;
  }
}
int datcmp(void *dst,void *src,int blen,short fmt,float tol)
{
  float delta;
  if (tol == 0) return memcmp(dst,src,blen);
  switch (fmt)
  {
    case CF_BYTE:   delta = (float)abs((*(BYTE *)dst) - (*(BYTE *)src)); break;
    case CF_SHORT:  delta = (float)abs((*(short *)dst) - (*(short *)src)); break;
    case CF_LONG:   delta = (float)abs((*(long *)dst) - (*(long *)src)); break;
    case CF_FLOAT:  delta = (float)fabs((*(float *)dst) - (*(float *)src)); break;
    case CF_DOUBLE: delta = (float)fabs((*(double *)dst) - (*(double *)src)); break;
    default: return memcmp(dst,src,blen); break;
  }
  return delta > tol ? -1 : 0;
}
void tine_bkg(void)
{
  int i;
  time_t t = time(NULL);
  for (i=0; i<nglobals; i++) 
  {
    if ((gblk[i].compCode=GETGLOBAL(gblk[i].name,gblk[i].data,gblk[i].dArrayLength,gblk[i].dFormat)) != 0)
    {
      gblk[i].acquire = TRUE;
      (*callBack)(MAKELONGLINKID(gblk[i].id,_TINE_));
      gblk[i].timestamp = t;
    }
    else if (datcmp(gblk[i].dataStatic,gblk[i].data,gblk[i].dByteLength,gblk[i].dFormat,gblk[i].tolerance))
    {
      gblk[i].acquire = TRUE;
      memcpy(gblk[i].dataStatic,gblk[i].data,gblk[i].dByteLength);
      (*callBack)(MAKELONGLINKID(gblk[i].id,_TINE_));
      gblk[i].timestamp = t;
    }
    else if (t > gblk[i].timestamp + GLBHEARTBEAT)
    {
      gblk[i].acquire = TRUE;
      (*callBack)(MAKELONGLINKID(gblk[i].id,_TINE_));
      gblk[i].timestamp = t;
    }
    else if (gblk[i].acquire == TRUE)  // keep firing until it gets read
    {
      (*callBack)(MAKELONGLINKID(gblk[i].id,_TINE_));
 	  }
 	}
}
#endif  // _TINE_

short ftoi(char *fmt)
{
  char *c,fmtstr[64];
  short ifmt;
  strncpy(fmtstr,fmt,64);
  if ((c=strchr(fmtstr,'.')) != NULL) *c = 0;
  strupr(fmtstr);
  if ((ifmt=getSimpleFormatType(fmtstr)) != CF_NULL) return ifmt;
  return getCompoundFormatType(fmtstr);
}
short getFormatFromString(char *fmtstr,char *fmttag,int *fmtsiz)
{
  // TODO:  return meaningful format code according to in-coming
  //        format string parameter
  char *c = NULL,fs[48];
  short fmt, siz = 0;
# ifdef _MKI3_
  short i;
# endif
  if (fmtstr == NULL) return CF_NULL;
  memset(fs,0,48); strncpy(fs,fmtstr,47);
# ifdef _HW_
  if ((fmt=hwGetFormat(fs,fmtsiz)) > 0) return fmt;
# endif
  if ((c=strpbrk(fs,DELIMITER)) != NULL && fmttag != NULL) 
  {
    *c++ = 0;
    strncpy(fmttag,c,8);
    fmttag[8] = 0;
  }
  strupr(fs);
  fmt = getFormatType(fs);
  siz = getFormatSize(fmt);
# ifdef _MKI3_
  if (fmt == CF_NULL ) 
  {
    for (i=0; i<6; i++) if (!strcmpi(fs,mki3Format[i])) break;
    if (i < 6) fmt = i + 1;
    siz = 1;
  }
# endif // _MKI3_
  if (fmtsiz != NULL) *fmtsiz = siz;
  return fmt;
}

void WINAPI InitAcopDll(void)
{
	// TODO: place any ACOP instance initialization here.
}
char ACOPQueryMemPool[ACOPQUERYMEMPOOLSIZE];
long WINAPI QueryDeviceGroupInformation(LPSTR aProtocol,LPSTR *dContexts,LPSTR *dGroups,LPSTR *dNames,LPSTR *dProps)
{
  int cc = 0;
  int i,siz,p;
  char *c;
# ifdef _TINE_
  NAME16 *shortname,*dname = NULL;
# endif
# ifdef _MKI3_
# define MKI_NAME_LENGTH	22
  char *mkiname;
# endif

  if (aProtocol == NULL || strlen(aProtocol) == 0)
  {
    return -1;
  }
# ifdef _TINE_
  else if (!strcmp(aProtocol,"TINE"))
  {
    if (dContexts == NULL) return -1;
    siz = 500; shortname = &((NAME16 *)ACOPQueryMemPool)[500];
    if ((cc=GetDeviceServers((NAME16 *)*dContexts,shortname,&siz)) == 0) // success
    {
      c = ACOPQueryMemPool;
      for (i=0,p=0; i<siz; i++)
      {
        if (strlen(shortname[i].name) == 0) continue; // blank entry?
        strncpy(c,shortname[i].name,16); c[16] = 0;
        c += strlen(c) + 1;
      }
      *dGroups = ACOPQueryMemPool;
      return siz;
    }
	  return -1;
  }
# endif // _TINE_
# ifdef _MKI3_
  else if (!strcmp(aProtocol,"MKI3"))
  {
    if (dContexts == NULL) return -1;
    siz = 300; mkiname = (char *)ACOPQueryMemPool;
    for (i=0; i<siz; i++) if ( GetPcName((short)i, (char *)&mkiname[i*MKI_NAME_LENGTH]) != 0 ) break;
    if ( i != 0 ) // success
    {
      c = ACOPQueryMemPool;
      for (i=0,p=0; i<siz; i++)
      {
        if (strlen((char *)&mkiname[i*MKI_NAME_LENGTH]) == 0) continue; // blank entry?
        strncpy(c,(char *)&mkiname[i*MKI_NAME_LENGTH],MKI_NAME_LENGTH);
        c += strlen(c) + 1;
      }
      *dGroups = ACOPQueryMemPool;
      return siz;
    }
  return -1;
  }
# endif // _MKI3_
  else
  {
    if (dContexts != NULL) *dContexts = NULL;
    if (dGroups != NULL) *dGroups = NULL;
    if (dProps != NULL) *dProps = NULL;
    if (dNames != NULL) *dNames = NULL;
  }
  return 0;
}
long WINAPI QueryDeviceNameInformation(LPSTR aProtocol,LPSTR *dContexts,LPSTR *dGroups,LPSTR *dNames,LPSTR *dProps)
{
  int cc = 0;
  int i,siz,p;
  char *c;
# ifdef _TINE_
  NAME16 *shortname,*dname = NULL;
  NAME32 *pname = NULL;
# endif

  if (aProtocol == NULL || strlen(aProtocol) == 0)
  {
    return -1;
  }
# ifdef _TINE_
  else if (!strcmp(aProtocol,"TINE"))
  {
    if (dContexts == NULL) return -1;
    if (dGroups == NULL) return -1;
    siz = 500; shortname = &((NAME16 *)ACOPQueryMemPool)[500];
    if (dProps != NULL) pname = (NAME32 *)*dProps;
    if ((cc=GetDeviceNames((NAME16 *)*dContexts,(NAME16 *)*dGroups,shortname,pname,&siz)) == 0) // success
    {
      c = ACOPQueryMemPool;
      for (i=0,p=0; i<siz; i++)
      {
        if (strlen(shortname[i].name) == 0) continue; // blank entry?
        strncpy(c,shortname[i].name,16);
        if (shortname[i].name[15] == '&') 
        {
          i++;
          strncpy(&c[15],shortname[i].name,15);
          c[30] = 0;
        }
        else c[16] = 0;
        c += strlen(c) + 1;
      }
      *dNames = ACOPQueryMemPool;
      return siz;
    }
    return -1;
  }
# endif // _TINE_
  else
  {
    if (dContexts != NULL) *dContexts = NULL;
    if (dGroups != NULL) *dGroups = NULL;
    if (dProps != NULL) *dProps = NULL;
    if (dNames != NULL) *dNames = NULL;
  }
  return 0;
}
long WINAPI QueryDevicePropertyInformation(LPSTR aProtocol,LPSTR *dContexts,LPSTR *dGroups,LPSTR *dNames,LPSTR *dProps)
{
  int cc = 0;
  int i,siz,p;
  char *c;
# ifdef _TINE_
  NAME16 *dname = NULL;
  NAME32 *longname,*pname = NULL;
# endif

  if (aProtocol == NULL || strlen(aProtocol) == 0)
  {
    return -1;
  }
# ifdef _TINE_
  else if (!strcmp(aProtocol,"TINE"))
  {
    if (dContexts == NULL) return -1;
    if (dGroups == NULL) return -1;
    siz = 500; longname = (NAME32 *)ACOPQueryMemPool;
    if (dNames != NULL) dname = (NAME16 *)*dNames;
    if ((cc=GetDeviceProperties((NAME16 *)*dContexts,(NAME16 *)*dGroups,dname,longname,&siz)) == 0) // success
    {
      c = ACOPQueryMemPool;
      for (i=0,p=0; i<siz; i++)
      {
        if (strlen(longname[i].name) == 0) continue; // blank entry?
        strncpy(c,longname[i].name,32); c[31] = 0;
        c += strlen(c) + 1;
      }
      *dProps = ACOPQueryMemPool;
      return siz;
    }
    return -1;
  }
# endif // _TINE_
  else
  {
    if (dContexts != NULL) *dContexts = NULL;
    if (dGroups != NULL) *dGroups = NULL;
    if (dProps != NULL) *dProps = NULL;
    if (dNames != NULL) *dNames = NULL;
  }
  return 0;
}
long WINAPI QueryDeviceContextInformation(LPSTR aProtocol,LPSTR *dContexts,LPSTR *dGroups,LPSTR *dNames,LPSTR *dProps)
{
  int cc = 0;
  int i,siz,p;
  char *c;
# ifdef _TINE_
  NAME16 *shortname,*dname = NULL;
  NAME32 *pname = NULL;
# endif
  if (aProtocol == NULL || strlen(aProtocol) == 0)
  {
    return -1;
  }
# ifdef _TINE_
  else if (!strcmp(aProtocol,"TINE"))
  {
    if (dContexts == NULL) return 0;
    siz = 100;  shortname = &((NAME16 *)ACOPQueryMemPool)[500];
    if ((cc=GetContexts(shortname,&siz)) == 0) // success
    {
      c = ACOPQueryMemPool;
      for (i=0,p=0; i<siz; i++)
      {
        if (strlen(shortname[i].name) == 0) continue; // blank entry?
        strncpy(c,shortname[i].name,16); c[16] = 0;
        c += strlen(c) + 1;
      }
      *dContexts = ACOPQueryMemPool;
      return siz;
    }
    return -1;
  }
# endif
  else
  {
    if (dContexts != NULL) *dContexts = NULL;
    if (dGroups != NULL) *dGroups = NULL;
    if (dProps != NULL) *dProps = NULL;
    if (dNames != NULL) *dNames = NULL;
  }
  return 0;
}
long WINAPI QueryDeviceEGUInformation(LPSTR aProtocol,LPSTR dContext,LPSTR dGroup,LPSTR dName,LPSTR dProp,
									  float *yMin,float *yMax,LPSTR egu,int *graphStyle)
{
  int cc = 0;

  if (aProtocol == NULL || strlen(aProtocol) == 0)
  {
    return -1;
  }
# ifdef _TINE_
  else if (!strcmp(aProtocol,"TINE"))
  {
    if ((cc=GetDeviceEGU((NAME16 *)dContext,(NAME16 *)dGroup,(NAME16 *)dName,(NAME32 *)dProp,yMin,yMax,egu)) == 0) // success
    {
      if (fabs((double)(*yMax-*yMin)) > 1E5) *graphStyle = 1; else *graphStyle = 0;
      return 0;
    }
    return -1;
  }
# endif // _TINE_
  return 0;
}

long WINAPI QueryDefaultAccessProtocol(LPSTR *protArray)
{
  static LPSTR prots = NULL;
  static int len = 0;
  int i,s;

  if (prots == NULL)
  {
    for (s=1,i=0; i<(len=sizeof(AccessProtocols)/sizeof(LPSTR)); i++)
    {
      s += strlen(AccessProtocols[i])+1;
    }
    prots = GlobalAlloc(GPTR,s);
  }
  *protArray = prots;
	
  for (s=0,i=0; i<len; i++)
  {
    strcpy(*protArray+s,AccessProtocols[i]);
    s += strlen(AccessProtocols[i])+1;
  }
  return i;
}
long WINAPI QueryDataBlockSize(long dimension,char *fstr)
{
  int ssize = 1, fmtsiz = 0;
# ifdef _MKI3_
  int i;
# endif
  char *c,fmttag[9],fmtstr[128];
  short format;

  strncpy(fmtstr,fstr,128);
  if ((c=strstr(fmtstr,"/USE_ON_ERROR")) != NULL) *c = 0;
# ifdef _MKI3_
  for (i=0; i<6; i++) if (!strcmpi(mki3Format[i],fmtstr)) return dimension;
# endif
  fmttag[0] = 0; format = getFormatFromString(fmtstr,fmttag,&fmtsiz);
# ifdef _TINE_
  if (format == CF_STRUCT && strlen(fmttag)) ssize = GETSTRUCTSIZE(fmttag);
  if (format == CF_SPECTRUM) dimension += 24;
# endif
  return dimension*ssize*fmtsiz;
}
long lastDataSize = 0;
__declspec(dllexport) long PASCAL GetCompletionDataSize(void)
{
  return lastDataSize;
}
// prepare devName with a look toward backward compatibility:
#define MAX_DEVSTRLEN 64
void prepDeviceName(char *dn,DevStruct *ds)
{
  char *c,delimiter[2];
  // which style delimiter has been used (if any)?
  if ((c=strpbrk(ds->DeviceName,DELIMITER)) != NULL) delimiter[0] = *c; else delimiter[0] = '/';
  delimiter[1] = 0; dn[0] = 0;
  if (strlen(ds->DeviceContext) > 0) sprintf(dn,"%s%s",delimiter,ds->DeviceContext);
  if (strlen(ds->DeviceGroup) > 0) 
  {
    if (strlen(dn) > 0) sprintf(&dn[strlen(dn)],"%s%s",delimiter,ds->DeviceGroup);
    else strncpy(dn,ds->DeviceGroup,MAX_DEVSTRLEN);
  }
  if (strlen(ds->DeviceName) > 0) 
  {
    if (strlen(dn) > 0) sprintf(&dn[strlen(dn)],"%s%s",delimiter,ds->DeviceName);
    else strncpy(dn,ds->DeviceName,MAX_DEVSTRLEN);
 	}
}
short CompletionOrigin = CompletionRemote;
long WINAPI DevRequest(DevStruct * ds)
{
  short i, compCode=0,useOnError = 0;
  short format,formatEx,mode,scope,access;
  long  rate;
  static int busy = 0;
  char *c,stagIn[9],stagOut[9],accessMode[MAX_DEVSTRLEN],devName[MAX_DEVSTRLEN];
  int id,dupflg=0;
  short linkTag;
  long dt;  // for getDTYPEFormat macro
  double dts = 0;  // double time stamp 
  time_t ts = time(NULL);
# ifdef _MKI3_
  short formatMKI3, cc;
# endif // _MKI3_
# ifdef _TINE_
  DBLK dout,din;
# endif // _TINE_
# ifdef _CA_
  char ca_name[128];
  void *ca_dat;
# endif // _CA_

  CompletionOrigin = CompletionLocal;
# ifdef _HW_
  if (hwIsDevice(ds->DeviceName)) strcpy(ds->AccessProtocol,"HW");
# endif
# ifdef _TINE_
  if ((c=getGlobalDataTag(ds->AccessProtocol)) != NULL)
  {
    memset(dataTagOut,0,9); strncpy(dataTagOut,c,8); 
    memset(dataTagIn,0,9);  strncpy(dataTagIn,c,8);
  }
  memcpy(stagOut,dataTagOut,9); memcpy(stagIn,dataTagIn,9);
# endif // _TINE_
# ifdef _MKI3_
  if ((c=getGlobalDataTag(ds->AccessProtocol)) != NULL)
  {
    memset(fromID,0,16); strncpy(fromID,c,8);
 	}
# endif // _MKI3_

  if (ds->Data.custFormat != NULL && (c=strstr(ds->Data.custFormat,"/USE_ON_ERROR")) != NULL)
  {
    useOnError = -1;
    *c = 0; // remove from the custom format string
  }
  switch (format=getDTYPEformat(ds->Data,stagOut))
  {
    case CF_NULL: return SetCompletion(ds,illegal_format); // should have been there!
    case -1:  format = CF_NULL;                            // no data set 
    default:  break;                                       // data set with format okay
  }
  switch (formatEx=getDTYPEformat(ds->DataEx,stagIn))
  {
    case CF_NULL: return SetCompletion(ds,illegal_format); // should have been there!
    case -1:  formatEx = CF_NULL;                          // no data set 
    default:  break;                                       // data set with format okay
  }

  // TODO:  interpret device request parameters:
  //        DeviceName, DeviceProperty, AccessMode, AccessRate
  //        Switch on AccessProtocol if necessary
  rate = atol(ds->AccessRate);
  strncpy(accessMode,ds->AccessMode,MAX_DEVSTRLEN); strupr(accessMode);
  mode = getAccessMode(accessMode);
  scope = getAccessScope(accessMode);
  access = getAccessType(accessMode);
  mode |= scope;
  if (IsDiscardable(accessMode)) access |= CA_DISCARD;
# ifdef _TINE_
  dout.dArrayLength = format == CF_TEXT ? ds->Data.blockSize : ds->Data.arraySize;
  dout.dFormat = format;
  memcpy(dout.dTag,stagOut,8);
  din.dArrayLength = formatEx == CF_TEXT ? ds->DataEx.blockSize : ds->DataEx.arraySize;
  din.dFormat = formatEx;
  memcpy(din.dTag,stagIn,8);
  if (!strcmpi(ds->AccessProtocol,"TINE")) if (dout.dArrayLength) access |= CA_READ;
# endif // _TINE_

  switch (ds->MethodType)
  {
    case EXECUTE_METHOD:
      // TODO:  respond to EXECUTE_METHOD requests from ACOP.OCX  
      if (BASEMODE(mode) > CM_SINGLE) return SetCompletion(ds,illegal_mode);
      prepDeviceName(devName,ds);
#     ifdef _HW_
      if (!strcmpi(ds->AccessProtocol,"HW"))
      {
        if (rate < 50) rate = 50; else if (rate > 60000) rate = 60000;  // timeout parameter
        compCode =  hwExecLink(devName,ds->DeviceProperty,&dout,ds->Data.hMem,&din,ds->DataEx.hMem,access,rate);
        gDataTimeStamp = makeDataTimeStamp();
        ds->TimeStamp = (long)gDataTimeStamp;
        ds->TimeStampMSec = (short)((gDataTimeStamp-(double)ds->TimeStamp)*1000.0);
        lastDataSize = dout.dArrayLength;
        break;
      }
#     endif // _HW_
#     ifdef _TINE_
      if (!strcmpi(ds->AccessProtocol,"TINE")) // TINE
      {
        if (scope & CM_CONNECT) access |= CM_CONNECT; // let the access carry this
        if (rate < 100) rate = 100; else if (rate > 60000) rate = 60000;  // timeout parameter
        compCode = EXECLINKEx(devName,ds->DeviceProperty,&dout,ds->Data.hMem,&din,ds->DataEx.hMem,access,rate);
        gDataTimeStamp = T_GETDATATIMESTAMP();
        ds->TimeStamp = (long)gDataTimeStamp;
        ds->TimeStampMSec = (short)((gDataTimeStamp-(double)ds->TimeStamp)*1000.0);
        ds->Data.arraySize = dout.dArrayLength;
        lastDataSize = dout.dArrayLength;
        break;
      }
#     endif // _TINE_
#     ifdef _MKI3_
      if (!strcmpi(ds->AccessProtocol,"MKI3"))
      {
        if (rate < 100) rate = 100; else if (rate > 60000) rate = 60000;  // timeout parameter
        if ( access & CA_WRITE) 
        {
          compCode = illegal_format;
          if ((formatMKI3=getMKI3Format(formatEx)) < 0) break;
          if ( ds->DataEx.blockSize < mki3Size[formatMKI3-1] ) compCode = dimension_error;		// check memory size
          else 
          {
            // DeviceName -> e.g. "<toID>/<rpackName>.<rpackIndex>"
            cc = (short)ExecMKI3Link(hWndMKI,fromID,devName,ds->DeviceProperty,ds->DataEx.hMem,ds->DataEx.blockSize,formatMKI3,(formatEx >= CF_DOUBLE ? 1 : 0),access,rate);
            compCode = getMKI3Error(cc);
          }
        }
        if (compCode != 0 || ds->Data.blockSize == 0) break;
        compCode = illegal_format;
        if ((formatMKI3=getMKI3Format(format)) < 0) break;
        if ( ds->Data.blockSize < mki3Size[formatMKI3-1] ) compCode = dimension_error;		// check memory size
        else 
        {
          // exeMKI3Link has to treat single READ calls like a one-time OPENLink
          cc = (short)ExecMKI3Link(hWndMKI,fromID,devName,ds->DeviceProperty,ds->Data.hMem,ds->Data.blockSize,formatMKI3,(format >= CF_DOUBLE ? 1 : 0),access,rate);
          compCode = getMKI3Error(cc);
          lastDataSize = ds->Data.arraySize;
        }
        break;
      }
#     endif // _MKI3_
#     ifdef _CA_
      if (!strcmpi(ds->AccessProtocol,"CA"))
      {
        strncpy(ca_name,devName,128);
        if ((i=strcspn(ca_name,DELIMITER)) < (short)strlen(ca_name)) ca_name[i] = ':';
        if ((i=strlen(ds->DeviceProperty)) != 0 && strlen(ca_name) + i < 128) 
        sprintf(&ca_name[strlen(ca_name)],":%s",ds->DeviceProperty);
        ca_dat = access == CA_WRITE ? ds->DataEx.hMem : ds->Data.hMem;
        compCode = (short)CA_accessValue(ca_name,ca_dat,format,ds->Data.arraySize,access,rate);
        ds->TimeStamp = time(NULL);
        lastDataSize = ds->Data.arraySize;
        break;
      }
#     endif // _CA_
#     ifdef _SIMULATE_
      if (!strcmpi(ds->AccessProtocol,"SIMULATE"))
      {
        simulate(ds->Data.hMem,ds->Data.arraySize,format); 
        lastDataSize = ds->Data.arraySize;
        break;
      }
#     endif // _SIMULATE_
      break;

    case ATTACH_METHOD:
      // TODO:  respond to ATTACH_METHOD requests from ACOP.OCX  
      if (!ds->callBack) return SetCompletion(ds,illegal_operator);
      prepDeviceName(devName,ds);
      callBack=ds->callBack;
#     ifdef _HW_
      if (!strcmpi(ds->AccessProtocol,"HW"))
      {
        id = hwOpenLink (hWndHw,devName,ds->DeviceProperty,&dout,&din,ds->DataEx.hMem,access,rate,mode);
        linkTag = _HW_;
        goto attach_break;
      }
#     endif // _HW_
#     ifdef _TINE_
      if (!strcmpi(ds->AccessProtocol,"TINE")) 
      {
        if (mode == CM_RECEIVE)  // someone is interested in a network global
        {
          if ((id=findGlbID(ds->DeviceProperty)) == -1) 
          {
            if (dout.dArrayLength == 0) return SetCompletion(ds,argument_list_error);
            if (dout.dFormat == CF_NULL) return SetCompletion(ds,argument_list_error);
            if ((gblk[nglobals].data=calloc(dout.dArrayLength,fmtsizeof(format))) == NULL) 
              return SetCompletion(ds,out_of_local_memory);
            if ((gblk[nglobals].dataStatic=calloc(dout.dArrayLength,fmtsizeof(format))) == NULL) 
              return SetCompletion(ds,out_of_local_memory);
            id = nglobals;
            gblk[nglobals].tolerance = getGlobalsTolerance(ds->Data.hMem,dout.dFormat);
            strncpy(gblk[nglobals].name,ds->DeviceProperty,16);
            gblk[nglobals].dArrayLength = (UINT16)dout.dArrayLength;
            gblk[nglobals].dFormat = dout.dFormat;
            gblk[nglobals].id = id;
            gblk[nglobals].dByteLength = dout.dArrayLength*fmtsizeof(dout.dFormat);
            gblk[nglobals].acquire = TRUE;
            id = nglobals;
            nglobals++;
            if ((gblk[id].compCode=GETGLOBAL(gblk[id].name,gblk[id].data,gblk[id].dArrayLength,gblk[id].dFormat)) != 0)
              return SetCompletion(ds,gblk[id].compCode);
          }
        }
        else                    // normal data link
        {
          if (ds->DataCheck) mode |= CM_NODUPLICATES;
          id = OPENLINK (hWnd,devName,ds->DeviceProperty,&dout,&din,ds->DataEx.hMem,access,rate,mode);
          if (useOnError)
          {
            PutRPCErrorValue((short)id,ds->Data.hMem);
          }
          if (ds->DataCheck && (ds->toleranceAbs > 0 || ds->tolerancePct >0)) 
            SetNotificationTolerance(id,ds->toleranceAbs,ds->tolerancePct);
        }
        linkTag = _TINE_;
        goto attach_break;
      }
#     endif // _TINE_
#     ifdef _MKI3_
      if (!strcmpi(ds->AccessProtocol,"MKI3"))
      {
        if (mode == CM_REFRESH) return SetCompletion(ds,illegal_mode);
        linkTag = _MKI3_;
        id = -illegal_format;
        if (access & CA_WRITE)
        {
          if (mode != CM_SINGLE) return SetCompletion(ds,illegal_mode);
          if ((formatMKI3=getMKI3Format(formatEx)) < 0) goto attach_break;
          if ( ds->DataEx.blockSize < mki3Size[formatMKI3-1] ) id = -dimension_error;
          else 
          {
            cc = (short)ExecMKI3Link(hWndMKI,fromID,devName,ds->DeviceProperty,ds->DataEx.hMem,ds->DataEx.blockSize,formatMKI3,(formatEx >= CF_DOUBLE ? 1 : 0),access,rate);
            id = -getMKI3Error(cc);
          }
          goto attach_break;  // finished -> no event
        }
        if ((formatMKI3=getMKI3Format(format)) < 0) goto attach_break;
        if ( access == CA_READ ) 
        {
          if (ds->Data.blockSize < mki3Size[formatMKI3-1] ) id = -dimension_error; // check memory size
          else 
          {
            id = OpenMKI3Link(hWndMKI, ds->DeviceProperty, ds->Data.hMem, ds->Data.blockSize, rate/1000, formatMKI3, (format >= CF_DOUBLE ? 1 : 0));
            if (id < 0) id = -getMKI3Error(id);
          }
        }
        goto attach_break;
      }
#     endif // _MKI3_
#     ifdef _CA_
      if (!strcmpi(ds->AccessProtocol,"CA"))
      {
        strncpy(ca_name,devName,128);
        // allow CA names or a TINE-style name to pass thru
        if ((i=strcspn(ca_name,DELIMITER)) < (short)strlen(ca_name)) ca_name[i] = ':';
        if ((i=strlen(ds->DeviceProperty)) != 0 && strlen(ca_name) + i < 128) 
          sprintf(&ca_name[strlen(ca_name)],":%s",ds->DeviceProperty);
        id = CA_monitorValue(ca_name,ds->Data.hMem,format,ds->Data.arraySize,mode);
        ds->TimeStamp = time(NULL);
        linkTag = _CA_;
        goto attach_break;
      }
#     endif // _CA_
#     ifdef _SIMULATE_
      if (!strcmpi(ds->AccessProtocol,"SIMULATE"))
      {
        simulatedPollingRate = rate;
        id =  ++numlinks;
        linkTag = _SIMULATE_;
        goto attach_break;
      }
#     endif // _SIMULATE_
attach_break:
      if (id < 0) compCode = -id; 
      else ds->hLink = MAKELONGLINKID(id,linkTag);
      break;

    case DETACH_METHOD:
      // TODO:  respond to DETACH_METHOD requests from ACOP.OCX  
#     ifdef _HW_
      if (!strcmpi(ds->AccessProtocol,"HW")) 
      {
        if (mode != CM_RECEIVE && mode > CM_SINGLE)
        {
          compCode = hwCloseLink((short)GETSHORTLINKID(ds->hLink));
        }
        else compCode = 0;
        break;
      }
#     endif // _HW_
#     ifdef _TINE_
      if (!strcmpi(ds->AccessProtocol,"TINE")) 
      {
        if (mode == CM_RECEIVE || mode == CM_SINGLE) compCode = 0;
        else compCode = CLOSERPC(hWnd,(short)GETSHORTLINKID(ds->hLink));
        break;
      }
#     endif // _TINE_
#     ifdef _MKI3_
      if (!strcmpi(ds->AccessProtocol,"MKI3"))
      {
        compCode = (short)CloseMKI3Link(GETSHORTLINKID(ds->hLink));
        break;
      }
#     endif // _MKI3_
#     ifdef _CA_
      if (!strcmpi(ds->AccessProtocol,"CA"))
      {
        ca_quit(GETSHORTLINKID(ds->hLink));
        compCode = 0;
        break;
      }
#     endif // _CA_
#     ifdef _SIMULATE_
      if (!strcmpi(ds->AccessProtocol,"SIMULATE"))
      {
        simulatedPollingRate = 0;
        compCode = 0; 
        numlinks--;		
        break;
      }
#     endif // _SIMULATE_
      break;

    case GETDATA_METHOD:
      // TODO:  respond to GETDATA_METHOD requests from ACOP.OCX  
      if (busy) { compCode = not_ready; break; }
      busy = 1;
#     ifdef _HW_
      if (!strcmpi(ds->AccessProtocol,"HW"))
      {
        compCode = hwGetDataEx((short)GETSHORTLINKID(ds->hLink),ds->Data.hMem,&ts,&lastDataSize);
        gDataTimeStamp = makeDataTimeStamp();
        ds->Data.arraySize = lastDataSize;
        goto getdata_break;
      }
#     endif // _HW_
#     ifdef _TINE_
      if (!strcmpi(ds->AccessProtocol,"TINE")) 
      {
        if (mode == CM_RECEIVE)
        {
          i = findGlbEntry((short)GETSHORTLINKID(ds->hLink));
          compCode = gblk[i].compCode;
          gblk[i].acquire = FALSE;
          memcpy(ds->Data.hMem,gblk[i].dataStatic,gblk[i].dByteLength);
          gDataTimeStamp = makeDataTimeStamp();
        }
        else
        {
          compCode = GetDataAndTime((short)GETSHORTLINKID(ds->hLink),ds->Data.hMem,&gDataTimeStamp,&lastDataSize,&dupflg);
          if (ds->StatusCode == -91) dupflg = 0; // first call is returning -> always fire the event !
          ds->Data.arraySize = lastDataSize;
        }
        goto getdata_break;
      }
#     endif // _TINE_
#     ifdef _MKI3_
      if (!strcmpi(ds->AccessProtocol,"MKI3"))
      {
        compCode = (short)GetMKI3Data(GETSHORTLINKID(ds->hLink), &ts);
        compCode = getMKI3Error(compCode);
        if (mode == CM_SINGLE) CloseMKI3Link(GETSHORTLINKID(ds->hLink));
        lastDataSize = ds->Data.arraySize;
        gDataTimeStamp = (double)ts;
        goto getdata_break;
      }
#     endif // _MKI3_
#     ifdef _CA_
      if (!strcmpi(ds->AccessProtocol,"CA"))
      {
        compCode = (short)ca_getdata(GETSHORTLINKID(ds->hLink),ds->Data.hMem);
        lastDataSize = ds->Data.arraySize;
        gDataTimeStamp = makeDataTimeStamp();
        goto getdata_break;
      }
#     endif // _CA_
#     ifdef _SIMULATE_
      if (!strcmpi(ds->AccessProtocol,"SIMULATE"))
      {
        simulate(ds->Data.hMem,ds->Data.arraySize,format); 
        lastDataSize = ds->Data.arraySize;
        gDataTimeStamp = makeDataTimeStamp();
        goto getdata_break;
      }
#     endif // _SIMULATE_
getdata_break:
      ds->DataChanged = dupflg ? 0 : -1;
      ds->TimeStamp = (long)gDataTimeStamp;
      ds->TimeStampMSec = (short)((gDataTimeStamp-(double)ds->TimeStamp)*1000.0);
      busy = 0;
      break;
 	}
  CompletionOrigin = CompletionRemote;
  return SetCompletion(ds,compCode);
}

long WINAPI QueryStockAccessModes(LPSTR * modeArray)
{
  static LPSTR modes = NULL;
  static int len = 0;
  int i,s;

  if (modes == NULL)
  {
    for (s=1,i=0; i<(len=sizeof(AccessModes)/sizeof(LPSTR)); i++)
    {
      s += strlen(AccessModes[i])+1;
 	  }
    modes = GlobalAlloc(GPTR,s);
 	}
  *modeArray = modes;

  for (s=0,i=0; i<len; i++)
  {
    strcpy(*modeArray+s,AccessModes[i]);
    s += strlen(AccessModes[i])+1;
  }
  return i;
}

LRESULT CALLBACK ACOPCallback(HWND hwnd,UINT uMsg,WPARAM wParam,LPARAM lParam)
{
  int i;
  // TODO:  notify ACOP.OCX of in-coming events by calling (*callback)(hLink)
  //        as needed
  switch (uMsg)
  {
#   ifdef _TINE_
    case WM_SYSCOMMAND:
	    if ((i=wParam) > 0) (*callBack)(MAKELONGLINKID(i,_TINE_));
	    return 0;
#   endif
    case WM_TIMER:
#     ifdef _TINE_
      tine_bkg();
#     endif
#     ifdef _CA_
      ca_bkg();
#     endif
#     ifdef _SIMULATE_
      simulate_bkg();
#     endif
      break;
 	}
  return (DefWindowProc(hwnd,uMsg,wParam,lParam));
}
#ifdef _HW_
LRESULT CALLBACK hwCallback(HWND hwnd,UINT uMsg,WPARAM wParam,LPARAM lParam)
{
  int i;
  // TODO:  notify ACOP.OCX of in-coming events by calling (*callback)(hLink)
  //        as needed
  switch (uMsg)
  {
    case WM_SYSCOMMAND:
      if ((i=wParam) >= 0) (*callBack)(MAKELONGLINKID(i,_HW_));
      return 0;
    case WM_TIMER:
      break;
  }
  return (DefWindowProc(hwnd,uMsg,wParam,lParam));
}
#endif // _HW_
#ifdef _MKI3_
#define MKI3_FILE "servers.dat"
char MKI3DBpath[64],mki3path[128];
int mki3loaded = 0;
LRESULT CALLBACK MKICallback(HWND hwnd,UINT uMsg,WPARAM wParam,LPARAM lParam)
{
  // TODO:  notify ACOP.OCX of in-coming events by calling (*callback)(hLink)
  //        as needed
  switch (uMsg)
  {
    case WM_SYSCOMMAND:
      if ( lParam >= 0) (*callBack)(MAKELONGLINKID(wParam,_MKI3_));
      return 0;
  }
  return (DefWindowProc(hwnd,uMsg,wParam,lParam));
}
#endif

BOOL WINAPI DllMain(HANDLE hInst,ULONG dwReason,LPVOID lpvReserved)
{
  WNDCLASS wClass;
# ifdef _MKI3_
  char *ptr;
# endif

  switch(dwReason) 
  {
    case DLL_PROCESS_ATTACH:
      memset(&wClass,0,sizeof(WNDCLASS));

      wClass.lpfnWndProc=(WNDPROC)ACOPCallback;
      wClass.hInstance=hInst;
      wClass.lpszClassName="ACOPDLLCLASS";

      if (!RegisterClass(&wClass)) 
      {
        MessageBox(0,"Cannot register ACOPDLLCLASS","ACOP.DLL",0);
        return FALSE;
      }
      // process specific callback window used for event notification:
      hWnd=CreateWindow((LPSTR)"ACOPDLLCLASS",(LPSTR)"ACOPDLLWND",0,0,0,0,0,NULL,NULL,hInst,(LPSTR)NULL);
      if (!hWnd) 
      {
        MessageBox(0,"Cannot create ACOPDLLWND","ACOP.DLL",0);
        return FALSE;
      }
#     ifdef _HW_
      if ((hwInitialized=hwIsInitialized()) != 0)
      {
        wClass.lpfnWndProc=(WNDPROC)hwCallback;
        wClass.hInstance=hInst;
        wClass.lpszClassName="HWDLLCLASS";

        if (!RegisterClass(&wClass)) 
        {
          MessageBox(0,"Cannot register HWDLLCLASS","ACOP.DLL",0);
          return FALSE;
        }
        // process specific callback window used for event notification:
        hWndHw=CreateWindow((LPSTR)"HWDLLCLASS",(LPSTR)"HWDLLWND",0,0,0,0,0,NULL,NULL,hInst,(LPSTR)NULL);
        if (!hWndHw) 
        {
          MessageBox(0,"Cannot create HWDLLWND","ACOP.DLL",0);
          return FALSE;
        }
      }
#     endif // _HW_
#     ifdef _MKI3_
      MKI3DBpath[0] = 0;
      if ((ptr=getenv("MKI3DB")) != NULL) strcpy(MKI3DBpath,ptr);
      strcat(mki3path,MKI3DBpath); strcat(mki3path,MKI3_FILE);
      mki3loaded = MKI3Loaded();
      if (mki3loaded)
      {
        //MKI3...
        wClass.lpfnWndProc=(WNDPROC)MKICallback;
        wClass.hInstance=hInst;
        wClass.lpszClassName="MKIDLLCLASS";
        if (!RegisterClass(&wClass)) 
        {
          MessageBox(0,"Cannot register MKIDLLCLASS","ACOP.DLL",0);
          return FALSE;
        }
        // process specific callback window used for event notification:
        hWndMKI=CreateWindow((LPSTR)"MKIDLLCLASS",(LPSTR)"MKIDLLWND",0,0,0,0,0,NULL,NULL,hInst,(LPSTR)NULL);
        if (!hWndMKI) 
        {
          MessageBox(0,"Cannot create MKIDLLWND","ACOP.DLL",0);
          return FALSE;
        }
      }
      if ((ptr=getenv("COMPUTERNAME")) != NULL) strncpy(computerName,ptr,48);
      else strcpy(computerName,"UNKNOWN");
#     endif // _MKI3_
#     ifdef _TINE_
      memset(gblk,0,sizeof(gblk));
      memset(dataTagIn,0,9); memset(dataTagOut,0,9);
#     endif // _TINE_
      if (!SetTimer(hWnd,1,SYSTEM_DELAY,NULL))
      {
        MessageBox(0, "Can't set Timer !","ACOP.DLL",0);
        return FALSE;
      }
#     ifdef _CA_
      ca_initTable();
#     endif // _CA_
      break;
    case DLL_PROCESS_DETACH:
      if (hWnd)
      {
        KillTimer(hWnd,1);
        DestroyWindow(hWnd);
        UnregisterClass("ACOPDLLCLASS",hInst);
      }
#     ifdef _HW_
      if (hWndHw)
      {
        DestroyWindow(hWndHw);
        UnregisterClass("HWDLLCLASS",hInst);
      }
#     endif // _HW_
#     ifdef _MKI3_
      if (mki3loaded && hWndMKI) 
      {
        DestroyWindow(hWndMKI);
        UnregisterClass("MKIDLLCLASS",hInst);
      }
#     endif // _MKI3_
      break;
 	}

  return TRUE;
}
long SetCompletion(DevStruct *ds,short errorcode)
{
  //  TODO:  fill in ds->Status with a string desciption of the errorcode
  //         fill in ds->StatusCode with (probably) errorcode
  static char str[500];

  if (CompletionOrigin == CompletionLocal)
  {
    sprintf(str,"[OCX] ");
    RPCERR(errorcode,&str[6]);
  }
# ifdef _HW_
  else if (!strcmpi(ds->AccessProtocol,"HW"))
  {
    if (errorcode == -1) sprintf(str,"[LCL] ???");
    else hwGetError(errorcode,str);
  }
# endif // _HW_
# ifdef _TINE_
  else if (!strcmpi(ds->AccessProtocol,"TINE"))
  {
    if (errorcode == -1) sprintf(str,"[LCL] data already reported");
    else RPCGETLASTERROR(str);
  }
# endif // _TINE_
# ifdef _CA_
  else if (!strcmpi(ds->AccessProtocol,"CA"))
  {
    ca_getLastError(errorcode,str);
  }
# endif // _CA_
# ifdef _MKI3_
  else
  {
    RPCERR(errorcode,str);
 	}
# endif // _MKI3_
  ds->StatusCode = errorcode;
  ds->Status=str;
  return errorcode;
}

#ifdef _SIMULATE_
#define NPOINTS 512
double sinedata[NPOINTS];
double randomdata[NPOINTS];
int fillsinedata(void)
{
  int i;
  for (i=0; i<NPOINTS; i++) sinedata[i] = ((1.0 + sin(i*2*6.2832/NPOINTS))*200 + 55);
  srand( (unsigned)time( NULL ) );
  for (i=0; i<NPOINTS; i++) randomdata[i] = 102 * rand()/RAND_MAX - 51;
  return 1;
}

void simulate(void *data,int arraysize,int format)
{
  static int initdone = 0;
  static double phi = 0.0;
  int i,ptr = 0; //((int)(phi*NPOINTS/6.2832))%NPOINTS;
  int randptr = ((int)(rand()*NPOINTS/RAND_MAX))%NPOINTS;
  if (!initdone) initdone = fillsinedata();

  switch (format)
  {
    case CF_BYTE:
      for (i=0; i<arraysize; i++)  ((BYTE *)data)[i] = (BYTE)(sinedata[(ptr+i)%NPOINTS]+randomdata[(randptr+i)%NPOINTS]);
      phi += .1; break;
    case CF_SHORT:
      for (i=0; i<arraysize; i++)  ((short *)data)[i] = (short)(sinedata[(ptr+i)%NPOINTS]+randomdata[(randptr+i)%NPOINTS]);
      phi += .1; break;
    case CF_LONG:
      for (i=0; i<arraysize; i++)  ((long *)data)[i] = (long)(sinedata[(ptr+i)%NPOINTS]+randomdata[(randptr+i)%NPOINTS]);
      phi += .1; break;
    case CF_FLOAT:
      for (i=0; i<arraysize; i++)  ((float *)data)[i] = (float)(sinedata[(ptr+i)%NPOINTS]+randomdata[(randptr+i)%NPOINTS]);
      phi += .1; break;
    case CF_DOUBLE:
      for (i=0; i<arraysize; i++)  ((double *)data)[i] = (double)(sinedata[(ptr+i)%NPOINTS]+randomdata[(randptr+i)%NPOINTS]);
      phi += .1; break;
    default: // do nothing
      break;
  }
}
void simulate_bkg(void)
{
  static int cnt = 0;
  int i;
  cnt = ++cnt % 1000;
  if (simulatedPollingRate && (simulatedPollingRate/SYSTEM_DELAY) < cnt)
  {
	  cnt = 0;
	  for (i=0; i<numlinks; i++) (*callBack)(MAKELONGLINKID(i+1,_SIMULATE_));
  }
}
#endif // _SIMULATE_
