#define CA_ERROR_GLBLSOURCE
#include <malloc.h>
#include <string.h>
#include <cadef.h>			// Epics specific
#include "rpcdefs.h"
#include "errors.h"
#include "caccess.h"

#define _TINE_     0x00
#define _MKI3_     0x01
#define _CA_       0x02
//#define _SIMULATE_ 0x03
#define MAKELONGLINKID(lid,tg) (((tg)<<16) | ((lid)&0xffff) )
#define GETSHORTLINKID(lid) ((lid)&0xffff)

typedef struct
{
  chid id;
  int  hid;
  BYTE *data;
  int  fmt;
  int  len;
  int  cc;
} HID;

HID hID[MAXCA_ID];
void (*callBack)(long hLink);

void ca_getLastError(int errorcode,char *errstr)
{
  int code = CA_EXTRACT_MSG_NO(errorcode);
  if (code < 0 || code > 50) strcpy(errstr,"ERROR undefined");
  else strncpy(errstr,ca_message_text[code],500);
}
int ca_getdata(int hid,void *data)
{
  int len = hID[hid].len*fmtsizeof(hID[hid].fmt);
  if (len > 0) memcpy(data,hID[hid].data,len);
  return hID[hid].cc;
}
void ca_callback(struct event_handler_args arg)
{
  int i = *(int *)arg.usr;
  hID[i].cc = arg.status;
  if (arg.type != getCAformat(hID[i].fmt)) 
  {
	hID[i].cc = ECA_NOCONVERT;
	hID[i].fmt = CF_NULL;
  }
  if (arg.count > hID[i].len) 
  {
	 hID[i].cc = ECA_TOLARGE;
	 hID[i].len = 0;
  }
  hID[i].data = arg.dbr;
  callBack(MAKELONGLINKID(i,_CA_));
}
int ca_bkg(void)
{
  int cc=0,status;
  if ((status=ca_pend_io (.05)) != ECA_NORMAL) return status;
  return 0;
}
void ca_quit(int hid)
{
  if (hid < 0 || hid > MAXCA_ID) return;
  if (hID[hid].hid >= 0) ca_clear_channel(hID[hid].id);
  hID[hid].hid = -un_allocated;
}
void ca_initTable(void)
{
  int i;
  for (i=0; i<MAXCA_ID; i++) 
  {
	  hID[i].hid = -un_allocated;
	  hID[i].data = NULL;
  }
}
int getNextIDHandle(short format,long arraysize)
{
  int i,cc=0;
  for (i=0; hID[i].hid >= 0 && i<MAXCA_ID; i++);
  if (i == MAXCA_ID) ccerr(buffer_too_small);
  hID[i].hid = i;
  hID[i].fmt = format;
  hID[i].len = arraysize;
err:
  if (cc) hID[i].hid = -cc;
  return hID[i].hid;
}
int CA_monitorValue(char *name,void *data,short format,long arraysize,short mode)
{
  int cc=0;
  int hid,ca_fmt;
  
  if ((hid=getNextIDHandle(format,arraysize)) < 0) ccerr(ECA_ALLOCMEM);
  if ((ca_fmt=getCAformat(format)) < 0) ccerr(ECA_BADTYPE);
  if ((cc=ca_search (name, &hID[hid].id)) != ECA_NORMAL) ccerr(cc);
  if ((cc=ca_pend_io (1.0)) != ECA_NORMAL)  ccerr(cc);
  cc = mode == CM_SINGLE ? 
	   ca_array_get_callback(ca_fmt,arraysize,hID[hid].id,ca_callback,&hID[hid].hid) :
       ca_add_array_event(ca_fmt,arraysize,hID[hid].id,ca_callback,&hID[hid].hid,0.0,0.0,0.0,NULL);
  if (cc != ECA_NORMAL) ccerr(cc);
err:
  if (cc != ECA_NORMAL) 
  {
	ca_quit(hid);
	return -cc;
  }
  return hid;
}
int CA_accessValue(char *name,void *data,short format,long arraysize,short access,long timeout)
{
  int cc = 0;
  chid	mychid;
  int ca_fmt;
  float ca_to = (float)((float)timeout)/(float)1000.0;

  if (ca_to < 500.0) ca_to = 500.0;
  if ((ca_fmt=getCAformat(format)) < 0) ccerr(ECA_BADTYPE);
  if((cc=ca_search(name, &mychid)) != ECA_NORMAL) ccerr(cc);
  if((cc=ca_pend_io(ca_to)) != ECA_NORMAL) ccerr(cc);
  cc = access == CA_WRITE ?
	  ca_array_put(ca_fmt,arraysize,mychid,data) :
      ca_array_get(ca_fmt,arraysize,mychid,data);
  if((cc=ca_pend_io(.5)) != ECA_NORMAL) ccerr(cc);
err:
  return cc == ECA_NORMAL ? 0 : cc; 
}
