/**
 * \mainpage TINE API for Server ActiveX Control
 *
 * The Methods, Events, and Properties documented here are refer to the TINE
 * Server ActiveX control 'Srv.ocx'.  Although the ActiveX control is language neutral,
 * and can therefore be used in any container which supports ActiveX (such as Visual C++,
 * LabView, HPVee, etc.) it is optimized for ease-of-use with Visual Basic.
 *
 * \image html srvproperties.jpg
 *
 */
/** \file SRVCTL.CPP
 * \brief TINE Server ActiveX Control documentation
 *
 *
 */
// SrvCtl.cpp : Implementation of the CSrvCtrl ActiveX Control class.

#include "stdafx.h"
#include "srv.h"
#include "SrvCtl.h"
#include "SrvPpg.h"
#include "rpcdefs.h"
#include "errors.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

typedef union
{
  void   *vptr;
  char   *cptr;
  UINT8  *bptr;
  UINT16 *sptr;
  UINT32 *lptr;
  float  *fptr;
  double *dptr;
} DUNION;

typedef struct
{
  UINT32 dArrayLength;
  short  dFormat;
  short  reserved;
  char   dTag[16];
  DUNION data;
} DTYPE;

#ifdef __cplusplus
#define F_PREFIX extern "C"
#else
#define F_PREFIX extern
#endif

F_PREFIX long WINAPI QueryDataBlockSize(long dimension,char *fmtstr);

extern "C" __declspec(dllimport) short PASCAL ScheduleProperty(char *eqm, char *prp);
extern "C" __declspec(dllimport) double PASCAL makeDataTimeStamp(void);
extern "C" __declspec(dllimport) void PASCAL setDataTimeStamp(double ts);
extern "C" __declspec(dllimport) double PASCAL getDataTimeStamp(void);
extern "C" __declspec(dllimport) void PASCAL GetCallerInfo(char *ClientName,char *ClientAddress);
extern "C" __declspec(dllimport) short AdjustContract(char *devName, char *devProperty, long dataLengthOut,long dataLengthIn,short eqpidx);
extern "C" __declspec(dllimport) void GetContract(short *eqno,char *prop,short *size,short index, time_t *time);
extern "C" __declspec(dllimport) void RemoveEqpfcn(char *name,HWND hVBWnd);
extern "C" __declspec(dllimport) short GetData(void *data,short index);
extern "C" __declspec(dllimport) short GetEqpData(VARIANT FAR * data, short index);
extern "C" __declspec(dllimport) short GetCallerFormat(short index);
extern "C" __declspec(dllimport) short RegisterEQP(char *exportName,char *eqpName,short number,HWND hVBWnd);
extern "C" __declspec(dllimport) short RegisterProperty(char *eqm,char *prop,short siz,short fmt,short acc,char *dsc);
extern "C" __declspec(dllimport) short GetPropertyId(char *eqm,char *prpName);
extern "C" __declspec(dllimport) short RegisterModuleName(char *eqname,char *devname,short devnr);
extern "C" __declspec(dllimport) short GetDeviceNumber(char *eqname,char *devname);
extern "C" __declspec(dllimport) short GetDeviceNumberEx(char *eqname,char *devname,char *prpname);
extern "C" __declspec(dllimport) void GetDeviceName(char *eqname,int devnr,char *devname);
extern "C" __declspec(dllimport) short SetAlarm(char *name,short devNr,long almCode,BYTE *almData);
extern "C" __declspec(dllimport) short SetAlarmEx(char *name,short devNr,long almCode,BYTE *almData,BYTE flags);
extern "C" __declspec(dllimport) void ClearAlarm(char *name,short devNr);
extern "C" __declspec(dllimport) void PutStatus(short status);
extern "C" __declspec(dllimport) short PutData(void *data,time_t eqptime,short eqpidx);
extern "C" __declspec(dllimport) short PutDataEx(void *data,double dataTimeStamp,short eqpidx);
extern "C" __declspec(dllimport) void GetContractEx(char *devName, char *devProperty,DTYPE *dout,
                                                    DTYPE *din,short *access,short eqpidx);
extern "C" __declspec(dllimport) short RegisterPropertyEx(char *eqm,char *prop,long sizIn,short fmtIn,char *tagIn,
                                                          long sizOut,short fmtOut,char *tagOut,short acc,char *dsc);
extern "C" __declspec(dllimport) short SetDeviceAlarm(char *name,char *devname,long almCode,BYTE *almData,BYTE flags);
extern "C" __declspec(dllimport) void ClearDeviceAlarm(char *name,char *devname);
extern "C" __declspec(dllimport) void RemoveDeviceAlarm(char *name,char *devname,long code);
extern "C" __declspec(dllimport) void RemoveAlarm(char *name,short devNr,long code);
extern "C" __declspec(dllimport) short SETRPCERROREx(short errcode,char *errstr,short eqpidx);

IMPLEMENT_DYNCREATE(CSrvCtrl, COleControl)


/////////////////////////////////////////////////////////////////////////////
// Message map

BEGIN_MESSAGE_MAP(CSrvCtrl, COleControl)
	//{{AFX_MSG_MAP(CSrvCtrl)
	ON_WM_SYSCOMMAND()
	//}}AFX_MSG_MAP
	ON_OLEVERB(AFX_IDS_VERB_EDIT, OnEdit)
	ON_OLEVERB(AFX_IDS_VERB_PROPERTIES, OnProperties)
END_MESSAGE_MAP()


/////////////////////////////////////////////////////////////////////////////
// Dispatch map

BEGIN_DISPATCH_MAP(CSrvCtrl, COleControl)
	//{{AFX_DISPATCH_MAP(CSrvCtrl)
	DISP_PROPERTY_NOTIFY(CSrvCtrl, "EqpName", m_eqpName, OnEqpNameChanged, VT_BSTR)
	DISP_PROPERTY_NOTIFY(CSrvCtrl, "ExportName", m_exportName, OnExportNameChanged, VT_BSTR)
	DISP_PROPERTY_NOTIFY(CSrvCtrl, "EqpStatus", m_eqpStatus, OnEqpStatusChanged, VT_I2)
	DISP_PROPERTY_NOTIFY(CSrvCtrl, "EqpTime", m_eqpTime, OnEqpTimeChanged, VT_I4)
	DISP_PROPERTY_NOTIFY(CSrvCtrl, "EqpNumberModules", m_eqpNumberModules, OnEqpNumberModulesChanged, VT_I2)
	DISP_FUNCTION(CSrvCtrl, "EqpSend", EqpSend, VT_I2, VTS_I4)
	DISP_FUNCTION(CSrvCtrl, "EqpRecv", EqpRecv, VT_I2, VTS_I4)
	DISP_FUNCTION(CSrvCtrl, "EqpClearAlarm", EqpClearAlarm, VT_EMPTY, VTS_I2)
	DISP_FUNCTION(CSrvCtrl, "EqpSetAlarm", EqpSetAlarm, VT_I2, VTS_I2 VTS_I4 VTS_I4 VTS_I2)
	DISP_FUNCTION(CSrvCtrl, "EqpRegisterProperty", EqpRegisterProperty, VT_I2, VTS_BSTR VTS_I2 VTS_I2 VTS_I2 VTS_BSTR)
	DISP_FUNCTION(CSrvCtrl, "EqpRegisterModule", EqpRegisterModule, VT_I2, VTS_BSTR VTS_I2)
	DISP_FUNCTION(CSrvCtrl, "EqpRegisterPropertyEx", EqpRegisterPropertyEx, VT_I2, VTS_BSTR VTS_I4 VTS_I2 VTS_BSTR VTS_I4 VTS_I2 VTS_BSTR VTS_I2 VTS_BSTR)
	DISP_FUNCTION(CSrvCtrl, "EqpSetDeviceAlarm", EqpSetDeviceAlarm, VT_I2, VTS_BSTR VTS_I4 VTS_I4 VTS_I2)
	DISP_FUNCTION(CSrvCtrl, "EqpClearDeviceAlarm", EqpClearDeviceAlarm, VT_EMPTY, VTS_BSTR)
	DISP_FUNCTION(CSrvCtrl, "EqpSetCompletion", EqpSetCompletion, VT_EMPTY, VTS_I2 VTS_BSTR)
	DISP_FUNCTION(CSrvCtrl, "EqpGetModuleNumber", EqpGetModuleNumber, VT_I2, VTS_BSTR)
	DISP_FUNCTION(CSrvCtrl, "EqpGetModuleName", EqpGetModuleName, VT_BSTR, VTS_I2)
	DISP_FUNCTION(CSrvCtrl, "EqpGetPropertyId", EqpGetPropertyId, VT_I2, VTS_BSTR)
	DISP_FUNCTION(CSrvCtrl, "EqpRecvData", EqpRecvData, VT_I2, VTS_VARIANT)
	DISP_FUNCTION(CSrvCtrl, "EqpSendData", EqpSendData, VT_I2, VTS_VARIANT)
	DISP_FUNCTION(CSrvCtrl, "EqpSendDataEx", EqpSendDataEx, VT_I2, VTS_VARIANT VTS_VARIANT)
	DISP_FUNCTION(CSrvCtrl, "GetCallerInformation", GetCallerInformation, VT_I2, VTS_VARIANT VTS_VARIANT)
	DISP_FUNCTION(CSrvCtrl, "GetDataTimeStamp", GetDataTimeStamp, VT_R8, VTS_NONE)
	DISP_FUNCTION(CSrvCtrl, "SetDataTimeStamp", SetDataTimeStamp, VT_EMPTY, VTS_R8)
	DISP_FUNCTION(CSrvCtrl, "ScheduleProperty", _ScheduleProperty, VT_I2, VTS_BSTR)
	DISP_FUNCTION(CSrvCtrl, "EqpGetModuleNumberEx", EqpGetModuleNumberEx, VT_I2, VTS_BSTR VTS_VARIANT)
	DISP_FUNCTION(CSrvCtrl, "EqpRemoveAlarm", EqpRemoveAlarm, VT_EMPTY, VTS_I2 VTS_I4)
	DISP_FUNCTION(CSrvCtrl, "EqpRemoveDeviceAlarm", EqpRemoveDeviceAlarm, VT_EMPTY, VTS_BSTR VTS_I4)
	DISP_DEFVALUE(CSrvCtrl, "EqpName")
	DISP_PROPERTY_NOTIFY_ID(CSrvCtrl, "Enabled", DISPID_ENABLED, m_enabled, OnEnabledChanged, VT_BOOL)
	//}}AFX_DISPATCH_MAP
	DISP_FUNCTION_ID(CSrvCtrl, "AboutBox", DISPID_ABOUTBOX, AboutBox, VT_EMPTY, VTS_NONE)
END_DISPATCH_MAP()


/////////////////////////////////////////////////////////////////////////////
// Event map

BEGIN_EVENT_MAP(CSrvCtrl, COleControl)
	//{{AFX_EVENT_MAP(CSrvCtrl)
	EVENT_CUSTOM("EqpFcn", FireEqpFcn, VTS_BSTR  VTS_BSTR  VTS_I4  VTS_I4  VTS_I2)
	EVENT_CUSTOM("Receive", FireReceive, VTS_BSTR  VTS_I2  VTS_I2  VTS_I2  VTS_I2)
	//}}AFX_EVENT_MAP
END_EVENT_MAP()


/////////////////////////////////////////////////////////////////////////////
// Property pages

// TODO: Add more property pages as needed.  Remember to increase the count!
BEGIN_PROPPAGEIDS(CSrvCtrl, 1)
	PROPPAGEID(CSrvPropPage::guid)
END_PROPPAGEIDS(CSrvCtrl)


/////////////////////////////////////////////////////////////////////////////
// Initialize class factory and guid

IMPLEMENT_OLECREATE_EX(CSrvCtrl, "SRV.SrvCtrl.1",
	0x4727fed3, 0xd8ec, 0x11d0, 0xb1, 0x4, 0, 0, 0xe8, 0xd, 0x9c, 0xc1)


/////////////////////////////////////////////////////////////////////////////
// Type library ID and version

IMPLEMENT_OLETYPELIB(CSrvCtrl, _tlid, _wVerMajor, _wVerMinor)


/////////////////////////////////////////////////////////////////////////////
// Interface IDs

const IID BASED_CODE IID_DSrv =
		{ 0x4727fed1, 0xd8ec, 0x11d0, { 0xb1, 0x4, 0, 0, 0xe8, 0xd, 0x9c, 0xc1 } };
const IID BASED_CODE IID_DSrvEvents =
		{ 0x4727fed2, 0xd8ec, 0x11d0, { 0xb1, 0x4, 0, 0, 0xe8, 0xd, 0x9c, 0xc1 } };


/////////////////////////////////////////////////////////////////////////////
// Control type information

static const DWORD BASED_CODE _dwSrvOleMisc =
	OLEMISC_ACTIVATEWHENVISIBLE |
	OLEMISC_INVISIBLEATRUNTIME |
	OLEMISC_SETCLIENTSITEFIRST |
	OLEMISC_INSIDEOUT |
	OLEMISC_CANTLINKINSIDE |
	OLEMISC_RECOMPOSEONRESIZE;

IMPLEMENT_OLECTLTYPE(CSrvCtrl, IDS_SRV, _dwSrvOleMisc)



/////////////////////////////////////////////////////////////////////////////
// CSrvCtrl::CSrvCtrlFactory::UpdateRegistry -
// Adds or removes system registry entries for CSrvCtrl

BOOL CSrvCtrl::CSrvCtrlFactory::UpdateRegistry(BOOL bRegister)
{
	// TODO: Verify that your control follows apartment-model threading rules.
	// Refer to MFC TechNote 64 for more information.
	// If your control does not conform to the apartment-model rules, then
	// you must modify the code below, changing the 6th parameter from
	// afxRegInsertable | afxRegApartmentThreading to afxRegInsertable.

	if (bRegister)
		return AfxOleRegisterControlClass(
			AfxGetInstanceHandle(),
			m_clsid,
			m_lpszProgID,
			IDS_SRV,
			IDB_SRV,
			afxRegInsertable | afxRegApartmentThreading,
			_dwSrvOleMisc,
			_tlid,
			_wVerMajor,
			_wVerMinor);
	else
		return AfxOleUnregisterClass(m_clsid, m_lpszProgID);
}


/////////////////////////////////////////////////////////////////////////////
// CSrvCtrl::CSrvCtrl - Constructor

CSrvCtrl::CSrvCtrl()
{
	InitializeIIDs(&IID_DSrv, &IID_DSrvEvents);

	// TODO: Initialize your control's instance data here.
	m_devname[0] = 0;
	m_property[0] = 0;
	m_strTagIn[0] = 0;
	m_strTagOut[0] = 0;
	m_busy = 0;
	m_bEnabled = 0;
	m_enabled = 0;
	m_eqpStatus = 0;
    m_eqpIndex = 0;
	m_eqpTime = 0;
	m_eqpNumberModules = 0;
	EqpFcnEventHead = 0;
	EqpFcnEventTail = 0; 
}


/////////////////////////////////////////////////////////////////////////////
// CSrvCtrl::~CSrvCtrl - Destructor

CSrvCtrl::~CSrvCtrl()
{
	// TODO: Cleanup your control's instance data here.
}


/////////////////////////////////////////////////////////////////////////////
// CSrvCtrl::OnDraw - Drawing function

void CSrvCtrl::OnDraw(
			CDC* pdc, const CRect& rcBounds, const CRect& rcInvalid)
{
	// TODO: Replace the following code with your own drawing code.
//	pdc->FillRect(rcBounds, CBrush::FromHandle((HBRUSH)GetStockObject(WHITE_BRUSH)));
//	pdc->Ellipse(rcBounds);
	CBitmap bitmap;
	BITMAP bmp;
	CPictureHolder picHolder;
	CRect rcSrcBounds;

	// Load bitmap
	if (bitmap.LoadBitmap(IDB_SRV) == 0) return;
	bitmap.GetObject(sizeof(BITMAP),&bmp);
	rcSrcBounds.right = bmp.bmWidth;
	rcSrcBounds.bottom = bmp.bmHeight;

	picHolder.CreateFromBitmap((HBITMAP)bitmap.m_hObject,NULL,FALSE);
	picHolder.Render(pdc,rcBounds,rcSrcBounds);
}


/////////////////////////////////////////////////////////////////////////////
// CSrvCtrl::DoPropExchange - Persistence support

void CSrvCtrl::DoPropExchange(CPropExchange* pPX)
{
	ExchangeVersion(pPX, MAKELONG(_wVerMinor, _wVerMajor));
	COleControl::DoPropExchange(pPX);

	// TODO: Call PX_ functions for each persistent custom property.

	PX_String(pPX, _T("ExportName"),m_exportName,"");
	PX_String(pPX, _T("EqpName"),m_eqpName,"");
	PX_Short(pPX, _T("EqpNumberModules"),m_eqpNumberModules,1);

}


/////////////////////////////////////////////////////////////////////////////
// CSrvCtrl::OnResetState - Reset control to default state

void CSrvCtrl::OnResetState()
{
	COleControl::OnResetState();  // Resets defaults found in DoPropExchange

	// TODO: Reset any other control state here.

	m_eqpName = "";
	m_exportName = "";
	m_enabled = FALSE;
	m_bEnabled = FALSE;
	m_eqpStatus = 0;
}

/////////////////////////////////////////////////////////////////////////////
// CSrvCtrl::AboutBox - Display an "About" box to the user

void CSrvCtrl::AboutBox()
{
	CDialog dlgAbout(IDD_ABOUTBOX_SRV);
	dlgAbout.DoModal();
}


/////////////////////////////////////////////////////////////////////////////
// CSrvCtrl message handlers

void CSrvCtrl::OnEqpNameChanged() 
{
	// TODO: Add notification handler code

	SetModifiedFlag();
}

void CSrvCtrl::OnExportNameChanged() 
{
	// TODO: Add notification handler code

	SetModifiedFlag();
}

void CSrvCtrl::OnEqpStatusChanged() 
{
	// TODO: Add notification handler code
	EqpFcnEventTail = (EqpFcnEventTail+1)%MAXLISTDEPTH;
	if (EqpFcnEventTail > EqpFcnEventHead) EqpFcnEventTail = EqpFcnEventHead; // 10.8.06 don't pass the head
	SetModifiedFlag();
}

void CSrvCtrl::OnEqpTimeChanged() 
{
	// TODO: Add notification handler code
  m_dataTimeStamp = (double)m_eqpTime;
	SetModifiedFlag();
}

void CSrvCtrl::OnSysCommand(UINT nID, LPARAM lParam) 
{
	int i,k,index = (short) nID;  // n.b. in Win16 this was a short !
	DTYPE dout,din;
  static char lcl_devname[32], lcl_devprop[64];
	// TODO: Add your message handler code here and/or call default
	
	if (abs(index) <= 0x8fff)
	{
		if (m_busy) return; else m_busy = 1;
		for (i=0; i<MAXLISTDEPTH; i++)
		{
		  if ((k=(EqpFcnEventTail+i)%MAXLISTDEPTH) >= EqpFcnEventHead) break;
		  if (EqpFcnEventIndexList[k] == index) break;  // found it !
		}
		if (k == EqpFcnEventHead)  // don't increment unless it's new !
		{
		  EqpFcnEventIndexList[EqpFcnEventHead] = index;
		  EqpFcnEventHead = (EqpFcnEventHead+1)%MAXLISTDEPTH;
		}
    m_eqpIndex = index;
    if (lParam & 0x10000) // signal a SetCompletion from below !
    {
      EqpSetCompletion(91,"");
      return;
    }
    m_rpcProtocol = (short)(lParam & 0xffff);
    m_eqpTime = 0;
    m_dataTimeStamp = 0;
    m_eqpStatus = not_posted;
    m_eqpErrorString[0] = 0;
    // get the contract
    GetContractEx(m_devname,m_property,&dout,&din,&m_access,m_eqpIndex);
    m_eqpFormatIn = din.dFormat;
    m_eqpSizeIn = din.dArrayLength;
    strncpy(m_strTagIn,din.dTag,8);
    m_eqpFormatOut = dout.dFormat;
    m_eqpSizeOut = dout.dArrayLength;
    strncpy(m_strTagOut,dout.dTag,8);
    m_eqpNumber = m_devname[0] == '#' ? atoi(&m_devname[1]) : 0;		
    strncpy(lcl_devname,m_devname,16); lcl_devname[16] = 0;
    strncpy(lcl_devprop,m_property,32); lcl_devprop[32] = 0;
    FireEqpFcn(lcl_devname,lcl_devprop, m_eqpSizeOut, m_eqpSizeIn,m_access);
    if (m_eqpStatus == not_posted)
      FireReceive(m_property,m_eqpNumber,(short)MAX(m_eqpSizeIn,m_eqpSizeOut),m_access,m_eqpFormatOut);
    //		PutStatus(m_eqpStatus);
    SETRPCERROREx(m_eqpStatus,m_eqpErrorString,m_eqpIndex);
    m_busy = 0;
	}
	else COleControl::OnSysCommand(nID, lParam);
}

void CSrvCtrl::OnEnabledChanged() 
{
	static char localExport[17],localEqpname[17];

	if (!m_bEnabled && m_enabled)
	{
		strncpy(localExport,(LPCTSTR) m_exportName,16);
		strncpy(localEqpname,(LPCTSTR) m_eqpName,16);
	    m_eqpStatus = RegisterEQP(localExport,localEqpname,m_eqpNumberModules,m_hWnd);
	}
	else if (m_bEnabled && !m_enabled)
	{
		strncpy(localEqpname,(LPCTSTR) m_eqpName,16);
		RemoveEqpfcn(localEqpname,m_hWnd);
	}

	m_bEnabled = m_enabled;
	
	SetModifiedFlag();
}

void CSrvCtrl::OnSetClientSite()
{
	RecreateControlWindow();
}

/* EqpSend() */
/**
 * \b Visual \b Basic: 
 *
 *    \b EqpSend( \e data \b As \b Long) \b As \b Integer
 *
 * \brief Submits outgoing data to the TINE server engine
 * 
 * A server must specify the data it wishes to send to the caller inside the 
 * EqpFcn event routine.  It is assumed that the data to be returned for the property 
 * in question have the format specified in the call to the RegisterProperty() or 
 * RegisterPropertyEx() methods.  EqpSend should be passed the address of the 
 * Visual Basic variable containing the data to be returned.   
 * This generally requires the helper function CADDR() which returns a pointer address
 * cast as a long.  CADDR() in term requires the first element.
 *
 * \param data is the pointer to the data object to be submitted
 *
 * \note With Visual Basic 3.0 VBX controls there are no methods.  Therefore if
 * you are using the TINE server-side WINSRV.VBX in Visual Basic 3.0, \b EqpSend
 * shows up as a property of the control (\b EqpSend \b As \b Long), whose value
 * should be set to the address of the data object to send, again using the CADDR()
 * function.
 *
 * \return 0 if successful, otherwise a TINE completion code 
 *
 * \see EqpSendDataEx(), EqpSendData()
 *
 * \include eg_EqpSend.txt
 */
short CSrvCtrl::EqpSend(long data) 
{
	if (data == NULL) return -1;
//  return PutData((void *)data,m_eqpTime,m_eqpIndex);
  return PutDataEx((void *)data,m_dataTimeStamp,m_eqpIndex);
}

/* EqpRecv() */
/**
 * \b Visual \b Basic: 
 *
 *    \b EqpRecv( \e data \b As \b Long) \b As \b Integer
 *
 * \brief Retrieves incoming data from the TINE server engine
 * 
 * When data are sent to the server, the inArrayLen parameter will be greater than
 * '0' and the server must retrieve the data sent by the caller inside the 
 * EqpFcn event routine.  It is assumed that the data to be retrieved for the property 
 * in question have the format specified in the call to the RegisterProperty() or 
 * RegisterPropertyEx() methods.  EqpRecv should be passed the address of the 
 * Visual Basic variable containing the data to be returned.   
 * This generally requires the helper function CADDR() which returns a pointer address
 * cast as a long.  CADDR() in term requires the first element.
 *
 * \param data is the pointer to the data object to be submitted
 *
 * \note With Visual Basic 3.0 VBX controls there are no methods.  Therefore if
 * you are using the TINE server-side WINSRV.VBX in Visual Basic 3.0, \b EqpRecv
 * shows up as a property of the control (\b EqpRecv \b As \b Long), whose value
 * should be set to the address of the data object into which the incoming data 
 * should be written, again using the CADDR() function.
 *
 * \return 0 if successful, otherwise a TINE completion code 
 *
 * \see EqpRecvData()
 *
 * \include eg_EqpRecv.txt
 */
short CSrvCtrl::EqpRecv(long data) 
{
	if (data == NULL) return -1;
	return GetData((void *)data,m_eqpIndex);
}

/* EqpClearAlarm() */
/**
 * \b Visual \b Basic: 
 *
 *    \b EqpClearAlarm( \e devNumber \b As \b Integer)
 *
 * \brief Clears all alarms associated with the device number given
 * 
 * A server can clear alarms for specific device by giving its device number.  
 * Note that all alarms for the device specified are cleared.  Also note, that 
 * "clearing" an alarm means that its "clear" counter is incremented.  It is not 
 * removed immediately from the local alarm list. For further information, please 
 * see the discussion of the local alarm server. Note that in Win32, this is a method of 
 * the Server ActiveX control, and consequently the "eqm" argument used in the Win16 VBX 
 * is not necessary
 *
 * \param devNumber is the device number for which associated alarms are to be cleared.
 *
 * \note With Visual Basic 3.0 VBX controls there are no methods.  Therefore this
 * functionality is not present in the TINE server-side WINSRV.VBX.  Instead you will have
 * to call this function directly from the tinesv16.dll, passing an extra argument
 * containing the local EqpName.
 *
 * \return none
 *
 * \see EqpClearDeviceAlarm()
 *
 * \include eg_EqpClearAlarm.txt
 */
void CSrvCtrl::EqpClearAlarm(short devNumber) 
{
  static char localEqpname[17];

  strncpy(localEqpname,(LPCTSTR) m_eqpName,16);
  ClearAlarm(localEqpname,devNumber);
}
/* EqpSetAlarm() */
/**
 * \b Visual \b Basic: 
 *
 *    \b EqpSetAlarm( \e devNumber \b As \b Integer, \e almCode \b As \b Long, 
                      \e almData \b As \b Long, \e almFlags \b As \b Long)
 *
 * \brief Sets an alarm with the given alarm code, alarm data, and alarm flags 
 * associated with the device number given
 * 
 * A server can set an alarm for a specific device by giving its device number.  
 * For further information, please see the discussion of the local alarm server.  
 * Note that in Win32, this is a method of 
 * the Server ActiveX control, and consequently the "eqm" argument used in the Win16 VBX 
 * is not necessary
 *
 * \param devNumber is the device number for which the alarm is to be set
 * \param almCode is the alarm code identifying the alarm which is being set
 * \param almData is a pointer which if non-NULL points to the optional 6-byte set of 
 * alarm data associated with the alarm which has been set. As the address is passed, you 
 * must make use of the CADDR() routine to get the address of a Visual Basic array.
 * \param almFlags is a bit pattern containing the alarm flags (such as almINSTANT, etc) 
 * to be applied to the alarm being set. (See discussion on the Local Alarm Server.)
 * Under most circumstances you should simply pass a 0 here.
 *
 * \note With Visual Basic 3.0 VBX controls there are no methods.  Therefore this
 * functionality is not present in the TINE server-side WINSRV.VBX.  Instead you will have
 * to call this function directly from the tinesv16.dll, passing an extra argument
 * containing the local EqpName.
 *
 * \return 0 upon success, otherwise a TINE completion code.
 *
 * \see EqpSetDeviceAlarm()
 *
 * \include eg_EqpSetAlarm.txt
 */
short CSrvCtrl::EqpSetAlarm(short devNumber, long almCode, long almData, short almFlags) 
{
  static char localEqpname[17];
  static BYTE localAlmData[6] = {0,0,0,0,0,0};

  strncpy(localEqpname,(LPCTSTR) m_eqpName,16);

  return SetAlarmEx(localEqpname,devNumber,almCode,almData ? (BYTE *)almData : localAlmData,(BYTE)almFlags);
}

/* EqpRegisterProperty() */
/**
 * \b Visual \b Basic: 
 *
 *    \b EqpRegisterProperty(\e property \b As \b String, \e size \b As \b Integer,
 *                           \e fmt \b As \b Integer, \e access \b As \b Integer,
 *                           \e dsc \b As \b String) \b As \b Integer
 *
 * \brief Simple property registration call
 * 
 * Servers must register their properties by calling either RegisterProperty() or 
 * RegisterPropertyEx().  RegisterProperty() should be called after the enabled property of 
 * the Server ActiveX control has been set to TRUE.  Note that the format parameter 'fmt' in 
 * RegisterProperty() defines the native format for both input and output data sets.  Incoming 
 * data which reaches the EqpFcn event is guaranteed have the format specified.  Note also that 
 * the maximum array size parameter 'siz' is used for both input and output data sets.  If it 
 * is necessary to differentiate between input and output data sets regarding either native 
 * format or array size, then RegisterPropertyEx() should be used.  Note that in Win32, this 
 * is a method of the Server ActiveX control, and consequently the 'eqm' argument used in the 
 * Win16 VBX is not necessary.
 *
 * \param property is the property name to be registered
 * \param size is the maximum array size supported by this property
 * \param fmt is the native format supported by this property
 * \param access is the device access indicater (CA_READ or CA_WRITE or both)
 * \param dsc is the property description.  Note that the maximum and minimum values of the
 * property along with its engineering units and preferred graph type can be submitted with
 * this parameter if this information is containing within brackets '[]'.  This meta 
 * information will be parsed out of the description.
 *
 * \note With Visual Basic 3.0 VBX controls there are no methods.  Therefore this
 * functionality is not present in the TINE server-side WINSRV.VBX.  Instead you will have
 * to call this function directly from the tinesv16.dll, passing an extra argument
 * containing the local EqpName.
 *
 * \return A property indentifier, i.e. a positive number, upon success or 
 * the negative of a TINE completion code if the call fails.
 *
 * \see EqpRegisterProperty()
 *
 * \include eg_EqpRegisterProperty.txt
 */
short CSrvCtrl::EqpRegisterProperty(LPCTSTR property, short size, short fmt, short access, LPCTSTR dsc) 
{
  static char localEqpname[17],localProperty[33],localDescription[48];

  strncpy(localEqpname,(LPCTSTR) m_eqpName,16);
  strncpy(localProperty,property,32);
  strncpy(localDescription,dsc,47);
  return RegisterProperty(localEqpname,localProperty,size,fmt,access,localDescription);
}

/* EqpRegisterModule() */
/**
 * \b Visual \b Basic: 
 *
 *    \b EqpRegisterModule(\e devName \b As \b String, \e devNumber \b As \b Integer) \b As \b Integer
 *
 * \brief Device name registration call
 * 
 * Servers can also register their device names through a startup file. If used, the startup 
 * file is specified by the 6-char local (internal) equipment module name concatenated with 
 * "-device.csv" (e.g. "eqm1-devices.csv"). In cases where this is not practical, a call to 
 * RegisterModuleName() can be made. RegisterModuleName() can be called at any time after 
 * the enabled property of the Server ActiveX control has be set to TRUE. Note that in Win32, 
 * this is a method of the ActiveX control, and consequently the 'eqm' argument used in the 
 * Win16 VBX is not necessary.
 *
 * \param devName is the device name to be assigned to the device number passed
 * \param devNumber is sthe device number whose name is being assigned.
 *
 * \note the devNumber passed in this call must lie between (or equal to) 0 and the number
 * of devices set via the EqpNumberModules ActiveX control property.
 *
 * \note With Visual Basic 3.0 VBX controls there are no methods.  Therefore this
 * functionality is not present in the TINE server-side WINSRV.VBX.  Instead you will have
 * to call this function directly from the tinesv16.dll, passing an extra argument
 * containing the local EqpName.
 *
 * \return 0 upon success or a TINE completion code if the call fails.
 *
 * \see EqpRegisterProperty(), EqpRegisterPropertyEx()
 *
 * \include eg_EqpRegisterModule.txt
 */
short CSrvCtrl::EqpRegisterModule(LPCTSTR devName, short devNumber) 
{
  static char localEqpname[17],localDevname[17];

  strncpy(localEqpname,(LPCTSTR) m_eqpName,13);
  strncpy(localDevname,devName,16);
  return RegisterModuleName(localEqpname,localDevname,devNumber);
}

void CSrvCtrl::OnEqpNumberModulesChanged() 
{
	// TODO: Add notification handler code

	SetModifiedFlag();
}

/* EqpRegisterPropertyEx() */
/**
 * \b Visual \b Basic: 
 *
 *    \b EqpRegisterPropertyEx(\e devProperty \b As \b String, \e inMaxArrayLength \b As \b Long,
 *                           \e inFormat \b As \b Integer, \e inStructureTag \b As \b String,
 *                           \e outMaxArrayLength \b As \b Long, \e outFormat \b As \b Integer, 
 *                           \e outStructureTag \b As \b String, \e access \b As \b Integer,
 *                           \e description \b As \b String) \b As \b Integer
 *
 * \brief Extended property registration call
 * 
 * Servers must register their properties by calling either RegisterProperty() or 
 * RegisterPropertyEx().  RegisterPropertyEx() should be called after the enabled property of 
 * the Server ActiveX control has been set to TRUE.  RegisterPropertyEx() differs from 
 * RegisterProperty() in that the input and output data sets are treated separately.   
 * Also structure tag registration is only possible through this call. 
 * Note that in Win32, this 
 * is a method of the Server ActiveX control, and consequently the 'eqm' argument used in the 
 * Win16 VBX is not necessary.
 *
 * \param devProperty is the property name to be registered
 * \param inMaxArrayLength is the maximum array size for incoming data supported by this property
 * \param inFormat is the native format for incoming data supported by this property
 * \param inStructureTag (if not empty) is a structure 'tag' to be applied to associated with the incoming
 * data in case the inFormat is of type CF_STRUCT (i.e. a user-defined structure).
 * \param outMaxArrayLength is the maximum array size for outgoing data supported by this property
 * \param outFormat is the native format for outgoing data supported by this property
 * \param outStructureTag (if not empty) is a structure 'tag' to be applied to associated with the outgoing
 * data in case the outFormat is of type CF_STRUCT (i.e. a user-defined structure).
 * \param access is the device access indicater (CA_READ or CA_WRITE or both)
 * \param description is the property description.  Note that the maximum and minimum values of the
 * property along with its engineering units and preferred graph type can be submitted with
 * this parameter if this information is containing within brackets '[]'.  This meta 
 * information will be parsed out of the description.
 *
 * \note With Visual Basic 3.0 VBX controls there are no methods.  Therefore this
 * functionality is not present in the TINE server-side WINSRV.VBX.  Instead you will have
 * to call this function directly from the tinesv16.dll, passing an extra argument
 * containing the local EqpName.
 *
 * \return A property indentifier, i.e. a positive number, upon success or 
 * the negative of a TINE completion code if the call fails.
 *
 * \see EqpRegisterProperty()
 *
 * \include eg_EqpRegisterPropertyEx.txt
 */
short CSrvCtrl::EqpRegisterPropertyEx(LPCTSTR devProperty, long inMaxArrayLength, short inFormat, 
									  LPCTSTR inStructureTag, long outMaxArrayLength, short outFormat, 
									  LPCTSTR outStructureTag, short access, LPCTSTR description) 
{
  static char localEqpname[17],localProperty[33],localDescription[48];
  static char localStrTagIn[9],localStrTagOut[9];

  strncpy(localEqpname,(LPCTSTR) m_eqpName,16);
  strncpy(localProperty,devProperty,32);
  strncpy(localDescription,description,47);
  strncpy(localStrTagIn,inStructureTag,8);
  strncpy(localStrTagOut,outStructureTag,8);
  return RegisterPropertyEx(localEqpname,localProperty,inMaxArrayLength,
	                        inFormat,localStrTagIn,outMaxArrayLength,outFormat,
							localStrTagOut,access,localDescription);
}

/* EqpSetDeviceAlarm() */
/**
 * \b Visual \b Basic: 
 *
 *    \b EqpSetDeviceAlarm( \e devName \b As \b String, \e almCode \b As \b Long, 
                      \e almData \b As \b Long, \e almFlags \b As \b Long)
 *
 * \brief Sets an alarm with the given alarm code, alarm data, and alarm flags 
 * associated with the device name given
 * 
 * A server can set an alarm for a specific device by giving its device name.  
 * For further information, please see the discussion of the local alarm server.  
 * Note that in Win32, this is a method of 
 * the Server ActiveX control, and consequently the "eqm" argument used in the Win16 VBX 
 * is not necessary
 *
 * \param devName is the device name for which the alarm is to be set
 * \param almCode is the alarm code identifying the alarm which is being set
 * \param almData is a pointer which if non-NULL points to the optional 6-byte set of 
 * alarm data associated with the alarm which has been set. As the address is passed, you 
 * must make use of the CADDR() routine to get the address of a Visual Basic array.
 * \param almFlags is a bit pattern containing the alarm flags (such as almINSTANT, etc) 
 * to be applied to the alarm being set. (See discussion on the Local Alarm Server.)
 * Under most circumstances you should simply pass a 0 here.
 *
 * \note With Visual Basic 3.0 VBX controls there are no methods.  Therefore this
 * functionality is not present in the TINE server-side WINSRV.VBX.  Instead you will have
 * to call this function directly from the tinesv16.dll, passing an extra argument
 * containing the local EqpName.
 *
 * \return 0 upon success, otherwise a TINE completion code.
 *
 * \see EqpSetAlarm()
 *
 * \include eg_EqpSetDeviceAlarm.txt
 */
short CSrvCtrl::EqpSetDeviceAlarm(LPCTSTR devName, long almCode, long almData, short almFlags) 
{
  static char l_eqpname[17],l_devname[17];
  static BYTE l_AlmData[6] = {0,0,0,0,0,0};

  strncpy(l_eqpname,(LPCTSTR) m_eqpName,16);
  strncpy(l_devname,devName,16);

  return SetDeviceAlarm(l_eqpname,l_devname,almCode,almData ? (BYTE *)almData : l_AlmData,(BYTE)almFlags);
}

/* EqpClearDeviceAlarm() */
/**
 * \b Visual \b Basic: 
 *
 *    \b EqpClearDeviceAlarm( \e devName \b As \b String)
 *
 * \brief Clears all alarms associated with the device name given
 * 
 * A server can clear alarms for specific device by giving its device name.  
 * Note that all alarms for the device specified are cleared.  Also note, that 
 * "clearing" an alarm means that its "clear" counter is incremented.  It is not 
 * removed immediately from the local alarm list. For further information, please 
 * see the discussion of the local alarm server. Note that in Win32, this is a method of 
 * the Server ActiveX control, and consequently the "eqm" argument used in the Win16 VBX 
 * is not necessary
 *
 * \param devName is the device name for which associated alarms are to be cleared.
 *
 * \note With Visual Basic 3.0 VBX controls there are no methods.  Therefore this
 * functionality is not present in the TINE server-side WINSRV.VBX.  Instead you will have
 * to call this function directly from the tinesv16.dll, passing an extra argument
 * containing the local EqpName.
 *
 * \return none
 *
 * \see EqpClearAlarm()
 *
 * \include eg_EqpClearDeviceAlarm.txt
 */
void CSrvCtrl::EqpClearDeviceAlarm(LPCTSTR devName) 
{
  static char l_eqpname[17],l_devname[17];

  strncpy(l_eqpname,(LPCTSTR) m_eqpName,16);
  strncpy(l_devname,devName,16);
  ClearDeviceAlarm(l_eqpname,l_devname);
}

/* EqpSetCompletion() */
/**
 * \b Visual \b Basic: 
 *
 *    \b EqpSetCompletion( \e CompCode \b As \b Integer, \e CompString \b As \b String)
 *
 * \brief Set the completion code for the currently active call and advances the 
 * tail pointer in the local event queue.
 *
 * A server must set the completion code for every call into its equipment module handler
 * routine.  Failing to do so will result in client calls failing with 'command not posted'
 * as the return message.  The completion code can be set to any of the system standard 
 * error codes or to a user-define error code beginning at 512.
 * A server can also return a 32-character error status string to accompany a user-defined error. 
 * To do so, you should use this method.  Note that supplying an error string to a TINE-system
 * error code will have no effect, as the system error string will always be used in this case.
 *
 * \param CompCode is the completion code to be applied to the call
 * \param CompString is the string description of the completion to accompany the integer
 * value.
 *
 * \note The legacy property 'EqpStatus' will also apply a completion code to a call, but not
 * be able to apply a completion string.  Do NOT use both EqpStatus AND EqpSetCompletion()
 * or call EqpSetCompletion() more than once per equipment module handler event.  This call
 * serves to advance the tail of the event queue maintained by the Server ActiveX control.  
 * This method is used in lieu of a return code (which ActiveX events do not have!) to notify
 * the TINE server engine that the call has completed.
 *
 * \note With Visual Basic 3.0 VBX controls there are no methods.  Therefore this
 * functionality is not present in the TINE server-side WINSRV.VBX.  Instead you will have
 * to call this function directly from the tinesv16.dll, passing an extra argument
 * containing the local EqpName.  Or you can use the \b EqpStatus \b As \b Integer property 
 * to set the completion code, and use the \b ErrorString \b As \b String property to set an
 * accompanying completion error string.
 *
 * \return none
 *
 * \include eg_EqpSetCompletion.txt
 */
void CSrvCtrl::EqpSetCompletion(short CompCode, LPCTSTR CompString) 
{
	short index;
	index = EqpFcnEventIndexList[EqpFcnEventTail];
	EqpFcnEventTail = (EqpFcnEventTail+1)%MAXLISTDEPTH;
	if (EqpFcnEventTail > EqpFcnEventHead) EqpFcnEventTail = EqpFcnEventHead; // 10.8.06 don't pass the head
  m_eqpStatus = CompCode;
	strncpy(m_eqpErrorString,CompString,32);
	SETRPCERROREx(m_eqpStatus,m_eqpErrorString,index);
}

/* EqpGetModuleNumber() */
/**
 * \b Visual \b Basic: 
 *
 *    \b EqpGetModuleNumber(\e DevName \b As \b String ) \b As \b Integer
 *
 * \brief retrieves the device number assigned to the device name passed
 * 
 * A server can register module names (i.e. device names) at initialization time, thereby
 * assigning a number to a device name.  As the number will actually be the index into the
 * local device name table held by the TINE engine, it will always be between 0 and the number 
 * of device modules.  It is stored as a hash table, so that lookups are fast as possible.
 * In this simple version of the call, it is assumed that all registered properties share 
 * the same device list.
 *
 * \param DevName is the device name for which the assigned module number (i.e. device number) is
 * to be returned
 *
 * \note With Visual Basic 3.0 VBX controls there are no methods.  Therefore this
 * functionality is not present in the TINE server-side WINSRV.VBX.  Instead you will have
 * to call this function directly from the tinesv16.dll, passing an extra argument
 * containing the local EqpName.
 *
 * \return The device number associated with the device name if successful, otherwise 
 * the negative of a TINE completion code.
 *
 * \see EqpGetModuleNumberEx()
 *
 * \include eg_EqpGetModuleNumber.txt
 */
short CSrvCtrl::EqpGetModuleNumber(LPCTSTR DevName) 
{
	static char l_eqpname[17],l_devname[17];
	strncpy(l_eqpname,(LPCTSTR) m_eqpName,16);
  strncpy(l_devname,DevName,16);
	return GetDeviceNumber(l_eqpname,l_devname);
}

/* EqpGetModuleName() */
/**
 * \b Visual \b Basic: 
 *
 *    \b EqpGetModuleName(\e DevNumber \b As \b Integer ) \b As \b String
 *
 * \brief retrieves the device name assigned to the device number passed
 * 
 * A server can register module names (i.e. device names) at initialization time, thereby
 * assigning a number to a device name.  As the number will actually be the index into the
 * local device name table held by the TINE engine, it will always be between 0 and the number 
 * of device modules.  It is stored as a hash table, so that lookups are fast as possible.
 * In this simple version of the call, it is assumed that all registered properties share 
 * the same device list.  Under some circumstance the server might wish to query what name was
 * assigned to what number during initialization.  For instance, device names can also be 
 * assigned via the startup configuration file \<equipment name\>-devices.csv, in which case
 * the running server does not have device name hard coded and has no apriori knowledge as to
 * what name has been matched to what number.
 *
 * \param DevNumber is the device number for which the assigned device name is
 * to be returned
 *
 * \note With Visual Basic 3.0 VBX controls there are no methods.  Therefore this
 * functionality is not present in the TINE server-side WINSRV.VBX.  Instead you will have
 * to call this function directly from the tinesv16.dll, passing an extra argument
 * containing the local EqpName.
 *
 * \return The device name associated with the device number if successful, otherwise 
 * an empty string.
 *
 * \see EqpGetModuleNumberEx(), EqpGetModuleNumber()
 *
 * \include eg_EqpGetModuleName.txt
 */
BSTR CSrvCtrl::EqpGetModuleName(short DevNumber) 
{
	CString strResult;
	static char l_eqpname[17],l_devname[17];
	strncpy(l_eqpname,(LPCTSTR) m_eqpName,16);
  GetDeviceName(l_eqpname,DevNumber,l_devname);
	//strResult.Insert(0,l_devname);
  strResult = (CString)l_devname;
	//strncpy((char *)(LPCTSTR)strResult.GetBuffer(16),l_devname,16);
	return strResult.AllocSysString();
}

short CSrvCtrl::EqpGetPropertyId(LPCTSTR devProperty) 
{
  static char l_eqpname[17],l_devprop[33];
  strncpy(l_eqpname,(LPCTSTR) m_eqpName,16);
  strncpy(l_devprop,devProperty,32);
  return GetPropertyId(l_eqpname,l_devprop);
}

/* EqpRecvData() */
/**
 * \b Visual \b Basic: 
 *
 *    \b EqpRecvData( \e data \b As \b Variant) \b As \b Integer
 *
 * \brief Retrieves incoming data from the TINE server engine
 * 
 * When data are sent to the server, the inArrayLen parameter will be greater than
 * '0' and the server must retrieve the data sent by the caller inside the 
 * EqpFcn event routine.  It is assumed that the data to be retrieved for the property 
 * in question have the format specified in the call to the RegisterProperty() or 
 * RegisterPropertyEx() methods.  EqpRecvData should be passed the Visual Basic variable 
 * which is to be filled with the incoming data.   The variable passed must be a fundemental 
 * OLE data type (i.e. short (integer), long, float (single), double, byte, or string.  
 * If composite or user-defined types are passed, then the first element of the user-type 
 * should be passed.
 *
 * \param data is the data object to be filled with the incoming data.
 *
 * \note With Visual Basic 3.0 VBX controls there are no methods.  Therefore this
 * functionality is not present in the TINE server-side WINSRV.VBX. Use \b EqpRecv.
 *
 * \return 0 if successful, otherwise a TINE completion code 
 *
 * \see EqpRecv()
 *
 * \include eg_EqpRecvData.txt
 */
short CSrvCtrl::EqpRecvData(const VARIANT FAR& data) 
{
	void *vp, *strPtr = NULL;
	SAFEARRAY FAR *psa;
	int cc = 0;
	short index;
	index = EqpFcnEventIndexList[EqpFcnEventTail];
	if( cc = VariantToVoid( (VARIANT FAR&)data, &vp, &psa, &strPtr ) ) return cc;
	if( cc = GetData( vp, index) ) return cc;
	if( psa ) SafeArrayUnaccessData(psa);   // unlock memory	
	cc = VoidToVariant( vp, (VARIANT FAR&)data );
	if( strPtr ) GlobalFree( strPtr );
	return cc;
}

short CSrvCtrl::VariantToVoid( VARIANT& vData, void ** vp, SAFEARRAY **psaRet, void ** strPtr )
{
  SAFEARRAY FAR *psa;
  CString str;
  void * vptr;
  long datasize;
  VARIANT *vD;

  *psaRet = NULL;
  *strPtr = NULL;

  if(LOBYTE(vData.vt)==VT_VARIANT) 
    vD = vData.pvarVal;	
  else 
    vD =&vData;

  if(vD->vt==VT_EMPTY || vD->vt==VT_NULL )  
  {
    *vp = NULL; 
    return -1;
	}	

  if(vD->vt & VT_ARRAY)
  {
    if(LOBYTE(vD->vt)==VT_BSTR) 
    {
      *vp = NULL;
      return -1; // not supported yet
    }
    psa = vD->vt & VT_BYREF ? *(SAFEARRAY FAR * FAR *)vD->parray : vD->parray;
    if(SafeArrayAccessData(psa,(void HUGEP* FAR*)vp)) 
    {
      *vp = NULL;
      return -1;
    }			
    *psaRet=psa;
    return 0;       
  }
  else 
  {		
    vptr = vD->vt & VT_BYREF ? vD->byref : &(vD->byref);
    if( LOBYTE(vD->vt)==VT_BSTR ) 
    {
      if( (datasize = SysStringLen(*(BSTR *)vptr)) == 0) 
      {
        *vp = NULL;
        return -1;
      }
      if((*vp = GlobalAlloc( GPTR,abs(datasize)+1 )) == NULL ) return -1;
      if( datasize > 0 )
      {
        str = CString((LPCWSTR)*(BSTR *)vptr);
        memcpy( *vp, str.GetBuffer(0), datasize );
        *strPtr = *vp;
      }			
    }
    else *vp = vptr;
  }
  return 0;
}


short CSrvCtrl::VoidToVariant( void * vp,VARIANT& vData )
{
	SAFEARRAY FAR *psa;
	void HUGEP* ppvdata;
	void *vptr;
	CString str;
	long blockSize,lb,ub;

	if (vData.vt==VT_ERROR && vData.scode==DISP_E_PARAMNOTFOUND) return -1;
	if (vData.vt==VT_EMPTY || vData.vt==VT_NULL) return -1;

	if (LOBYTE(vData.vt)==VT_BSTR)
	{
		vptr = vData.vt & VT_BYREF ? vData.byref : &vData.byref;

		str=CString((LPCSTR)vp);
		str.SetSysString((BSTR *)vptr);		
		return 0;
	}

	if(vData.vt & VT_ARRAY)
	{
		if(LOBYTE(vData.vt)==VT_BSTR) return -1; // not supported yet
		psa = vData.vt & VT_BYREF ? *(SAFEARRAY FAR * FAR *)vData.parray : vData.parray;
		if(SafeArrayAccessData(psa,(void HUGEP* FAR*)&ppvdata)) return -1;
		if(SafeArrayGetDim(psa) != 1 ) return -1;  // don't handle multiple arrays
		if(SafeArrayGetUBound(psa, 1, &ub) != S_OK ) return -1;
		if(SafeArrayGetLBound(psa, 1, &lb) != S_OK ) return -1;
		blockSize = SafeArrayGetElemsize(psa) * (ub - lb +1);
		if(IsBadWritePtr(ppvdata, blockSize)) return -1;
		memcpy(ppvdata,vp,blockSize);
		SafeArrayUnaccessData(psa);
	}		
	else 
	{
		if( vData.vt & VT_BYREF )
		{
			blockSize = sizeof(void *);
			vptr = vData.byref;
		}
		else
		{
			vptr = &vData.byref;
			switch(LOBYTE(vData.vt))
			{
			case VT_UI1: blockSize = 1; break;		
			case VT_I2:  blockSize = 2; break;
			case VT_I4:  blockSize = 4; break;
			case VT_R4:  blockSize = 4; break;
			case VT_R8:  blockSize = 8; break;
			case VT_CY:  blockSize = 8; break;
			default:	return -1;   // not expected type
			break;
			}
		}
		if (IsBadWritePtr(vptr,blockSize)) return -1;
		memcpy(vptr,vp,blockSize);
	}
	
	return 0;
}

/* EqpSendData() */
/**
 * \b Visual \b Basic: 
 *
 *    \b EqpSendData( \e data \b As \b Variant) \b As \b Integer
 *
 * \brief Submits outgoing data to the TINE server engine
 * 
 * A server must specify the data it wishes to send to the caller inside the 
 * EqpFcn event routine.  It is assumed that the data to be returned for the property 
 * in question have the format specified in the call to the RegisterProperty() or 
 * RegisterPropertyEx() methods.  EqpSendData should be passed the Visual Basic variable 
 * containing the data to be returned.   The variable passed must be a fundemental 
 * OLE data type (i.e. short (integer), long, float (single), double, byte, or string.  
 * If composite or user-defined types are passed, then the first element of the user-type 
 * should be passed.
 *
 * \param data is the data object to be submitted.
 *
 * \note With Visual Basic 3.0 VBX controls there are no methods.  Therefore this
 * functionality is not present in the TINE server-side WINSRV.VBX.  Use \b EqpSend.
 *
 * \return 0 if successful, otherwise a TINE completion code 
 *
 * \see EqpSendDataEx(), EqpSend()
 *
 * \include eg_EqpSendData.txt
 */
short CSrvCtrl::EqpSendData(const VARIANT FAR& data) 
{	
	int cc = 0;
  void *vp, *strPtr = NULL;
	SAFEARRAY FAR *psa;
	short index;
	index = EqpFcnEventIndexList[EqpFcnEventTail];
	if( cc = VariantToVoid((VARIANT FAR&)data, &vp, &psa, &strPtr ) ) return cc;
//  cc = PutData((void *)vp,m_eqpTime,index);
  cc = PutDataEx((void *)vp,m_dataTimeStamp,index);
  SafeArrayUnaccessData(psa);
	if( strPtr ) GlobalFree( strPtr );
	return cc;
}

char * getVariantAsString(VARIANT& vData)  
{
	void *vptr,*dptr=NULL;
	CString str;
	long datasize;
	char tmp[128];

	if (vData.vt == VT_EMPTY || vData.vt == VT_ERROR) return NULL;
	switch(LOBYTE(vData.vt))
	{
		case VT_UI1: sprintf(tmp,"%d",vData.bVal); 
			 datasize = strlen(tmp); str = tmp; break;
		case VT_I2:  sprintf(tmp,"%d",vData.iVal); 
			 datasize = strlen(tmp); str = tmp; break;
		case VT_I4:  sprintf(tmp,"%d",vData.lVal); 
			 datasize = strlen(tmp); str = tmp; break;
		case VT_R4:  sprintf(tmp,"%g",vData.fltVal); 
			 datasize = strlen(tmp); str = tmp; break;
		case VT_R8:  sprintf(tmp,"%g",vData.dblVal); 
			 datasize = strlen(tmp); str = tmp; break;
		case VT_BSTR:
	     vptr = vData.vt & VT_BYREF ? vData.byref : &vData.byref;
		   if ((datasize=SysStringLen(*(BSTR *)vptr)) == 0) return NULL;
		   str=CString((LPCWSTR)*(BSTR *)vptr);
       break;
		default:     return NULL;
	}
	if ((dptr=GlobalAlloc(GPTR,abs(datasize)+1)) == NULL) return NULL;
	memcpy(dptr,str.GetBuffer(0),datasize);
	str.ReleaseBuffer();

	return (char*)dptr;
}
#define LONGNOTFOUND 2147483648
long getVariantAsLong(const VARIANT& V)  
{
	if (V.vt == VT_EMPTY || V.vt == VT_ERROR) return LONGNOTFOUND;
	switch(LOBYTE(V.vt))
	{
		case VT_BOOL: return V.boolVal ? 1 : 0;
		case VT_UI1: return (long)V.bVal;
		case VT_I2:  return (long)V.iVal;
		case VT_I4:  return (long)V.lVal;
		case VT_R4:  return (long)V.fltVal;
		case VT_R8:  return (long)V.dblVal;
		default:     return 0;
	}
	return 0;
}

/* EqpSendDataEx() */
/**
 * \b Visual \b Basic: 
 *
 *    \b EqpSendDataEx( \e data \b As \b Variant [, \e size \b As \b Long] ) \b As \b Integer
 *
 * \brief Submits outgoing data to the TINE server engine
 * 
 * A server must specify the data it wishes to send to the caller inside the 
 * EqpFcn event routine.  It is assumed that the data to be returned for the property 
 * in question have the format specified in the call to the RegisterProperty() or 
 * RegisterPropertyEx() methods.  EqpSendDataEx should be passed the Visual Basic variable 
 * containing the data to be returned.   The variable passed must be a fundemental 
 * OLE data type (i.e. short (integer), long, float (single), double, byte, or string.  
 * If composite or user-defined types are passed, then the first element of the user-type 
 * should be passed.  Using this call, you can truncate the amount of data being returned
 * by specifying the size parameter to be less than the outArrayLen demanded by the client.
 *
 * \param data is the data object to be submitted.
 * \param size [optional] is the size of the data object to be returned.  If not supplied, then
 * the value of 'outArrayLen' is assumed
 *
 * \note With Visual Basic 3.0 VBX controls there are no methods.  Therefore this
 * functionality is not present in the TINE server-side WINSRV.VBX.  Use \b EqpSend.
 * If you wish to truncate the number of elements returned in the call to less than what
 * was asked for with the calling parameter outArrayLen in the Equipment Module handler routine,
 * the you should set outArrayLen to the truncated size before quitting the subroutine.
 *
 * \return 0 if successful, otherwise a TINE completion code 
 *
 * \see EqpSendData(), EqpSend()
 *
 * \include eg_EqpSendDataEx.txt
 */
short CSrvCtrl::EqpSendDataEx(const VARIANT FAR& data, const VARIANT FAR& size) 
{
  void *vp, *strPtr;
  SAFEARRAY FAR *psa;
  int cc = 0;
  long doutlen;
  short index,tmpAcc;
  DTYPE dout,din;
  char tmpDev[17],tmpPrp[33];

  index = EqpFcnEventIndexList[EqpFcnEventTail];
  if ((doutlen=getVariantAsLong(size)) != LONGNOTFOUND)
  {
    GetContractEx(tmpDev,tmpPrp,&dout,&din,&tmpAcc,index);  // make sure to get the proper input length
    if (doutlen < (long)dout.dArrayLength)
      AdjustContract(NULL,NULL, doutlen,din.dArrayLength,index);
	}
  if( cc = VariantToVoid((VARIANT FAR&)data, &vp, &psa, &strPtr ) ) return cc;
  cc = PutDataEx((void *)vp,m_dataTimeStamp,index);
  SafeArrayUnaccessData(psa);
  if( strPtr ) GlobalFree( strPtr );
  return cc;
}

/* GetCallerInformation() */
/**
 * \b Visual \b Basic: 
 *
 *    \b GetCallerInformation( [\e ClientName \b As \b String], [ \e ClientAddress \b As \b String] ) \b As \b Integer
 *
 * \brief retrieves the identity of the caller accessing the property in question
 * 
 * A server can retrieve the caller name and address at the time the EqpFcn() 
 * equipment function event service routine is fired.  The information returned 
 * corresponds to the first caller in the list of callers interested in the given 
 * contract.  For single commands and reads, this is typically a list of one caller.
 *
 * \param ClientName [optional] is the string variable into which the caller's client name is to
 * be written
 * \param ClientAddress [optional] is the string variable into which the caller's network address
 * is to be written
 *
 * \note if the variables ClientName and ClientAddress are used, they must be pre-dimensioned to 
 * at least 16 characters.
 * \see EqpRecvData(), EqpRecv()
 *
 * \note With Visual Basic 3.0 VBX controls there are no methods.  Therefore this
 * functionality is not present in the TINE server-side WINSRV.VBX.  Use the \b ClientName
 * \b As \b String and \b ClientAddress \b As \b String properties instead.
 *
 * \return 0 if successful, otherwise a TINE completion code 
 *
 * \include eg_GetCallerInformation.txt
 */
short CSrvCtrl::GetCallerInformation(const VARIANT FAR& ClientName,const VARIANT FAR& ClientAddress) 
{
	void *vptr,*dptr=NULL;
	CString str;
	int len;
	char clnnam[17],clnaddr[17];

  GetCallerInfo(clnnam,clnaddr);
	if (!(ClientName.vt == VT_EMPTY || ClientName.vt == VT_ERROR))
	{
		if (LOBYTE(ClientName.vt) == VT_BSTR && (ClientName.vt & VT_BYREF))
		{
      vptr = ClientName.byref;
		  if ((len=SysStringLen(*(BSTR *)vptr)) < 16) return -1;
//		  memcpy(vptr,clnnam,16);
		  str=CString(clnnam);
		  str.SetSysString((BSTR *)vptr);
		}
		else return -2;
	}
	if (!(ClientAddress.vt == VT_EMPTY || ClientAddress.vt == VT_ERROR))
	{
		if (LOBYTE(ClientAddress.vt) == VT_BSTR && (ClientAddress.vt & VT_BYREF))
		{
      vptr = ClientAddress.byref;
		  if ((len=SysStringLen(*(BSTR *)vptr)) < 16) return -1;
//		  memcpy(vptr,clnaddr,16);
		  str=CString(clnaddr);
		  str.SetSysString((BSTR *)vptr);
		}
		else return -2;
	}
	return 0;
}

/* GetDataTimeStamp() */
/**
 * \b Visual \b Basic: 
 *
 *    \b GetDataTimeStamp() As \b Double 
 *
 * \brief Returns the current system time as a TINE data time stamp.
 * 
 * A server can apply its own timestamp to a data request, for instance to
 * indicate when the data were actually read out.  
 * A valid data time stamp can be obtained by calling GetDataTimeStamp().
 *
 * \note With Visual Basic 3.0 VBX controls there are no methods.  Therefore this
 * functionality is not present in the TINE server-side WINSRV.VBX, and instead must be
 * called as a function from the DLL interface.
 *
 * \return The current system time as a TINE data time stamp.
 *
 */
double CSrvCtrl::GetDataTimeStamp() 
{
  return makeDataTimeStamp();
}

/* SetDataTimeStamp() */
/**
 * \b Visual \b Basic: 
 *
 *    \b SetDataTimeStamp(\e dataTime \b As \b Double )
 *
 * \brief Sets the data time stamp according to the value given
 * 
 * A server can apply its own timestamp to a data request, for instance to
 * indicate when the data were actually read out.  
 * If used, this method should be called prior to submitting the data via
 * a call to one of the EqpSendData() methods.  If not used, then the data
 * time stamp will be the current time read from the system clock.
 *
 * \param dataTime is the data timestamp to be assigned to the returned data.
 * A valid data time stamp can be obtained by calling GetDataTimeStamp().
 *
 * \note With Visual Basic 3.0 VBX controls there are no methods.  Therefore this
 * functionality is not present in the TINE server-side WINSRV.VBX, and instead must be
 * called as a function from the DLL interface.
 *
 * \note Data time stamp values under 1000 are ignored by the TINE engine.
 *
 * \return 0 if successful, otherwise a TINE completion code 
 *
 */
void CSrvCtrl::SetDataTimeStamp(double dataTime) 
{
  m_dataTimeStamp = dataTime;
}

/* _ScheduleProperty() */
/**
 * \b Visual \b Basic: 
 *
 *    \b ScheduleProperty(\e PropertyList \b As \b String ) \b As \b Integer
 *
 * \brief Schedules all contracts attached the any of the properties in the PropertyList
 *  for immediate delivery.
 * 
 * Under some circumstances, a server might want to reduce the latency between local
 * data acquistion and delivery to clients interested in the data.  By calling 'ScheduleProperty()
 * you can ensure that all contracts involving the 
 * properties supplied in the argument list will be called immediately
 * and the results sent immediately to all clients in the associated client lists.
 *
 * \param PropertyList is a single string containing a comma separated list of properties
 * which are to be 'scheduled' for immediate access.
 *
 * \note With Visual Basic 3.0 VBX controls there are no methods.  Therefore this
 * functionality is not present in the TINE server-side WINSRV.VBX, and instead must be
 * called as a function from the DLL interface.
 *
 * \return 0 if successful, otherwise a TINE completion code 
 *
 * \include eg_ScheduleProperty.txt
 */
short CSrvCtrl::_ScheduleProperty(LPCTSTR PropertyList) 
{
  static char localEqpname[17];
  static char localPropertyList[512];
  strncpy(localEqpname,(LPCTSTR) m_eqpName,16);
  strncpy(localPropertyList,PropertyList,512);
  return ScheduleProperty(localEqpname,localPropertyList);
}

/* EqpGetModuleNumberEx() */
/**
 * \b Visual \b Basic: 
 *
 *    \b EqpGetModuleNumberEx(\e DevName \b As \b String [, \e DevProperty \b As \b String] ) \b As \b Integer
 *
 * \brief retrieves the device number assigned to the device name and device property passed
 * 
 * A server can register module names (i.e. device names) at initialization time, thereby
 * assigning a number to a device name.  Under some circumstances, some registered properties 
 * will have a different property list than others.  When this is the case for specific property,
 * it requires a registered 'meta' property to be handled by the server with the same name as 
 * the property but appended with ".NAM".  The TINE engine will then make a one time query at 
 * initialization time of this property-specific device list and use it to ascertain name to
 * number cross-references.
 *
 * \param DevName is the device name for which the assigned module number (i.e. device number) is
 * to be returned.
 * \param DevProperty [optional] is the device property for which the assigned device number is
 * to be returned.  If DevProperty is omitted, the call reverts to the behavior of 
 * EqpGetModuleNumber().
 *
 * \note With Visual Basic 3.0 VBX controls there are no methods.  Therefore this
 * functionality is not present in the TINE server-side WINSRV.VBX.  
 *
 * \return The device number associated with the device name if successful, otherwise 
 * the negative of a TINE completion code.
 *
 * \see EqpGetModuleNumber()
 *
 * \include eg_EqpGetModuleNumberEx.txt
 */
short CSrvCtrl::EqpGetModuleNumberEx(LPCTSTR DevName, const VARIANT FAR& DevProperty) 
{
	static char l_eqpname[17],l_devname[17],l_prpname[33];
	CString str;
  int strsize, devnr = -1;
	char *prp=NULL;
  void *vptr;

	if (!(DevProperty.vt == VT_EMPTY || DevProperty.vt == VT_ERROR))
	{
		if (LOBYTE(DevProperty.vt) == VT_BSTR)
    {
      vptr = DevProperty.vt & VT_BYREF ? DevProperty.byref : (void *)&(DevProperty.byref);
      //vptr = DevProperty.byref;
      if ((strsize=SysStringLen(*(BSTR *)vptr)) > 0) 
      {
        prp = l_prpname;
        str = CString((LPCWSTR)*(BSTR *)vptr);
        memcpy(prp, str.GetBuffer(0), strsize );
        prp[strsize] = 0;
      }			
    }
	}
	strncpy(l_eqpname,(LPCTSTR) m_eqpName,16);
  strncpy(l_devname,DevName,16);
	devnr = GetDeviceNumberEx(l_eqpname,l_devname,prp);
  return devnr;
}
/** \var BOOL CSrvCtrl::m_enabled
 * \b Visual \b Basic: 
 * \n \n \b enabled
 *
 * \brief Enables the Server ActiveX control when set to TRUE.  This action
 * begins server intialization.
 * 
 * \note The correct point in the server initialization to set this value to TRUE is important.
 * The property values for EqpName, ExportName and EqpNumberModules must be set 
 * prior to enabling the ActiveX control. Likewise any system calls which affect the performance,
 * the resources, etc. of the TINE server engine should be made prior to enabling it.
 * And only after the server has been enabled will the Property Registration and Device 
 * Registration calls succeed.
 *
 */
/** \var BOOL CSrvCtrl::m_exportName
 * \b Visual \b Basic: 
 * \n \n \b ExportName \b As \b String
 *
 * \brief Sets the "exported" device server name bound to the Equipment Module.
 * 
 * This is the
 * system-wide name used by all clients to address your server.  This name therefore need to
 * be unique system-wide.  Note however that two different server can export the same device
 * server name if they are registered on different contexts (see FEC registration). It is
 * none the less a good idea to try to keep the export name unique even across diffent contexts
 * in order to avoid confusion.  Although the context is a part of the name space, clients may
 * omit it, in which case the address resolution for a given device server name will return the
 * first "match".
 *
 * \note The Export Name can be up to 16 characters long.
 * 
 * \note This property must be set prior to enabling the ActiveX control otherwise the 
 * server's information will not be correctly exported.
 *
 */
/** \var BOOL CSrvCtrl::m_eqpName
 * \b Visual \b Basic: 
 * \n \n \b EqpName \b As \b String
 *
 * \brief Sets the "local" Equipment Module Name for this ActiveX control.  
 *
 * This is the
 * process-specific name used by the TINE server engine to distinguish this device server
 * from others bound to the same process.  This name need only be unique within the 
 * application.  
 *
 * \note Only the first 6 characters of this name will be utilized, and it is these first
 * 6 characters which must be unique.
 * 
 * \note This property must be set prior to enabling the ActiveX control otherwise the 
 * server's information will not be correctly exported.
 *
 */
/** \var BOOL CSrvCtrl::m_eqpNumberModules
 * \b Visual \b Basic: 
 * \n \n \b EqpNumberModules \b As \b Integer
 *
 * \brief Sets the size of the module device table bound to the server.
 * 
 * \note This property must be set prior to enabling the ActiveX control otherwise the 
 * server's device information will not be correctly exported.  For instance, it will then
 * be impossible to assign a device name "DEVICE13" to device number 13.
 *
 */

/** \fn void CSrvCtrl::FireEqpFcn(LPCTSTR devName, LPCTSTR devProperty, long outArrayLen, long inArrayLen, short devAccess)
 * \b Visual \b Basic: 
 *
 *    \b _EqpFcn(\e DevName \b As \b String, \e DevProperty \b As \b String, 
 *              \e outArrayLen \b As \b Long, \e inArrayLen \b As \b Long,
 *              \e devAccess \b As \b Integer)
 *
 * \brief is the Equipment Function handler routine (ActiveX event) which processes all command and 
 * requests for data.  Inside Visual Basic, the name of the event will pre-pend the module's
 * name, producing event routines such as Srv1_EqpFcn(), etc.
 *
 * This corresponds to the Equipment Module handler routine (See the discussion on Equipment Modules).
 * This routine handles all client user requests (either directly or indirectly).  Note that 
 * when N different clients establish a data link for the same data request, only one event is 
 * generated (at the designated polling rate).
 *
 * \param devName is a pointer to the 16-character device name requested by the caller.  In 
 * many cases, equipment modules make use of device number rather than (or in addition to) 
 * device names.  In such cases, a legitimate device "name" might be "#37" for example. 
 * \param devProperty is a pointer to the 32-character device property requested by the caller.
 * In order for the device property to be passed in this routine, it must be exported.  A 
 * request from a caller for a property which has not been exported will automatically 
 * receive an "illegal_property" return code. 
 * \param outArrayLen is the array length of the requested output data set.  This is generally 
 * the number of elements in units of the native format specified in the call to the 
 * EqpRegisterProperty() or EqpRegisterPropertyEx()  methods.  Note that for format type 
 * CF_STRUCT (i.e. a user-defined structure), if the structure is un-tagged, then the value of 
 * outArrayLen& gives the total number of bytes.  Likewise, for format CF_TEXT, it gives the 
 * string length.
 * \param inArrayLen is the array length of the incoming input data set. This is generally the 
 * number of elements in units of the native format specified in the call to the 
 * EqpRegisterProperty() or EqpRegisterPropertyEx() methods.  (See above remarks for 
 * outArrayLen&).  The incoming data set is guaranteed to have the format given in the 
 * property registration.
 * \param devAccess is the access code for the request (see the discussion on Data 
 * Access Codes).  A server can check this value against CA_WRITE for instance to see if a 
 * write request is being made.  The bits CA_FIRST and CA_LAST are also of some interest to 
 * determine if a data link is coming into or going out of scope.
 *
 * \include eg_EqpFcn.txt
*/

/* EqpRemoveAlarm() */
/**
 * \b Visual \b Basic: 
 *
 *    \b EqpRemoveAlarm( \e devNumber \b As \b Integer, \e almCode \b As \b Long)
 *
 * \brief Removes the specified alarm if present in the local alarm list
 * 
 * A server can remove alarms for a specific device number and alarm code 
 * by using this call. The alarm is then marked immediately as terminated and 
 * is scheduled for removal.
 *
 * \param devNumber is the device number for which the alarm is to be removed
 * \param almCode is the alarm code for which the alarm is to be removed.
 *
 * \note With Visual Basic 3.0 VBX controls there are no methods.  Therefore this
 * functionality is not present in the TINE server-side WINSRV.VBX.  Instead you will have
 * to call this function directly from the tinesv16.dll, passing an extra argument
 * containing the local EqpName.
 *
 * \return none
 *
 * \see EqpClearAlarm()
 */
void CSrvCtrl::EqpRemoveAlarm(short devNumber, long almCode) 
{
  static char localEqpname[17];

  strncpy(localEqpname,(LPCTSTR) m_eqpName,16);
  RemoveAlarm(localEqpname,devNumber,almCode);
}

/* EqpRemoveDeviceAlarm() */
/**
 * \b Visual \b Basic: 
 *
 *    \b EqpRemoveDeviceAlarm( \e devName \b As \b String, \e almCode \b As \b Long)
 *
 * \brief Removes the specified alarm if present in the local alarm list
 * 
 * A server can remove alarms for a specific device name and alarm code 
 * by using this call. The alarm is then marked immediately as terminated and 
 * is scheduled for removal.
 *
 * \param devName is the device name for which associated alarms are to be cleared.
 * \param almCode is the alarm code for which the alarm is to be removed.
 *
 * \note With Visual Basic 3.0 VBX controls there are no methods.  Therefore this
 * functionality is not present in the TINE server-side WINSRV.VBX.  Instead you will have
 * to call this function directly from the tinesv16.dll, passing an extra argument
 * containing the local EqpName.
 *
 * \return none
 *
 * \see EqpClearAlarm()
 *
 * \include eg_EqpClearDeviceAlarm.txt
 */
void CSrvCtrl::EqpRemoveDeviceAlarm(LPCTSTR devName, long almCode) 
{
  static char l_eqpname[17],l_devname[17];

  strncpy(l_eqpname,(LPCTSTR) m_eqpName,16);
  strncpy(l_devname,devName,16);
  RemoveDeviceAlarm(l_eqpname,l_devname,almCode);
}
