/** \file TBufferedSrv.c
 * \brief TINE Buffered Server DLL documentation 
 *
 */
#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "rpcdefs.h"
#include "errors.h"

typedef struct
{
  char name[32];
} NAME32;

typedef struct
{
  char name[16];
} NAME16;

typedef union
{
  void   *vptr;
  char   *cptr;
  UINT8  *bptr;
  UINT16 *sptr;
  UINT32 *lptr;
  float  *fptr;
  double *dptr;
} DUNION;

typedef struct
{
  UINT32 dArrayLength;
  short  dFormat;
  short  reserved;
  char   dTag[16];
  DUNION data;
} DTYPE;

typedef struct PropertyEGU
{
  char units[16];
  double min;
  double max;
  BYTE graph;
  BYTE reserved[3];
} PropertyEGUStruct;

typedef struct PrpRedirBlkStruct
{
  NAME16 prpRedirection[4];  
  struct PrpRedirBlkStruct *next;
} PrpRedirBlk;

typedef struct ExportPropertyListTag
{
  char  prpName[32];       /* property name */
  char  prpAlias[32];      /* property alias name */
  UINT32 prpSizeIn;        /* maximum supported array size */
  UINT32 prpSizeOut;       /* maximum supported array size */
  int   prpId;             /* property id */
  short prpFormatIn;       /* native format */
  short prpFormatOut;      /* native format */
  short prpAccessMode;     /* bit 0: read access, bit 1: write access */
  short prpArrayType;      /* property array type if array */
  short numRows;           /* num rows if array */
  short rowSize;           /* row size if array */
  char  strTagIn[8]; 	     /* structure tag */
  char  strTagOut[8];      /* structure tag */
  int   structSizeIn;      /* input structure size */
  int   structSizeOut;     /* output structure size */
  PrpRedirBlk *prd;        /* re-director if not NULL */
  NAME16 *devNames;        /* device list if not NULL */
  int numDevices;          /* number of devices in device list */
  int hasQueriedForDeviceNames; /* just ask once !*/
  char  prpDescription[64];/* short description of property */
  PropertyEGUStruct egu;   /* EGU information for display */
  PropertyEGUStruct xegu;  /* x-axis EGU information for spectrum display */
  struct ExportPropertyListTag *next;
} ExportPropertyListStruct;

__declspec(dllimport) short PASCAL ScheduleProperty(char *eqm, char *prp);
__declspec(dllimport) double PASCAL makeDataTimeStamp(void);
__declspec(dllimport) void PASCAL setDataTimeStamp(double ts);
__declspec(dllimport) double PASCAL getDataTimeStamp(void);
__declspec(dllimport) void PASCAL GetCallerInfo(char *ClientName,char *ClientAddress);
__declspec(dllimport) short AdjustContract(char *devName, char *devProperty, long dataLengthOut,long dataLengthIn,short eqpidx);
__declspec(dllimport) void GetContract(short *eqno,char *prop,short *size,short index, time_t *time);
__declspec(dllimport) void RemoveEqpfcn(char *name,HWND hVBWnd);
__declspec(dllimport) short GetData(void *data,short index);
__declspec(dllimport) short GetEqpData(VARIANT FAR * data, short index);
__declspec(dllimport) short GetCallerFormat(short index);
__declspec(dllimport) short RegisterEQP(char *exportName,char *eqpName,short number,HWND hVBWnd);
__declspec(dllimport) short RegisterProperty(char *eqm,char *prop,short siz,short fmt,short acc,char *dsc);
__declspec(dllimport) short GetPropertyId(char *eqm,char *prpName);
__declspec(dllimport) short RegisterModuleName(char *eqname,char *devname,short devnr);
__declspec(dllimport) short GetDeviceNumber(char *eqname,char *devname);
__declspec(dllimport) short GetDeviceNumberEx(char *eqname,char *devname,char *prpname);
__declspec(dllimport) void GetDeviceName(char *eqname,int devnr,char *devname);
__declspec(dllimport) short SetAlarm(char *name,short devNr,long almCode,BYTE *almData);
__declspec(dllimport) short SetAlarmEx(char *name,short devNr,long almCode,BYTE *almData,BYTE flags);
__declspec(dllimport) void ClearAlarm(char *name,short devNr);
__declspec(dllimport) void PutStatus(short status);
__declspec(dllimport) short PutData(void *data,time_t eqptime,short eqpidx);
__declspec(dllimport) short PutDataEx(void *data,double dataTimeStamp,short eqpidx);
__declspec(dllimport) void GetContractEx(char *devName, char *devProperty,DTYPE *dout,
                                                    DTYPE *din,short *access,short eqpidx);
__declspec(dllimport) short RegisterPropertyEx(char *eqm,char *prop,long sizIn,short fmtIn,char *tagIn,
                                                          long sizOut,short fmtOut,char *tagOut,short acc,char *dsc);
__declspec(dllimport) short SetDeviceAlarm(char *name,char *devname,long almCode,BYTE *almData,BYTE flags);
__declspec(dllimport) void ClearDeviceAlarm(char *name,char *devname);
__declspec(dllimport) void RemoveDeviceAlarm(char *name,char *devname,long code);
__declspec(dllimport) short SETRPCERROREx(short errcode,char *errstr,short eqpidx);
__declspec(dllimport) short AttachToDatabase(char *expname,char *lclname,HWND hwnd);
__declspec(dllimport) short PASCAL GetRegisteredPropertyList(char *eqm,NAME32 *prpNames,int *nprps);
__declspec(dllimport) short PASCAL GetRegisteredPropertyListStruct(char *eqm,char *prpName,ExportPropertyListStruct *prpLstStruct);
void OnSysCommand(UINT nID, LPARAM lParam);
__declspec(dllimport) short PASCAL GetNumberRegisteredDevices(char *eqm);
__declspec(dllimport) short PASCAL feclog(char *text);
__declspec(dllimport) short PASCAL hasDevicePropertyList(char *eqm);

#define ccerr(i) {cc = i; goto err; }
#define MAX_BUFFERED_PROPERTIES 512
HANDLE hBufSrv;
HANDLE hDLLWnd = NULL;
char gServerWindow[64];
char gServerName[64];
char gEqpName[16];
int gNumDevices = 0;
int gInitialized = 0;

typedef struct
{
  char eqm[8];
  char nam[32];
  DTYPE dout;
  DTYPE din;
  DTYPE blk;
  float max;
  float min;
  double dataTimeStamp;
  int inputChanged;
  int arrayType;
  void (*callback)(void);
  void (*notifier)(int);
  int notifierId;
} BufferedProperty;

BufferedProperty bufferedProperties[MAX_BUFFERED_PROPERTIES];
int nBufferedProperties = 0;
NAME32 registeredProperties[MAX_BUFFERED_PROPERTIES];

__declspec(dllexport) LRESULT CALLBACK BufSrvWndProc(HWND hWnd, UINT iMessage, WPARAM wParam, LPARAM lParam)
{
  switch (iMessage)
  {
    case WM_SYSCOMMAND:
      OnSysCommand(wParam,lParam);
      break;
    default: 
      return DefWindowProc (hWnd, iMessage, wParam, lParam);
  }
  return 0;
}
int CreateDLLWindow(void)
{
  static WNDCLASS wndclass;
  static char lpszWN[32];
  static int done = 0;

  if (done) return 0;
  done = TRUE;
  feclog("Creating buffered server window");
  
  sprintf(gServerWindow,"TBS:%d",GetCurrentProcessId());
  strncpy(lpszWN,gServerWindow,16); lpszWN[16] = 0;
  wndclass.style         = CS_GLOBALCLASS|CS_SAVEBITS;  // essential style!!
  wndclass.lpfnWndProc   = (WNDPROC) BufSrvWndProc;
  wndclass.cbClsExtra    = 0;
  wndclass.cbWndExtra    = 0;
  wndclass.hInstance     = hBufSrv;
  wndclass.hIcon         = LoadIcon (NULL, IDI_APPLICATION);
  wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW);
  wndclass.hbrBackground = GetStockObject (WHITE_BRUSH);
  wndclass.lpszMenuName  = NULL;
  wndclass.lpszClassName = lpszWN;

  if (!RegisterClass(&wndclass)) return -1;

  if (!(hDLLWnd = CreateWindow (
			lpszWN, lpszWN,               // lpClassName, lpWindowName
			WS_BORDER|WS_CAPTION,         // dwStyle
			CW_USEDEFAULT, CW_USEDEFAULT, // X, Y
			CW_USEDEFAULT, CW_USEDEFAULT, // nWidth, nHeight
			NULL, NULL,                   // hWndParent, hMenu
			hBufSrv,                      // hInstance
			(LPSTR)""                     // createstruct
      ) ) ) return -1;

  return 0;
}
int DestroyDLLWindow(void)
{
  static char lpszWN[32];
  if (hDLLWnd != NULL)
  {
    feclog("destroy buffered server window");
    strncpy(lpszWN,gServerWindow,16); lpszWN[16] = 0;
    DestroyWindow(hDLLWnd);
    UnregisterClass(lpszWN,hBufSrv);
  }
  return 0;
}
BOOL WINAPI DllMain(HINSTANCE hInstDll,DWORD fdwReason,LPVOID lpvReserved)
{
	switch (fdwReason)
	{
	 case DLL_PROCESS_ATTACH:  // DLL being loaded
			break;
	 case DLL_THREAD_ATTACH:
			break;
	 case DLL_THREAD_DETACH:
			break;
	 case DLL_PROCESS_DETACH:  // DLL being unloaded
      DestroyDLLWindow();
      feclog("unloading Buffered Server DLL");
			break;
  }
  return (TRUE);
}
int areValuesOutOfBounds(int id)
{
  int i;

  if (bufferedProperties[id].min == 0 && bufferedProperties[id].max == 0) return 0;
  switch (bufferedProperties[id].din.dFormat)
  {
    case CF_BYTE: 
      for (i=0; i<(int)bufferedProperties[id].din.dArrayLength; i++)
      {        
        if ((bufferedProperties[id].blk.data.bptr[i] < bufferedProperties[id].min) ||
            (bufferedProperties[id].blk.data.bptr[i] > bufferedProperties[id].max))
          return out_of_range;
      }
      break;
    case CF_SHORT: 
      for (i=0; i<(int)bufferedProperties[id].din.dArrayLength; i++)
      {        
        if ((bufferedProperties[id].blk.data.sptr[i] < bufferedProperties[id].min) ||
            (bufferedProperties[id].blk.data.sptr[i] > bufferedProperties[id].max))
          return out_of_range;
      }
      break;
    case CF_LONG: 
      for (i=0; i<(int)bufferedProperties[id].din.dArrayLength; i++)
      {        
        if ((bufferedProperties[id].blk.data.lptr[i] < bufferedProperties[id].min) ||
            (bufferedProperties[id].blk.data.lptr[i] > bufferedProperties[id].max))
          return out_of_range;
      }
      break;
    case CF_FLOAT: 
      for (i=0; i<(int)bufferedProperties[id].din.dArrayLength; i++)
      {        
        if ((bufferedProperties[id].blk.data.fptr[i] < bufferedProperties[id].min) ||
            (bufferedProperties[id].blk.data.fptr[i] > bufferedProperties[id].max))
          return out_of_range;
      }
      break;
    case CF_DOUBLE: 
      for (i=0; i<(int)bufferedProperties[id].din.dArrayLength; i++)
      {        
        if ((bufferedProperties[id].blk.data.dptr[i] < bufferedProperties[id].min) ||
            (bufferedProperties[id].blk.data.dptr[i] > bufferedProperties[id].max))
          return out_of_range;
      }
      break;
  }
  return 0;
}
short currentContractIndex = -1;
int currentPropertyId = 0;
int currentDeviceNumber = 0;
void OnSysCommand(UINT nID, LPARAM lParam) 
{
	int i,p,id,cc,dsiz;
  short access, index = (short)nID;
	DTYPE dout,din;
  UINT32 doutlen;
  static char lcl_devname[32], lcl_devprop[64];

  if (abs(index) <= 0x8fff)
	{
    // get the contract
    if (currentContractIndex == index) 
    {  // still waiting ...
       SETRPCERROREx((short)not_ready,NULL,index);
       return;
    }
    if (currentContractIndex >= 0) return; // refuse the request
    currentContractIndex = index;          // -1 is okay !
    GetContractEx(lcl_devname,lcl_devprop,&dout,&din,&access,index);
    if ((id=GetPropertyId(gEqpName,lcl_devprop)) < 0) ccerr(illegal_property);
    if ((i=GetDeviceNumberEx(gEqpName,lcl_devname,lcl_devprop)) < 0) ccerr(illegal_equipment_number);
    if (din.dArrayLength > 0)
    {
      if (bufferedProperties[id].inputChanged) ccerr(operation_busy);
      if (!(access & CA_WRITE)) ccerr(illegal_read_write);
      dsiz = bufferedProperties[id].din.dArrayLength * fmtsizeof(bufferedProperties[id].din.dFormat);
      if (bufferedProperties[id].din.dFormat == CF_TEXT) memset(bufferedProperties[id].blk.data.cptr,0,dsiz);
      p = fmtsizeof(bufferedProperties[id].din.dFormat) * i;
      if ((cc=GetData(&bufferedProperties[id].blk.data.bptr[p],index)) != 0) ccerr(cc);
      // do range checking
      if (areValuesOutOfBounds(id)) ccerr(out_of_range);
      memcpy(bufferedProperties[id].din.data.bptr,bufferedProperties[id].blk.data.bptr,dsiz);
      bufferedProperties[id].inputChanged = -1;
      currentPropertyId = id;
      currentDeviceNumber = i;
      if (bufferedProperties[id].notifier != NULL)
      {
        bufferedProperties[id].notifier(bufferedProperties[id].notifierId);
        SETRPCERROREx((short)not_ready,NULL,index);
        return;  // let the caller set the completion ...
      }
      if (bufferedProperties[id].callback != NULL) bufferedProperties[id].callback();
    }
    if (dout.dArrayLength > 0)
    {
      p = fmtsizeof(bufferedProperties[id].dout.dFormat) * i;
      doutlen = bufferedProperties[id].dout.dArrayLength;
      if (bufferedProperties[id].arrayType & AT_SPECTRUM) 
      {
        p *= bufferedProperties[id].dout.dArrayLength;
      }
      else
      {
        if ((doutlen -= i) < 1) ccerr(invalid_index);
      }
      if (doutlen < dout.dArrayLength) AdjustContract(NULL,NULL, doutlen,din.dArrayLength,index);
      if ((cc=PutDataEx(&bufferedProperties[id].dout.data.bptr[p],bufferedProperties[id].dataTimeStamp,index)) != 0) ccerr(cc);
    }
err:
    SETRPCERROREx((short)cc,NULL,index);
    currentContractIndex = -1;
	}
}
__declspec(dllexport) short PASCAL SetBufferedCompletion(int compCode)
{
  int cc=0,p,id = currentPropertyId,i = currentDeviceNumber;
  if (currentContractIndex < 0) ccerr(non_existent_client);
  p = fmtsizeof(bufferedProperties[id].dout.dFormat) * i;
  if ((cc=PutDataEx(&bufferedProperties[id].dout.data.bptr[p],bufferedProperties[id].dataTimeStamp,currentContractIndex)) != 0) ccerr(cc);
  SETRPCERROREx((short)compCode,NULL,currentContractIndex);
  currentContractIndex = -1;
err:
  return cc;
}
__declspec(dllexport) short PASCAL AttachDatabase()
{
  int cc;
	if (!gInitialized)
	{
    if ((cc=CreateDLLWindow()) != 0) return code_failure;
    if ((cc=AttachToDatabase(NULL,NULL,hDLLWnd)) != 0) return cc;
    gInitialized = -1;
	}
  return 0;
}
/** 
 * \brief Attaches the TINE server DLL according to the input given
 *
 * \b Visual \b Basic :
 * \n Declare Function \b AttachServer Lib "TBufferedSrv.dll" (ByVal srvExportName As String,ByVal srvEQPName As String,ByVal ndevices As Long) As Integer
 *
 * \param srvExportName is the desired export name.  If this is omitted (i.e. a zero-length 
 *        string is passed, then the export name must be obtained from an associated 'exports.csv'
 *        in the local FEC database repository.
 * \param srvEQPName is the desired local equipment module name.  If this is omitted (i.e. a zero-length
 *        string is passed, then the local equipment module name is extracted from an
 *        associated 'exports.csv' in the local FEC database repository.
 * \param ndevices is the number of devices (instances) associated with this device server.
 *        If this is omitted (i.e. 0 is passed), then the number of devices is extracted from an
 *        associated 'exports.csv' in the local FEC database repository.
 *
 * A call to AttachServer serves to initialize the device server.  It is 
 * analogous to setting the server ActiveX control's enabled property to 'true', where the
 * ActiveX interface is concerned.  Theoretically, you can call AttachServer() with all 'NULL'
 * parameters, in which case all registration information is obtained from the 'exports.csv' 
 * file (see <a href="csvFiles.html"><b><i>Server Startup csv-Files</i></b></a>.
 * Note however, that if you take this strategy, you should take care to avoid a lock-out
 * situation regarding the name resolution.  Namely, if all initialization names are to reside
 * in 'fecid.csv' (FEC name) and 'exports.csv' (device server name, local equipment module name,
 * and properties) then the fecid.csv should not make use of an 'Export_Name' column or a 
 * 'Local_Name' column.  That is, the fecid.csv file should only contain one entry, referring
 * to the server process in question.  In cases where more than one FEC process are running
 * on the same computer, each process should have its own fecid.csv in its working directory
 * when the export.csv file cannot make use of an 'Export_Name' column in order to latch
 * the appropriate device server.  Of course, if AttachServer() provides an 'Export Name' in
 * the first parameter, then there is no ambiguity.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \b Example:  Attach to server and set the server's export name (requires exports.csv file to
 *              set the local equipment module name and number of modules):
 *
 * \include eg_AttachServer0.txt
 *
 * \b Example:  Attach to server and set the server's export name, local equipment module name and
 *              number of modules (does not read exports.csv file!):
 *
 * \include eg_AttachServer1.txt
 */
__declspec(dllexport) short PASCAL AttachServer(char *srvExportName,char *srvEQPName,int ndevices) 
{
  int i, id, cc, dsiz, n, nelems;
  int nprps = MAX_BUFFERED_PROPERTIES, hasprplst = FALSE;
  ExportPropertyListStruct pls;

  currentContractIndex = -1;
	if (!gInitialized)
	{
    if ((cc=CreateDLLWindow()) != 0) return code_failure;
    nBufferedProperties = 0;
    gInitialized = -1;
  }

  memset(gEqpName,0,16);  memset(gServerName,0,32);
  if (srvExportName != NULL && strlen(srvExportName) > 0) strncpy(gServerName,srvExportName,16);
  if (srvEQPName != NULL && strlen(srvEQPName) != 0) strncpy(gEqpName,srvEQPName,6);

  if (strlen(gServerName) > 0 && strlen(gEqpName) > 0)
  { // handle all registration within the buffered server
    gNumDevices = ndevices;
    if ((cc=RegisterEQP(gServerName,gEqpName,(short)gNumDevices,hDLLWnd)) != 0) return cc;
  }
  else
  {
    if ((cc=AttachToDatabase(gServerName,gEqpName,hDLLWnd)) != 0) return cc;
    if ((cc=GetRegisteredPropertyList(gEqpName,registeredProperties,&nprps)) != 0) return cc;
    if ((n=GetNumberRegisteredDevices(gEqpName)) < 0) n = ndevices > 0 ? ndevices : 1;
    gNumDevices = n;
    hasprplst = hasDevicePropertyList(gEqpName);
    for (i=0; i<nprps; i++)
    {
      if ((id=GetPropertyId(gEqpName,registeredProperties[i].name)) < 0) return -id;
      if ((cc=GetRegisteredPropertyListStruct(gEqpName,registeredProperties[i].name,&pls)) != 0) return cc;
      if (hasprplst) pls.prpArrayType |= AT_SPECTRUM; // force separate storage for each device (can't be a channel array)
      strncpy(bufferedProperties[id].nam,pls.prpName,32);
      bufferedProperties[id].max = (float)pls.egu.max;
      bufferedProperties[id].min = (float)pls.egu.min;
      bufferedProperties[id].dout.dFormat = pls.prpFormatOut;
      bufferedProperties[id].dout.dArrayLength = pls.prpSizeOut;
      bufferedProperties[id].arrayType = pls.prpArrayType;
      nelems = pls.prpArrayType & AT_SPECTRUM ? n : 1;
      if ((dsiz=pls.prpSizeOut * fmtsizeof(pls.prpFormatOut)) > 0)
      {
        if (pls.prpFormatOut == CF_SPECTRUM) dsiz += 96;
        if ((bufferedProperties[id].dout.data.bptr=calloc(nelems,dsiz)) == NULL)
          return -out_of_server_memory;
      }
      bufferedProperties[id].din.dFormat = pls.prpFormatIn;
      bufferedProperties[id].din.dArrayLength = pls.prpSizeIn;
      if ((dsiz=pls.prpSizeIn * fmtsizeof(pls.prpFormatIn)) > 0)
      {
        if (pls.prpFormatIn == CF_SPECTRUM) dsiz += 96;
        if (((bufferedProperties[id].din.data.bptr=calloc(1,dsiz)) == NULL ||
            (bufferedProperties[id].blk.data.bptr=calloc(1,dsiz)) == NULL))
          return -out_of_server_memory;
      }
      bufferedProperties[id].dataTimeStamp = 0;
      nBufferedProperties++;
    }
  }

  return 0;
}

/** 
 * \brief Device name registration call
 * 
 * \b Visual \b Basic: 
 * \n Declare Function \b RegisterDeviceName Lib "TBufferedSrv.dll" (ByVal devName As String,ByVal devNumber As Integer) As Integer
 *
 * Servers can also register their device names through a startup file. If used, the startup 
 * file is specified by the 6-char local (internal) equipment module name concatenated with 
 * "-device.csv" (e.g. "eqm1-devices.csv"). In cases where this is not practical, a call to 
 * RegisterDeviceName() can be made. RegisterDeviceName() can be called at any time after 
 * the AttachServer() has been called. 
 *
 * \param devName is the device name to be assigned to the device number passed
 * \param devNumber is the device number whose name is being assigned.
 *
 * \note the devNumber passed in this call must lie between (or equal to) 0 and the number
 * of devices set via the EqpNumberModules ActiveX control property.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \b Example:
 *
 * \include eg_RegisterDeviceName.txt
 */
__declspec(dllexport) short PASCAL RegisterDeviceName(char *devName, short devNumber) 
{
  static char localEqpname[17],localDevname[17];

  strncpy(localEqpname,gEqpName,6);
  strncpy(localDevname,devName,16);
  return RegisterModuleName(localEqpname,localDevname,devNumber);
}
__declspec(dllexport) short PASCAL GetRegisteredProperties(NAME32 *prps, short *nprps)
{
  int i, n;
  if (prps == NULL || nprps == NULL) return -1;
  for (i=0, n=0; i<nBufferedProperties && n<*nprps; i++)
  {
    if (strlen(bufferedProperties[i].nam) == 0) continue;
    strncpy(prps[n++].name,bufferedProperties[i].nam,32);
  }
  *nprps = n;
  return nBufferedProperties;
}

/** 
 * \brief Buffered Property registration call
 * 
 * \b Visual \b Basic: 
 * \n Declare Function \b RegisterBufferedProperty Lib "TBufferedSrv.dll" _
 * \n (ByVal prpName As String, ByVal prpInSiz As Long, ByVal prpInFmt As Integer _
 * \n  ByVal prpOutSiz As Long, ByVal prpOutFmt As Integer, ByVal prpMax As Single, _
 * \n  ByVal prpMin As Single, ByVal prpEgu As String, ByVal access As Integer, _
 * \n  ByVal prpDesc As String) As Integer
 *
 * Servers can also register their 'buffered' properties through the 'exports.csv' file. 
 * If the FEC Name can be unambiguously determined at startup time (e.g. there's only one
 * entry in the fecid.csv, or the RegisterFecName() API call is used) then a startup 
 * file with the name '\<FECNAME\>-exports.csv' is first accessed for registration information,
 * where '\<FECNAME\>' refers the the registered FEC name. 
 * If this file does not exist then 'exports.csv' is accessed.
 * In cases where it is not practical to register property names in this manner, a call to 
 * RegisterBufferedProperty() can be made. RegisterBufferedProperty() can be called at any time after 
 * the AttachServer() has been called. 
 *
 * \param prpName is the property name to be exported and buffered
 * \param prpInSiz is the number of elements to be associated with the property when
 *        input data is sent to the server.
 * \param prpInFmt is the TINE data type associated with input data for this property
 * \param prpOutSiz is the number of elements to be associated with the property when
 *        output data is requested from the server.
 * \param prpOutFmt is the TINE data type associated with output data for this property
 * \param prpMax is the maximum valid value the property is allowed to contain.
 * \param prpMin is the minimum valid value the property is allowed to contain.
 * \param prpEgu is the engineering units associated with the propery
 * \param access is the bit-wise combination of TINE access codes (i.e. CA_READ and
 *        CA_WRITE) which determine the read/write access characteristics of the property.
 * \param prpDesc is a description of the property (up to 64 characters).
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \b Example:
 *
 * \include eg_RegisterBufferedProperty.txt
 *
 */
__declspec(dllexport) short PASCAL RegisterBufferedProperty
(
  char *prpName, long prpInSiz, short prpInFmt, long prpOutSiz, short prpOutFmt,
  float prpMax, float prpMin, char *prpEgu,	short access, char *prpDsc
) 
{
  int nelems=1,id,dsiz=0;
  static char localEqpname[17],localProperty[33],localDescription[96];
  static char nulltag[9];

  strncpy(localEqpname,gEqpName,6);
  strncpy(localProperty,prpName,32);
  sprintf(localDescription,"[%g:%g %s]%s",prpMin,prpMax,prpEgu,prpDsc);
  nulltag[0] = 0;
  id = RegisterPropertyEx(localEqpname,localProperty,prpInSiz,
	            prpInFmt,nulltag,prpOutSiz,prpOutFmt,
							nulltag,access,localDescription);
  if (id < 0 || id >= MAX_BUFFERED_PROPERTIES) return -illegal_property;
  strncpy(bufferedProperties[id].nam,prpName,32);
  bufferedProperties[id].max = prpMax;
  bufferedProperties[id].min = prpMin;
  bufferedProperties[id].dout.dFormat = prpOutFmt;
  bufferedProperties[id].dout.dArrayLength = prpOutSiz;
  if (prpOutFmt == CF_SPECTRUM)
  {
    bufferedProperties[id].arrayType = AT_SINGLE|AT_SPECTRUM;
    nelems = gNumDevices;
  }
  if ((dsiz=prpOutSiz * fmtsizeof(prpOutFmt)) > 0)
  {
    if (prpOutFmt == CF_SPECTRUM) dsiz += 96;
    if ((bufferedProperties[id].dout.data.bptr=calloc(nelems,dsiz)) == NULL)
      return -out_of_server_memory;
  }
  bufferedProperties[id].din.dFormat= prpInFmt;
  bufferedProperties[id].din.dArrayLength = prpInSiz;
  if ((dsiz=prpInSiz * fmtsizeof(prpInFmt)) > 0)
  {
    if (prpInFmt == CF_SPECTRUM) dsiz += 96;
    if (((bufferedProperties[id].din.data.bptr=calloc(1,dsiz)) == NULL ||
        (bufferedProperties[id].blk.data.bptr=calloc(1,dsiz)) == NULL))
      return -out_of_server_memory;
  }
  nBufferedProperties++;
  return id;
}
/** 
 * \brief Refreshes the contents of the data buffer associated with the given property
 * 
 * \b Visual \b Basic: 
 * \n Declare Function \b pushBufferedData Lib "TBufferedSrv.dll" _
 * \n (ByVal prpName As String, ByVal devName As String, _
 * \n prpData As Any, ByVal prpSiz As Long, ByVal prpSchedule As Integer) As Integer
 *
 *
 *
 * \param prpName is the property name into whose buffer the associated data are to be written
 * \param devName is the device name giving the starting address i.e. entry point into the
 *        property's buffer into which the associated data are to be written
 * \param prpData is a reference to the data to be written (pushed) into the property's data
 *        buffer.
 * \param prpSiz is the number of elements to be written into the property's data buffer.
 * \param prpSchedule is the schedule 'flag' determining whether the property is to be
 *        scheduled immediately ('true') or not.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \b Example:
 *
 * \include eg_pushBufferedData.txt
 */
__declspec(dllexport) short PASCAL pushBufferedData(char *prpName,char *devName,BYTE *prpData,long prpSiz,int prpSchedule)
{
  int id, i, p, fsz, dsz;

  if (prpData == NULL) return buffer_too_small;
  if ((id=GetPropertyId(gEqpName,prpName)) < 0) return illegal_property;
  if ((i=GetDeviceNumberEx(gEqpName,devName,prpName)) < 0) return illegal_equipment_number;
  if (prpSiz > i + (int)bufferedProperties[id].dout.dArrayLength) 
  {
    prpSiz = (int)bufferedProperties[id].dout.dArrayLength;
    if (!(bufferedProperties[id].arrayType & AT_SPECTRUM)) prpSiz -= i;
  }

  fsz = fmtsizeof(bufferedProperties[id].dout.dFormat);
  p = fsz * i;
  if (bufferedProperties[id].arrayType & AT_SPECTRUM) p *= bufferedProperties[id].dout.dArrayLength;
  dsz = prpSiz*fsz;
  if (bufferedProperties[id].dout.dFormat == CF_SPECTRUM) dsz += 96; // room for SPECTRUM header
  memcpy(&bufferedProperties[id].dout.data.bptr[p],prpData,dsz);
  bufferedProperties[id].dataTimeStamp = makeDataTimeStamp();
  if (prpSchedule) ScheduleProperty(gEqpName,prpName);
  return 0;
}
/** 
 * \brief Retrieves the contents of the input data buffer associated with the given property
 * 
 * \b Visual \b Basic: 
 * \n Declare Function \b pushBufferedData Lib "TBufferedSrv.dll" _
 * \n (ByVal prpName As String, ByVal devName As String, prpData As Any, ByVal prpSiz As Long) As Integer
 *
 *
 *
 * \param prpName is the property name from whose input buffer the associated data are to be read.
 * \param devName is the device name giving the starting address i.e. entry point into the
 *        property's buffer from which the associated data are to be read.
 * \param prpData is a reference to the data buffer into which the incoming data should be read 
 *        (pulled).
 * \param prpSiz is the number of elements to be read into the property's data buffer.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \b Example:
 *
 * \include eg_pullBufferedData.txt
 */
__declspec(dllexport) short PASCAL pullBufferedData(char *prpName,char *devName,BYTE *prpData,long prpSiz)
{
  int id, i, p, fsz, dsz;

  if (prpData == NULL) return buffer_too_small;
  if ((id=GetPropertyId(gEqpName,prpName)) < 0) return illegal_property;
  if ((i=GetDeviceNumberEx(gEqpName,devName,prpName)) < 0) return illegal_equipment_number;
  dsz = (int)bufferedProperties[id].din.dArrayLength;
  if (bufferedProperties[id].din.dFormat == CF_TEXT) i = 0;
  if (prpSiz > i + dsz) return dimension_error;

  fsz = fmtsizeof(bufferedProperties[id].din.dFormat);
  p = fsz * i;
  dsz = prpSiz*fsz;
  if (bufferedProperties[id].din.dFormat == CF_SPECTRUM) 
  {
    memcpy(prpData,bufferedProperties[id].din.data.bptr,96); // the SPECTRUM header
    p += 96; // skip over SPECTRUM header
  }
  memcpy(prpData,&bufferedProperties[id].din.data.bptr[p],dsz);
  bufferedProperties[id].inputChanged = 0;
  return 0;
}
/** 
 * \brief Checks whether there are new input data for the given property
 * 
 * \b Visual \b Basic: 
 * \n Declare Function \b hasInputChanged Lib "TBufferedSrv.dll" (ByVal prpName As String) As Integer
 *
 *
 *
 * \param prpName is the property name whose input data buffer is to be checked for changes.
 *
 * \return 0 (False) if the input data have not changed or -1 (True) if there are new input
 * data to be acquired by calling PullBufferedData().
 *
 * \b Example:
 *
 * \include eg_hasInputChanged.txt
 */
__declspec(dllexport) short PASCAL hasInputChanged(char *prpName)
{
  int id;

  if ((id=GetPropertyId(gEqpName,prpName)) < 0) return 0;
  return bufferedProperties[id].inputChanged;
}
/** 
 * \brief Returns the device number associated with the WRITE call.
 * 
 * \b Visual \b Basic: 
 * \n Declare Function \b GetInputDeviceNumber Lib "TBufferedSrv.dll" () As Integer
 *
 * This call can be made inside the server's callback routine along with
 * 'hasInputChanged' in order to ascertain which device was 'set'. 
 *
 * \return The device number (obtained from the registered device name) used
 * in the call.
 *
 * \b Example:
 *
 * \include eg_hasInputChanged.txt
 */
__declspec(dllexport) short PASCAL GetInputDeviceNumber(void)
{
  return currentDeviceNumber;
}
/** 
 * \brief Registers a callback routine to be called when a WRITE access property is
 * called.
 * 
 * \b Visual \b Basic: 
 * \n Declare Function \b RegisterServerCallback Lib "TBufferedSrv.dll" (ByVal prpName As String, cb As Any) As Integer
 *
 *
 *
 * \param prpName
 * \param cb
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \b Example:
 *
 * \include eg_RegisterServerCallback.txt
 */
__declspec(dllexport) short PASCAL RegisterServerCallback(char *prpName, void (*cb)(void))
{
  int id;

  if ((id=GetPropertyId(gEqpName,prpName)) < 0) return illegal_property;
  bufferedProperties[id].callback = cb;
  return 0;
}
/** 
 * \brief Registers a Notifier routine to be called when a WRITE access property is
 * called.
 * 
 * \b Visual \b Basic: 
 * \n Declare Function \b RegisterServerNotifier Lib "TBufferedSrv.dll" (ByVal prpName As String, cb As Any) As Integer
 *
 * The difference between a callback and a notifier is that it is assumed that the callback 
 * completes all activity associated with the transaction while inside the callback routine.
 * A notififier will call the notifier routine and expect the caller to call the 
 * SetBufferedCompletion() routine when the transaction is complete.  This scenario is useful 
 * for development platforms such as LabView.
 *
 * \param prpName the Property name for which the notifier is in effect
 * \param nf the notifier routine
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \b Example:
 *
 * \include eg_RegisterServerNotifier.txt
 */
__declspec(dllexport) short PASCAL RegisterServerNotifier(char *prpName, void (*nf)(int))
{
  int id;

  if ((id=GetPropertyId(gEqpName,prpName)) < 0) return illegal_property;
  bufferedProperties[id].notifier = nf;
  bufferedProperties[id].notifierId = 0;
  return 0;
}
__declspec(dllexport) short PASCAL RegisterServerNotifierEx(char *prpName, void (*nf)(int), int nid)
{
  int id;

  if ((id=GetPropertyId(gEqpName,prpName)) < 0) return illegal_property;
  bufferedProperties[id].notifier = nf;
  bufferedProperties[id].notifierId = nid;
  return 0;
}
