/** \file historyUtils.c
 * \brief TINE history utilities documentation 
 *
 */
#include "project.def"
#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <malloc.h>
#include <sys/timeb.h>
#include <wsipx.h>
#include <wsnwlink.h>
#include "errors.h"
#include "database.h"
#include "rpcsrv.h"

// parse these like /<context>/<server>/<device>[<property>]
__declspec(dllexport) short PASCAL GetArchivedDataAsFloat(char *devsrv,time_t start,time_t stop,FLTINT *fiDataArray,int *num)
{
  DBLK din,dout;
  UINT32 startstopArray[3];
  char *c,tmp[96],aServer[64],kwd[32],dev[32];
  int cc;

  if (num == NULL || *num <= 0) return buffer_too_small;

  strncpy(tmp,devsrv,96);
  if ((c=strchr(tmp,'[')) == 0) return illegal_delimiter;
  *c++ = 0;
  strncpy(kwd,c,32);
  if ((c=strchr(kwd,']')) != 0) *c = 0;
  strrev(tmp);
  if ((c=strchr(tmp,'/')) == 0) return illegal_delimiter;
  *c++ = 0;
  strncpy(dev,tmp,16); strrev(dev);
  strncpy(aServer,c,64); strrev(aServer);
  if (strstr(aServer,"HISTORY") == NULL && // not a central archiver
      strstr(kwd,".HIST") == NULL && strstr(kwd,".HST") == NULL) // meta property not yet appended
    strcat(kwd,".HST");
  strcat(aServer,"/"); strcat(aServer,dev);

  startstopArray[0] = start;
  startstopArray[1] = stop;
  dout.dArrayLength = *num;
  dout.dFormat = CF_FLTINT;
  din.dArrayLength = 2;
  din.dFormat = CF_LONG;

  if ((cc=dllExecLinkEx(aServer,kwd,&dout,fiDataArray,&din,startstopArray,CA_READ,200)) != 0) *num = 0;
  else *num = dout.dArrayLength;

  return cc;
}
__declspec(dllexport) short PASCAL GetArchivedDataAsText(char *devsrv,time_t start,time_t stop,NAME32I *niDataArray,int *num)
{
  DBLK din,dout;
  UINT32 startstopArray[3];
  char *c,tmp[96],aServer[64],kwd[32],dev[32];
  int cc;

  if (num == NULL || *num <= 0) return buffer_too_small;

  strncpy(tmp,devsrv,96);
  if ((c=strchr(tmp,'[')) == 0) return illegal_delimiter;
  *c++ = 0;
  strncpy(kwd,c,32);
  if ((c=strchr(kwd,']')) != 0) *c = 0;
  strrev(tmp);
  if ((c=strchr(tmp,'/')) == 0) return illegal_delimiter;
  *c++ = 0;
  strncpy(dev,tmp,16); strrev(dev);
  strncpy(aServer,c,64); strrev(aServer);
  if (strstr(aServer,"HISTORY") == NULL && // not a central archiver
      strstr(kwd,".HIST") == NULL && strstr(kwd,".HST") == NULL) // meta property not yet appended
    strcat(kwd,".HST");
  strcat(aServer,"/"); strcat(aServer,dev);

  startstopArray[0] = start;
  startstopArray[1] = stop;
  dout.dArrayLength = *num;
  dout.dFormat = CF_NAME32I;
  din.dArrayLength = 2;
  din.dFormat = CF_LONG;

  if ((cc=dllExecLinkEx(aServer,kwd,&dout,niDataArray,&din,startstopArray,CA_READ,200)) != 0) *num = 0;
  else *num = dout.dArrayLength;

  return cc;
}
__declspec(dllexport) short PASCAL GetArchivedDataAsSnapshot(char *devsrv,time_t *target,float *fDataArray,int size)
{
  DBLK din,dout;
  UINT32 startstopArray[3];
  char *c,tmp[96],aServer[64],kwd[32],dev[32];
  int cc;

  if (size <= 0) return buffer_too_small;
  if (target == NULL) return argument_list_error;
  strncpy(tmp,devsrv,96);
  if ((c=strchr(tmp,'[')) == 0) return illegal_delimiter;
  *c++ = 0;
  strncpy(kwd,c,32);
  if ((c=strchr(kwd,']')) != 0) *c = 0;
  strrev(tmp);
  if ((c=strchr(tmp,'/')) == 0) return illegal_delimiter;
  *c++ = 0;
  strncpy(dev,tmp,16); strrev(dev);
  strncpy(aServer,c,64); strrev(aServer);
  if (strstr(aServer,"HISTORY") == NULL && // not a central archiver
      strstr(kwd,".HIST") == NULL && strstr(kwd,".HST") == NULL) // meta property not yet appended
    strcat(kwd,".HST");
  strcat(aServer,"/"); strcat(aServer,dev);
  startstopArray[0] = *target;
  startstopArray[1] = *target;
  dout.dArrayLength = size;
  dout.dFormat = CF_FLOAT;
  din.dArrayLength = 2;
  din.dFormat = CF_LONG;

  if ((cc=dllExecLinkEx(aServer,kwd,&dout,fDataArray,&din,startstopArray,CA_READ,200)) != 0) *target = 0;
  else *target = (time_t)gDataTimeStamp;

  return cc;
}

double blob[8000];
time_t parseTimeString(char *ts)
{
  time_t t=time(NULL);
  int i, len;
  char *c,day[16],mon[64],year[64];
  char hr[64],mn[32],sec[16];
  struct tm tmts, *tnow;

  hr[0] = 0;
  tnow = localtime(&t);
  len = strlen(ts);
  strupr(ts);
  for (i=0; i<len && isdigit(ts[i]); i++);  
  if (i == len) t = atoi(ts);
  else if (!strcmp(ts,"NOW")) t = time(NULL);
  else if ((c=strstr(ts,".")) != NULL) /* day.month.year format */
  {
    *c++ = 0;
    strncpy(day,ts,16);
    strncpy(mon,c,64);
    if ((c=strstr(mon,".")) != NULL)
    { 
      *c++ = 0;
      strncpy(year,c,64);
      if ((c=strstr(year," ")) != NULL)
      {
        *c++ = 0;
        strncpy(hr,c,64);
      }
    }
    else
    { // no year given, use this year
      sprintf(year,"%d",1900+tnow->tm_year);
      if ((c=strstr(mon," ")) != NULL)
      {
        *c++ = 0;
        strncpy(hr,c,64);
      }
    }
    if (strlen(hr) > 0)
    {
      mn[0] = 0; sec[0] = 0;
      if ((c=strstr(hr,":")) != NULL)
      {     
        *c++ = 0;
        strncpy(mn,c,32);
        if ((c=strstr(mn,":")) != NULL)
        {
          *c++ = 0;
          strncpy(sec,c,16);
        }
      }
      tmts.tm_hour = atoi(hr);
      tmts.tm_min =  atoi(mn);
      tmts.tm_sec =  atoi(sec);
    }
    else
    {
      memset(&tmts,0,sizeof(struct tm));
    }
    tmts.tm_mon = (atoi(mon)+11)%12;
    tmts.tm_mday = atoi(day);
    tmts.tm_year = atoi(year);
    if (tmts.tm_year > 1900) tmts.tm_year -= 1900;
    tmts.tm_isdst = -1;
    t = mktime(&tmts);
  }
  return t;
}
__declspec(dllexport) int thistory(char *server,char *property,char *devname,char *stoptime,char *depth,float *values,time_t *times,int *bufsiz)
{
  char devsrv[256];
  long stop = time(NULL), start, mf = 1, dp = 0;
  int i, num = 5000, cc, offset = 0;
  FLTINT *fiDataArray = (FLTINT *)blob;
  NAME32I *niDataArray = (NAME32I *)blob;

  if (values == NULL || bufsiz == NULL) argument_list_error;
  if (property == NULL) return argument_list_error;
  if (devname == NULL) devname = "#0";

  strncpy(devsrv,server,36); // use it as it comes in !
  strcat(devsrv,"/");
  strcat(devsrv,devname);
  strcat(devsrv,"[");
  strcat(devsrv,property);
  if (strstr(server,"HISTORY") == NULL)  // not talking to a central archive server => local history
  { 
    if (strstr(property,".HIST") == NULL && strstr(property,".HST") == NULL)
    { // append it
      strcat(devsrv,".HIST");
    }
  }
  strcat(devsrv,"]");
 
  stop = parseTimeString(stoptime);

  if (!stricmp(depth,"SNAPSHOT"))
  { // snapshot request ?
    if (times == NULL) return argument_list_error;
    *times = stop;
    num = *bufsiz;
    return GetArchivedDataAsSnapshot(devsrv,times,values,num);
  }

  strupr(depth);
  if (strstr(depth,"DAY")) mf = 24 * 3600;
  else if (strstr(depth,"HOUR")) mf = 3600;
  else if (strstr(depth,"MINUTE")) mf = 60;
  else if (strstr(depth,"WEEK")) mf = 24 * 3600 * 7;
  dp = mf * atoi(depth);
  start = stop - dp;
  printf("%s/%s, Timestamp\n",property,devname);
  if ((cc=GetArchivedDataAsFloat(devsrv,start,stop,fiDataArray,&num)) == 0) 
  {
    if ((int)fiDataArray[0].ival <= num) offset = 1;
    for (i=offset; i<num && i<*bufsiz; i++)
    {
      values[i] = fiDataArray[i].fval;
      if (times != NULL) times[i] = fiDataArray[i].ival;
    }
    *bufsiz = num;
    return 0;
  }

  return cc;
}
