/** \file listener.c
 * \brief TINE static listener documentation 
 *
 */
#include "project.def"
#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <malloc.h>
#include <sys/timeb.h>
#include <wsipx.h>
#include <wsnwlink.h>
#include "errors.h"
#include "database.h"
#include "rpcsrv.h"

int messageLog(char *txt,int status);
#define ALLOWED_IDLE_TIME 300
int stopListener(int lid);
typedef struct mexListenerStruct
{
  char devName[64];
  char devProperty[32];
  DTYPE dout;
  DTYPE din;
  UINT16 prpArrayType;
  short prpFormat;
  UINT32 prpSize;
  BYTE *outbuf;
  BYTE *inbuf;
  int lid;         // link id
  int nid;         // notifier id
  void (*nf)(int); // notifier function
  int status;
  int idleTimeInSeconds;
  int pollingRate;
  double timestamp;
  struct mexListenerStruct *nxt;
} mexListener;

mexListener *mexListeners = NULL;
int nListeners = 0;

void listenerCb(int id,int cc)
{
  mexListener *ml;
  for (ml=mexListeners; ml != NULL; ml = ml->nxt)
  {
    if (ml->lid == id)
    {
      if (ml->nf != NULL) ml->nf(ml->nid);
      if (ml->idleTimeInSeconds++ > ALLOWED_IDLE_TIME) stopListener(id);
      return;
    }
  }
  return;
}
mexListener *startListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,int pollrate,int flags)
{
  int lid,siz,mode = CM_POLL|CM_WAIT;
  mexListener *ml=NULL;
  char s[128];
//  char *c,*sn,dname[128];
  UINT16 arrayType = AT_UNKNOWN;
  short prpFormat = CF_BYTE;
  UINT32 prpSize = dout->dArrayLength;
  XPropertyQueryStruct xpqs;

//  strncpy(dname,devName,128);
  siz = 1;
  if (GetDeviceXPropertyInformation(devName,devProperty,&xpqs,&siz) == 0)
  {
    arrayType = xpqs.prpArrayType;
    prpSize = xpqs.prpSize;
    prpFormat = CF_DOUBLE + xpqs.prpFormat;
    /*
    if (arrayType & AT_SPECTRUM)
    {
      if (strchr(DELIMITER,dname[0]) != NULL) sn = strchr(&dname[1],'/'); else sn = dname;
      if ((c=strchr(&sn[1],'/')) != NULL) *c = 0;
      strcat(sn,"/#0"); // get the whole array no matter what
      if (dout->dArrayLength < xpqs.prpSize) dout->dArrayLength = xpqs.prpSize;
    }
    */
  }

  siz = dout->dArrayLength * fmtsizeof(dout->dFormat);
  if (dout->dFormat == CF_SPECTRUM) siz += 96;
  if ((ml=(mexListener *)calloc(1,sizeof(mexListener))) == NULL) return NULL;
  if ((ml->outbuf=(BYTE *)calloc(siz,1)) == NULL) 
  {
    free(ml);
    return NULL;
  }
  dout->data.vptr = ml->outbuf;

  if (din != NULL && (siz=din->dArrayLength * fmtsizeof(din->dFormat)) > 0)
  {
    if (din->dFormat == CF_SPECTRUM) siz += 96;
    if ((ml->inbuf=(BYTE *)calloc(siz,1)) == NULL) 
    {
      if (ml->outbuf != NULL) free(ml->outbuf);
      free(ml);
      return NULL;
    }
    memcpy(ml->din.data.vptr,din->data.vptr,siz);
  }
  else
  {
    ml->din.dFormat = CF_NULL;
    ml->din.dArrayLength = 0;
    ml->din.data.vptr = NULL;
    memset(ml->din.dTag,0,8);
  }
  if (flags > 0) mode |= flags;
  if ((lid=AttachLink(devName,devProperty,dout,din,CA_READ,pollrate,listenerCb,mode)) < 0)
  {
    if (ml->outbuf != NULL) free(ml->outbuf);
    if (ml->inbuf != NULL) free(ml->inbuf);
    free(ml);
    return NULL;
  }
  sprintf(s,"Static Listener for %s %s started",devName,devProperty);
  messageLog(s,0);
  strncpy(ml->devName,devName,64);
  strncpy(ml->devProperty,devProperty,32);
  ml->dout.dArrayLength = dout->dArrayLength;
  ml->dout.dFormat = dout->dFormat;
  memcpy(ml->dout.dTag,dout->dTag,8);
  ml->prpArrayType = arrayType;
  ml->prpSize = prpSize;
  ml->prpFormat = prpFormat;
  if (ml->din.dArrayLength != 0)
  {
    ml->din.dArrayLength = din->dArrayLength;
    ml->din.dFormat = din->dFormat;
    memcpy(ml->din.dTag,din->dTag,8);
  }
  ml->lid = lid;
  ml->pollingRate = pollrate;
  ml->nxt = mexListeners;
  mexListeners = ml;
  return ml;
}
int stopListener(int lid)
{
  mexListener *ml, *lstml = NULL;
  char s[128];
  for (ml=mexListeners; ml != NULL; ml = ml->nxt)
  {
    if (ml->lid == lid)
    {
      CLOSELINK(NULL,(short)lid);
      sprintf(s,"Static Listener for %s %s stopped",ml->devName,ml->devProperty);
      messageLog(s,0);
      if (ml == mexListeners) mexListeners = ml->nxt;
      else lstml->nxt = ml->nxt;
      if (ml->outbuf != NULL) free(ml->outbuf);
      if (ml->inbuf != NULL) free(ml->inbuf);
      free(ml);
      return 0;
    }
    lstml = ml;
  }
  return non_existent_elem;
}
mexListener *getListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,int pollrate)
{
  mexListener *ml=NULL;
  UINT32 dsiz;
  for (ml=mexListeners; ml != NULL; ml = ml->nxt)
  {
    if (strncmp(devName,ml->devName,64)) continue;
    if (strlen(devProperty) > 0 && strncmp(devProperty,ml->devProperty,32)) continue;
    if (dout != NULL)
    {
      if (dout->dFormat != ml->dout.dFormat) continue;
      if (dout->dArrayLength != ml->dout.dArrayLength) continue;
    }
    if (din != NULL && ml->din.dArrayLength != 0)
    {
      if (din->dFormat != ml->din.dFormat) continue;
      if (din->dArrayLength != ml->din.dArrayLength) continue;
      dsiz = ml->din.dArrayLength*fmtsizeof(ml->din.dFormat);
      if (memcmp(din->data.vptr,ml->inbuf,dsiz)) continue;
    }
    break;
  }
  if (ml != NULL && pollrate > 0)
  {
    if (pollrate < ml->pollingRate)
    { // force restart of listener at faster rate
      stopListener(ml->lid);
      return NULL; 
    }
  }
  return ml;  
}
int getDataFromListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,double *dTime,int rate,int flags)
{
  mexListener *ml;
  long retsiz;
  int cc = 0;
  BYTE *data;
  // int dsiz;

  if (devName == NULL || devProperty == NULL || dout == NULL) return argument_list_error;
  data = dout->data.bptr;
  if ( (ml=getListener(devName,devProperty,dout,din,rate)) == NULL && 
       (ml=startListener(devName,devProperty,dout,din,rate,flags)) == NULL )
  {
    return un_allocated;
  }
  ml->idleTimeInSeconds = 0;
  cc = ml->status;
  cc = (int)GetRPCDataAndTime((short)ml->lid,data,&ml->timestamp,&retsiz);
  if (dTime != NULL) *dTime = ml->timestamp;
  if (cc == 0) dout->dArrayLength = retsiz;
  return cc;
}
__declspec(dllexport) int getPropertyInfoFromListener(char *dev, char *prp, int *siz, int *fmt)
{
  mexListener *ml;

  if ((ml=getListener(dev,prp,NULL,NULL,0)) != NULL)
  {
    if (siz != NULL) *siz = ml->dout.dArrayLength;
    if (fmt != NULL) *fmt = ml->dout.dFormat;
    return 0;
  }
  return un_allocated;
}
__declspec(dllexport) int getPropertyInfo(char *dev, char *prp, int *siz, int *fmt, int *atyp)
{
  XPropertyQueryStruct xpqs;
  int psiz = 1, cc = 0;

  if ((cc=GetDeviceXPropertyInformation(dev,prp,&xpqs,&psiz)) == 0)
  {
    if (atyp != NULL)  *atyp = xpqs.prpArrayType;
    if (siz != NULL) *siz = xpqs.prpSize;
    if (fmt != NULL) *fmt = CF_DOUBLE + xpqs.prpFormat;
    return 0;
  }
  return cc;
}
int isGlobal(char *devName)
{
  if (strstr(devName,"GLOBQRY") != NULL) return -1;
  if (strstr(devName,"GLOBAL") != NULL) return -1;
  return 0;
}
int isMetaCall(char *devName,char *devProperty)
{
  if (strstr(devName,"HISTORY") != NULL) return -1;
  if (strstr(devProperty,".HIST") != NULL) return -1;
  if (strstr(devProperty,".EGU") != NULL) return -1;
  if (strstr(devProperty,".NAM") != NULL) return -1;
  if (strstr(devProperty,".DESC") != NULL) return -1;
  return 0;
}
int isQueryProperty(char *devProperty)
{
  if (strstr(devProperty,"PROPS") != NULL) return -1;
  if (strstr(devProperty,"PROPERTIES") != NULL) return -1;
  if (strstr(devProperty,"DEVICES") != NULL) return -1;
  if (strstr(devProperty,"TAGS") != NULL) return -1;
  if (strstr(devProperty,"CONTEXTS") != NULL) return -1;
  if (strstr(devProperty,"SUBSYSTEMS") != NULL) return -1;
  if (strstr(devProperty,"USERS") != NULL) return -1;
  if (strstr(devProperty,"IPNETS") != NULL) return -1;
  if (strstr(devProperty,"CLIENTS") != NULL) return -1;
  if (strstr(devProperty,"CONTRACTS") != NULL) return -1;
  if (strstr(devProperty,"HISTORIES") != NULL) return -1;
  if (strstr(devProperty,"LOGFILE") != NULL) return -1;
  if (strstr(devProperty,"LOGDEPTH") != NULL) return -1;
  if (strstr(devProperty,"STRUCTFORMAT") != NULL) return -1;
  if (strncmp(devProperty,"SRV",3) == 0) return -1;
  if (strstr(devProperty,"APPDATE") != NULL) return -1;
  if (strstr(devProperty,"APPVERSION") != NULL) return -1;
  return 0;
}
/* alsnExecLinkEx() */
/**
 * \brief Non-Pascal version (Win32 Only). Executes a synchronous call within an asynchronous static listener.
 * 
 * Asynchronous/Synchronous data exchange. For query calls (for instance to obtain a property
 * list), calls to the local history server, or calls with a timeout parameter less than 50 
 * milliseconds, the call reverts to ExecLinkEx(), i.e. synchronous data transfer.  Otherwise,
 * the initial call starts an asynchronous static listener for the selected device name and
 * property, polling at the rate determined by the timeout parameter. 
 * The initial call waits for the first data set to be returned before it completes.  Subsequent
 * calls retrieve data from the listener's buffer and complete immediately.  If no subsequent
 * calls are made within a 5 minute interval, the asynchronous link is cancelled.
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". 
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server 
 * to the client.  This should refer to a DTYPE object giving the data size, format, and tag (if any)
 * and reference to the data of the output data set.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the server from the 
 * client.  This should refer to a DTYPE object giving the data size, format, and tag (if any)
 * and reference to the data of the input data set.
 * \param access is the data access mode.  This can be any combination of access codes ORed together (CA_READ,
 * CA_WRITE, CA_CONNECT, etc.).
 * \param timeout the timeout period in milliseconds for the call to complete.  When it is known a priori 
 * that a call takes a long time to complete than you can set this value accordingly.  This parameter
 * has a particular meaning when used in the static listener routines.  Namely, it is also used to
 * determine the polling interval for the asynchronous listener which is attached to the call.  You
 * can force the call to execute synchronously without starting a listener by passing a value less
 * than 50 for this parameter.
 * \param dTimeStamp will be filled with the data time stamp corresponding to the requested data.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see EXECLINK(), EXECLINKEx(), DTYPE
 */
__declspec(dllexport) int alsnExecLinkEx(char *devName,char *devProperty,
                          DTYPE *dout,DTYPE *din,short access,long timeout,double *dTimeStamp)
{
  int cc,retry=2,len,accessflags=0;
  PropertyQueryStruct svrProp;
  DTYPE lcldin, *dinptr;

  if (din != NULL) 
  {
    dinptr = din;
  }
  else
  {
    lcldin.data.vptr = NULL;
    lcldin.dArrayLength = 0;
    lcldin.dFormat = CF_NULL;
    memset(lcldin.dTag,0,8);
    dinptr = &lcldin;
  }
  if (isGlobal(devName))
  {
    if (dout == NULL) return argument_list_error;
    if (!strncmp(devProperty,"NPROPS",6))
    {
      dout->data.sptr[0] = 1;
      return 0;
    }
    if (!strncmp(devProperty,"PROPS",5))
    {
      if ((len=dinptr->dArrayLength) > 31) len = 31; 
      strncpy(svrProp.prpName,dinptr->data.cptr,len); svrProp.prpName[len] = 0;
      svrProp.prpSize = 1;
      if (strstr(dinptr->data.cptr,"OPTICS") || strstr(dinptr->data.cptr,"STATE"))
      {
        svrProp.prpFormat = BFMT(CF_NAME16);
      }
      else
      {
        svrProp.prpFormat = BFMT(CF_FLOAT);
      } 
      memcpy(dout->data.vptr,&svrProp,sizeof(PropertyQueryStruct));
      return 0;
    }
    cc = GetNetGlobal(devProperty,dout->data.vptr,(short)dout->dArrayLength,dout->dFormat);
    if (dTimeStamp != NULL) *dTimeStamp = gDataTimeStamp;
    return cc;
  }
  if (!isQueryProperty(devProperty) && !isMetaCall(devName,devProperty) && timeout >= 50)
  {
    if (access & CA_NETWORK) accessflags = CM_NETWORK;
    cc = getDataFromListener(devName,devProperty,dout,dinptr,dTimeStamp,(int)timeout,accessflags);
    if (cc != use_stream_transport) return cc;
  }
  while (retry--)
  {
    if (timeout < 10) timeout *= 1000;
    cc = ExecLinkEx(devName,devProperty,dout,dinptr,access,timeout);
    if (dTimeStamp != NULL) *dTimeStamp = gDataTimeStamp;
    if (cc & CE_SENDDATA) cc = 0;
    if (cc == 0) break;
  }
  return cc;
}
/* alsnExecLink() */
/**
 * \brief Non-Pascal version (Win32 Only). Executes a synchronous call within an asynchronous static listener.
 * 
 * Asynchronous/Synchronous data exchange. For query calls (for instance to obtain a property
 * list), calls to the local history server, or calls with a timeout parameter less than 50 
 * milliseconds, the call reverts to ExecLinkEx(), i.e. synchronous data transfer.  Otherwise,
 * the initial call starts an asynchronous static listener for the selected device name and
 * property, polling at the rate determined by the timeout parameter. 
 * The initial call waits for the first data set to be returned before it completes.  Subsequent
 * calls retrieve data from the listener's buffer and complete immediately.  If no subsequent
 * calls are made within a 5 minute interval, the asynchronous link is cancelled.
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". 
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server 
 * to the client.  This should refer to a DTYPE object giving the data size, format, and tag (if any)
 * and reference to the data of the output data set.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the server from the 
 * client.  This should refer to a DTYPE object giving the data size, format, and tag (if any)
 * and reference to the data of the input data set.
 * \param access is the data access mode.  This can be any combination of access codes ORed together (CA_READ,
 * CA_WRITE, CA_CONNECT, etc.).
 * \param timeout the timeout period in milliseconds for the call to complete.  When it is known a priori 
 * that a call takes a long time to complete than you can set this value accordingly.  This parameter
 * has a particular meaning when used in the static listener routines.  Namely, it is also used to
 * determine the polling interval for the asynchronous listener which is attached to the call.  You
 * can force the call to execute synchronously without starting a listener by passing a value less
 * than 50 for this parameter.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see EXECLINK(), EXECLINKEx(), DTYPE
 */
__declspec(dllexport) int alsnExecLink(char *devName,char *devProperty,
                          DTYPE *dout,DTYPE *din,short access,long timeout)
{
  return alsnExecLinkEx(devName,devProperty,dout,din,access,timeout,NULL);
}
/* vBalsnExecLinkEx() */
/**
 * \brief Executes a synchronous call within an asynchronous static listener.  Uses PASCAL
 * calling convention for Visual Basic.
 * 
 * Asynchronous/Synchronous data exchange. For query calls (for instance to obtain a property
 * list), calls to the local history server, or calls with a timeout parameter less than 50 
 * milliseconds, the call reverts to ExecLinkEx(), i.e. synchronous data transfer.  Otherwise,
 * the initial call starts an asynchronous static listener for the selected device name and
 * property, polling at the rate determined by the timeout parameter. 
 * The initial call waits for the first data set to be returned before it completes.  Subsequent
 * calls retrieve data from the listener's buffer and complete immediately.  If no subsequent
 * calls are made within a 5 minute interval, the asynchronous link is cancelled.
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". 
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dbout is a pointer to the output data set, that is, the data set to be returned from the server 
 * to the client.  This should refer to a DBLK object giving the data size, format, and tag (if any)
 * of the output data set.
 * \param dataout is a reference to the data buffer which is to receive the returned data.
 * \param dbin is a pointer to the input data set, that is, the data set to be sent to the server from the 
 * client.  This should refer to a DBLK object giving the data size, format, and tag (if any)
 * of the input data set.
 * \param datain is a reference to the input data buffer.
 * \param access is the data access mode.  This can be any combination of access codes ORed together (CA_READ,
 * CA_WRITE, CA_CONNECT, etc.).
 * \param timeout the timeout period in milliseconds for the call to complete.  When it is known a priori 
 * that a call takes a long time to complete than you can set this value accordingly.  This parameter
 * has a particular meaning when used in the static listener routines.  Namely, it is also used to
 * determine the polling interval for the asynchronous listener which is attached to the call.  You
 * can force the call to execute synchronously without starting a listener by passing a value less
 * than 50 for this parameter.
 * \param dTimeStamp will be filled with the data time stamp corresponding to the requested data.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see EXECLINK(), EXECLINKEx(), DBLK
 */
__declspec(dllexport) short PASCAL vBalsnExecLinkEx(char *devName,char *devProperty,
             DBLK *dbout,void *dataout,DBLK *dbin,void *datain,short access,long timeout,double *dTimeStamp)
{
  DTYPE dout,din;
  DTYPE *doutptr = &dout, *dinptr = &din;

  if (dbout != NULL)
  {  
    dout.dFormat = dbout->dFormat;
    dout.dArrayLength = dbout->dArrayLength;
    memcpy(dout.dTag,dbout->dTag,8);
    dout.data.vptr = dataout;
  }
  else
  {
    doutptr = NULL;
  }
  if (dbin != NULL)
  {
    din.dFormat = dbin->dFormat;
    din.dArrayLength = dbin->dArrayLength;
    memcpy(din.dTag,dbin->dTag,8);
    din.data.vptr = datain;
  }
  else
  { 
    dinptr = NULL;
  }
  return alsnExecLinkEx(devName,devProperty,doutptr,dinptr,access,timeout,dTimeStamp);
}
/* vBalsnExecLink() */
/**
 * \brief Executes a synchronous call within an asynchronous static listener.  Uses PASCAL
 * calling convention for Visual Basic.
 * 
 * Asynchronous/Synchronous data exchange. For query calls (for instance to obtain a property
 * list), calls to the local history server, or calls with a timeout parameter less than 50 
 * milliseconds, the call reverts to ExecLinkEx(), i.e. synchronous data transfer.  Otherwise,
 * the initial call starts an asynchronous static listener for the selected device name and
 * property, polling at the rate determined by the timeout parameter. 
 * The initial call waits for the first data set to be returned before it completes.  Subsequent
 * calls retrieve data from the listener's buffer and complete immediately.  If no subsequent
 * calls are made within a 5 minute interval, the asynchronous link is cancelled.
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". 
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dbout is a pointer to the output data set, that is, the data set to be returned from the server 
 * to the client.  This should refer to a DBLK object giving the data size, format, and tag (if any)
 * of the output data set.
 * \param dataout is a reference to the data buffer which is to receive the returned data.
 * \param dbin is a pointer to the input data set, that is, the data set to be sent to the server from the 
 * client.  This should refer to a DBLK object giving the data size, format, and tag (if any)
 * of the input data set.
 * \param datain is a reference to the input data buffer.
 * \param access is the data access mode.  This can be any combination of access codes ORed together (CA_READ,
 * CA_WRITE, CA_CONNECT, etc.).
 * \param timeout the timeout period in milliseconds for the call to complete.  When it is known a priori 
 * that a call takes a long time to complete than you can set this value accordingly.  This parameter
 * has a particular meaning when used in the static listener routines.  Namely, it is also used to
 * determine the polling interval for the asynchronous listener which is attached to the call.  You
 * can force the call to execute synchronously without starting a listener by passing a value less
 * than 50 for this parameter.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see EXECLINK(), EXECLINKEx(), DBLK
 */
__declspec(dllexport) short PASCAL vBalsnExecLink(char *devName,char *devProperty,
             DBLK *dbout,void *dataout,DBLK *dbin,void *datain,short access,long timeout)
{
  return vBalsnExecLinkEx(devName,devProperty,dbout,dataout,dbin,datain,access,timeout,NULL);
}
__declspec(dllexport) short PASCAL tputgetArrayEx(char *fullDeviceNameAndProperty, short access,
                                                short putdatafmt,void *putdata,int putsiz,
                                                short getdatafmt, void *getdata, int getsiz,double *dTimeStamp)
{
  DTYPE dout, din;
  char *c;
  static char sbuf[256],dev[64],prp[48];
  int cc, i, rate = 1000;
  static int inside = FALSE;

  if (inside) return operation_busy;
  inside = TRUE;
  for (c=fullDeviceNameAndProperty; *c == ' ' && *c != 0; c++); // remove leading blanks
  strncpy(sbuf,c,256);
  if ((c=strchr(sbuf,'@')) != NULL)
  {
    *c++ = 0; // peel it off before we start tokenizing
    strupr(c);
    if (strstr(c,"NETWORK") != NULL || strstr(c,"MCAST") != NULL || strstr(c,"MULTICAST") != NULL)
      access |= CA_NETWORK;
    rate = atoi(c);
  }
  if ((c=strchr(sbuf,'[')) == NULL) return illegal_property;
  *c++ = 0; while (*c == ' ' && *c != 0) c++; // remove leading blanks
  strncpy(prp,c,32);  prp[32] = 0; // record property
  if ((c=strchr(prp,']')) != NULL) *c = 0;
  for (i=strlen(prp)-1; i > 0 && prp[i] == ' '; i--) prp[i] = 0; // remove trailing blanks
	strncpy(dev,sbuf,64);
  for (i=strlen(dev)-1; i > 0 && dev[i] == ' '; i--) dev[i] = 0; // remove trailing blanks
  
  din.dFormat = putdatafmt;
  din.dArrayLength = putsiz;
  memset(din.dTag,0,8);
  din.data.vptr = putdata;

  dout.dFormat = getdatafmt;
  dout.dArrayLength = getsiz;
  memset(dout.dTag,0,8);
  dout.data.vptr = getdata;

  cc = alsnExecLinkEx(dev,prp,&dout,&din,access,1,dTimeStamp);
  inside = FALSE;
  return cc;
}
__declspec(dllexport) short PASCAL tputgetArray(char *fullDeviceNameAndProperty, short access,
                                                short putdatafmt,void *putdata,int putsiz,
                                                short getdatafmt, void *getdata, int getsiz)
{
  return tputgetArrayEx(fullDeviceNameAndProperty,access,putdatafmt,putdata,putsiz,getdatafmt,getdata,getsiz,NULL);
}
__declspec(dllexport) short PASCAL tputString(char *fullDeviceNameAndProperty,char *putdata)
{
  char *c,sbuf[256];
  int len = 0;

  if (fullDeviceNameAndProperty == NULL || putdata == NULL) return argument_list_error;
  if ((len=strlen(putdata)) == 0) return argument_list_error;
  strncpy(sbuf,fullDeviceNameAndProperty,256);
  if ((c=strchr(sbuf,'@')) != NULL) *c = 0; // peel it off and replace it
  strcat(sbuf,"@1"); // make sure StringArray calls don't start a listener !
  return tputgetArray(sbuf,CA_WRITE|CA_READ,CF_TEXT,putdata,len,CF_NULL,NULL,0);
}
__declspec(dllexport) short PASCAL tputDoubleArray(char *fullDeviceNameAndProperty,double *putdata,int putsiz)
{
  void *outdata = NULL;
  int outsiz = 0;
  short outfmt = CF_NULL;

  return tputgetArray(fullDeviceNameAndProperty,CA_WRITE|CA_READ,CF_DOUBLE,putdata,putsiz,outfmt,outdata,outsiz);
}
__declspec(dllexport) short PASCAL tputSingleArray(char *fullDeviceNameAndProperty,float *putdata,int putsiz)
{
  void *outdata = NULL;
  int outsiz = 0;
  short outfmt = CF_NULL;

  return tputgetArray(fullDeviceNameAndProperty,CA_WRITE|CA_READ,CF_FLOAT,putdata,putsiz,outfmt,outdata,outsiz);
}
__declspec(dllexport) short PASCAL tputLongArray(char *fullDeviceNameAndProperty,int *putdata,int putsiz)
{
  void *outdata = NULL;
  int outsiz = 0;
  short outfmt = CF_NULL;

  return tputgetArray(fullDeviceNameAndProperty,CA_WRITE|CA_READ,CF_LONG,putdata,putsiz,outfmt,outdata,outsiz);
}
__declspec(dllexport) short PASCAL tputShortArray(char *fullDeviceNameAndProperty,short *putdata,int putsiz)
{
  void *outdata = NULL;
  int outsiz = 0;
  short outfmt = CF_NULL;

  return tputgetArray(fullDeviceNameAndProperty,CA_WRITE|CA_READ,CF_SHORT,putdata,putsiz,outfmt,outdata,outsiz);
}
__declspec(dllexport) void tDetachNotifier(void (*nf)(int))
{
  mexListener *ml;
  for (ml=mexListeners; ml != NULL; ml = ml->nxt)
  {
    if (ml->nf == nf) ml->nf = NULL;
  }
}
__declspec(dllexport) short tAttachNotifier(char *fullDeviceNameAndProperty, void (*nf)(int), int nid)
{
  char *c,sbuf[256],dev[64],prp[48];
  int i, rate = 1000;
  mexListener *ml=NULL;

  for (c=fullDeviceNameAndProperty; *c == ' ' && *c != 0; c++); // remove leading blanks
  strncpy(sbuf,c,256);
  if ((c=strchr(sbuf,'@')) != NULL)
  {
    *c++ = 0; // peel it off before we start tokenizing
    rate = atoi(c);
  }
  if ((c=strchr(sbuf,'[')) == NULL) return illegal_property;
  *c++ = 0; while (*c == ' ' && *c != 0) c++; // remove leading blanks
  strncpy(prp,c,32);  prp[32] = 0; // record property
  if ((c=strchr(prp,']')) != NULL) *c = 0;
  for (i=strlen(prp)-1; i > 0 && prp[i] == ' '; i--) prp[i] = 0; // remove trailing blanks
	strncpy(dev,sbuf,64);
  for (i=strlen(dev)-1; i > 0 && dev[i] == ' '; i--) dev[i] = 0; // remove trailing blanks

  for (ml=mexListeners; ml != NULL; ml = ml->nxt)
  {
    if (strncmp(dev,ml->devName,64)) continue;
    if (strlen(prp) > 0 && strncmp(prp,ml->devProperty,32)) continue;
    break;
  }
  if (ml == NULL) return link_not_open;
  ml->nf = nf;
  ml->nid = nid;
  return 0;
}
__declspec(dllexport) short PASCAL tgetArray(char *fullDeviceNameAndProperty,short datafmt,void *dataout,int buffersiz)
{
  DTYPE dout;
  char *c,sbuf[256],dev[64],prp[48];
  int i, rate = 1000;
  short access = CA_READ;

  for (c=fullDeviceNameAndProperty; *c == ' ' && *c != 0; c++); // remove leading blanks
  strncpy(sbuf,c,256);
  if ((c=strchr(sbuf,'@')) != NULL)
  {
    *c++ = 0; // peel it off before we start tokenizing
    strupr(c);
    if (strstr(c,"NETWORK") != NULL || strstr(c,"MCAST") != NULL || strstr(c,"MULTICAST") != NULL)
      access |= CA_NETWORK;
    rate = atoi(c);
  }
  if ((c=strchr(sbuf,'[')) == NULL) return illegal_property;
  *c++ = 0; while (*c == ' ' && *c != 0) c++; // remove leading blanks
  strncpy(prp,c,32);  prp[32] = 0; // record property
  if ((c=strchr(prp,']')) != NULL) *c = 0;
  for (i=strlen(prp)-1; i > 0 && prp[i] == ' '; i--) prp[i] = 0; // remove trailing blanks
	strncpy(dev,sbuf,64);
  for (i=strlen(dev)-1; i > 0 && dev[i] == ' '; i--) dev[i] = 0; // remove trailing blanks
  
  dout.dFormat = datafmt;
  dout.dArrayLength = buffersiz;
  memset(dout.dTag,0,8);
  dout.data.vptr = dataout;
  return alsnExecLink(dev,prp,&dout,NULL,access,rate);
}
/* tgetSingleArray() */
/**
 * \brief Executes a synchronous call within an asynchronous static listener.
 * 
 * Asynchronous/Synchronous data exchange. This is a convenience
 * routine which accesses alsnExecLink() explicitly for an array of Single (float) data type
 * without data input.
 *
 * \param fullDeviceNameAndProperty is the full device name 
 * appended with the requested property parsed according to 
 * '/\<Context\>/\<Server\>/\<Device\>[\<Property\>]@\<rate\>'.
 * For example: "/HERA/BPM/WL167[ORBIT.X]@500". If the '@\<rate\>' is omitted, a polling
 * rate of 1000 milliseconds is assumed.
 * \param dataout is a reference to the float array to receive the output data.
 * \param buffersiz is the size of (i.e. number of elements in) the float array.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see alsnExecLink(), vBalsnExecLink().
 */
__declspec(dllexport) short PASCAL tgetSingleArray(char *fullDeviceNameAndProperty,float *dataout,int buffersiz)
{
  return tgetArray(fullDeviceNameAndProperty,CF_FLOAT,dataout,buffersiz);
}
__declspec(dllexport) short PASCAL tputgetSingleArray(char *fullDeviceNameAndProperty,float *datain,int insize,float *dataout,int outsize,double *datatime)
{
  return tputgetArrayEx(fullDeviceNameAndProperty,CA_READ,CF_FLOAT,datain,insize,CF_FLOAT,dataout,outsize,datatime);
}
/* tgetDoubleArray() */
/**
 * \brief Executes a synchronous call within an asynchronous static listener.
 * 
 * Asynchronous/Synchronous data exchange. This is a convenience
 * routine which accesses alsnExecLink() explicitly for an array of double data type
 * without data input.
 *
 * \param fullDeviceNameAndProperty is the full device name 
 * appended with the requested property parsed according to 
 * '/\<Context\>/\<Server\>/\<Device\>[\<Property\>]@\<rate\>'.
 * For example: "/HERA/BPM/WL167[ORBIT.X]@500". If the '@\<rate\>' is omitted, a polling
 * rate of 1000 milliseconds is assumed.
 * \param dataout is a reference to the double array to receive the output data.
 * \param buffersiz is the size of (i.e. number of elements in) the array.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see alsnExecLink(), vBalsnExecLink().
 */
__declspec(dllexport) short PASCAL tgetDoubleArray(char *fullDeviceNameAndProperty,double *dataout,int buffersiz)
{
  return tgetArray(fullDeviceNameAndProperty,CF_DOUBLE,dataout,buffersiz);
}
__declspec(dllexport) short PASCAL tputgetDoubleArray(char *fullDeviceNameAndProperty,double *datain,int insize,double *dataout,int outsize,double *datatime)
{
  return tputgetArrayEx(fullDeviceNameAndProperty,CA_READ,CF_DOUBLE,datain,insize,CF_DOUBLE,dataout,outsize,datatime);
}
/* tgetLongArray() */
/**
 * \brief Executes a synchronous call within an asynchronous static listener.
 * 
 * Asynchronous/Synchronous data exchange. This is a convenience
 * routine which accesses alsnExecLink() explicitly for an array of long data type
 * without data input.
 *
 * \param fullDeviceNameAndProperty is the full device name 
 * appended with the requested property parsed according to 
 * '/\<Context\>/\<Server\>/\<Device\>[\<Property\>]@\<rate\>'.
 * For example: "/HERA/BPM/WL167[ORBIT.X]@500". If the '@\<rate\>' is omitted, a polling
 * rate of 1000 milliseconds is assumed.
 * \param dataout is a reference to the long integer array to receive the output data.
 * \param buffersiz is the size of (i.e. number of elements in) the array.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see alsnExecLink(), vBalsnExecLink().
 */
__declspec(dllexport) short PASCAL tgetLongArray(char *fullDeviceNameAndProperty,long *dataout,int buffersiz)
{
  return tgetArray(fullDeviceNameAndProperty,CF_LONG,dataout,buffersiz);
}
__declspec(dllexport) short PASCAL tputgetLongArray(char *fullDeviceNameAndProperty,long *datain,int insize,long *dataout,int outsize,double *datatime)
{
  return tputgetArrayEx(fullDeviceNameAndProperty,CA_READ,CF_LONG,datain,insize,CF_LONG,dataout,outsize,datatime);
}
/* tgetShortArray() */
/**
 * \brief Executes a synchronous call within an asynchronous static listener.
 * 
 * Asynchronous/Synchronous data exchange. This is a convenience
 * routine which accesses alsnExecLink() explicitly for an array of short integer data type
 * without data input.
 *
 * \param fullDeviceNameAndProperty is the full device name 
 * appended with the requested property parsed according to 
 * '/\<Context\>/\<Server\>/\<Device\>[\<Property\>]@\<rate\>'.
 * For example: "/HERA/BPM/WL167[ORBIT.X]@500". If the '@\<rate\>' is omitted, a polling
 * rate of 1000 milliseconds is assumed.
 * \param dataout is a reference to the short integer array to receive the output data.
 * \param buffersiz is the size of (i.e. number of elements in) the array.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see alsnExecLink(), vBalsnExecLink().
 */
__declspec(dllexport) short PASCAL tgetShortArray(char *fullDeviceNameAndProperty,short *dataout,int buffersiz)
{
  return tgetArray(fullDeviceNameAndProperty,CF_SHORT,dataout,buffersiz);
}
__declspec(dllexport) short PASCAL tputgetShortArray(char *fullDeviceNameAndProperty,short *datain,int insize,short *dataout,int outsize,double *datatime)
{
  return tputgetArrayEx(fullDeviceNameAndProperty,CA_READ,CF_SHORT,datain,insize,CF_SHORT,dataout,outsize,datatime);
}
/* tgetStringArray() */
/**
 * \brief Executes a synchronous call within an asynchronous static listener.
 * 
 * Synchronous data exchange. This is a convenience
 * routine which accesses ExecLink() explicitly for an array of NAME32 data type
 * without data input.  This is useful for retrieving a list of properties or devices from a 
 * server for instance.
 *
 * \param fullDeviceNameAndProperty is the full device name 
 * appended with the requested property parsed according to 
 * '/\<Context\>/\<Server\>/\<Device\>[\<Property\>]'.
 * For example: "/HERA/BPM/WL167[PROPERTIES]".
 * \param dataout is a reference to the NAME32 array to receive the output data.
 * \param buffersiz is the size of (i.e. number of elements in) the NAME32 array.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see alsnExecLink(), vBalsnExecLink().
 */
__declspec(dllexport) short PASCAL tgetStringArray(char *fullDeviceNameAndProperty,char *dataout,int buffersiz)
{
  int cc, len, i, j;
  char *c,sbuf[256];

  if (buffersiz <=0 || fullDeviceNameAndProperty == NULL || dataout == NULL) return argument_list_error;

  strncpy(sbuf,fullDeviceNameAndProperty,256);
  if ((c=strchr(sbuf,'@')) != NULL) *c = 0; // peel it off and replace it
  strcat(sbuf,"@1"); // make sure StringArray calls don't start a listener !
  if ((c=alloca(buffersiz)) == NULL) return out_of_client_memory;
  if ((cc=tgetArray(sbuf,CF_NAME32,c,buffersiz/32)) != 0) return cc;

  for (j=0, i=0; i<buffersiz && j<buffersiz; j += 32)
  {
    len = MIN(strlen(&c[j]),32);
    strncpy(&dataout[i],&c[j],len);
    i += len;
    dataout[i++] = '\n';
  }
  return 0;
}
__declspec(dllexport) short PASCAL tgetString(char *fullDeviceNameAndProperty,char *dataout,int buffersiz)
{
  int cc;
  char *c,sbuf[256];

  if (buffersiz <=0 || fullDeviceNameAndProperty == NULL || dataout == NULL) return argument_list_error;

  strncpy(sbuf,fullDeviceNameAndProperty,256);
  if ((c=alloca(buffersiz)) == NULL) return out_of_client_memory;
  if ((cc=tgetArray(sbuf,CF_TEXT,c,buffersiz)) != 0) return cc;
  strncpy(dataout,c,buffersiz);
  return 0;
}
__declspec(dllexport) short PASCAL tputgetString(char *fullDeviceNameAndProperty,char *datain,char *dataout,int buffersiz)
{
  int cc, inputlength = 0;
  char *c,sbuf[256],*cin = NULL;

  if (buffersiz <=0 || fullDeviceNameAndProperty == NULL || dataout == NULL) return argument_list_error;
  if (datain != NULL) 
  {
    cin = datain;
    inputlength = strlen(cin);
  }
  strncpy(sbuf,fullDeviceNameAndProperty,256);
  if ((c=alloca(buffersiz)) == NULL) return out_of_client_memory;
  if ((cc=tputgetArray(sbuf,CA_READ,CF_TEXT,cin,inputlength,CF_TEXT,c,buffersiz)) != 0) return cc;
  strncpy(dataout,c,buffersiz);
  return 0;
}
__declspec(dllexport) short PASCAL tputgetStringEx(char *fullDeviceNameAndProperty,char *datain,char *dataout,int buffersiz,double *datatime)
{
  int cc, inputlength = 0;
  char *c,sbuf[256],*cin = NULL;

  if (buffersiz <=0 || fullDeviceNameAndProperty == NULL || dataout == NULL) return argument_list_error;
  if (datain != NULL) 
  {
    cin = datain;
    inputlength = strlen(cin);
  }
  strncpy(sbuf,fullDeviceNameAndProperty,256);
  if ((c=alloca(buffersiz)) == NULL) return out_of_client_memory;
  if ((cc=tputgetArrayEx(sbuf,CA_READ,CF_TEXT,cin,inputlength,CF_TEXT,c,buffersiz,datatime)) != 0) return cc;
  strncpy(dataout,c,buffersiz);
  return 0;
}
