/** \file tinecln32.h
 * \brief TINE Client DLL documentation 
 *
 */
#ifndef _TINECLN32_H_
#define _TINECLN32_H_

#include <windows.h>
#include <winsock.h>
#include <wsipx.h>
#include <wsnwlink.h>
#include "rpcdefs.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#define INTSWAP(w)       ((w<<8)|(w>>8))
#define hNULL            0

#define FEC_NAME_SIZE    16
#define ELEM_NAME_SIZE   16
#define PROPERTY_SIZE    32
#define EQNAME_SIZE	      6
#define DEVICE_SIZE      16
#define CONTEXT_SIZE     16

#define RPCERR_SIZE 32

typedef void (__stdcall *cbTLink)(int id,int cc);
#define USE_STDCALL_CALLBACK 0x1111

#define ccerr(e) {cc = e; goto err; }
#define OFFSETIN(struc,field)	((unsigned short)&(((struc *)0L)->field))

typedef struct
{
  char name[8];
} NAME8;

typedef struct
{
  char name[16];
} NAME16;

typedef struct
{
  char name[32];
} NAME32;

typedef struct
{
  char name[64];
} NAME64;

typedef struct
{
  char comment[80];
  time_t tm;
  float s_start;
  float s_inc;
  UINT32 status;
  float d_spect_array[1024];
} DSPECTRUM;

typedef union
{
  void   *vptr;
  char   *cptr;
  UINT8  *bptr;
  UINT16 *sptr;
  UINT32 *lptr;
  float  *fptr;
  double *dptr;
} DUNION;

typedef struct
{
  UINT32 dArrayLength;
  short  dFormat;
  short  reserved;
  char   dTag[16];
  DUNION data;
} DTYPE;

typedef struct
{
  UINT32 dArrayLength;
  short  dFormat;
  short  reserved;
  char   dTag[8];
} DBLK;

typedef struct
{
  long   ival;
  float f1val;
  float f2val;
  time_t   ts;
  char str[80];
} DUSTRING;

// =================================================================

typedef struct
{
  char  prpName[32];
  char  prpDescription[32];
  unsigned short prpSize;
  BYTE prpFormat;
  BYTE prpAccess;
} PropertyQueryStruct;
#define PROPERTYQUERYSTRUCT_SIZE (32+32+2+1+1)

typedef struct
{
  char  prpName[32];
  char  prpDescription[64];
  char  prpRedirection[32];
  char  prpTag[8];
  char  prpTagIn[8];
  char  prpUnits[16];
  float prpMinValue;
  float prpMaxValue;
  UINT32 prpSize;
  UINT32 prpSizeIn;
  UINT32 prpNumOverloads;
  UINT16 prpHistoryDepthShort;
  UINT16 prpHistoryDepthLong;
  BYTE prpFormat;
  BYTE prpFormatIn;
  BYTE prpAccess;
  BYTE prpGraphType;
} PropertyQueryStructEx;
#define PROPERTYQUERYSTRUCTEX_SIZE (32+64+32+8*2+16+4*5+2*2+4)

typedef struct
{
  char  prpName[32];
  char  prpDescription[64];
  char  prpRedirection[32];
  char  prpTag[8];
  char  prpTagIn[8];
  char  prpUnits[16];
  float prpMinValue;
  float prpMaxValue;
  UINT32 prpSize;
  UINT32 prpSizeIn;
  UINT32 prpNumOverloads;
  UINT16 prpHistoryDepthShort;
  UINT16 prpHistoryDepthLong;
  BYTE prpFormat;
  BYTE prpFormatIn;
  BYTE prpAccess;
  BYTE prpGraphType;
  char  rngUnits[16];
  float rngMinValue;
  float rngMaxValue;
  UINT16 numRows;
  UINT16 rowSize;
  UINT16 prpArrayType;
  UINT16 reserved[3];
} XPropertyQueryStruct;
#define XPROPERTYQUERYSTRUCT_SIZE (PROPERTYQUERYSTRUCTEX_SIZE + 16+2*4+6*2)

typedef struct
{
  char FecName[FEC_NAME_SIZE];
  char reserved[4];
  time_t localtime;
  time_t starttime;
  long   systemPollingRate;
  short  numBkgTsks;
  short  numTotalContracts;
  short  numTotalClients;
  short  numTargetContracts;
  short  numTargetClients;
  short  numConnections;
  UINT32 numConnectionTimeouts;
  UINT32 numConnectionArrivals;
  UINT32 numUDPpkts;
  UINT32 numTCPpkts;
  UINT32 numIPXpkts;
  UINT32 numSPXpkts;
} ActivityQueryStruct;
#define ACTIVITYQUERYSTRUCT_SIZE (FEC_NAME_SIZE+4+4*3+2*6+4*6)

typedef struct
{
  char desc[64];
  BYTE dataFmt;
  BYTE dataSiz;
  BYTE data[64];
  time_t timeEntered;
  UINT32 codeNow;
  UINT32 codePrv;
  UINT32 codeNxt;
} ServerStateStruct;
#define SERVERSTATESTRUCT_SIZE (64+2+64+4*4)

#define IPX			    0x01
#define SPX         0x02
#define TCP         0x04
#define UDP         0x08
#define GLB         0x10
#define UDPBCST			0x20
#define IPXBCST			0x40
#define LCL         0x80

#define BCAST_ID 255

#define WORK_BUFFER_SIZE 0x1000
// IPX/SPX specific:
#define RPC_SOCKET      0x45 // listening SPX socket
#define BRDCST_SOCKET   0x67 // listening broadcast socket
#define REG_SOCKET      0x55 // listening registration socket
#define MCAST_SOCKET    0x85 // listening mcast socket
#define NUM_RCV_ECB     3              // associated incoming contracts
#define NUM_GLB_ECB     2              // incoming broadcasts
#define NUM_REG_ECB     2              // incoming registration requests
#define NUM_CLN_ECB     3              // client RPC ecbs
#define NUM_CNS_ECB     5              // consumer ecbs
#define IPX_DATA_SIZE   512
#define IPX_BUFFER_SIZE 1466 // for Client32 for DOS

// TCP/UDP specific:
#define NumTCPChan	      5
#ifndef UDP_PORT
# define UDP_PORT	      9003
#endif
#ifndef TCP_PORT
# define TCP_PORT	      8503
#endif
# define TCP_LEGACYPORT 8003
#ifndef GCAST_PORT
# define GCAST_PORT     8004
#endif
#ifndef MCAST_PORT
# define MCAST_PORT     9503
#endif
#ifndef UDP_CLN_PORT
# define UDP_CLN_PORT   8050
#endif
#ifndef UDP_CLN_RANGE
# define UDP_CLN_RANGE 400
#endif

#ifndef MCAST_ADDR
# define MCAST_ADDR "238.1.1.1"
#endif
#ifndef GCAST_ADDR
# define GCAST_ADDR "238.1.1.0"
#endif
#ifndef MCAST_TTL
# define MCAST_TTL 16
#endif

#define UDP_BUFFER_SIZE 1472
#define WORK_SIZE  (WORK_BUFFER_SIZE*sizeof(double))

// ==========================================================
// ======            prototypes:                    =========
// ==========================================================

__declspec(dllimport) short PASCAL RPC(char *name, short EqNo, char *prop, short size, void *data, short datasize, short format);
__declspec(dllimport) short PASCAL RPCEx(char *name, short EqNo, char *prop, short size, void *data, short datasize, short format,long timeout);
__declspec(dllimport) short PASCAL HOTRPC(HWND hWnd, char *name, short EqNo, char *prop, long timer, short size, short format);
__declspec(dllimport) short PASCAL CLOSERPC(HWND hWnd, short ID);
__declspec(dllimport) short PASCAL GETRPC(short ID, void *buf);
__declspec(dllimport) short PASCAL GETGLOBAL(char *keyword, void *data, short size, short format);
__declspec(dllimport) short PASCAL PUTRPC (short i, void  *VBbuf);

/** 
 * \brief Use with \b Visual \b Basic \b 3.0 or \b 6.0. Retrieves data associated with the given link index into the calling process.
 *
 * \b Visual \b Basic :
 * \n Declare Function \b GETRPCDATA Lib "tinecln32.dll" (ByVal linkId As Integer, databuf As Any, timestamp As Long) As Integer
 * \n Declare Function \b GETRPCSTRDATA Lib "tinecln32.dll" Alias "GETRPCDATA" (ByVal linkId As Integer, ByVal sdatabuf As String, timestamp As Long) As Integer
 *
 * \param linkId is the link identifier for which the data is desired
 * \param databuf is a reference to the data buffer into which the link data are to be copied.  This
 *        buffer must be pre-alloced with enough room to store the data.
 * \param timestamp is the data timestamp as UTC (seconds since Jan 1, 1970)
 *
 * It is only necessary to call this routine directly when using the 'Simple TINE Client ActiveX 
 * interface.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see GETRPCDATAEx(), GetRPCDataAndTime()
 *
 * \b Example:
 *
 * \include eg_GETRPCDATA.txt
 */
__declspec(dllimport) short PASCAL GETRPCDATA(short linkId, void *databuf,time_t *timestamp);

/** 
 * \brief Use with \b Visual \b Basic \b 3.0 or \b 6.0. Retrieves data associated with the given link index into the calling process.
 *
 * \b Visual \b Basic :
 * \n Declare Function \b GETRPCDATAEx Lib "tinecln32.dll" (ByVal linkId As Integer, databuf As Any, timestamp As Long, datasize As Long) As Integer
 * \n Declare Function \b GETRPCSTRDATAEx Lib "tinecln32.dll" Alias "GETRPCDATA" (ByVal linkId As Integer, ByVal sdatabuf As String, timestamp As Long, datasize As Long) As Integer
 *
 * \param linkId is the link identifier for which the data is desired
 * \param databuf is a reference to the data buffer into which the link data are to be copied.  This
 *        buffer must be pre-alloced with enough room to store the data.
 * \param timestamp is the data timestamp as UTC (seconds since Jan 1, 1970)
 * \param datasize contains the data size (as array dimension) of the data set returned from
 *        the server. It is assumed that the size of the data buffer can hold the data returned.
 *        The data returned will never be greater than the size demanded by the initial
 *        link parameters as given for example in AttachLink().
 *
 * It is only necessary to call this routine directly when using the 'Simple TINE Client ActiveX 
 * interface.  This call differs from GETRPCDATA() in that it returns in the fourth parameter
 * the size of the data set actually returned from the server (possibly less than what was asked for.).
 * This call differs from GetRPCDataAndTime(), in that only the 'seconds' part of the data timestamp
 * is returned in the third parameter.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see GETRPCDATA(), GetRPCDataAndTime()
 */
__declspec(dllimport) short PASCAL GETRPCDATAEx(short linkId, void *databuf,time_t *timestamp,long *datasize);

/** 
 * \brief Use with \b Visual \b Basic \b 6.0. Retrieves data associated with the given link 
 * index into the calling process
 *
 * \b Visual \b Basic :
 * \n Declare Function \b GetRPCDataAndTime Lib "tinecln32.dll" (ByVal linkId As Integer, databuf As Any, timestamp As Long, datasize As Long) As Integer
 * \n Declare Function \b GetRPCStrDataAndTime Lib "tinecln32.dll" Alias "GetRPCDataAndTime" (ByVal linkId As Integer, ByVal sdatabuf As String, timestamp As Long, datasize As Long) As Integer
 *
 * \param linkId is the link identifier for which the data is desired
 * \param databuf is a reference to the data buffer into which the link data are to be copied.  This
 *        buffer must be pre-alloced with enough room to store the data.
 * \param dtimestamp is the TINE data timestamp as UTC + fractional seconds.
 * \param datasize contains the data size (as array dimension) of the data set returned from
 *        the server. It is assumed that the size of the data buffer can hold the data returned.
 *        The data returned will never be greater than the size demanded by the initial
 *        link parameters as given for example in AttachLink().
 *
 * It is only necessary to call this routine directly when using the 'Simple TINE Client ActiveX 
 * interface.  This call differs from GETRPCDATA() in that it returns the full TINE timestamp
 * (UTC + millisecons as fractional part) as well as the size of the data set actually returned
 * from the server (possibly less than what was asked for.).
 *
 * \note This call is unavailable in Visual Basic 3.0 (use GETRPCDATAEx()).
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see GETRPCDATA(), GETRPCDATAEx()
 */
__declspec(dllimport) short PASCAL GetRPCDataAndTime(short linkId, void *databuf,double *dtimestamp,long *datasize);
__declspec(dllimport) short PASCAL GetDataAndTime(short i, void *buf,double *dtimestamp,long *datasize,int *status);
void IncomingIPX(void);
void IncomingGlobalIPX(void);
int  IncomingUDP(unsigned long i, int err);
__declspec(dllimport) short GetDeviceServersFromFile(NAME16 *context,NAME16 *servers,int *bufsiz);
__declspec(dllimport) short GetContextsFromFile(NAME16 *contexts,int *bufsiz);
__declspec(dllimport) short PASCAL ListenForMulticasts(void);
__declspec(dllimport) short PASCAL ListenForNetGlobals(void);

/**
 * \brief Non-Pascal version (Win32 Only). Establishes an asynchronous link with the given calling parameters.
 * 
 * Asynchronous data exchange. Calls in the AttachLink() family attempt to establish the requested data link
 * with the given calling parameters and complete immediately.  If the address of the requested 
 * server can be resolved and the calling parameters are otherwise okay, the call will complete
 * successfully and return a valid (positive) link handle.  This does not mean that the 
 * requested server will respond to the request.  The data and status information from the server
 * will be returned asynchronously inside the callback routine passed in the call.
 * In Visual Basic 3.0, there is no equivalent to this call.
 * In the Win32 world, use T_ATTACHLINK() with Visual Basic.
 * Use AttachLink() with interfaces which do not require the Pascal calling convention (e.g. VC++,
 * LabView, MatLab, etc.).
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". Note that, if devName references a locally registered equipment module,
 * an internal call to ExecLocalLink() is made.  For instance: "/LOCAL/BPMEQM/WL167" sets the context to 
 * "LOCAL", and "BPMEQM" must be a locally registered equipment module. In such a case, the call can only 
 * be successful when the client and server are running together in the same process). 
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server 
 * to the client.  The structure element dout->data should be a pointer to a pre-allocated data buffer,
 * which is to receive the returned data.  If dout is a NULL pointer, it is assumed that no data are to 
 * be returned.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the server from the 
 * client.  If din is a NULL pointer, it is assumed that no data are to be sent to the server
 * \param access is the data access mode.  This can be any combination of access codes ORed together (CA_READ,
 * CA_WRITE, CA_CONNECT, etc.).
 * \param pollrate is the polling interval in milliseconds at which the server is requested to
 * acquire data. This value also signals a timeout if there is no notification within this amount
 * of time plus the system heartbeat (1000 msec for POLLing links).  One can use AttachLink()
 * for asynchronous single READ and WRITE requests by specifying a pollmode of CM_SINGLE.
 * \param callback is a reference to the a callback function which is to be called by the TINE
 * engine when either incoming data have arrived or an error condition has arised.  This should
 * be a pointer to a funtion with prototype \b void \b callback(\b int linkId, \b int statusCode)
 * \param pollmode is the TINE poll mode, i.e. one of CM_CANCEL, CM_SINGLE, CM_REFRESH, CM_POLL,
 * which can be ORed with one of CM_BCAST, CM_NETWORK, CM_MCAST, CM_USE_ON_ERROR, CM_RETRY,
 * CM_GROUPED, CM_CONNECT, CM_WAIT, or CM_SYNCNOTIFY. (see the discussion of Data Access Modes).
 *
 * \return A positive link handle ID if successful, otherwise the negative of a TINE completion 
 * code which can be interpreted by a call to GetRPCLastError().  If the TINE engine can resolve
 * the address of the requested server and all of the calling parameters are otherwise
 * correct the call should complete successfully.  Possible errors might be 'non existent element'
 * or 'non existent fec', etc.
 *
 * \note This form of the call cannot be used with Visual Basic. Use T_ATTACHLINK() instead.
 *
 * \see AttachLinkEx(), T_ATTACHLINK(), DTYPE, EXECLINK(), EXECLINKEx()
 */
__declspec(dllimport) int AttachLink(char *devName,char *devProperty,
				DTYPE *dout,DTYPE *din,short access,int pollrate,
        void (*callback)(int,int),int pollmode);

/**
 * \brief Non-Pascal version (Win32 Only). Establishes an asynchronous link with the given calling parameters.
 * 
 * Asynchronous data exchange. Calls in the AttachLink() family attempt to establish the requested data link
 * with the given calling parameters and complete immediately.  If the address of the requested 
 * server can be resolved and the calling parameters are otherwise okay, the call will complete
 * successfully and return a valid (positive) link handle.  This does not mean that the 
 * requested server will respond to the request.  The data and status information from the server
 * will be returned asynchronously inside the callback routine passed in the call.
 * This call differs from AttachLink() in that it offers an additional parameter, callbackID, whereby
 * the caller can specify which callback identifier he would like to be passed when the callback routine is
 * called.
 * In Visual Basic 3.0, there is no equivalent to this call.
 * In the Win32 world, use T_ATTACHLINK() with Visual Basic.
 * Use AttachLinkEx() with interfaces which do not require the Pascal calling convention (e.g. VC++,
 * LabView, MatLab, etc.).
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". Note that, if devName references a locally registered equipment module,
 * an internal call to ExecLocalLink() is made.  For instance: "/LOCAL/BPMEQM/WL167" sets the context to 
 * "LOCAL", and "BPMEQM" must be a locally registered equipment module. In such a case, the call can only 
 * be successful when the client and server are running together in the same process). 
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server 
 * to the client.  The structure element dout->data should be a pointer to a pre-allocated data buffer,
 * which is to receive the returned data.  If dout is a NULL pointer, it is assumed that no data are to 
 * be returned.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the server from the 
 * client.  If din is a NULL pointer, it is assumed that no data are to be sent to the server
 * \param access is the data access mode.  This can be any combination of access codes ORed together (CA_READ,
 * CA_WRITE, CA_CONNECT, etc.).
 * \param pollrate is the polling interval in milliseconds at which the server is requested to
 * acquire data. This value also signals a timeout if there is no notification within this amount
 * of time plus the system heartbeat (1000 msec for POLLing links).  One can use AttachLink()
 * for asynchronous single READ and WRITE requests by specifying a pollmode of CM_SINGLE.
 * \param callback is a reference to the a callback function which is to be called by the TINE
 * engine when either incoming data have arrived or an error condition has arised.  This should
 * be a pointer to a funtion with prototype \b void \b callback(\b int linkId, \b int statusCode)
 * \param pollmode is the TINE poll mode, i.e. one of CM_CANCEL, CM_SINGLE, CM_REFRESH, CM_POLL,
 * which can be ORed with one of CM_BCAST, CM_NETWORK, CM_MCAST, CM_USE_ON_ERROR, CM_RETRY,
 * CM_GROUPED, CM_CONNECT, CM_WAIT, or CM_SYNCNOTIFY. (see the discussion of Data Access Modes).
 * \param callbackID is a user-specified callback identifier, which will passed to the callback
 * routine when it is called.  Passing a '-1' for this parameter has the result of defaulting to
 * a call to AttachLink() where the returned link identifier is used.
 *
 * \return A positive link handle ID if successful, otherwise the negative of a TINE completion 
 * code which can be interpreted by a call to GetRPCLastError().  If the TINE engine can resolve
 * the address of the requested server and all of the calling parameters are otherwise
 * correct the call should complete successfully.  Possible errors might be 'non existent element'
 * or 'non existent fec', etc.
 *
 * \note This form of the call cannot be used with Visual Basic. Use T_ATTACHLINK() instead.
 *
 * \see AttachLink(), T_ATTACHLINK(), DTYPE, EXECLINK(), EXECLINKEx()
 */
__declspec(dllimport) int AttachLinkEx(char *devName,char *devProperty,
				DTYPE *dout,DTYPE *din,short access,int pollrate,
        void (*callback)(int,int),int pollmode,UINT32 callbackID);

/**
 * \brief Use with \b Visual \b Basic \b 6.0. Establishes an asynchronous link with the given calling parameters.
 * 
 * Asynchronous data exchange. Calls in the AttachLink() family attempt to establish the requested data link
 * with the given calling parameters and complete immediately.  If the address of the requested 
 * server can be resolved and the calling parameters are otherwise okay, the call will complete
 * successfully and return a valid (positive) link handle.  This does not mean that the 
 * requested server will respond to the request.  The data and status information from the server
 * will be returned asynchronously inside the callback routine passed in the call.
 * In Visual Basic 3.0, there is no equivalent to this call.
 * In Visual Basic 6.0, you must declare the callback routine as a Subroutine 
 * (not a Function) inside a .BAS module and make use of the AddressOf routine to pass the
 * address of the this subroutine to T_ATTACHLINK().  This works in Visual Basic only to the
 * extent that no debug break points are set inside the callback routine.
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". Note that, if devName references a locally registered equipment module,
 * an internal call to ExecLocalLink() is made.  For instance: "/LOCAL/BPMEQM/WL167" sets the context to 
 * "LOCAL", and "BPMEQM" must be a locally registered equipment module. In such a case, the call can only 
 * be successful when the client and server are running together in the same process). 
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server 
 * to the client.  The structure element dout->data should be a pointer to a pre-allocated data buffer,
 * which is to receive the returned data.  If dout is a NULL pointer, it is assumed that no data are to 
 * be returned.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the server from the 
 * client.  If din is a NULL pointer, it is assumed that no data are to be sent to the server
 * \param access is the data access mode.  This can be any combination of access codes ORed together (CA_READ,
 * CA_WRITE, CA_CONNECT, etc.).
 * \param pollrate is the polling interval in milliseconds at which the server is requested to
 * acquire data. This value also signals a timeout if there is no notification within this amount
 * of time plus the system heartbeat (1000 msec for POLLing links).  One can use AttachLink()
 * for asynchronous single READ and WRITE requests by specifying a pollmode of CM_SINGLE.
 * \param callback is a reference to the a callback function which is to be called by the TINE
 * engine when either incoming data have arrived or an error condition has arised.  This should
 * be a pointer to a funtion with prototype \b Sub \b sinCallback(\b ByVal id \b As \b Long, \b ByVal cc \b As \b Long)
 * \param pollmode is the TINE poll mode, i.e. one of CM_CANCEL, CM_SINGLE, CM_REFRESH, CM_POLL,
 * which can be ORed with one of CM_BCAST, CM_NETWORK, CM_MCAST, CM_USE_ON_ERROR, CM_RETRY,
 * CM_GROUPED, CM_CONNECT, CM_WAIT, or CM_SYNCNOTIFY. (see the discussion of Data Access Modes).
 *
 * \return A positive link handle ID if successful, otherwise the negative of a TINE completion 
 * code which can be interpreted by a call to GetRPCLastError().  If the TINE engine can resolve
 * the address of the requested server and all of the calling parameters are otherwise
 * correct the call should complete successfully.  Possible errors might be 'non existent element'
 * or 'non existent fec', etc.
 *
 * \note Unless you have a good reason to use this routine, it is better to make use of the
 * ACOP Receive Event for asynchronous data links.
 *
 * \see AttachLink(), DTYPE, EXECLINK(), EXECLINKEx()
 *
 * \b Example:
 * 
 * \include eg_TATTACHLINK.txt
 */
__declspec(dllexport) short PASCAL T_ATTACHLINK(char *devName,char *devprop,
				DTYPE *dout,DTYPE *din,short access,int pollrate,
        cbTLink callback,int mode);

__declspec(dllimport) short OPENLINK (HWND hWnd,char *devname,char *devprop,
						    DBLK *dout,DBLK *din,void *datain,
							short access,long timer,short mode);

__declspec(dllimport) short PASCAL RPCERR(short err, char *errstr);

/** 
 * \brief Retrieves the TINE status string for the given completion code for the most
 * recent data link if cc < 0.
 *
 * \b Visual \b Basic :
 * \n Declare Function \b GetRPCLastError Lib "tinecln32.dll" (ByVal cc As Integer, ByVal ErrString As String) As Integer
 *
 * \param cc is the TINE completion code (i.e. status code) for which the string representation 
 * is desired.
 * \param errString is a (pre-allocated!) string buffer into which the most recent completion
 * status will be copied.  This must point to a string buffer with enough room to hold the 
 * completion status (at least 48 characters).
 *
 * It is only necessary to call this routine directly when using the 'Simple TINE Client ActiveX 
 * interface.  If this the routine is called passing a negative number as the first
 * parameter, then
 * this routine should be called immediately following any synchronous link or 
 * asynchronous Receive event.  Used in this manner, the call is clearly not thread-safe, but
 * as Visual Basic builds single-threaded applications, this will
 * most likely give reasonable results. 
 * As a synchronous call will return the completion code directly as a return code, you can
 * safely use this call by passing the returned completion code as the first parameter.
 *
 * \return 0 if successful, otherwise a TINE completion code.
 *
 * \b Example:
 *
 * \include eg_GETRPCDATA.txt
 */
__declspec(dllimport) short PASCAL GetRPCLastError(short cc, char *ErrString);

__declspec(dllimport) short PASCAL RPCGETLASTERROR(char *errString);

/**
 * \brief Non-Pascal version (Win32 Only). Closes an asynchronous link with the given link id.
 * 
 * Asynchronous data exchange. Calls in the AttachLink() family attempt to establish the requested data link
 * with the given calling parameters and complete immediately.  If the address of the requested 
 * server can be resolved and the calling parameters are otherwise okay, the call will complete
 * successfully and return a valid (positive) link handle.  
 * The link will remain active until the client application quits or until it explicitly closes
 * the active link using this call.
 * In Visual Basic 3.0, there is no equivalent to this call.
 * In the Win32 world, use CLOSERPC() with Visual Basic.
 * Use CloseLink() with interfaces which do not require the Pascal calling convention (e.g. VC++,
 * LabView, MatLab, etc.).
 *
 * \param linkId is the link identifier originally returned by a call to AttachLink() or AttachLinkEx().
 *
 * \see AttachLinkEx(), AttachLink()
 */
__declspec(dllimport) void CloseLink(int linkId);
__declspec(dllimport) short CLOSELINK (HWND hWnd, short i);
__declspec(dllimport) int GetNetGlobal(char *keyword, void *buf, short size, short format);

/**
 * \brief Use with \b Visual \b Basic \b 3.0 or \b 6.0. Executes a synchronous link. 
 * 
 * Synchronous data exchange. Calls in the ExecLink() family do not complete until the data transfer has completed or 
 * a timeout has ensued. EXECLINK() differs from EXECLINKEx() in that the default timeout 
 * of 1000 msec is assumed.
 * Use ExecLink() with interfaces which do not require the Pascal calling convention.
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". Note that, if devName references a locally registered equipment module,
 * an internal call to ExecLocalLink() is made.  For instance: "/LOCAL/BPMEQM/WL167" sets the context to 
 * "LOCAL", and "BPMEQM" must be a locally registered equipment module. In such a case, the call can only 
 * be successful when the client and server are running together in the same process). 
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a reference to the output data block, that is, the data set to be returned from the server 
 * to the client.  This parameter must be present and should specifiy a format data type of
 * CF_NULL and data size of 0 if no return data are requested.
 * \param dataout is a reference to the array into which the server's output data should be copied.
 * This must be consistent with the data block parameters passed in the dout parameter,
 * otherwise the call will fail.
 * \param din is a pointer to the input data block, that is, the data set to be sent to the server from the 
 * client.  This parameter is required and should contain a format data type of CF_NULL and
 * data size of 0 if no data are to be sent to the server.
 * \param datain is a reference to the array from which the input data should be copied prior to
 * sending to the server.  If a 0 is passed as this parameter, no data are submitted to the
 * server.  This must be consistent with the data block parameters passed in the din parameter,
 * otherwise the call will fail.
 * \param access is the data access mode.  This can be any set of access codes ORed together (CA_READ,
 * CA_WRITE, CA_CONNECT, etc.)
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see EXECLINKEx(), DBLK  
 *
 * \b Example:
 *
 * \include eg_EXECLINK.txt
 */
__declspec(dllimport) short PASCAL EXECLINK (char *devName,char *devProperty,DBLK *dout,void *dataout,DBLK *din,void *datain,short access);

/**
 * \brief Use with \b Visual \b Basic \b 3.0 or \b 6.0. Extended call. Executes a synchronous link with timeout parameter.
 * 
 * Synchronous data exchange. Calls in the ExecLink() family do not complete until the data transfer has completed or 
 * a timeout has ensued. EXECLINK() differs from EXECLINKEx() in that the default timeout 
 * of 1000 msec is assumed.
 * Use ExecLinkEx() with interfaces which do not require the Pascal calling convention.
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". Note that, if devName references a locally registered equipment module,
 * an internal call to ExecLocalLink() is made.  For instance: "/LOCAL/BPMEQM/WL167" sets the context to 
 * "LOCAL", and "BPMEQM" must be a locally registered equipment module. In such a case, the call can only 
 * be successful when the client and server are running together in the same process). 
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a reference to the output data block, that is, the data set to be returned from the server 
 * to the client.  This parameter must be present and should specifiy a format data type of
 * CF_NULL and data size of 0 if no return data are requested.
 * \param dataout is a reference to the array into which the server's output data should be copied.
 * This must be consistent with the data block parameters passed in the dout parameter,
 * otherwise the call will fail.
 * \param din is a pointer to the input data block, that is, the data set to be sent to the server from the 
 * client.  This parameter is required and should contain a format data type of CF_NULL and
 * data size of 0 if no data are to be sent to the server.
 * \param datain is a reference to the array from which the input data should be copied prior to
 * sending to the server.  If a 0 is passed as this parameter, no data are submitted to the
 * server.  This must be consistent with the data block parameters passed in the din parameter,
 * otherwise the call will fail.
 * \param access is the data access mode.  This can be any set of access codes ORed together (CA_READ,
 * CA_WRITE, CA_CONNECT, etc.)
 * \param timeout the timeout period in milliseconds for the call to complete.  When it is known a priori 
 * that a call takes a long time to complete than you can set this value accordingly
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see EXECLINK(), DBLK, ExecLinkEx()
 *
 * \b Alias: T_EXECLINK()
 *
 * \b Example:
 *
 * \include eg_EXECLINKEx.txt
 */
__declspec(dllimport) short PASCAL EXECLINKEx (char *devName,char *devProperty,DBLK *dout,void *dataout,DBLK *din,void *datain,short access,long timeout);

__declspec(dllimport) short PASCAL T_EXECLINK (char *devName,char *devProperty,DBLK *dout,void *dataout,DBLK *din,void *datain,short access,long timeout);

/**
 * \brief Non-Pascal version (Win32 Only). Executes a synchronous link with timeout parameter.
 * 
 * Synchronous data exchange. Calls in the ExecLink() family do not complete until the data transfer has completed or 
 * a timeout has ensued. In the Win32 world, use EXECLINK() or T_EXECLINK() with Visual Basic.
 * Use ExecLinkEx() with interfaces which do not require the Pascal calling convention (e.g. VC++,
 * LabView, MatLab, etc.). 
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". Note that, if devName references a locally registered equipment module,
 * an internal call to ExecLocalLink() is made.  For instance: "/LOCAL/BPMEQM/WL167" sets the context to 
 * "LOCAL", and "BPMEQM" must be a locally registered equipment module. In such a case, the call can only 
 * be successful when the client and server are running together in the same process). 
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server 
 * to the client.  The structure element dout->data should be a pointer to a pre-allocated data buffer,
 * which is to receive the returned data.  If dout is a NULL pointer, it is assumed that no data are to 
 * be returned.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the server from the 
 * client.  If din is a NULL pointer, it is assumed that no data are to be sent to the server
 * \param access is the data access mode.  This can be any combination of access codes ORed together (CA_READ,
 * CA_WRITE, CA_CONNECT, etc.).
 * \param timeout the timeout period in milliseconds for the call to complete.  When it is known a priori 
 * that a call takes a long time to complete than you can set this value accordingly
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \note This form of the call cannot be used with Visual Basic.  ExecLinkEx() has the same prototype
 * as the console API call, namely it makes use of the DTYPE object (which is cumbersome to use in
 * Visual Basic) instead of the DBLK object + data reference used in EXECLINK().
 *
 * \see ExecLink(), DTYPE
 */
__declspec(dllimport) short ExecLinkEx(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access,long pollrate);

/**
 * \brief Non-Pascal version (Win32 Only). Executes a synchronous link.
 * 
 * Synchronous data exchange. Calls in the ExecLink() family do not complete until the data transfer has completed or 
 * a timeout has ensued. In the Win32 world, use EXECLINK() or T_EXECLINK() with Visual Basic.
 * Use ExecLinkEx() with interfaces which do not require the Pascal calling convention (e.g. VC++,
 * LabView, MatLab, etc.). 
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". Note that, if devName references a locally registered equipment module,
 * an internal call to ExecLocalLink() is made.  For instance: "/LOCAL/BPMEQM/WL167" sets the context to 
 * "LOCAL", and "BPMEQM" must be a locally registered equipment module. In such a case, the call can only 
 * be successful when the client and server are running together in the same process). 
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server 
 * to the client.  The structure element dout->data should be a pointer to a pre-allocated data buffer,
 * which is to receive the returned data.  If dout is a NULL pointer, it is assumed that no data are to 
 * be returned.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the server from the 
 * client.  If din is a NULL pointer, it is assumed that no data are to be sent to the server
 * \param access is the data access mode.  This can be any combination of access codes ORed together (CA_READ,
 * CA_WRITE, CA_CONNECT, etc.).
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \note This form of the call cannot be used with Visual Basic.  ExecLink() has the same prototype
 * as the console API call, namely it makes use of the DTYPE object (which is cumbersome to use in
 * Visual Basic) instead of the DBLK object + data reference used in EXECLINK().
 *
 * \see ExecLinkEx(), DTYPE
 */
__declspec(dllimport) short ExecLink(char *devname,char *devprop,DTYPE *dout,DTYPE *din,short access);
__declspec(dllimport) short PASCAL GetDevicePropertyInformation(char *fullDeviceName,char *property,PropertyQueryStructEx *pqsx,int *bufsiz);
__declspec(dllimport) short PASCAL FILLSTRUCT (char *tag, short addr, short size, short fmt);
__declspec(dllimport) short PASCAL SetNotificationTolerance(int i, float toleranceAbsolute, float tolerancePercent);

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif // _TINECLN32_H_

