#include "project.def"
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <malloc.h>
#include <time.h>
#include <winsock.h>
#include "errors.h"
#include "database.h"
#include "winsrv.h"
#include "almsrv.h"

extern char txtbuf[128];
extern int debug;
extern int isHistoryCall;
HANDLE hAlmList = 0;
ALARM *almList = NULL;
int gRequireReadFromCAS = FALSE;
int AlarmServerInitialized = 0;
int numStoredAlarms = 0;
int AlarmWindow = DEFAULT_ALARM_WINDOW;
int AlarmHeartbeat = DEFAULT_ALARM_HEARTBEAT;
int MaxNumberAlarms = MAXALARMS;
int AlarmInstantWindow = DEFAULT_INSTANT_WINDOW;
int TerminationWindow = DEFAULT_TERMINATION_WINDOW;
ExportListStruct *findTag(char *name);

#define NEW_ALARM_WINDOW 3

ALARM_WATCH *almWatchList = NULL;
int nAlmWatchList = 0;

#define idx_too_high      0
#define idx_warn_too_high 1
#define idx_too_low       2
#define idx_warn_too_low  3
#define idx_invalid_data  4

ADS almWatchTableADS[] = 
{
  { "Value Too HIGH",value_too_high,0,14,BFMT(CF_FLOAT),1,0,0,0,"measured value is above threshold","",0,"last measured value","",NULL },
  { "Value Too High",warn_too_high,0,12,BFMT(CF_FLOAT),1,0,0,0,"measured value is above warn threshold","",0,"last measured value","",NULL },
  { "Value Too LOW",value_too_low,0,14,BFMT(CF_FLOAT),1,0,0,0,"measured value is below threshold","",0,"last measured value","",NULL },
  { "Value Too Low",warn_too_low,0,12,BFMT(CF_FLOAT),1,0,0,0,"measured value is below warn threshold","",0,"last measured value","",NULL },
  { "Invalid Data",invalid_data,0,12,BFMT(CF_LONG),1,0,0,0,"measured value not at set point","",0,"last measured value","",NULL }
};
int csvAlmWatchFileGetFormat(char *fmtstr);
CSVCOLUMN csvAlmWatchFileColumns[] = 
{
  { "LOCAL_NAME","",0,CF_TEXT,6,(BYTE *)OFFSETIN(ALARM_WATCH,eqm),NULL },
  { "DEVICE_NAME","",0,CF_TEXT,32,(BYTE *)OFFSETIN(ALARM_WATCH,dev),NULL },
  { "PROPERTY","",0,CF_TEXT,32,(BYTE *)OFFSETIN(ALARM_WATCH,prp),NULL },
  { "SIZE","",0,CF_LONG,1,(BYTE *)OFFSETIN(ALARM_WATCH,siz),NULL },
  { "FORMAT","",0,CF_NULL,1,NULL,csvAlmWatchFileGetFormat },
  { "SEVERITY","",0,CF_LONG,1,(BYTE *)OFFSETIN(ALARM_WATCH,sev),NULL },
  { "ALARM_SYSTEM","0",0,CF_LONG,1,(BYTE *)OFFSETIN(ALARM_WATCH,asys),NULL },
  { "MASK","0",0,CF_LONG,1,(BYTE *)OFFSETIN(ALARM_WATCH,mask),NULL },
  { "NORMAL","0",0,CF_LONG,1,(BYTE *)OFFSETIN(ALARM_WATCH,normal),NULL},
  { "COUNT_THRESHOLD","0",0,CF_LONG,1,(BYTE *)OFFSETIN(ALARM_WATCH,cntThreshold),NULL },
  { "HIGH","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(ALARM_WATCH,hi),NULL },
  { "LOW","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(ALARM_WATCH,lo),NULL },
  { "HIGHWARN","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(ALARM_WATCH,hiwarn),NULL },
  { "LOWWARN","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(ALARM_WATCH,lowarn),NULL }
};
CSVDB csvAlmWatchFileDb =
{
  csvAlmWatchFileColumns,
  sizeof(csvAlmWatchFileColumns)/sizeof(csvAlmWatchFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(ALARM_WATCH)
};
int csvAlmWatchFileGetFormat(char *fmtstr)
{
  almWatchList[csvAlmWatchFileDb.cnt].fmt = BFMT(ftoi(fmtstr));
  almWatchList[csvAlmWatchFileDb.cnt].atyp = getArrayType(fmtstr);
  // set the next pointer :
  if (csvAlmWatchFileDb.cnt < csvAlmWatchFileDb.siz - 1)
    almWatchList[csvAlmWatchFileDb.cnt].nxt = &almWatchList[csvAlmWatchFileDb.cnt+1];
  return 0;
}
int GetAlarmWatchTable(void)
{
  int i,cc;
  char str[256];
  char almfn[64];

  strncpy(almfn,ALMARM_WATCH_FILE,64);
  if (strlen(gFecDBSubDir) > 0) sprintf(almfn,"%s%s",gFecDBSubDir,ALMARM_WATCH_FILE);
  if ((cc=csvReadFile(FecDBpath,almfn,&csvAlmWatchFileDb,(void **)&almWatchList)) != 0) ccerr(cc);
  nAlmWatchList = csvAlmWatchFileDb.siz;
  for (i=0; i<nAlmWatchList; i++)
  {
    almWatchList[i].data = (BYTE *)calloc(almWatchList[i].siz,fmtsizeof(almWatchList[i].fmt));
    if (almWatchList[i].data == NULL) ccerr(out_of_server_memory);
    if (almWatchList[i].mask != 0) almWatchList[i].chkNormal = TRUE;
  }
err:
  sprintf(str,"ALM: watch file %s : %s",almfn, cc ? erlst[cc] : "found");
  feclog(str);
  return cc;
}
__declspec(dllexport) short PASCAL appendAlarmWatchTable(char *eqm,char *prp,char *dev,int siz,int fmt,int atyp,int sev,int sys,ALARM_WATCH_THRESHOLDS *thr)
{
  short cc = 0;
  ALARM_WATCH *aw;

  if (eqm == NULL || prp == NULL || dev == NULL) ccerr(argument_list_error);
  if ((aw=(ALARM_WATCH *)calloc(1,sizeof(ALARM_WATCH))) == NULL) ccerr(out_of_server_memory);
  strncpy(aw->eqm,eqm,EQP_NAME_SIZE);
  strncpy(aw->prp,eqm,EQP_PROP_SIZE);
  strncpy(aw->dev,dev,32);
  aw->siz = siz;
  aw->fmt = fmt;
  aw->atyp = atyp;
  aw->sev = sev;
  aw->asys = sys;
  aw->cntThreshold = thr->cnt;
  if (thr->mask != 0)
  {
    aw->mask = thr->mask;
    aw->normal = thr->normal;
    aw->chkNormal = TRUE;
  }
  aw->hi = thr->hi;
  aw->hiwarn = thr->hiwarn;
  aw->lo = thr->lo;
  aw->lowarn = thr->lowarn;
  aw->nxt = almWatchList;
  almWatchList = aw;
  nAlmWatchList++;
err:
  return cc;
}
void checkAlarmsInWatchTable(void)
{
  ALARM_WATCH *aw;
  int lid,i,j,cnt;
  short devNr;
  ExportListStruct *el;
  float fval;
  UINT32 ival;
  static BYTE almdata[6];

  //for (i=0; i<nAlmWatchList; i++)
  for (i=0, aw=almWatchList; aw != NULL; i++, aw = aw->nxt)
  {
    //aw = &almWatchList[i];
    if ((el=findTagID(aw->eqm)) == NULL) continue;
    if (el->hEqpFcn == NULL) continue;
    lid = (long)MAKEPROTID(i,ALM_PROTOCOL);
    strcpy(clientName,"LOCAL ALARM"); strcpy(clientaddr,"THIS STATION");
    if (aw->interimCounter == 0) aw->interimStatus = not_posted;
    if (aw->interimStatus == not_posted)
    {
      interimStatus = not_posted;  // let the VB app. clear this
      isHistoryCall = TRUE;
      SendMessage(el->hEqpFcn,WM_SYSCOMMAND,lid,ALM_PROTOCOL);
      isHistoryCall = FALSE;
    }
    if (aw->interimStatus != 0) 
    {
      if (aw->interimStatus == not_posted || aw->interimStatus == not_ready)
      {
        aw->interimStatus = not_ready;
        if (aw->interimCounter++ > 10) aw->interimCounter = 0;
      }
      continue;
    }
    devNr = GetDeviceNumber(aw->eqm,aw->dev);
    cnt = 0;
    for (j=0; j<aw->siz; j++)
    {
      if (aw->atyp == AT_CHANNEL) devNr = j;
      if (aw->atyp != AT_WAVEFORM && aw->siz == el->EqpNumModules) devNr = j; /* assume array of devices */
      switch (LFMT(aw->fmt))
      {
        case CF_BYTE: 
          ival = (UINT32)((BYTE *)aw->data)[j]; 
          fval = (float)((BYTE *)aw->data)[j]; 
          break;
        case CF_SHORT:
          ival = (UINT32)((short *)aw->data)[j]; 
          fval = (float)((short *)aw->data)[j]; 
          break;
        case CF_LONG: 
          ival = (UINT32)((int *)aw->data)[j]; 
          fval = (float)((int *)aw->data)[j]; 
          break;
        case CF_FLOAT:
          ival = (UINT32)((float *)aw->data)[j]; 
          fval = (float)((float *)aw->data)[j]; 
          break;
        case CF_DOUBLE:
          ival = (UINT32)((double *)aw->data)[j]; 
          fval = (float)((double *)aw->data)[j]; 
          break;
        default: fval = 0;
      }
      if (aw->chkNormal)
      {
        if (aw->normal != (ival & aw->mask))
        {
          cnt++;
          memcpy(almdata,&ival,4);
          aw->cnt++;
          almWatchTableADS[idx_invalid_data].alarmSystem = aw->asys;
          almWatchTableADS[idx_invalid_data].severity = aw->sev;
          if (aw->cnt > aw->cntThreshold) SetAlarm(aw->eqm,devNr,invalid_data,almdata);
        }
      }
      else if (fval > aw->hiwarn) 
      {
        cnt++;
        memcpy(almdata,&fval,4);
        aw->cnt++;
        almWatchTableADS[idx_too_high].alarmSystem = aw->asys;
        almWatchTableADS[idx_too_high].severity = aw->sev;
        almWatchTableADS[idx_warn_too_high].alarmSystem = aw->asys;
        almWatchTableADS[idx_warn_too_high].severity = aw->sev - 2;
        if (aw->cnt > aw->cntThreshold) 
          SetAlarm(aw->eqm,devNr,fval > aw->hi ? value_too_high : warn_too_high,almdata);
      }
      else if (fval < aw->lowarn) 
      {
        cnt++;
        memcpy(almdata,&fval,4);
        aw->cnt++;
        almWatchTableADS[idx_too_low].alarmSystem = aw->asys;
        almWatchTableADS[idx_too_low].severity = aw->sev;
        almWatchTableADS[idx_warn_too_low].alarmSystem = aw->asys;
        almWatchTableADS[idx_warn_too_low].severity = aw->sev - 2;
        if (aw->cnt > aw->cntThreshold) 
          SetAlarm(aw->eqm,devNr,fval < aw->lo ? value_too_low : warn_too_low,almdata);
      }
      else 
      {
        if (aw->cnt) /* if there is or was an alarm somewhere, run through RemoveAlarm() for the others */
        {
          RemoveAlarm(aw->eqm,devNr,invalid_data);
          RemoveAlarm(aw->eqm,devNr,value_too_high);
          RemoveAlarm(aw->eqm,devNr,value_too_low);
          RemoveAlarm(aw->eqm,devNr,warn_too_high);
          RemoveAlarm(aw->eqm,devNr,warn_too_low);
        }
      }
    }
    /* only clear the watch table counter when a complete pass has no alarms */
    if (cnt == 0) aw->cnt = 0;
  }
  return;
}

__declspec(dllexport) short PASCAL SetMaximumAlarms(short MaxAlarms)
{
  if (MaxAlarms < 0) return out_of_range;
  MaxNumberAlarms = MaxAlarms;
  return 0;
}

ExportListStruct *getExportFromList(char *eqm)
{
  ExportListStruct *export;

  for (export = ExportList; export; export = export->next)
    if (!strncmp(eqm,export->EqpName,6)) break;
  return export;
}
void getAlarmInfo(void)
{
  ExportListStruct *export;

  for (export = ExportList; export; export = export->next)
    getAlarmInfoTable(export->EqpName);
}
__declspec(dllexport) short PASCAL appendAlarmInfoTable(char *eqm,ADS *ads)
{
  ADS *g_ads,*adsTable = NULL;
  ExportListStruct *export;
  int cc = 0;

  if (eqm == NULL || ads == NULL) ccerr(argument_list_error);
  if ((export=getExportFromList(eqm)) == NULL) ccerr(non_existent_elem);
  if ((g_ads=(ADS *)calloc(1,sizeof(ADS))) == NULL) ccerr(out_of_server_memory);
  memcpy(g_ads,ads,sizeof(ADS));
  g_ads->next = export->adsTable;
  export->adsTable  = g_ads;
  if (!export->alarmSystem) export->alarmSystem = 1;
err:
  return cc;
}
FILE *alarmFileOpen(char *eqm)
{
  char almFile[64];
  int i;
  FILE *fp=NULL;

  for (i=0; i<2; i++)
  {
    memset(almFile,0,sizeof(almFile));
    strncpy(almFile,eqm,6);
    if (strlen(gFecDBSubDir) > 0) sprintf(almFile,"%s%s",gFecDBSubDir,eqm);
    strcat(almFile,i == 0 ? "-alarms.csv" : "al.csv");
    FecPath[0] = 0; strcat(FecPath,FecDBpath); strcat(FecPath,almFile);
    if ((fp=fopen(FecPath,"r")) != NULL) return fp;
  }
  return NULL;
}
int getAlarmInfoTable(char *eqname)
{
  ADS *ads,*adsTable = NULL;
  FILE *fp = NULL;
  ExportListStruct *export;
  int cc=0,done=0,fmt;
  short alarmTag_col,alarmCode_col,alarmMask_col,severity_col,
        format_col,size_col,timeRes_col,hitsChannel_col,
        hitsModule_col,alarmText_col,deviceText_col,dataText_col,
        url_col,alarmSystem_col;
  char str[256],scratch[16];

  if ((fp=alarmFileOpen(eqname)) == NULL) ccerr(no_such_file);
  while (fgets(str,255,fp) != NULL)   /* scan file       */
  {
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((alarmTag_col=findcol(str,"ALARM_TAG")) < 0) ccerr(no_such_line);
      if ((alarmCode_col=findcol(str,"ALARM_CODE")) < 0) ccerr(no_such_line);
      alarmMask_col = findcol(str,"ALARM_MASK");
      if ((severity_col=findcol(str,"SEVERITY")) < 0) ccerr(no_such_line);
      format_col = findcol(str,"DATA_FORMAT");
      size_col = findcol(str,"DATA_ARRAYSIZE");
      timeRes_col = findcol(str,"TIME_RESOLUTION");
      hitsChannel_col = findcol(str,"HITS_CHANNEL");
      hitsModule_col = findcol(str,"HITS_MODULE");
      alarmText_col = findcol(str,"ALARM_TEXT");
      deviceText_col = findcol(str,"DEVICE_TEXT");
      dataText_col = findcol(str,"DATA_TEXT");
      url_col = findcol(str,"URL");
      if ((alarmSystem_col=findcol(str,"ALARM_SYSTEM")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    if ((ads=(ADS *)calloc(1,sizeof(ADS))) == NULL) ccerr(out_of_server_memory);
    strncpy(ads->alarmTag,colptr(alarmTag_col,str),16);
    if (strlen(ads->alarmTag) == 0) strcpy(ads->alarmTag,"undefined");
    strncpy(scratch,colptr(alarmCode_col,str),16);
    ads->alarmCode = atoi(scratch);
    if (alarmMask_col < 0) ads->alarmMask = 0;
    else
    {
      strncpy(scratch,colptr(alarmMask_col,str),16);
      ads->alarmMask = atoi(scratch);
    }
    strncpy(scratch,colptr(severity_col,str),16);
    ads->severity = atoi(scratch);
    if (format_col < 0) fmt = CF_NULL;
    else
    {
      strncpy(scratch,colptr(format_col,str),16);
      fmt = ftoi(scratch);
    }
    ads->alarmDataFormat = BFMT(fmt);
    if (size_col < 0) ads->alarmDataArraySize = 0;
    else
    {
      strncpy(scratch,colptr(size_col,str),16);
      ads->alarmDataArraySize = MIN(atoi(scratch),6);
    }
    if (timeRes_col < 0) ads->timeRes = 1;
    else
    {
      strncpy(scratch,colptr(timeRes_col,str),16);
      ads->timeRes = (float)atof(scratch);
    }
    if (hitsChannel_col < 0) ads->hitsChannel = 0;
    else
    {
      strncpy(scratch,colptr(hitsChannel_col,str),16);
      ads->hitsChannel = atoi(scratch);
    }
    if (hitsModule_col < 0) ads->hitsModule = 0;
    else
    {
      strncpy(scratch,colptr(hitsModule_col,str),16);
      ads->hitsModule = atoi(scratch);
    }
    if (alarmText_col >= 0)
    {
      strncpy(ads->alarmText,colptr(alarmText_col,str),40);
    }
    if (deviceText_col >= 0)
    {
      strncpy(ads->deviceText,colptr(deviceText_col,str),38);
    }
    if (dataText_col >= 0)
    {
      strncpy(ads->dataText,colptr(dataText_col,str),40);
    }
    if (url_col >= 0)
    {
      strncpy(ads->url,colptr(url_col,str),40);
    }
    strncpy(scratch,colptr(alarmSystem_col,str),16);
    ads->alarmSystem = atoi(scratch);
    ads->next = adsTable;
    adsTable  = ads;
    cc |= ads->alarmSystem ? 1 : 0;
  }
  fclose(fp); fp = NULL;
  if ((export=getExportFromList(eqname)) != NULL)
  {
    export->adsTable = adsTable;
    export->alarmSystem = cc;
  }
  return cc;
err: if (fp) fclose(fp);
  return -cc;
}
static char gCASName[16] = "";
CSVCOLUMN csvCASNameColumns[] =
{
  { "CAS","",0,CF_TEXT,16,(BYTE *)gCASName,NULL,0 }
};
CSVDB csvCASNameDb =
{
  csvCASNameColumns,
  sizeof(csvCASNameColumns)/sizeof(csvCASNameColumns[0]),
  0,
  NULL,
  NULL,
  NULL,
  0,
  NULL,
  0
};
char *getCASUserName(void)
{
  int cc = 0;
  FILE *fp = NULL;

  if ((cc=csvReadFile(FecDBpath,CASHOST_FILE,&csvCASNameDb,NULL)) != 0) ccerr(cc);
  sprintf(txtbuf,"ALM: recognizing %s as Central Alarm Server",gCASName);
  feclog(txtbuf);
err:
  return gCASName;
}
int isCASInUserList(void)
{
  NAME16 usr;
  if (strlen(gCASName) == 0) return 0;
  GetCallerInfo(usr.name,NULL);
  return strncmp(gCASName,usr.name,16) == 0 ? -1 : 0;
}

void enforceAlarm(short cc)
{
 /* method to make sure this alarm goes out */
  return;
}
short initAlarmServer(void)
{
 /* initialization ? */
  getCASUserName();
  GetAlarmWatchTable();
  AlarmServerInitialized = 1;
  return 0;
}
short verifyDeviceNr(ExportListStruct *export,short devNr)
{
 /* verify that the device number passed in the API is valid */
  if (devNr >= export->EqpNumModules || devNr < 0) return -1;
  return 0;
}
int findDeviceNr(char *name,char *devname)
{
  int i;
  ExportListStruct *el;

  if (name == NULL || devname == NULL) return -argument_list_error;
  if ((el=findTagID(name)) == NULL) return -non_existent_elem;
  if (el->EqpModuleName == NULL) return -un_allocated;

  for (i=0; i<el->EqpNumModules; i++)
  {
    if (!strncmp(devname,el->EqpModuleName[i].name,16)) return i;
  }
  return -device_not_connected;
}
__declspec(dllexport) short SetDeviceAlarm(char *name,char *devname,long almCode,BYTE *almData,BYTE flags)
{
  short devNr = findDeviceNr(name,devname);
  return SetAlarmEx(name,devNr,almCode,almData,flags);
}

__declspec(dllexport) short SetAlarm(char *name,short devNr,long almCode,BYTE *almData)
{
  return SetAlarmEx(name,devNr,almCode,almData,almNEWALARM);
}

#define RefreshMRT(t,e) {\
  if ((t) == (e)->almMRT) (e)->nAlmMRT++;\
  else {\
	 (e)->almMRCT = (e)->almMRT;\
	 (e)->nAlmMRCT = (e)->nAlmMRT;\
	 (e)->almMRT = t; (e)->nAlmMRT = 1;\
  }\
}
#define RefreshHSV(s,e) {\
  if ((s) == (e)->almHSV) (e)->nAlmHSV++;\
  else if ((s) > (e)->almHSV)\
  { (e)->almHSV = s; (e)->nAlmHSV = 1; }\
}

__declspec(dllexport) short SetAlarmEx(char *name,short devNr,long almCode,BYTE *almData,BYTE flags)
{
  short adatasize = 0,cc = 0;
  ALARM *alm = NULL;
  ALARM *lsa = NULL;  /* least significant alarm */
  short lsc,sc = findSeverity(name,almCode),tmpsc = sc;
  ADS *ads = NULL;
  ExportListStruct *el;
  time_t almTime = time(NULL);

  // fill in ALARMStruct for the tag via link list:

  /* Instant alarms pickup the TERMINATE signal immediately */

  if (flags & almINSTANT) flags |= almTERMINATE;

  /* fill in ALARMStruct for the tag via link list: */

  if ((el=findTagID(name)) == NULL) ccerr(0);
  if (verifyDeviceNr(el,devNr)) ccerr(out_of_range);
  if ((ads=getAlarmTable(name,almCode)) != NULL && almData != NULL)
  {
    adatasize = ads->alarmDataArraySize*getFormatSize(LFMT(ads->alarmDataFormat));
    if (adatasize > 6) adatasize = 6;
  }
  for (alm=el->almLst[devNr]; alm != NULL; alm=alm->next)
  {
    if ((lsc=findSeverity(name,alm->alarmCode)) < tmpsc) { lsa = alm; tmpsc = lsc; }
    if (almCode == alm->alarmCode &&          /* it's in the list */
        !(alm->alarmStatus & almTERMINATE))   /* and not marked for termination */
    {
      if (alm->alarmStatus & almNEWALARM && 
          almTime <= (time_t)alm->timestamp + NEW_ALARM_WINDOW) /* a double SetAlarm() ? */
      {
        alm->clrCount = 0;                     /* reset clear count      */
        return 0;
      }
      if (almData != NULL && adatasize > 0 && memcmp(almData,alm->alarmData,adatasize))
      {                                        /* alarm data have changed */
        alm->alarmStatus &= ~((BYTE)almNEWALARM); /* remove NEWALARM */
        alm->alarmStatus |= almDATACHANGE;
        alm->timestamp = almTime;
        RefreshMRT(alm->timestamp,el);
        memcpy(alm->alarmData,almData,adatasize);      /* copy changed data !! */
        if (debug)
        {
          wsprintf(txtbuf,"ALM: data change set for %s <%d>\n>",el->tagName,alm->alarmCode);
          output(txtbuf);
        }
      }
      else
      {                                        /* alarm has been reset */
        if (alm->clrCount > 1)
        {
          alm->alarmStatus &= ~((BYTE)almNEWALARM); /* remove NEWALARM */
          alm->alarmStatus |= almOSCILLATION;  /* appears to be oscillating */
          alm->timestamp = almTime;
          RefreshMRT(alm->timestamp,el);
          if (debug)
          {
            wsprintf(txtbuf,"ALM: alarm oscillation set for %s <%d>\n>",el->tagName,alm->alarmCode);
            output(txtbuf);
          }
        }
      }
      alm->clrCount = 0;                     /* reset clear count      */
      return 0;                              /* already in list, so return */
    }
  }
  
  flags |= almNEWALARM;                        /* alarm not in list */

  if (el->nalarms < MaxNumberAlarms)          /* still room to put the alarm in list */
  {
    if ((alm=(ALARM *)calloc(1,sizeof(ALARM))) == NULL) ccerr(out_of_server_memory);
    if (gRequireReadFromCAS) alm->removalBlocked = TRUE; /* must be cleared by CAS readout */
    alm->timestamp = almTime;
    alm->alarmCode = almCode;
    alm->alarmStatus |= flags;
    if (adatasize > 0) memcpy(alm->alarmData,almData,adatasize);
    alm->next = (ALARM *)el->almLst[devNr];
    el->almLst[devNr] = alm;
    el->nalarms++;
    RefreshMRT(alm->timestamp,el);
    RefreshHSV(sc,el);
    if (debug)
    {
      wsprintf(txtbuf,"ALM: alarm set for %s <%d>\n>",el->tagName,alm->alarmCode);
      output(txtbuf);
    }
    return 0;
  }
  else if (lsa != NULL)  /* replace the least significant alarm */
  {
    alm = lsa;
    if (debug)
    {
      wsprintf(txtbuf,"ALM: replace alarm %s (%ld) with %s (%ld)",
                el->EqpName,alm->alarmCode,el->EqpName,almCode);
      output(txtbuf);
    }
    alm->timestamp = almTime;
    alm->alarmCode = almCode;
    alm->alarmStatus = flags;
    if (adatasize > 0)
    {
      memset(alm->alarmData,0,6);
      memcpy(alm->alarmData,almData,adatasize);
    }
    RefreshMRT(alm->timestamp,el);
    RefreshHSV(sc,el);
  }
  ccerr(max_alarms_exceeded);
err:
  if (cc) enforceAlarm(cc);
  switch (cc)
  {
    case 0: return 0;
    case out_of_server_memory:
    case max_alarms_exceeded:
      enforceAlarm(cc);
      break;
    default: if (alm) free(alm);
      break;
  }
  return 0;
}
__declspec(dllexport) void ClearDeviceAlarm(char *name,char *devname)
{
  short devNr = -1;
  if (strcmp(devname,"ALL")) devNr = findDeviceNr(name,devname);
  ClearAlarm(name,devNr);
}

__declspec(dllexport) void ClearAlarm(char *name,short devNr)
{
  ALARM *alm;
  ExportListStruct *el;
  int start,stop,i;

  if (!AlarmServerInitialized) return;

  if ((el=findTagID(name)) == NULL) return;
  if (devNr == -1)                        /* clear them all */
  {
    start = 0; stop = el->EqpNumModules;
  }
  else
  {
    if (verifyDeviceNr(el,devNr)) return;
    start = devNr; stop = devNr + 1;
  }
  for (i=start; i<stop; i++)
  {
     for (alm=el->almLst[i]; alm != NULL; alm=alm->next)
     {
      if (alm->alarmStatus & almINSTANT) alm->clrCount = 8; /* it'll be cleared immediately */
      else alm->clrCount = MIN(alm->clrCount+1,8); /* increment the clear count */
     }
  }
  return;                                 /* wasn't set */
}
__declspec(dllexport) void RemoveDeviceAlarm(char *name,char *devname,long code)  /* clear it immediately */
{
  short devNr = findDeviceNr(name,devname);
  RemoveAlarm(name,devNr,code);
}

__declspec(dllexport) void RemoveAlarm(char *name,short devNr,long code)  /* clear it immediately */
{
  ALARM *alm;
  ExportListStruct *el;

  if (!AlarmServerInitialized) return;
  if ((el=findTagID(name)) == NULL) return;
  if (verifyDeviceNr(el,devNr)) return;

  for (alm=el->almLst[devNr]; alm != NULL; alm=alm->next)
  {
    if (code == alm->alarmCode)
    {
      alm->clrCount = 8;
    }
  }
  return;                                 /* wasn't set */
}

// and in the background...

ALARM *RemoveAlarmFromList(ExportListStruct *el, int n, ALARM *alm)
{
  ALARM *a0, *a1, *atmp = alm;
  short sc;

  if (el == NULL) return alm->next;
  if (n >= el->EqpNumModules) return alm->next;
  for (a1=a0=el->almLst[n]; a1 != NULL; a0=a1,a1=a1->next)
  {
    if (a1 == alm) /* found it */
    {
      if (a1 == el->almLst[n]) el->almLst[n] = a1->next;
      else a0->next = a1->next;
      atmp = a1->next;
      if (a1->timestamp == el->almMRT) el->nAlmMRT--;
	    if ((sc=findSeverity(el->EqpName,a1->alarmCode)) == el->almHSV) el->nAlmHSV--;
      free(a1); a1 = NULL;
      el->nalarms--;
      break;
    }
  }
  return atmp;    /* return the next link in the list */
}

void SurveyAlarms(void)
{
  int n;
  ALARM *alm;
  ExportListStruct *el;
  time_t t = time(NULL);
  static time_t interim = 0;

  if (!AlarmServerInitialized) return;
  if (t == interim) return;
  interim = t;

  if (nAlmWatchList > 0) checkAlarmsInWatchTable();
  for (el=ExportList; el != NULL; el=el->next)
  {
    if (t > el->almMRT) /* adjust most recent complete timestamp */
    {
      el->almMRCT = el->almMRT;
      el->nAlmMRCT = el->nAlmMRT;
    }
    for (n=0;el->nalarms && n<el->EqpNumModules; n++)
    {
      for (alm=el->almLst[n]; alm != NULL;)
      {
        if ((alm->alarmStatus & almTERMINATE)) /*  time to clear the alarm ? */
        {
          if (alm->removalBlocked) 
          {
            alm = alm->next;
            continue;
          }
          if (alm->timestamp + TerminationWindow < t) /* okay to clear the alarm */
          {
            if (debug)
            {
              wsprintf(txtbuf,"ALM: clearing alarm for %s <%d>\n>",el->tagName,alm->alarmCode);
              output(txtbuf);
            }
            alm = RemoveAlarmFromList(el,n,alm);
            continue;
          }
          else if ((alm->alarmStatus & almINSTANT) && alm->timestamp + AlarmInstantWindow < t)
          {
            if (debug)
            {
              wsprintf(txtbuf,"ALM: Instant clear %s <%ld> (0x%x)",el->EqpName,alm->alarmCode,alm->alarmStatus);
              output(txtbuf);
            }
            alm = RemoveAlarmFromList(el,n,alm);
            continue;
          }
        }
        else                                   /* check alarm status */
        {
          if ((alm->clrCount > 3) && alm->timestamp + AlarmWindow < t)
          {
            if (debug)
            {
              wsprintf(txtbuf,"ALM: termination for %s <%ld> (0x%x)",el->EqpName,alm->alarmCode,alm->alarmStatus);
              output(txtbuf);
            }
            alm->alarmStatus &= ~((BYTE)almNEWALARM); /* remove NEWALARM */
            alm->alarmStatus |= almTERMINATE;
            alm->timestamp = t;
            RefreshMRT(alm->timestamp,el);
          }
          else if (alm->timestamp + AlarmHeartbeat < t) /*  set the heartbeat */
          {
            if (debug)
            {
              wsprintf(txtbuf,"ALM: heartbeat set for %s <%d>\n>",el->tagName,alm->alarmCode);
              output(txtbuf);
            }
            alm->alarmStatus &= ~((BYTE)almNEWALARM); /* remove NEWALARM */
            alm->alarmStatus |= almHEARTBEAT;
            alm->timestamp = t;
            RefreshMRT(alm->timestamp,el);
          }
        }
        alm=alm->next;
      }
    }
  }
  return;
}

// caller subscribes to NALARMS which calls (REFRESH mode)

short numAlarmsEx(char *eqpname,time_t *almTimeStamp,short *almSeverity,short *NrAtAlmTimeStamp,short *NrAtAlmSeverity)
{
  ExportListStruct *el;

  *almTimeStamp = 0; *almSeverity = 0;

  if ((el=findTagID(eqpname)) == NULL) return 0;
  *almTimeStamp = el->almMRCT;
  *almSeverity  = el->almHSV;
  *NrAtAlmTimeStamp = el->nAlmMRCT;
  *NrAtAlmSeverity = el->nAlmHSV;
  return el->nalarms;
}

short numAlarms(char *eqpname,time_t *almTimeStamp,short *almSeverity)
{
  return numAlarmsEx(eqpname,almTimeStamp,almSeverity,NULL,NULL);
}

int numAlarmDefs(char *eqm)
{
  int n;
  ADS *ads;
  ExportListStruct *el;

  if ((el=findTagID(eqm)) == NULL) return 0;
  for (n=0,ads=el->adsTable; ads; ads=ads->next,n++);
  if (nAlmWatchList > 0) n += sizeof(almWatchTableADS)/sizeof(ADS);
  return n;
}
int getAlarmDefs(char *eqm,ADS *rads,int num)
{
  int j,n,nw;
  ADS *ads;
  ExportListStruct *el;

  if ((el=findTagID(eqm)) == NULL) return 0;
  n = numAlarmDefs(eqm);
  for (j=0,ads=el->adsTable; ads && j<num; ads=ads->next,j++)
  {
    memcpy(&rads[j],ads,sizeof(ADS));
  }
  if (nAlmWatchList > 0)
  {
    n = j;
    nw = sizeof(almWatchTableADS)/sizeof(ADS);
    for (j=0; j<nw && j<num-n; j++)
    {
      memcpy(&rads[j+n],&almWatchTableADS[j],sizeof(ADS));
    }
    n += nw;
  }
  return n;
}

short findSeverity(char *eqm, long cc)
{
  ADS *ads;
  ExportListStruct *start,*stop,*el;

  if (eqm == NULL)
  {
    start = ExportList; stop = NULL;
  }
  else
  {
    if ((start=findTagID(eqm)) == NULL) return 0;
    stop = start->next;
  }
  for (el=start; el != stop; el=el->next)
  {
    for (ads=el->adsTable; ads && ads->alarmCode != cc; ads = ads->next);
    if (ads) return ads->severity;
  }
  switch (cc)
  {
    case invalid_data: return almWatchTableADS[idx_invalid_data].severity;
    case value_too_high: return almWatchTableADS[idx_too_high].severity;
    case value_too_low: return almWatchTableADS[idx_too_low].severity;
    case warn_too_high: return almWatchTableADS[idx_warn_too_high].severity;
    case warn_too_low: return almWatchTableADS[idx_warn_too_low].severity;
  }
  return 0;
}

ADS *getAlarmTable(char *eqm,long code)
{
  ADS *ads;
  ExportListStruct *el;

  if ((el=findTagID(eqm)) != NULL)
  {
    for (ads=el->adsTable; ads; ads=ads->next)
    {
      if (code == ads->alarmCode) return ads;
    }
    switch (code)
    {
      case invalid_data: return &almWatchTableADS[idx_invalid_data];
      case value_too_high: return &almWatchTableADS[idx_too_high];
      case value_too_low: return &almWatchTableADS[idx_too_low];
      case warn_too_high: return &almWatchTableADS[idx_warn_too_high];
      case warn_too_low: return &almWatchTableADS[idx_warn_too_low];
    }
  }
  return NULL;
}

short getAlarmsEx(char *eqpname,AMS *ams,short num,short minSeverity,time_t starttime,time_t stoptime,short startnum)
{
  int sc,dv,i=0,n=0;
  ALARM *alm;
  ADS *ads;
  ExportListStruct *el;
  char defaultDeviceName[36];

  if ((el=findTagID(eqpname)) == NULL) return 0;
  for (dv=0; el->nalarms && dv<el->EqpNumModules; dv++)
  {
    for (alm=el->almLst[dv]; alm != NULL; alm=alm->next)
    {
      if (alm->removalBlocked && isCASInUserList()) alm->removalBlocked = FALSE;
      if (alm->timestamp >= starttime && alm->timestamp <= stoptime &&
          (sc=findSeverity(eqpname,alm->alarmCode)) >= minSeverity )
      {
        if (i++ < startnum) continue;
        /* fill in ams[n] here ... */
        strncpy(ams[n].tag,el->tagName,16);
        sprintf(defaultDeviceName,"DEVICE %d",dv);
        if (el->EqpModuleName) strncpy(ams[n].device,el->EqpModuleName[dv].name,16);
        if (strlen(ams[n].device) == 0) strncpy(ams[n].device,defaultDeviceName,16);
        if ((ads=getAlarmTable(eqpname,alm->alarmCode)) != NULL)
        {
          strcpy(ams[n].alarmTag,ads->alarmTag);
          ams[n].alarmMask = ads->alarmMask;
          ams[n].alarmDataFormat = ads->alarmDataFormat;
          ams[n].alarmDataArraySize = ads->alarmDataArraySize;
          ams[n].alarmSystem = ads->alarmSystem;
        }
        ams[n].timestamp = alm->timestamp;
        ams[n].alarmCode = alm->alarmCode;
        memcpy(ams[n].alarmData,alm->alarmData,6);
        ams[n].severity = sc;
        ams[n].descriptor = alm->alarmStatus;
        if (ams[n].alarmCode < numErr && ams[n].alarmCode > 0 && strlen(ams[n].alarmTag) == 0)
          strncpy(ams[n].alarmTag,erlst[ams[n].alarmCode],16);
        if (++n >= num) break;
      }
    }
  }
  return n;
}

short getAlarms(char *eqpname,AMS *ams,short num,short minSeverity)
{
  return getAlarmsEx(eqpname,ams,num,minSeverity,0,0x7fffffffL,0);
}
__declspec(dllexport) void PASCAL setAlarmWindow(int newAlarmWindow)
{
  if (newAlarmWindow < 1) return;
  AlarmWindow = newAlarmWindow;
}
__declspec(dllexport) int PASCAL getAlarmWindow(void)
{
  return AlarmWindow;
}
__declspec(dllexport) void PASCAL setTerminationWindow(int newTerminationWindow)
{
  if (newTerminationWindow < 1) return;
  TerminationWindow = newTerminationWindow;
}
__declspec(dllexport) int PASCAL getTerminationWindow(void)
{
  return TerminationWindow;
}
__declspec(dllexport) void PASCAL setAlarmHeartbeat(int newAlarmHeartbeat)
{
  if (newAlarmHeartbeat < 1) return;
  AlarmHeartbeat = newAlarmHeartbeat;
}
__declspec(dllexport) int PASCAL getAlarmHeartbeat(void)
{
  return AlarmHeartbeat;
}

