#include "project.def"
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <io.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <malloc.h>
#include "direct.h"
#include "database.h"
#include "winsrv.h"
#include "errors.h"
#include "history.h"

#define O_mode O_RDONLY|O_BINARY

extern char txtbuf[128];
extern char FecDBpath[256];
char ArchiveLocation[128];
HistoryRecord *historyRecords = NULL;
int nHistoryRecords = 0;
int maxHistoryRecords = MAX_HISTORY_RECORDS;
int isHistoryCall = FALSE;
int minHistoryPollingRate = 1000;
int useMonthlyHistoryFiles = FALSE;
int useHighResolutionTimeStamp = FALSE;
extern CONSUMER *EQPconsumerlist;                      /* global reference  */

int makeDirectoryPath(char *dname)
{
  char fn[128];
  char *tok = "\\";
  struct _finddata_t fp;
  char *c,*cp=NULL,tmpname[128],dirname[128];

  strncpy(tmpname,dname,128); dirname[0] = 0;
  for (c=strtok(tmpname,tok); c != NULL; c=strtok(NULL,tok))
  {
    if (strlen(c) == 0) break; /* terminating delimiter gives the last token empty */
    strcat(dirname,c);
    strcat(dirname,tok);
    sprintf(fn,"%s*.*",dirname);
    if (_findfirst(fn, &fp) == -1)
    {
      if (mkdir(dirname)) return 0;
    }
  }
  return -1;
}

#define MakeTargetFileName(nm,yy,mm,dd,id) (\
  sprintf((nm),"%s%02d%02d%02d.%0x",\
  useHighResolutionTimeStamp ? "ha" : "ar",((yy)%100),(mm),\
  useMonthlyHistoryFiles ? 1 : (dd),historyRecords[(id)].recordIndex)\
)
__declspec(dllexport) int PASCAL makeMonthlyHistoryFiles(int value)
{
  return useMonthlyHistoryFiles = value;
}
__declspec(dllexport) int PASCAL useMsecHistoryTimeStamps(int value)
{
  return useHighResolutionTimeStamp = value;
}
int removeOutdatedLTSFile(char *fn)
{
  char fileName[128],lfn[32];
  int i, filesRemoved = 0;
  static int stagger = -1;

  if (stagger < 0) stagger = time(NULL); /* seed the start up */
  /* stagger file removal to avoid blocking the file system */
  stagger = (stagger + 1) % 100;
  if (stagger > 0) return -1; 

  for (i=1; i<32; i++)
  {
    strncpy(lfn,fn,6);
    sprintf(&lfn[6],"%02d",i);
    strcat(lfn,&fn[8]);
    sprintf(fileName,"%s%s",ArchiveLocation,lfn);
    if (unlink(fileName) == 0) filesRemoved++;
  }
  if (filesRemoved > 0)
  {
    sprintf(fileName,"HST: removed %d files from long term storage",filesRemoved);
    feclog(fileName);
  }
  return filesRemoved;
}
int WriteRecordToLTS(int i)  /* write record #i to file */
{
  int hFile,ern,ptr;
  HistoryRecord *hst;
  char fileName[128];
  char s[256];
  BYTE *tmpHistoryDataBuffer;
  int timestampLength = useHighResolutionTimeStamp ? sizeof(double) : sizeof(UINT32);
  time_t lTimeStamp;

  if (i < 0 || i >= nHistoryRecords) return invalid_index;
  hst = &historyRecords[i];
  ptr = hst->dataPtr;
  sprintf(fileName,"%s%s",ArchiveLocation,hst->ArchiveFileName);

  if ( (hFile=open(fileName,O_RDWR|O_CREAT|O_APPEND|O_BINARY,_S_IREAD|_S_IWRITE)) == -1 )
  {
    ern = errno;
    sprintf(s,"HST: error opening history file %s (%d)",fileName,ern);
    feclog(s);
    return file_error;
  }
  lseek(hFile,0L,SEEK_END);

  tmpHistoryDataBuffer = (BYTE *)alloca(historyRecords[i].recordLength+timestampLength);

  if (useHighResolutionTimeStamp)
  {
    memcpy(tmpHistoryDataBuffer,(char *)&hst->timestamp[ptr],sizeof(double));
  }
  else
  {
    lTimeStamp = (time_t)hst->timestamp[ptr];
    memcpy(tmpHistoryDataBuffer,(char *)&lTimeStamp,sizeof(time_t));
  }
  memcpy(&tmpHistoryDataBuffer[timestampLength],hst->data + ptr * hst->recordLength,hst->recordLength);

  write(hFile,(char *)tmpHistoryDataBuffer,hst->recordLength+timestampLength);
  
  close(hFile);
  hst->lastArchiveTime = hst->timestamp[ptr];
  return 0;
}
long locateRecordFromLTS(char *fileName,int i,double dt)
{
  time_t t,t0,t1,t2;
  double dt0, dt1, dt2;
  long nrecs,p0,p1,p2;
  double f;
  int hFile,recsize,it=0;
  int timestampLength = useHighResolutionTimeStamp ? sizeof(double) : sizeof(time_t);
  HistoryRecord *hst;
  struct tm *lt=localtime(&t);

  if (historyRecords == NULL) return -un_allocated;
  if (i < 0 || i >= nHistoryRecords) return -invalid_index;
  hst = &historyRecords[i];
  recsize = hst->recordLength + timestampLength;

  if ( (hFile=open(fileName,O_mode)) == -1 ) return -file_error;

  t = (time_t)dt;
  nrecs = lseek(hFile,0,SEEK_END)/recsize;
  p0 = p1 = 0; p2 = (nrecs-1)*recsize;
  lseek(hFile,p0,SEEK_SET);
  if (useHighResolutionTimeStamp)
  {
    read(hFile,(char *)&dt0,timestampLength);
    t0 = (time_t)dt0;
  }
  else
  {
    read(hFile,(char *)&t0,timestampLength);
    dt0 = (double)t0;
  }
  if (dt < dt0) goto out;
  lseek(hFile,p2,SEEK_SET); 
  if (useHighResolutionTimeStamp)
  {
    read(hFile,(char *)&dt2,timestampLength);
    t2 = (time_t)dt2;
  }
  else
  {
    read(hFile,(char *)&t2,timestampLength);
    dt2 = (double)t2;
  }
  while (p2 > p0 + recsize)
  {
    if (dt > (dt1=dt2)) { p1 = p2; break; }
    while (dt1 > dt0)   /* seek <--  */
    {
      f = ((dt-dt0)/(dt1-dt0))*((double)(p2-p0)/recsize);
      if ((p1=p0+recsize*(time_t)f) >= p2) break;
	    lseek(hFile,p1,SEEK_SET);
      if (useHighResolutionTimeStamp)
      {
        read(hFile,(char *)&dt1,timestampLength);
        t1 = (time_t)dt1;
      }
      else
      {
        read(hFile,(char *)&t1,timestampLength);
        dt1 = (double)t1;
      }
      if (dt1 < dt) break;
      p2 = p1;
      t2 = t1;
      dt2 = dt1;
      it++;
    }
    dt0 = dt1; t0 = t1; p0 = p1;
    while (dt2 > dt1)  /* seek --> */
    {
      f = ((dt2-dt)/(dt2-dt1))*((double)(p2-p0)/recsize);
      if ((p1=p2-recsize*(time_t)f) <= p0) break;
	    lseek(hFile,p1,SEEK_SET); 
      if (useHighResolutionTimeStamp)
      {
        read(hFile,(char *)&dt1,timestampLength);
        t1 = (time_t)dt1;
      }
      else
      {
        read(hFile,(char *)&t1,timestampLength);
        dt1 = (double)t1;
      }
      if (dt1 > dt) break;
      p0 = p1; 
      t0 = t1;
      dt0 = dt1;
      it++;
    }
    p2 = p1; t2 = t1; dt2 = dt1;
  }
out: 
  close(hFile);
  return p1;
}
int needToArchiveRecord(int i)
{
  double t, ts = makeDataTimeStamp();
  BYTE *d0,*d1;
  HistoryRecord *hst;
  int p,ptr,delta_t,doarchive = 0;

  if (i < 0 || i >= nHistoryRecords) return FALSE;
  hst = &historyRecords[i];
  if (hst->archiveRate < 0 || 
      hst->heartbeat < 0   || 
      hst->depthLong < 0) return FALSE; // no long term storage
  ptr = hst->dataPtr;

  if (hst->timestamp[ptr] <= hst->lastArchiveTime) return FALSE;/* data timestamp hasn't changed */
  if (!useHighResolutionTimeStamp && (time_t)hst->timestamp[ptr] == (time_t)hst->lastArchiveTime) return FALSE;
  if (hst->lastArchiveTime == 0) hst->lastArchiveTime = ts - hst->archiveRate/1000.0;
  delta_t = (int)((ts - hst->lastArchiveTime)*1000);
  if (delta_t < hst->archiveRate) return FALSE;
  if (delta_t >= hst->heartbeat*1000) return TRUE; /* archive no matter what */

  d0 = hst->data + ptr * hst->recordLength;
  d1 = ptr > 0 ? hst->data + (ptr-1) * hst->recordLength :
                 hst->data + (hst->depthShort-1) * hst->recordLength;

  switch (LFMT(hst->c.EqpFormatOut))
  {
    case CF_DOUBLE:
         for (p=0; !doarchive && p<(int)hst->c.EqpSizeOut; p++)
         {
           t = hst->aTolerance +
               hst->pTolerance * fabs(((double *)d1)[p]);
           if ( fabs(((double *)d0)[p] - ((double *)d1)[p]) > t) doarchive = 1;
         }
         break;
    case CF_FLOAT:
         for (p=0; !doarchive && p<(int)hst->c.EqpSizeOut; p++)
         {
           t = hst->aTolerance +
               hst->pTolerance * fabs(((float *)d1)[p]);
           if ( fabs(((float *)d0)[p] - ((float *)d1)[p]) > t) doarchive = 1;
         }
         break;
    case CF_INTEGER:
         for (p=0; !doarchive && p<(int)hst->c.EqpSizeOut; p++)
         {
           t = hst->aTolerance +
               hst->pTolerance * abs(((short *)d1)[p]);
           if ( abs(((short *)d0)[p] - ((short *)d1)[p]) > t) doarchive = 1;
         }
         break;
      case CF_LONG:
         for (p=0; !doarchive && p<(int)hst->c.EqpSizeOut; p++)
         {
           t = hst->aTolerance +
               hst->pTolerance * labs(((long *)d1)[p]);
           if ( labs(((long *)d0)[p] - ((long *)d1)[p]) > t) doarchive = 1;
         }
         break;
      case CF_BYTE:
         for (p=0; !doarchive && p<(int)hst->c.EqpSizeOut; p++)
         {
           t = hst->aTolerance +
               hst->pTolerance * abs(((BYTE *)d1)[p]);
           if ( abs(((BYTE *)d0)[p] - ((BYTE *)d1)[p]) > t) doarchive = 1;
         }
         break;
      default:
         if (!doarchive && memcmp(d0,d1,hst->recordLength)) doarchive = 1;
         break;
  }
  return doarchive ? TRUE : FALSE;
}
extern char tRedirectedServer[64];
int getRemoteHistoryServer(char *eqp,char *prp)
{
  int i;
  if (nHistoryRecords <= 0 || historyRecords == NULL) return un_allocated;
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strncmp(historyRecords[i].c.EqpName,eqp,EQP_NAME_SIZE)) continue;
    if (strncmp(historyRecords[i].c.EqpProperty,prp,EQP_PROP_SIZE)) continue;
    break;
  }
  if (i == nHistoryRecords) return un_allocated;
  if (historyRecords[i].prd == NULL) return not_defined;
  memcpy(tRedirectedServer,historyRecords[i].prd->prpRedirection[0].name,16*4);
  return server_redirection;
}
__declspec(dllexport) int PASCAL AppendHistoryInformation(char *eqp,char *prp,char *dev,int len,int fmt,int idx,HistorySpecification *hspec)
{
  int i, cc = 0;
  time_t t = time(NULL);
  struct tm *lt=localtime(&t);
  char str[256];

  if (eqp == NULL || prp == NULL || dev == NULL) ccerr(argument_list_error);
  if (historyRecords == NULL)
  {
    if ((historyRecords=(HistoryRecord *)calloc(maxHistoryRecords,sizeof(HistoryRecord))) == NULL)
      ccerr(out_of_local_memory);
  }
  if ((i=nHistoryRecords) >= maxHistoryRecords) ccerr(out_of_local_memory);
  historyRecords[i].recordIndex = idx;
  strncpy(historyRecords[i].c.EqpName,eqp,6);
  strncpy(historyRecords[i].c.EqpProperty,prp,32);
  strncpy(historyRecords[i].c.EqpDeviceName,dev,16);
  historyRecords[i].c.EqpSizeOut = len;
  historyRecords[i].c.EqpFormatOut = BFMT(fmt);
  historyRecords[i].heartbeat = hspec->heartbeat;
  historyRecords[i].pollingRate = hspec->pollingRate;
  if (historyRecords[i].pollingRate < 20) historyRecords[i].pollingRate *= 1000; /* was thinking in seconds ! */
  if (historyRecords[i].pollingRate < minHistoryPollingRate) minHistoryPollingRate = historyRecords[i].pollingRate;
  historyRecords[i].archiveRate = hspec->archiveRate;
  if (historyRecords[i].archiveRate < 20) historyRecords[i].archiveRate *= 1000; /* was thinking in seconds ! */
  historyRecords[i].pTolerance = hspec->pTolerance;
  if (historyRecords[i].pTolerance > 1.0) historyRecords[i].pTolerance /= 100.0;
  historyRecords[i].aTolerance = hspec->aTolerance;
  historyRecords[i].depthShort = hspec->depthShort;
  historyRecords[i].depthLong = hspec->depthLong;
  historyRecords[i].recordLength =
    (int)historyRecords[i].c.EqpSizeOut * fmtsizeof(historyRecords[i].c.EqpFormatOut);
  MakeTargetFileName(historyRecords[i].ArchiveFileName,lt->tm_year,lt->tm_mon+1,lt->tm_mday,i);
  if (strlen(hspec->rhsServerName) > 0)
  {
    if ((historyRecords[i].prd=(PrpRedirBlk *)calloc(1,sizeof(PrpRedirBlk))) == NULL) 
      ccerr(out_of_server_memory);

    strncpy(historyRecords[i].prd->prpRedirection[0].name,hspec->rhsServerName,16);
    strncpy(historyRecords[i].prd->prpRedirection[1].name,hspec->rhsPropertyName,16);
  }
  if ((historyRecords[i].timestamp=(double *)calloc(historyRecords[i].depthShort,sizeof(double))) == NULL)
    ccerr(out_of_local_memory);
  if ((historyRecords[i].data=(BYTE *)calloc(historyRecords[i].depthShort,historyRecords[i].recordLength)) == NULL)
    ccerr(out_of_local_memory);
  sprintf(str,"HIST: %s %s %s (%d s %d mon)",
          historyRecords[i].c.EqpName,
          historyRecords[i].c.EqpProperty,
          historyRecords[i].c.EqpDeviceName,
          historyRecords[i].depthShort,
          historyRecords[i].depthLong);
  feclog(str);
  nHistoryRecords++;
err:
  return cc;
}
int LoadHistoryInformation(char *fname,char *expName,char *lclName) /* read history record info from file */
{
  FILE *fp = NULL;
  char fn[256],str[256],scratch[48];
  int done = 0,i,cc = 0;
  short name_col,property_col,device_col,datalen_col,fmt_col,
      heartbeat_col,poll_col,rate_col,tol_col,recidx_col,
      sdepth_col,mdepth_col,exp_col,rhs_col=-1;
  time_t t = time(NULL);
  struct tm *lt=localtime(&t);
  static int nTotalRecords = 0;

  sprintf(fn,"%s%s",FecDBpath,fname);
  if (historyRecords == NULL)
  {
    if ((i=GetDataBaseSize(fn)) < 0) ccerr(-i);
    if ((nHistoryRecords=i) > maxHistoryRecords) maxHistoryRecords = nHistoryRecords;
    if ((historyRecords=(HistoryRecord *)calloc(maxHistoryRecords,sizeof(HistoryRecord))) == NULL)
      ccerr(out_of_local_memory);
  }
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  sprintf(str,"HISTORY HOME : %s",ArchiveLocation);
  feclog(str);
  for (i=nTotalRecords; fgets(str,255,fp) != NULL; )
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((recidx_col=findcol(str,"INDEX")) < 0) ccerr(no_such_line);
      name_col = findcol(str,"LOCAL NAME");
      if ((property_col=findcol(str,"PROPERTY")) < 0) ccerr(no_such_line);
      if ((device_col=findcol(str,"DEVICE")) < 0) ccerr(no_such_line);
      if ((datalen_col=findcol(str,"DATA LENGTH")) < 0) ccerr(no_such_line);
      if ((fmt_col=findcol(str,"FORMAT")) < 0) ccerr(no_such_line);
      if ((heartbeat_col=findcol(str,"HEARTBEAT")) < 0) ccerr(no_such_line);
      if ((poll_col=findcol(str,"POLLING RATE")) < 0) ccerr(no_such_line);
      if ((rate_col=findcol(str,"ARCHIVE RATE")) < 0) ccerr(no_such_line);
      if ((tol_col=findcol(str,"TOLERANCE")) < 0) ccerr(no_such_line);
      if ((sdepth_col=findcol(str,"SHORT DEPTH")) < 0) ccerr(no_such_line);
      if ((mdepth_col=findcol(str,"LONG DEPTH")) < 0) ccerr(no_such_line);
      exp_col = findcol(str,"EXPORT NAME");
      rhs_col = findcol(str,"REMOTE SERVER");
      done = 1;
      continue;
    }
    if (name_col < 0 && exp_col < 0) ccerr(no_such_line);
    if (exp_col >= 0 && _strnicmp(expName,colptr(exp_col,str),16)) continue;
    strncpy(scratch,colptr(recidx_col,str),16);
    historyRecords[i].recordIndex = atoi(scratch);
    if (name_col >= 0) strncpy(historyRecords[i].c.EqpName,colptr(name_col,str),6);
    else strncpy(historyRecords[i].c.EqpName,lclName,6);
    strncpy(historyRecords[i].c.EqpProperty,colptr(property_col,str),32);
    strncpy(historyRecords[i].c.EqpDeviceName,colptr(device_col,str),16);
    strncpy(scratch,colptr(datalen_col,str),16);
    historyRecords[i].c.EqpSizeOut = atoi(scratch);
    strncpy(scratch,colptr(fmt_col,str),16);
    historyRecords[i].c.EqpFormatOut = BFMT(ftoi(scratch));
    strncpy(scratch,colptr(heartbeat_col,str),16);
    historyRecords[i].heartbeat = atoi(scratch);
    strncpy(scratch,colptr(poll_col,str),16);
    historyRecords[i].pollingRate = atoi(scratch);
    if (historyRecords[i].pollingRate < 20) historyRecords[i].pollingRate *= 1000; /* was thinking in seconds ! */
    if (historyRecords[i].pollingRate < minHistoryPollingRate) minHistoryPollingRate = historyRecords[i].pollingRate;
    strncpy(scratch,colptr(rate_col,str),16);
    historyRecords[i].archiveRate = atoi(scratch);
    if (historyRecords[i].archiveRate < 20) historyRecords[i].archiveRate *= 1000; /* was thinking in seconds ! */
    strncpy(scratch,colptr(tol_col,str),16);
    if (strchr(scratch,'%')) historyRecords[i].pTolerance = (float)(atof(scratch) / 100.0);
    else historyRecords[i].aTolerance = (float)atof(scratch);
    strncpy(scratch,colptr(sdepth_col,str),16);
    historyRecords[i].depthShort = atoi(scratch);
    strncpy(scratch,colptr(mdepth_col,str),16);
    historyRecords[i].depthLong = atoi(scratch);
    historyRecords[i].recordLength =
      (int)historyRecords[i].c.EqpSizeOut * fmtsizeof(historyRecords[i].c.EqpFormatOut);
    MakeTargetFileName(historyRecords[i].ArchiveFileName,lt->tm_year,lt->tm_mon+1,lt->tm_mday,i);
    if (rhs_col >= 0)
    {
      strncpy(scratch,colptr(rhs_col,str),48);
      if (strlen(scratch) > 0)
      {
        if ((historyRecords[i].prd=(PrpRedirBlk *)calloc(1,sizeof(PrpRedirBlk))) == NULL) 
          ccerr(out_of_server_memory);
        ParseRedirector(scratch,historyRecords[i].prd);
      }
    }
    if ((historyRecords[i].timestamp=(double *)calloc(historyRecords[i].depthShort,sizeof(double))) == NULL)
      ccerr(out_of_local_memory);
    if ((historyRecords[i].data=(BYTE *)calloc(historyRecords[i].depthShort,historyRecords[i].recordLength)) == NULL)
      ccerr(out_of_local_memory);
    sprintf(str,"HIST: %s %s %s (%d s %d mon)",
            historyRecords[i].c.EqpName,
            historyRecords[i].c.EqpProperty,
            historyRecords[i].c.EqpDeviceName,
            historyRecords[i].depthShort,
            historyRecords[i].depthLong);
    feclog(str);
    if (i++ >= maxHistoryRecords) break;
  }
  nTotalRecords = i;
  nHistoryRecords = nTotalRecords;
err:
  if (fp != NULL) fclose(fp);
  if (cc && cc != no_such_file)
  {
    nHistoryRecords = 0;
    sprintf(str,"HIST: %s : %s\n",fn,erlst[cc]);
    feclog(str);
  }
  if (nHistoryRecords == 0)
  {
    sprintf(str,"HIST: no history records");
    feclog(str);
  }  
  return cc;
}
void historyInit(char *expName,char *lclName)
{
  char *ptr, hstfn[64];
  int len;

  if ((ptr=getenv("HISTORY_HOME")) == NULL && (ptr=getenv("HISTORYDB")) == NULL) 
  {
    ptr = FecDBpath;
  }
  strncpy(ArchiveLocation,ptr,80);
  if ((len=strlen(ArchiveLocation)) > 0 && ArchiveLocation[len-1] != '\\') strcat(ArchiveLocation,"\\");
  sprintf(txtbuf,"Archive Location: [%s]",ArchiveLocation);
  dbllog(txtbuf);
  makeDirectoryPath(ArchiveLocation);
  // try the local name subdirectory first
  _snprintf(hstfn,64,"%s\\%s",lclName,HISTORY_FILENAME);
  if (LoadHistoryInformation(hstfn,expName,lclName) != 0)
  {
    strncpy(hstfn,HISTORY_FILENAME,64);
    LoadHistoryInformation(hstfn,expName,lclName);
  }
}
void historyCycle(void)
{
  int lid,i,p,n,filesRemoved=0;
  ExportListStruct *el;
  char rmvfn[16];
  struct tm *tm;
  int yy,mm,delta_t;
  static struct timeval lasttime = {0, 0};
  static time_t lt = 0;
  static DTYPE din,dout;
  static int mday = -1;
  
  if (lasttime.tv_sec == 0) memset(&din,0,sizeof(DTYPE));
  if (MSECS(gThisCycleTime,lasttime) < minHistoryPollingRate) return;
  lasttime = gThisCycleTime;
  lt = gThisCycleTime.tv_sec;
  tm = localtime(&lt);
  for (i=0; i<nHistoryRecords; i++)
  {
    if ((el=findTagID(historyRecords[i].c.EqpName)) == NULL) continue;
    if (el->hEqpFcn == hNULL) continue;
    if (historyRecords[i].depthShort <= 0) continue;
    if (mday != tm->tm_mday)
    {
      MakeTargetFileName(historyRecords[i].ArchiveFileName,tm->tm_year,tm->tm_mon+1,tm->tm_mday,i);
      if (filesRemoved == 0)
      { // careful of the modulo arithmetic here (months go from 0 to 11!)
        mm = ((tm->tm_mon+12) - ((historyRecords[i].depthLong+1)%12))%12;
        yy = tm->tm_year - historyRecords[i].depthLong/12;
        if (mm > tm->tm_mon) yy--;
        MakeTargetFileName(rmvfn,yy,mm+1,0,i);
        filesRemoved = removeOutdatedLTSFile(rmvfn);
      }
    }
    gDataTimeStamp = putDataTimeStamp(gDataTimeStampOffset,gThisCycleTime.tv_sec,gThisCycleTime.tv_usec/1000);
    n = historyRecords[i].dataPtr;
    if (historyRecords[i].timestamp[n] == 0) 
      historyRecords[i].timestamp[n] = gDataTimeStamp - (double)historyRecords[i].pollingRate/1000.0;
    delta_t = (int)((gDataTimeStamp - historyRecords[i].timestamp[n])*1000);
    if (delta_t < (int)(historyRecords[i].pollingRate-gSystemDelay)) continue;
    n = (historyRecords[i].dataPtr + 1) % historyRecords[i].depthShort;
    lid = (long)MAKEPROTID(i,HST_PROTOCOL);
    strcpy(clientName,"LOCAL HISTORY"); strcpy(clientaddr,"THIS STATION");
    if (historyRecords[i].interimCounter == 0) historyRecords[i].interimStatus = not_posted;
    if (historyRecords[i].interimStatus == not_posted)
    {
      isHistoryCall = TRUE;
      SendMessage(el->hEqpFcn,WM_SYSCOMMAND,lid,HST_PROTOCOL);
      isHistoryCall = FALSE;
    }
    if (historyRecords[i].interimStatus != 0) 
    {
      if (historyRecords[i].interimStatus == not_posted || historyRecords[i].interimStatus == not_ready)
      {
        historyRecords[i].interimStatus = not_ready;
        if (historyRecords[i].interimCounter++ > 10) historyRecords[i].interimCounter = 0;
      }
      continue;
    }
    p = historyRecords[i].dataPtr;
    delta_t = (int)((gDataTimeStamp - historyRecords[i].timestamp[p])*1000);
    if (delta_t < (int)(historyRecords[i].pollingRate-gSystemDelay)) continue;
    historyRecords[i].timestamp[n] = gDataTimeStamp;
    historyRecords[i].dataPtr = n;
    if (needToArchiveRecord(i)) WriteRecordToLTS(i);
  }
  if (filesRemoved == 0) mday = tm->tm_mday;
  return;
}
long loadLTSData(int i,int index,double dstarttime,double dendtime,
                 float lwrLimit, float uprLimit,
                 void *datbuf,int datfmt,int bufsize,int sample)
{
  int hFile,nrecs,raster = 1,isSnapshot = 0;
  time_t starttime = (time_t)dstarttime;
  time_t endtime = (time_t)dendtime;
  long pstart,pstop,cnt = 0,rlen;
  char fn[16],fileName[128];
  DTYPE d;
  short number = 0,fmtsize;
  int timestampLength = useHighResolutionTimeStamp ? sizeof(double) : sizeof(time_t);
  time_t timestamp;
  double dtimestamp, datatime = 0, dtnow = makeDataTimeStamp();
  static double timebin = -1;
  struct tm *lt=localtime((time_t *)&starttime);
  BYTE *tmpHistoryDataBuffer, tsBuffer[8];
  UINT32 strsiz;

  if (timebin < 0) timebin = useHighResolutionTimeStamp ? 0.01 : 1.0;
  if (bufsize < 1) return 0;
  if (historyRecords == NULL) return FALSE;
  if (i < 0 || i >= nHistoryRecords) return FALSE;
  if (historyRecords[i].depthLong < 0) return FALSE;
  if ((dtnow - dstarttime) < historyRecords[i].depthShort) return FALSE;
  if (historyRecords[i].recordLength > WORK_SIZE) goto err;
  if (index < 0 || index >= (int)historyRecords[i].c.EqpSizeOut) goto err;
  if (((dendtime-dstarttime) < 0.01) && datfmt <= CF_FLOAT && bufsize > 1) isSnapshot = TRUE;
  if (isSnapshot) 
  {
    index = 0; /* snapshot requests always have index = 0 */
    datatime = dstarttime - timebin;
  }

  MakeTargetFileName(fn,lt->tm_year,lt->tm_mon+1,lt->tm_mday,i);
  sprintf(fileName,"%s%s",ArchiveLocation,fn);

  if (dendtime < dstarttime) dendtime = dtnow;
  if ((pstart=locateRecordFromLTS(fileName,i,dstarttime)) < 0) goto err;
  if ((pstop=locateRecordFromLTS(fileName,i,dendtime)) < 0) goto err;
  rlen = historyRecords[i].recordLength + timestampLength;
  nrecs = (pstop-pstart)/rlen;
  if (bufsize == 1) 
  {
    setDataTimeStamp(dendtime);
    return MAX(nrecs,0);
  }
  if (bufsize < nrecs) raster += nrecs/(bufsize+1);
  if (sample > 0) raster = sample;

  if ((hFile=open(fileName,O_mode)) == -1) goto err;

  number = 0; cnt = 0; datatime = 0;
  lseek(hFile,pstart,SEEK_SET);

  tmpHistoryDataBuffer = (BYTE *)alloca(historyRecords[i].recordLength);
  fmtsize = fmtsizeof(historyRecords[i].c.EqpFormatOut);

  d.dFormat = LFMT(historyRecords[i].c.EqpFormatOut);
  d.dArrayLength = 1;
  d.data.bptr = tmpHistoryDataBuffer + index*fmtsize;

  while(read(hFile,(char *)tsBuffer,timestampLength) == timestampLength)   /* read through file */
  {
    if (useHighResolutionTimeStamp) 
    {
      memcpy(&dtimestamp,tsBuffer,timestampLength); 
      timestamp = (time_t)dtimestamp;
    }
    else 
    {
      memcpy(&timestamp,tsBuffer,timestampLength);
      dtimestamp = (double)timestamp;
    }
    cnt++;
    if (isSnapshot && dtimestamp > dendtime) dendtime = dtimestamp;
    if (dtimestamp > dendtime) goto err;
    if (dtimestamp < datatime + timebin)
    {
      if (lseek(hFile,historyRecords[i].recordLength,SEEK_CUR) < 0) goto err;
      continue;
    }
    if (read(hFile,(char *)tmpHistoryDataBuffer,historyRecords[i].recordLength)
        != historyRecords[i].recordLength) goto err;

    datatime = dtimestamp;

    switch (datfmt)
    {
      case CF_FLOAT:
        d.dArrayLength = MIN(bufsize,(int)historyRecords[i].c.EqpSizeOut);
        number = (int)d.dArrayLength;
        getValuesAsFloat(&d,datbuf,number);
        goto err;
      case CF_FLTINT:
        ((FLTINT *)datbuf)[number].ival = timestamp;
        getValuesAsFloat(&d,&((FLTINT *)datbuf)[number].fval,1);
        if (sample < 0 && !isInRange(datbuf,datfmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_INTFLTINT: /* CF_TDS */
        ((INTFLTINT *)datbuf)[number].i1val = timestamp;
        getValuesAsFloat(&d,&((INTFLTINT *)datbuf)[number].fval,1);
        ((INTFLTINT *)datbuf)[number].i2val = 0;
        if (sample < 0 && !isInRange(datbuf,datfmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_LONG:
        d.dArrayLength = MIN(bufsize,(int)historyRecords[i].c.EqpSizeOut);
        number = (int)d.dArrayLength;
        getValuesAsLong(&d,datbuf,number);
        goto err;
      case CF_LNGINT:
        ((INTINT *)datbuf)[number].i2val = timestamp;
        getValuesAsLong(&d,&((INTINT *)datbuf)[number].i1val,1);
        if (sample < 0 && !isInRange(datbuf,datfmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_DOUBLE:
        d.dArrayLength = MIN(bufsize,(int)historyRecords[i].c.EqpSizeOut);
        number = (int)d.dArrayLength;
        getValuesAsDouble(&d,datbuf,number);
        goto err;
      case CF_DBLDBL:
        ((DBLDBL *)datbuf)[number].d2val = dtimestamp;
        getValuesAsDouble(&d,&((DBLDBL *)datbuf)[number].d1val,1);
        if (sample < 0 && !isInRange(datbuf,datfmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_NAME16I:
        ((NAME16I *)datbuf)[number].ival = timestamp; strsiz = 16;
        getValuesAsString(&d,((NAME16I *)datbuf)[number].name,&strsiz);
        break;
      case CF_NAME32I:
        ((NAME32I *)datbuf)[number].ival = timestamp; strsiz = 32;
        getValuesAsString(&d,((NAME32I *)datbuf)[number].name,&strsiz);
        break;
      case CF_NAME48I:
        ((NAME48I *)datbuf)[number].ival = timestamp; strsiz = 48;
        getValuesAsString(&d,((NAME48I *)datbuf)[number].name,&strsiz);
        break;
      case CF_NAME64I:
        ((NAME64I *)datbuf)[number].ival = timestamp; strsiz = 64;
        getValuesAsString(&d,((NAME64I *)datbuf)[number].name,&strsiz);
        break;
      case CF_SHORT:
        d.dArrayLength = MIN(bufsize,(int)historyRecords[i].c.EqpSizeOut);
        number = (int)d.dArrayLength;
        getValuesAsShort(&d,datbuf,number);
        goto err;
      default: goto err;
    }
    number++;
    if (number >= bufsize) goto err;
    if (raster > 1) // move along through file
    {
      if (lseek(hFile,rlen*(raster-1),SEEK_CUR) < 0) goto err;
    }
  }
err:
  if (hFile != -1) close(hFile);

  if (datatime > 0) setDataTimeStamp(datatime);
  return number;
}
int IsLiteralDevice(char *eqp,char *prp, char *dev)
{
  int i;
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strncmp(historyRecords[i].c.EqpName,eqp,6)) continue;
    if (strncmp(historyRecords[i].c.EqpProperty,prp,32)) continue;
    if (strncmp(historyRecords[i].c.EqpDeviceName,dev,16)) continue;
    break;
  }
  return i < nHistoryRecords ? TRUE : FALSE;
}
long getSTSData(char *eqp,char *prp, char *dev,int index,
		 double dstarttime,double dendtime,
     float lwrLimit, float uprLimit,
     BYTE *data,int fmt,int arraysize,int sample,int recsreadFlag)
{
  int i,n,rnumber=0,number,num,raster,cnt,fmtsiz,literalDevice;
  time_t timestamp;
  time_t starttime = (time_t)dstarttime, endtime = (time_t)dendtime;
  int sttidx = recsreadFlag ? 1 : 0;
  int isSnapshot=0;
  double dtimestamp,lastreadtimestamp=0;
  DTYPE d;
  UINT32 strsiz;

  if (historyRecords == NULL || nHistoryRecords <= 0) return -un_allocated;
  literalDevice = IsLiteralDevice(eqp,prp,dev);
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strncmp(historyRecords[i].c.EqpName,eqp,6)) continue;
    if (strncmp(historyRecords[i].c.EqpProperty,prp,32)) continue;
    if (strncmp(historyRecords[i].c.EqpDeviceName,dev,16))
    {
      if (index == 0 && !literalDevice) index = GetDeviceNumberEx(eqp,dev,prp);
      else continue;
    }
    break;
  }
  if (i == nHistoryRecords) return -un_allocated;
  if (historyRecords[i].depthShort == 0) return -getRemoteHistoryServer(eqp,prp);
  if (index < 0) return index;
  if (index >= (int)historyRecords[i].c.EqpSizeOut) return -out_of_range;
  if (arraysize < 1) return -buffer_too_small;
  if (dstarttime < dendtime - (double)historyRecords[i].depthShort) 
    dstarttime = dendtime - (double)historyRecords[i].depthShort;

  d.dFormat = LFMT(historyRecords[i].c.EqpFormatOut);
  d.dArrayLength = 1;
  if ((raster=(int)((dendtime-dstarttime)/(double)arraysize)) < 1) raster = 1;
  if (arraysize == 1) 
  {
    raster = 1;
    sttidx = 0;
  }
  isSnapshot = ((dendtime-dstarttime) < .01) ? TRUE : FALSE;
  if (isSnapshot && arraysize > 1) index = 0; /* snapshot requests always have index = 0 */
  fmtsiz = fmtsizeof(d.dFormat);
  n = historyRecords[i].dataPtr;
  /* move one past the current position to start at the beginning */
  n = (n + raster + historyRecords[i].depthShort) % historyRecords[i].depthShort;
  if (arraysize == (int)(dendtime-dstarttime)) /* caller wants the array filled */
  {
    if (arraysize <= (num=historyRecords[i].depthShort))
    {
      n = (n + (num-arraysize) + historyRecords[i].depthShort) % historyRecords[i].depthShort;
      dstarttime = historyRecords[i].timestamp[n];
      dendtime = makeDataTimeStamp();
    }
  }
  lastreadtimestamp = dstarttime;
  for (number=sttidx,cnt=0; number < arraysize &&
                   number <= historyRecords[i].depthShort &&
                   cnt < historyRecords[i].depthShort; cnt += raster)
  {
    d.data.vptr = historyRecords[i].data + index*fmtsiz + n*historyRecords[i].recordLength;
    dtimestamp = historyRecords[i].timestamp[n];
    n = (n + raster + historyRecords[i].depthShort) % historyRecords[i].depthShort;
    if (dtimestamp < lastreadtimestamp || dtimestamp > dendtime) continue;
    lastreadtimestamp = dtimestamp;
    timestamp = (time_t)dtimestamp;
    rnumber++;
    if (arraysize < 2) continue;
    switch (fmt)
    {
      case CF_FLOAT:
        if (isSnapshot)
        {
          d.dArrayLength = MIN(arraysize,(int)historyRecords[i].c.EqpSizeOut);
          rnumber = number = (int)d.dArrayLength;
          getValuesAsFloat(&d,(float *)data,(short)number);
          goto err;
        }
        getValuesAsFloat(&d,&((float *)data)[number],1);
        break;
      case CF_FLTINT:
        ((FLTINT *)data)[number].ival = timestamp;
        getValuesAsFloat(&d,&((FLTINT *)data)[number].fval,1);
        if (sample < 0 && !isInRange(data,fmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_INTFLTINT: /* CF_TDS */
        ((INTFLTINT *)data)[number].i1val = timestamp;
        getValuesAsFloat(&d,&((INTFLTINT *)data)[number].fval,1);
        ((INTFLTINT *)data)[number].i2val = 0;
        if (sample < 0 && !isInRange(data,fmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_FLTINTINT: 
        getValuesAsFloat(&d,&((FLTINTINT *)data)[number].fval,1);
        ((FLTINTINT *)data)[number].i1val = timestamp;
        ((FLTINTINT *)data)[number].i2val = (UINT32)((dtimestamp-(double)timestamp)*1000);
        if (sample < 0 && !isInRange(data,fmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_LONG:
      	if (isSnapshot)
        {
          d.dArrayLength = MIN(arraysize,(int)historyRecords[i].c.EqpSizeOut);
          rnumber = number = (int)d.dArrayLength;
          getValuesAsLong(&d,(UINT32 *)data,(short)number);
          goto err;
        }
        getValuesAsLong(&d,&((UINT32 *)data)[number],1);
        break;
      case CF_LNGINT:
        ((INTINT *)data)[number].i1val = timestamp;
        getValuesAsLong(&d,&((INTINT *)data)[number].i2val,1);
        if (sample < 0 && !isInRange(data,fmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_DOUBLE:
      	if (isSnapshot)
        {
          d.dArrayLength = MIN(arraysize,(int)historyRecords[i].c.EqpSizeOut);
          rnumber = number = (int)d.dArrayLength;
          getValuesAsDouble(&d,(double *)data,(short)number);
          goto err;
        }
        getValuesAsDouble(&d,&((double *)data)[number],1);
        break;
      case CF_DBLDBL:
        ((DBLDBL *)data)[number].d2val = dtimestamp;
        getValuesAsDouble(&d,&((DBLDBL *)data)[number].d1val,1);
        if (sample < 0 && !isInRange(data,fmt,number,lwrLimit,uprLimit)) continue;
        break;
      case CF_NAME16I:
        ((NAME16I *)data)[number].ival = timestamp; strsiz = 16;
        getValuesAsString(&d,((NAME16I *)data)[number].name,&strsiz);
        break;
      case CF_NAME32I:
        ((NAME32I *)data)[number].ival = timestamp; strsiz = 32;
        getValuesAsString(&d,((NAME32I *)data)[number].name,&strsiz);
        break;
      case CF_NAME48I:
        ((NAME48I *)data)[number].ival = timestamp; strsiz = 48;
        getValuesAsString(&d,((NAME48I *)data)[number].name,&strsiz);
        break;
      case CF_NAME64I:
        ((NAME64I *)data)[number].ival = timestamp; strsiz = 64;
        getValuesAsString(&d,((NAME64I *)data)[number].name,&strsiz);
        break;
      case CF_SHORT:
      	if (isSnapshot)
        {
          d.dArrayLength = MIN(arraysize,(int)historyRecords[i].c.EqpSizeOut);
          rnumber = number = (int)d.dArrayLength;
          getValuesAsShort(&d,(UINT16 *)data,(short)number);
          goto err;
        }
        return -illegal_format;
      default: goto err;
    }
    number++;
  }

  assignNumberOfPoints(data,fmt,arraysize,&number,&rnumber,recsreadFlag);

err:
  if (lastreadtimestamp > 0) setDataTimeStamp((double)lastreadtimestamp);
  return arraysize < 2 ? rnumber : number;
}
int isInRange(BYTE *data,int fmt,int number,float lwrLimit,float uprLimit)
{
  float fval;
  switch (fmt)
  {
    case CF_FLTINT:
      fval = ((FLTINT *)data)[number].fval;
      if (fval < lwrLimit || fval > uprLimit) 
      {
        ((FLTINT *)data)[number].ival = 0; /* flag the boundary */
        if (number > 0 && ((FLTINT *)data)[number-1].ival == 0) return 0;
      }
      return -1;
    case CF_FLTINTINT:
      fval = ((FLTINTINT *)data)[number].fval;
      if (fval < lwrLimit || fval > uprLimit) 
      {
        ((FLTINTINT *)data)[number].i1val = 0; /* flag the boundary */
        ((FLTINTINT *)data)[number].i2val = 0; /* flag the boundary */
        if (number > 0 && ((FLTINTINT *)data)[number-1].i1val == 0) return 0;
      }
      return -1;
    case CF_INTINT:
      fval = (float)((INTINT *)data)[number].i1val;
      if (fval < lwrLimit || fval > uprLimit) 
      {
        ((INTINT *)data)[number].i2val = 0; /* flag the boundary */
        if (number > 0 && ((INTINT *)data)[number-1].i2val == 0) return 0;
      }
      return -1;
    case CF_DBLDBL:
      fval = (float)((DBLDBL *)data)[number].d1val;
      if (fval < lwrLimit || fval > uprLimit) 
      {
        ((DBLDBL *)data)[number].d2val = 0; /* flag the boundary */
        if (number > 0 && ((DBLDBL *)data)[number-1].d2val == 0) return 0;
      }
      return -1;
    case CF_INTFLTINT:
      fval = (float)((INTFLTINT *)data)[number].fval;
      if (fval < lwrLimit || fval > uprLimit) 
      {
        ((INTFLTINT *)data)[number].i1val = 0; /* flag the boundary */
        if (number > 0 && ((INTFLTINT *)data)[number-1].i1val == 0) return 0;
      }
      return -1;
  }
  return -1;
}
void assignNumberOfPoints(BYTE *data,int fmt,int siz,int *npoints,int *nfound,int recsreadFlag)
{
  int padding = recsreadFlag ? 1 : 0;
  if (data == NULL || npoints == NULL || nfound == NULL) return;

  if (siz == 1) *npoints = *nfound + padding;
  else if (!recsreadFlag) return;

  switch (fmt)
  {
    case CF_FLTINT:
      (*npoints)--;
      ((FLTINT *)data)[0].ival = (UINT32)(*npoints);
      ((FLTINT *)data)[0].fval = (float)(*npoints);
      break;
    case CF_FLTINTINT:
      (*npoints)--;
      ((FLTINTINT *)data)[0].i1val = (UINT32)(*npoints);
      ((FLTINTINT *)data)[0].i2val = (UINT32)(*npoints);
      ((FLTINTINT *)data)[0].fval = (float)(*npoints);
      break;
    case CF_INTINT:
      (*npoints)--;
      ((INTINT *)data)[0].i1val = (UINT32)(*npoints);
      ((INTINT *)data)[0].i2val = (UINT32)(*npoints);
      break;
    case CF_DBLDBL:
      (*npoints)--;
      ((DBLDBL *)data)[0].d1val = (double)(*npoints);
      ((DBLDBL *)data)[0].d2val = (double)(*npoints);
      break;
    case CF_NAME16I:
      (*npoints)--;
      ((NAME16I *)data)[0].ival = (UINT32)(*npoints);
      break;
    case CF_NAME32I:
      (*npoints)--;
      ((NAME32I *)data)[0].ival = (UINT32)(*npoints);
      break;
    case CF_NAME48I:
      (*npoints)--;
      ((NAME48I *)data)[0].ival = (UINT32)(*npoints);
      break;
    case CF_NAME64I:
      (*npoints)--;
      ((NAME64I *)data)[0].ival = (UINT32)(*npoints);
      break;
    case CF_LONG:
      if (siz == 1) *(UINT32 *)data = (UINT32)(*nfound);
      break;
    case CF_FLOAT:
      if (siz == 1) *(float *)data = (float)(*nfound);
      break;
    case CF_INTFLTINT:
      if (siz == 1) 
      {
        ((INTFLTINT *)data)[0].i1val = (UINT32)(*nfound);
        ((INTFLTINT *)data)[0].i2val = (UINT32)(*nfound);
        ((INTFLTINT *)data)[0].fval = (float)(*nfound);
      }
      break;
    case CF_DOUBLE:
      if (siz == 1) *(double *)data = (double)(*nfound);
      break;
  }
}

int nDaysInMonthTable[2][12] = {
  {31,28,31,30,31,30,31,31,30,31,30,31},
  {31,29,31,30,31,30,31,31,30,31,30,31},
};
#define nDaysInMonth(yy,mm) (((yy) % 4) ? nDaysInMonthTable[0][mm] : nDaysInMonthTable[1][mm])

int nDaysInTimeSpan(struct tm *s,struct tm *e)
{
  int n = 0, nx, y, d, ms, me, i;

  d = nDaysInMonth(e->tm_year,e->tm_mon);
  nx = s->tm_mday + (d-e->tm_mday-1);
  for (y=s->tm_year; y<e->tm_year+1; y++)
  {
    ms = y == s->tm_year ? s->tm_mon : 0;
    me = y == e->tm_year ? e->tm_mon : 11;
    for (i=ms; i<me+1; i++) 
    {
      n += nDaysInMonth(y,i);
    }
  }
  return n - nx;
}

long getStartTimeFromDepth(char *eqp,char *prp, int depth)
{
  int recsize,i,n,k,siz;
  struct tm *lt,it,itprev;
  char fn[16],fileName[128];
  FILE *fp = NULL;
  time_t t=time(NULL);
  int timestampLength = useHighResolutionTimeStamp ? sizeof(double) : sizeof(time_t);

  if (nHistoryRecords <= 0 || historyRecords == NULL) return -un_allocated;
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strncmp(historyRecords[i].c.EqpName,eqp,EQP_NAME_SIZE)) continue;
    if (strncmp(historyRecords[i].c.EqpProperty,prp,EQP_PROP_SIZE)) continue;
    break;
  }
  if (i == nHistoryRecords) return -un_allocated;
  recsize = historyRecords[i].recordLength + timestampLength;
  if (historyRecords[i].depthLong == 0) return t - historyRecords[i].depthShort;
  lt = localtime(&t); 
  it = *lt; it.tm_sec = 0; it.tm_hour = 0; it.tm_min = 0; itprev = it;
  for (k=0,n=0; k<30 && n<depth; k++)
  {
    MakeTargetFileName(fn,it.tm_year,it.tm_mon+1,it.tm_mday,i);
    sprintf(fileName,"%s%s",ArchiveLocation,fn);
    if ((fp=fopen(fileName,"r")) != NULL)
    {
      fseek(fp,0,SEEK_END);
      siz = ftell(fp);
      fclose(fp);
      if (siz > 0) 
      {
        if (n + siz/recsize > depth) 
        {
          return n == 0 ? mktime(&it) : mktime(&itprev);
        }
        else n += siz/recsize;
      }
    }
    itprev = it;
    if (it.tm_mday > 1) it.tm_mday--;
    else 
    {
      if (it.tm_mon > 0)
      {
        it.tm_mon--;
      }
      else
      {
        it.tm_year--;
        it.tm_mon = 11;
      }
      it.tm_mday = nDaysInMonth(it.tm_year,it.tm_mon) + 1;
    }  
  }
  if (n > 0) return mktime(&itprev);
  return t - historyRecords[i].depthShort;
}
long getLTSData(char *eqp,char *prp, char *dev,int index,
                 double dstarttime,double dendtime,
                 float lwrLimit, float uprLimit,
                 BYTE *data,int fmt,int arraysize,int sample,int recsreadFlag)
{
  long ps;
  int recsread=0,ptr=0,recsfound=0,rnumber=0,recsize,i,asize,daystotal,daysinmonth;
  int literalDevice,finalPass=FALSE;
  struct tm st,et,it;
  time_t t = time(NULL),itime;
  time_t starttime = (time_t)dstarttime;
  time_t endtime = (time_t)dendtime;

  if (historyRecords == NULL || nHistoryRecords <= 0) return -un_allocated;
  literalDevice = IsLiteralDevice(eqp,prp,dev);
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strncmp(historyRecords[i].c.EqpName,eqp,6)) continue;
    if (strncmp(historyRecords[i].c.EqpProperty,prp,32)) continue;
    if (strncmp(historyRecords[i].c.EqpDeviceName,dev,16))
    {
      if (index == 0 && !literalDevice) index = GetDeviceNumberEx(eqp,dev,prp);
      else continue;
    }
    break;
  }
  if (i == nHistoryRecords) return -un_allocated;
  if (historyRecords[i].depthShort == 0) return -getRemoteHistoryServer(eqp,prp);
  if (historyRecords[i].depthLong < 0) return 0;
  if (fmt >= CF_FLTINT && sample >= 0 && starttime >= t - historyRecords[i].depthShort * STS_DEADBAND) return 0;
  if (sample >= 0 && starttime >= t - historyRecords[i].depthShort) return 0;
  if (index < 0) return index;
  if (index >= (int)historyRecords[i].c.EqpSizeOut) return -out_of_range;
  if (arraysize < 1) return -buffer_too_small;
  if (starttime == endtime && arraysize > 1) index = 0; /* snapshot requests always start at index 0 */

  if (sample < 0) sample = -(sample+1);
  recsize = fmtsizeof(fmt);
  memcpy(&et,localtime(&endtime), sizeof(struct tm));
  memcpy(&st,localtime(&starttime),sizeof(struct tm));
  if (et.tm_mon == st.tm_mon && et.tm_mday == st.tm_mday)
  {
    asize = arraysize;
    finalPass = TRUE;
  }
  else if (arraysize == 1)   /* just a number query */
  {
    asize = arraysize;
  }
  else                      /* spans multiple files */
  {
    asize = 0;
  }

  daystotal = nDaysInTimeSpan(&st,&et);

  if (recsreadFlag && arraysize > 1)
  {
    recsread = 1;
    ptr = recsread * recsize;
  }

  while (recsread < arraysize)
  {
    if (asize < arraysize || arraysize == 1)
    {
      it = st;
      if (it.tm_year == et.tm_year && it.tm_mon == et.tm_mon && 
          (it.tm_mday == et.tm_mday || useMonthlyHistoryFiles))
      {
        asize = arraysize;
        finalPass = TRUE;
      }
      else  /* move along */
      {
        it.tm_mday = (it.tm_mday+1)%(nDaysInMonth(it.tm_year,it.tm_mon)+1);
        if (it.tm_mday == 0)
        {
          it.tm_mday++;
          it.tm_mon = (it.tm_mon+1)%12;
          if (it.tm_mon == 0) it.tm_year++;
        }
        it.tm_hour = 0; it.tm_min = 0; it.tm_sec = 0;
        itime = mktime(&it);
        daysinmonth = MAX((itime-starttime)/(3600*24),1);
        if (arraysize > 1) 
        {
          if (fmt <= CF_FLOAT) 
          {
            asize = arraysize;
          }
          else 
          {
            asize += MAX((arraysize * daysinmonth)/daystotal,1);
          }
          if (asize >= arraysize) finalPass = TRUE; /* just in case */
        }
      }
    }
    if (asize > arraysize) asize = arraysize;
    ps = loadLTSData(i,index,dstarttime,dendtime,lwrLimit,uprLimit,&data[ptr],fmt,asize-recsread,sample);
    if (asize == 1) recsfound += (int)ps; else recsread += (int)ps;
    ptr = recsread * recsize;
    if (finalPass) break;
    starttime = itime;
    dstarttime = (double)starttime;
    memcpy(&st,localtime(&starttime),sizeof(struct tm));
  }
  rnumber = recsfound > 0 ? recsfound : recsread;
  if (arraysize == 1) recsread = recsfound;

  assignNumberOfPoints(data,fmt,arraysize,&recsread,&recsfound,recsreadFlag);
  
  return rnumber;
}
int getValuesAsFloat(DTYPE *d,float *fval,short num)
{
  int i,n;

  if (d == NULL) return argument_list_error;
  if (d->dArrayLength < 1 || num < 1) return dimension_error;
  n = MIN((short)d->dArrayLength,num);

  switch (d->dFormat)
  {
    case CF_BYTE:   for (i=0; i<n; i++) fval[i] = (float)d->data.bptr[i];
                    return 0;
    case CF_SHORT:  for (i=0; i<n; i++) fval[i] = (float)d->data.sptr[i];
                    return 0;
    case CF_LONG:   for (i=0; i<n; i++) fval[i] = (float)d->data.lptr[i];
                    return 0;
    case CF_FLOAT:  memcpy(fval,d->data.fptr,n*sizeof(float));
                    return 0;
    case CF_DOUBLE: for (i=0; i<n; i++) fval[i] = (float)d->data.dptr[i];
                    return 0;
    default:        return illegal_format;
  }
}
int getValuesAsLong(DTYPE *d,UINT32 *lval,short num)
{
  int i,n;

  if (d == NULL) return argument_list_error;
  if (d->dArrayLength < 1 || num < 1) return dimension_error;
  n = MIN((short)d->dArrayLength,num);

  switch (d->dFormat)
  {
    case CF_BYTE:   for (i=0; i<n; i++) lval[i] = (UINT32)d->data.bptr[i];
                    return 0;
    case CF_SHORT:  for (i=0; i<n; i++) lval[i] = (UINT32)d->data.sptr[i];
                    return 0;
    case CF_LONG:   memcpy(lval,d->data.lptr,n*sizeof(UINT32));
                    return 0;
    case CF_FLOAT:  for (i=0; i<n; i++) lval[i] = (UINT32)d->data.fptr[i];
                    return 0;
    case CF_DOUBLE: for (i=0; i<n; i++) lval[i] = (UINT32)d->data.dptr[i];
                    return 0;
    default:        return illegal_format;
  }
}
int getValuesAsShort(DTYPE *d,UINT16 *sval,short num)
{
  int i,n;

  if (d == NULL) return argument_list_error;
  if (d->dArrayLength < 1 || num < 1) return dimension_error;
  n = MIN((short)d->dArrayLength,num);

  switch (d->dFormat)
  {
    case CF_BYTE:   for (i=0; i<n; i++) sval[i] = (UINT16)d->data.bptr[i];
                    return 0;
    case CF_LONG:   for (i=0; i<n; i++) sval[i] = (UINT16)d->data.lptr[i];
                    return 0;
    case CF_SHORT:   memcpy(sval,d->data.sptr,n*sizeof(UINT16));
                    return 0;
    case CF_FLOAT:  for (i=0; i<n; i++) sval[i] = (UINT16)d->data.fptr[i];
                    return 0;
    case CF_DOUBLE: for (i=0; i<n; i++) sval[i] = (UINT16)d->data.dptr[i];
                    return 0;
    default:        return illegal_format;
  }
}
int getValuesAsDouble(DTYPE *d,double *dval,short num)
{
  int i,n;

  if (d == NULL) return argument_list_error;
  if (d->dArrayLength < 1 || num < 1) return dimension_error;
  n = MIN((short)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_BYTE:   for (i=0; i<n; i++) dval[i] = (double)d->data.bptr[i];
                    return 0;
    case CF_SHORT:  for (i=0; i<n; i++) dval[i] = (double)d->data.sptr[i];
                    return 0;
    case CF_LONG:   for (i=0; i<n; i++) dval[i] = (double)d->data.lptr[i];
                    return 0;
    case CF_FLOAT:  for (i=0; i<n; i++) dval[i] = (double)d->data.fptr[i];
                    return 0;
    case CF_DOUBLE: memcpy(dval,d->data.dptr,n*sizeof(double));
                    return 0;
    default:        return illegal_format;
  }
}
int getValuesAsString(DTYPE *d,char *str,UINT32 *dsiz)
{
  int i,n,len, siz, cc = 0;
  int valueOnly = FALSE;

  if (str == NULL || dsiz == NULL) return un_allocated;
  if ((siz = *dsiz) < 16) return dimension_error;
  if (siz <= 64) valueOnly = TRUE;

  str[0] = 0; len = 0;
  n = (int)d->dArrayLength;
  switch (d->dFormat)
  {
    case CF_TEXT:
      len = MIN(siz,n);
      strncpy(d->data.cptr,str,len);
      ccerr(0);
    case CF_BYTE:   
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%d",d->data.bptr[i]); goto err; }
        sprintf(&str[len],"%d\n",d->data.bptr[i]);
        if ((len=strlen(str)) > siz + 16) break;
      }
      ccerr(0);
    case CF_SHORT:  
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%d",d->data.sptr[i]); goto err; }
        sprintf(&str[len],"%d\n",d->data.sptr[i]);
        if ((len=strlen(str)) > siz + 16) break;
      }
      ccerr(0);
    case CF_LONG:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%ld",d->data.lptr[i]); goto err; }
        sprintf(&str[len],"%ld\n",d->data.lptr[i]);
        if ((len=strlen(str)) > siz + 16) break;
      }
      ccerr(0);
    case CF_FLOAT:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%g",d->data.fptr[i]); goto err; }
        sprintf(&str[len],"%g\n",d->data.fptr[i]);
        if ((len=strlen(str)) > siz + 16) break;
      }
      ccerr(0);
    case CF_DOUBLE:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%g",d->data.dptr[i]); goto err; }
        sprintf(&str[len],"%g\n",d->data.dptr[i]);
        if ((len=strlen(str)) > siz + 16) break;
      }
      ccerr(0);
    case CF_NAME8:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME8 *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s\n",((NAME8 *)d->data.vptr)[i].name);
        if ((len=strlen(str)) > siz + 16) break;
      }
      ccerr(0);
    case CF_NAME16:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME16 *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s\n",((NAME16 *)d->data.vptr)[i].name);
        if ((len=strlen(str)) > siz + 32) break;
      }
      ccerr(0);
    case CF_NAME32:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME32 *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s\n",((NAME32 *)d->data.vptr)[i].name);
        if ((len=strlen(str)) > siz + 48) break;
      }
      ccerr(0);
    case CF_NAME48:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME48 *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s\n",((NAME48 *)d->data.vptr)[i].name);
        if ((len=strlen(str)) > siz + 64) break;
      }
      ccerr(0);
    case CF_NAME64:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME64 *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s\n",((NAME64 *)d->data.vptr)[i].name);
        if ((len=strlen(str)) > siz + 80) break;
      }
      ccerr(0);
    case CF_NAME16I:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME16I *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s %ld\n",((NAME16I *)d->data.vptr)[i].name,((NAME16I *)d->data.vptr)[i].ival);
        if ((len=strlen(str)) > siz + 64) break;
      }
      ccerr(0);
    case CF_NAME32I:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME32I *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s %ld\n",((NAME32I *)d->data.vptr)[i].name,((NAME32I *)d->data.vptr)[i].ival);
        if ((len=strlen(str)) > siz + 64) break;
      }
      ccerr(0);
    case CF_NAME48I:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME48I *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s %ld\n",((NAME48I *)d->data.vptr)[i].name,((NAME48I *)d->data.vptr)[i].ival);
        if ((len=strlen(str)) > siz + 64) break;
      }
      ccerr(0);
    case CF_NAME64I:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME64I *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s %ld\n",((NAME64I *)d->data.vptr)[i].name,((NAME64I *)d->data.vptr)[i].ival);
        if ((len=strlen(str)) > siz + 80) break;
      }
      ccerr(0);
    case CF_FLTINT:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) goto err;
        sprintf(&str[len],"%g %ld\n",((FLTINT *)d->data.vptr)[i].fval,((FLTINT *)d->data.vptr)[i].ival);
        if ((len=strlen(str)) > siz + 32) break;
      }
      ccerr(0);
    case CF_LNGINT:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) goto err;
        sprintf(&str[len],"%ld %ld\n",((INTINT *)d->data.vptr)[i].i1val,((INTINT *)d->data.vptr)[i].i2val);
        if ((len=strlen(str)) > siz + 32) break;
      }
      ccerr(0);
    case CF_INTFLTINT:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) goto err;
        sprintf(&str[len],"%ld %g %ld\n",((INTFLTINT *)d->data.vptr)[i].i1val,
          ((INTFLTINT *)d->data.vptr)[i].fval,((INTFLTINT *)d->data.vptr)[i].i2val);
        if ((len=strlen(str)) > siz + 32) break;
      }
      ccerr(0);
    case CF_USTRING:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) goto err;
        sprintf(&str[len],"%g %g %ld %s %s",((DUSTRING *)d->data.vptr)[i].f1val,
          ((DUSTRING *)d->data.vptr)[i].f2val,((DUSTRING *)d->data.vptr)[i].ival,
          ((DUSTRING *)d->data.vptr)[i].str,ctime(&((DUSTRING *)d->data.vptr)[i].tm));
        if ((len=strlen(str)) > siz + 128) break;
      }
      ccerr(0);
    default:
      return illegal_format;
  }
err:
  *dsiz = len;
  return 0;
}
