#include "project.def"
#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <malloc.h>
#include <sys\timeb.h>
#include <winsock.h>
#include "errors.h"
#include "database.h"
#include "winsrv.h"
#include "prolog.h"
#include "history.h"

extern IPXAddress IPXFecAddress;

__declspec(dllexport) short PASCAL setSrvUDPPort(int port)
{
  if (port < 1024) return not_allowed;
  gtUDPPort = port;
  return 0;
}
__declspec(dllexport) short PASCAL setSrvTCPPort(int port)
{
  if (port < 1024) return not_allowed;
  gtTCPPort = port;
  return 0;
}
__declspec(dllexport) short PASCAL setSrvGCastPort(int port)
{
  if (port < 1024) return not_allowed;
  gtGCastPort = port;
  return 0;
}
__declspec(dllexport) short PASCAL setSrvMCastPort(int port)
{
  if (port < 1024) return not_allowed;
  gtMCastPort = port;
  return 0;
}
__declspec(dllexport) short PASCAL setSrvMCastAddr(char *addr)
{
  if (addr == NULL || strncmp(addr,"238.",4)) return not_allowed;
  if ((gtMCastAddr=(char *)malloc(strlen(addr)+1)) == NULL) return out_of_local_memory;
  strcpy(gtMCastAddr,addr);
  return 0;
}
__declspec(dllexport) short PASCAL setSrvGCastAddr(char *addr)
{
  if (addr == NULL || strncmp(addr,"238.",4)) return not_allowed;
  if ((gtGCastAddr=(char *)malloc(strlen(addr)+1)) == NULL) return out_of_local_memory;
  strcpy(gtGCastAddr,addr);
  return 0;
}
__declspec(dllexport) short PASCAL setSrvMCastTTL(int ttl)
{
  if (ttl < 1 || ttl > 64) return not_allowed;
  gtMCastTTL = ttl;
  return 0;
}
__declspec(dllexport) void PASCAL setAppVersion(int major,int minor,int revision)
{
  sprintf(appmkversion,"%d.%02d.%04d",major,minor,revision);
  return;
}
__declspec(dllexport) void PASCAL setAppDate(time_t appdate)
{
  appmktime = appdate;
  return;
}
__declspec(dllexport) void PASCAL setAppDateFromString(char *datestr)
{
  struct tm it;
  char *c=datestr;
  int len; 

  if (c == NULL) return;
  len = strlen(datestr); // "DD-MM-YYYY hh:mm:ss"
  it.tm_mday = atoi(c); c += 3;
  it.tm_mon = atoi(c) - 1; c += 3;
  it.tm_year = atoi(c) - 1900; c += 5;
  it.tm_hour = atoi(c); c += 3;
  it.tm_min = atoi(c);  c += 3;
  it.tm_sec = atoi(c);
  it.tm_isdst = -1;
  appmktime = mktime(&it);
  return;
}
int getIPControlNets(char *fn,int *nr,struct sockaddr_in **list)
{
  int size,n;
  FILE *fp = NULL;
  static DWORD haddr;
  int cc=0,done=0,col;
  char str[256],hdr[256],scratch[17],ipnetsfilename[256];
  struct sockaddr_in *lst = NULL;

  n = 0;
  if ((lst=*list) != NULL) free((char *)lst);

  /* look for registerd properties: (must be a local file) */
  strcpy(ipnetsfilename,FecDBpath); strcat(ipnetsfilename,fn);

  if ((size=GetDataBaseSize(ipnetsfilename)) > 0)
  {
    lst = (struct sockaddr_in *)calloc(sizeof(struct sockaddr_in),size);
    if (lst == NULL) return out_of_server_memory;
    if ((fp=fopen(ipnetsfilename,"r")) == NULL) ccerr(no_such_file);
    for (done=0; fgets(str,255,fp) != NULL;)   /* scan file       */
    {
      if (strlen(str) == 0) continue;
      if (strchr("%;#\n",str[0])) continue;    /* comment line(s) */
      if (!done)
      {
        strcpy(hdr,str);
        if ((col=findcol(hdr,"SUBNET")) < 0) ccerr(no_such_line);
        done = 1;
        continue;
      }
      strncpy(scratch,colptr((short)col,str),16);
      haddr = inet_addr(scratch);
      lst[n].sin_family = AF_INET;
      lst[n].sin_port = htons((UINT16)gtGCastPort); /* potential broadcasting */
      memcpy(&lst[n].sin_addr,&haddr,4);
      if (++n > size) break;
    }
  }
err:
  if (fp) fclose(fp);
  *nr = n;
  *list = lst;
  return cc;
}
__declspec(dllexport) short PASCAL hasDevicePropertyList(char *eqm)
{
  ExportListStruct *el;

  if (eqm == NULL) return FALSE;
  for (el=ExportList; el != NULL; el=el->next) 
    if (!strncmp(eqm,el->EqpName,EQP_NAME_SIZE) && el->EqpModulePrpList != NULL)
      return TRUE;
  return FALSE;
}
ExportListStruct *getExportListItem(char *eqm)
{
  ExportListStruct *el;

  if (eqm == NULL) return NULL;
  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(eqm,el->EqpName,EQP_NAME_SIZE)) return el;
  return NULL;
}
__declspec(dllexport) short PASCAL JoinEquipmentGroup(char *eqpname,char *groupname,int groupindex)
{
  ExportListStruct *el = getExportListItem(eqpname);

  if (el == NULL) return non_existent_elem;
  if (groupname == NULL || strlen(groupname) == 0) return argument_list_error;
  if (groupindex < 0) return out_of_range;
  strncpy(el->group,groupname,16);
  el->groupIndex = groupindex;
  return 0;
}
short RegisterExport(char *expname,char *lclname,short numModules)
{
  ExportListStruct *el;
  short i,cc=0;

  if (expname == NULL || strlen(expname) == 0) ccerr(argument_list_error);
  for (el = ExportList; el; el = el->next) if (!strncmp(el->EqpName,lclname,EQP_NAME_SIZE)) break;
  if (el == NULL)  // wasn't in the list
  {
    if ((el=calloc(1,sizeof(ExportListStruct))) == NULL) ccerr(out_of_server_memory);
    el->next = ExportList;
    el->EqpNumProperties = 0;
    el->exportId = NrOfEqpTypes++;
    ExportList = el;
    ExportsRegistered = 0;
  }
  strncpy(el->tagName,expname,EXPORT_NAME_SIZE);
  strncpy(el->EqpName,lclname,EQP_NAME_SIZE);
  strncpy(el->Context,gDeviceContext,16);
  strncpy(el->EqpSubSystem,gDeviceSubSystem,EQP_SUBSYSTEM_SHORTTAG_SIZE);
  if ((i=getAliasTableIndexFromName(el->EqpName)) >= 0)
    strncpy(el->EqpAlias,gAliasTable[i].alias,EQP_NAME_SIZE);
  el->hEqpFcn = SearchFunction(el->EqpName);
  el->EqpNumModules = numModules;
  if (el->EqpModuleName == NULL &&
      (el->EqpModuleName = (NAME16 *)calloc(numModules,sizeof(NAME16))) == hNULL) 
    ccerr(out_of_server_memory);
  if (el->almLst == NULL &&
      (el->almLst=(ALARM **)calloc(el->EqpNumModules,sizeof(ALARM *))) == NULL) 
    ccerr(out_of_server_memory);
  getAlarmInfoTable(lclname);
err:
  return cc;
}
int joinEnsGroup(ExportListStruct *el)
{
  int cc=0;
  char str[256];
  NAME16 n16data[5];

  if (el == NULL) return non_existent_elem;
  strncpy(n16data[0].name,el->tagName,16); 
  strncpy(n16data[1].name,el->group,16); 
  strncpy(n16data[2].name,el->Context,16); 
  sprintf(n16data[3].name,"%d",el->groupIndex); 
  strncpy(n16data[4].name,el->EqpSubSystem,16); 

  cc = tExecLink(1,"GRPEQM","#0","JOIN",-5,(char *)n16data,CF_NAME16,500);
	if (cc != 0) 
  {
    sprintf(str,"%s joining group %s failed : %s",el->tagName,el->group,erlst[cc & 0xfff]);
    feclog(str);
  }
  return cc;
}
short sndRegisteredExports(void)
{
  short cc=0;
  time_t t = time(NULL);
  static time_t lastAttempt = 0;
  RPCDataStruct *rpcdata;
  RPCFecStruct  *rpcfec;
  static BYTE fecAddressBuf[sizeof(RPCDataStruct)+sizeof(RPCFecStruct)+sizeof(RPCFecExStruct)];
  ExportListStruct *el;
  char s[256];

  if (!FecNameResolved)
  {
    dbllog("Fec Name not resolved");
    ExportsRegistered = 1;
    return cc;
  }
  if ((cc=GetNameServerAddress()) != 0)
  {
    // Name Server Address not available, so don't try again
    dbllog("Cound not resolve name server address");
    ExportsRegistered = 1;
    return cc;
  }
  if (t < lastAttempt + 30) return 0;
  lastAttempt = t;
  rpcdata = (RPCDataStruct *)fecAddressBuf;
  rpcfec = (RPCFecStruct *)&fecAddressBuf[RPCDATASTRUCT_SIZE];

  for (el = ExportList; el; el = el->next)
  {
    if (el->registered) continue;
    memset(rpcfec,0,sizeof(RPCFecStruct));
    strncpy(rpcfec->Name,gFecName,FEC_NAME_SIZE);
    memcpy(rpcfec->Network,&IPXFecAddress,12);
    rpcfec->PortOffset = (BYTE)PortOffset;
    rpcfec->Port[0] = PortOffset&0xff;
    rpcfec->Port[1] = (PortOffset>>8)&0xff;
    memset(rpcdata,0,sizeof(RPCDataStruct));
    strncpy(rpcdata->FecName,gFecName,FEC_NAME_SIZE);
    strncpy(rpcdata->FullName,el->tagName,EXPORT_NAME_SIZE);
    strncpy(rpcdata->EqName,el->EqpName,EQP_NAME_SIZE);
    strncpy(rpcdata->Context,el->Context,16);
    if (strlen(el->EqpSubSystem) == 0) strncpy(el->EqpSubSystem,gDeviceSubSystem,EQP_SUBSYSTEM_SIZE);
    strncpy(rpcdata->SubSystem,el->EqpSubSystem,EQP_SUBSYSTEM_SHORTTAG_SIZE);
    rpcdata->alarmSystem = el->alarmSystem;
    memcpy(&fecAddressBuf[RPCDATASTRUCT_SIZE+RPCFECSTRUCT_SIZE],&RPCFecEx,sizeof(RPCFecExStruct));
    cc = tExecLink(0,CSHOSTS,gDeviceContext,"EXPORT",-(short)(RPCFECSTRUCT_SIZE+RPCDATASTRUCT_SIZE+sizeof(RPCFecExStruct)),(char *)fecAddressBuf,CF_STRUCT,1);
    if (cc) ccerr(cc);
    el->registered = REGISTERED_ENS;
    sprintf(s,"FEC %s, Server %s (%s) registered with equipment name server",
      gFecName,el->tagName,el->EqpName);
    dbllog(s);
    sprintf(s,"Context is %s",strlen(gDeviceContext) > 0 ? gDeviceContext : "not given");
    dbllog(s);
  }
  for (el = ExportList; el; el = el->next)
  {
    if (el->registered != REGISTERED_ENS) continue;
    if (strlen(el->group) > 0) 
    {
      sprintf(s,"joining group %s",el->group);
      dbllog(s);
      cc = joinEnsGroup(el);
      if (cc) ccerr(cc);
    }
    el->registered = REGISTERED_ALL;
  }
  ExportsRegistered = 1;
err:
  return cc;
}
ExportListStruct *findTagID(char *nam)
{
  ExportListStruct *el;
  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(nam,el->EqpName,EQP_NAME_SIZE)) break;
  return el;
}
__declspec(dllexport) short RegisterModuleName(char *eqname,char *devname,short devnr)
{
  HashXRefStruct *xref;
  ExportListStruct *el = findTagID(eqname);
  char lcldevname[17];
  int cc = 0,idx;

  if (el == NULL) return non_existent_elem;

  if (devnr < 0 || devnr >= el->EqpNumModules) return out_of_range;
  strncpy(el->EqpModuleName[devnr].name,devname,16);
  strncpy(lcldevname,devname,16); lcldevname[16] = 0;

  idx = (int)(ElfHash(lcldevname) % MODULE_HASH_SIZE);
  for (xref=el->eqpModuleXRef[idx]; xref != NULL && xref->index != devnr; xref=xref->next);

  if (xref == NULL) /* it's a new one*/
  {
    if ((xref=calloc(1,sizeof(HashXRefStruct))) == NULL) ccerr(out_of_server_memory);
    xref->index = devnr;
    xref->next = el->eqpModuleXRef[idx];
    el->eqpModuleXRef[idx] = xref;
  }
err:
  return cc;
}

__declspec(dllexport) short GetDeviceNumberEx(char *eqname,char *devname,char *prpname)
{
  int idx;
  HashXRefStruct *xref;
  ExportListStruct *el = findTagID(eqname);
  char lcldevname[17];
  ExportPropertyListStruct *prp;

  if (prpname != NULL && (prp=AssignDeviceListFromProperty(eqname,prpname)) != NULL)
  {
    // TODO: replace this too with a hash table ...
    for (idx=0; idx<prp->numDevices; idx++) if (!strncmp(devname,prp->devNames[idx].name,16)) return idx;
  }

  if (el == NULL) return -non_existent_elem;
  strncpy(lcldevname,devname,16); lcldevname[16] = 0;
  if (lcldevname[0] == '#') return atoi(&lcldevname[1]);

  idx = (int)(ElfHash(lcldevname) % MODULE_HASH_SIZE);
  for (xref=el->eqpModuleXRef[idx]; xref != NULL; xref=xref->next)
  {
    if (!strncmp(devname,el->EqpModuleName[xref->index].name,16)) return xref->index;
  }
  return -device_not_connected;
}
__declspec(dllexport) short GetDeviceNumber(char *eqname,char *devname)
{
  return GetDeviceNumberEx(eqname,devname,NULL);
}
__declspec(dllexport) void GetDeviceName(char *eqname,int devnr,char *devname)
{
  ExportListStruct *el = findTagID(eqname);
  static char lclname[16];
  if (el == NULL) return;
  if (devnr < 0 || devnr >= el->EqpNumModules) return;
  if (strlen(el->EqpModuleName[devnr].name) != 0) strncpy(devname,el->EqpModuleName[devnr].name,16);
  else
  {
    sprintf(lclname,"#%d",devnr);
    strncpy(devname,lclname,16);
  }
  return;
}
void ParseRedirector(char *str,PrpRedirBlk *prd)
{
  char *c = str, *sstr = NULL, *pstr = str, *dstr = NULL;
  //char s[132];

  if (str == NULL) return;
  if (str[0] == '/')                                             /* context */
  {
    if ((c=strchr(++c,'/')) == NULL) return;/* device server must be there !*/
    c++;
  }
  sstr = c;                                                /* device server */
  if ((c=strchr(sstr,'/')) != NULL)
  {
    *c++ = 0; dstr = c;                                      /* device name */
  }
  else
  {
    c = sstr;                                       // put it back !
  }
  if ((pstr=strtok(c,"[")) != NULL) pstr = strtok(NULL,"]");
  strncpy(prd->prpRedirection[0].name,sstr,16);
  if (pstr != NULL) strncpy(prd->prpRedirection[1].name,pstr,16); /* redirected property (note 16-char len) */
  if (dstr != NULL) strncpy(prd->prpRedirection[2].name,dstr,16); /* device name */
}
void ParsePropertyDescription(char *str,ExportPropertyListStruct *prp)
{
  char *c=NULL, *c1, tmp[64], *s = str;

  if (str == NULL) return;
  if ((c=strchr(str,'[')) != NULL)
  {
    if ((c1=strchr(++c,']')) != NULL)               /* 1st token: egu */
    {
      *c1++ = 0;
      str = c1;         
      if ((c1=strstr(c,"!")) != NULL)                  /* graph type */
      {
        *c1++ = 0; prp->egu.graph = getGraphType(c1);
      }
      if ((c1=strchr(c,':')) != NULL)               /* prp egu given */
      {
        c1++;
        prp->egu.min = (float)strtod(c,&c); c = c1;
        prp->egu.max = (float)strtod(c,&c); c++;
        if (c != str)
        {
          strncpy(tmp,c,64);
          strncpy(prp->egu.units,tmp,16);
        }
      }
    }
  }
  if (str != NULL && (c=strchr(str,'[')) != NULL)
  {
    if ((c1=strchr(++c,']')) != NULL)             /* 2nd token: x egu */
    {
      *c1++ = 0;
      str = c1;
      if (!stricmp(c,"BOOLEAN")) prp->prpArrayType |= AT_BOOLEAN;
      if (!stricmp(c,"CHANNEL")) prp->prpArrayType |= AT_CHANNEL;
      else if (!stricmp(c,"STRUCTURED")) prp->prpArrayType |= AT_STRUCTURED;
      else if (!stricmp(c,"IMAGE")) prp->prpArrayType |= AT_IMAGE;
      else prp->prpArrayType |= AT_SPECTRUM;
      if ((c1=strstr(c,"!")) != NULL)                  /* graph type */
      {
        *c1++ = 0; prp->xegu.graph = getGraphType(c1);
      }
      if ((c1=strchr(c,':')) != NULL)                 /* x egu given */
      {
        c1++;
        prp->xegu.min = (float)strtod(c,&c); c = c1;
        prp->xegu.max = (float)strtod(c,&c); c++;
        if (c != str)
        {
          strncpy(tmp,c,64);
          strncpy(prp->xegu.units,tmp,16);
        }
      }
    }
  }
  if (str == NULL) str = s;
  strncpy(prp->prpDescription,str,64);  /* 3rd token: description part */
}
int GetPrpId(char *eqm,char *prpName)
{
  ExportListStruct *el = findTagID(eqm);
  ExportPropertyListStruct *prp = NULL;
  int idx;

  if (el == NULL || prpName == NULL) return -1;
  idx = (int)(ElfHash((unsigned char *)prpName) % PROPERTY_HASH_SIZE);

  for (prp=el->EqpProperty[idx]; prp; prp=prp->next)
  {
    if (!strncmp(prp->prpName,prpName,32)) return prp->prpId;
  }
  return -1;
}
// when enabled go thru list for el->hasPropertyDeviceNames == TRUE
// call AssignDeviceListFromProperty if necessary ...
// for getDeviceNumberEx, see if there's a dev list for the property ...
CONTRACT gLclContract;
ExportPropertyListStruct *AssignDeviceListFromProperty(char *eqm,char *prp)
{
  int cc = 0,pid,lid,nid;
  ExportListStruct *el;
  ExportPropertyListStruct *pl;
  char namprp[64];
  static int inside = 0;

  if ((el=findTagID(eqm)) == NULL) return NULL;
  if (el->hEqpFcn == NULL) return NULL;
  if (!el->hasPropertyDeviceNames) return NULL;  // just added this P.D. 6.2.04
  if ((pid=GetPrpId(eqm,prp)) < 0) return NULL;
  if (strlen(prp) > 28) return NULL;
  pid = PROPERTY_HASH_INDEX(pid);
  for (pl=el->EqpProperty[pid]; pl != NULL && strncmp(pl->prpName,prp,32); pl=pl->next);
  if (pl == NULL) return NULL;
  if (pl->devNames == NULL && !pl->hasQueriedForDeviceNames)
  {
    if (inside) return NULL;
    inside = TRUE;
    strncpy(namprp,prp,28); namprp[28] = 0; strcat(namprp,".NAM");
    if ((nid=GetPrpId(eqm,namprp)) < 0) ccerr(illegal_property);
    pl->hasQueriedForDeviceNames = TRUE;
    lid = (long)MAKEPROTID(pid,NAM_PROTOCOL);
    sprintf(gLclContract.EqpProperty,"%s.NAM",prp);
    strncpy(gLclContract.EqpName,eqm,6);
    strcpy(gLclContract.EqpDeviceName,"#0");
    gLclContract.EqpAccess = CA_READ;
    gLclContract.EqpFormatOut = (BYTE)BFMT(CF_NAME16);
    gLclContract.EqpFormatIn = (BYTE)BFMT(CF_NULL);
    gLclContract.EqpSizeIn = 0;
    gLclContract.EqpSizeOut = pl->prpSizeOut;
    strcpy(clientName,"LOCAL INIT"); strcpy(clientaddr,"THIS STATION");
    interimStatus = not_posted;  // let the VB app. clear this
    SendMessage(el->hEqpFcn,WM_SYSCOMMAND,lid,NAM_PROTOCOL);
    if (interimStatus) ccerr(interimStatus);
    if (gLclContract.EqpSizeOut == 0) ccerr(illegal_data_size);
    if ((pl->devNames=(NAME16 *)calloc(gLclContract.EqpSizeOut,16)) == NULL) ccerr(out_of_server_memory);
    memcpy(pl->devNames,WorkArea,gLclContract.EqpSizeOut*16);
    pl->numDevices = gLclContract.EqpSizeOut;
  }
err:
  inside = FALSE;
  return cc ? NULL : pl;
}

__declspec(dllexport) int PASCAL AssignPropertyList(char *eqm,char *devname,char *listname,int listsize,NAME32 *list)
{
  int i,cc = 0;
  ExportListStruct *el;
  if (eqm == NULL || devname == NULL || listname == NULL || list == NULL) ccerr(argument_list_error);
  for (el=ExportList; el != NULL; el=el->next)
    if (!strncmp(eqm,el->EqpName,EQP_NAME_SIZE)) break;
  if (el == NULL) ccerr(non_existent_elem);
  if ((i=GetDeviceNumberEx(eqm,devname,NULL)) < 0 || i >= el->EqpNumModules) ccerr(device_not_connected);
  if (el->EqpModulePrpList == NULL)
  {
    if ((el->EqpModulePrpList=(EqpModulePropertyList *)SystemCalloc(el->EqpNumModules,sizeof(EqpModulePropertyList))) == NULL) 
      ccerr(out_of_server_memory);
  }
  if (el->EqpModulePrpList[i].prpList == NULL && 
      (el->EqpModulePrpList[i].prpList=(NAME32 *)SystemCalloc(listsize,sizeof(NAME32))) == NULL)
    ccerr(out_of_server_memory);
  strncpy(el->EqpModulePrpList[i].filename,listname,64);
  memcpy(el->EqpModulePrpList[i].prpList,list,listsize*sizeof(NAME32));
  el->EqpModulePrpList[i].prpListSize = listsize;

err:
  return cc;
}
__declspec(dllexport) short PASCAL RedirectDeviceName(char *eqm,char *devname,char *rdr)
{
  int i,cc = 0;
  ExportListStruct *el;
  PrpRedirBlk prd;
  if (eqm == NULL || devname == NULL || rdr == NULL) ccerr(argument_list_error);
  for (el=ExportList; el != NULL; el=el->next)
    if (!strncmp(eqm,el->EqpName,EQP_NAME_SIZE)) break;
  if (el == NULL) ccerr(non_existent_elem);
  if ((i=GetDeviceNumberEx(eqm,devname,NULL)) < 0 || i >= el->EqpNumModules) ccerr(device_not_connected);
  if (el->EqpModuleRdr == NULL)
  {
    if ((el->EqpModuleRdr=(NAME16 *)calloc(el->EqpNumModules,sizeof(NAME16))) == NULL) ccerr(out_of_server_memory);
  }
  ParseRedirector(rdr,&prd); /* only the device server name is relevant */
  strncpy(el->EqpModuleRdr[i].name,prd.prpRedirection[0].name,16); 
err:
  return (short)cc;
}
__declspec(dllexport) short PASCAL RedirectProperty(char *eqm,char *prop,char *rdr) 
{
  char strbuf[128];
  ExportPropertyListStruct *prp;
  PrpRedirBlk *prd;
  if (rdr == NULL || strlen(rdr) == 0) return argument_list_error; // must be given !
  if ((prp=GetProperyListStruct(eqm,prop)) == NULL) return illegal_property;
  if ((prd=(PrpRedirBlk *)calloc(1,sizeof(PrpRedirBlk))) == NULL) return out_of_server_memory;
  prd->next = prp->prd;
  prp->prd = prd;
  strncpy(strbuf,rdr,127);
  ParseRedirector(strbuf,prp->prd);
  return 0;
}

__declspec(dllexport) int RegisterPropertyInformation(char *eqm,char *prop,DTYPE *dout,DTYPE *din,short acc,short atype,UINT16 rowlength,char *dsc,int propId,char *rdr) 
{
  int i,ssize,idx,cc=0,isNew=0;
  long prpId = propId, prpId0 = -1;
  ExportPropertyListStruct *prp = NULL;
  ExportListStruct *el;
  PrpRedirBlk *prd;
  char strbuf[128];

  if (eqm == NULL || prop == NULL) ccerr(argument_list_error);
  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(eqm,el->EqpName,EQP_NAME_SIZE)) break;
  if (el == NULL) ccerr(non_existent_elem);
  if ((i=strlen(prop)) > 4 && !strcmp(&prop[i-4],".NAM")) 
  {
    el->hasPropertyDeviceNames = TRUE;
    acquireDeviceNamesFromProperties = TRUE;
  }
  if (strchr(prop,'*')) idx = PROPERTY_HASH_SIZE;
  else idx = (int)(ElfHash((unsigned char *)prop) % PROPERTY_HASH_SIZE);

  for (prp=el->EqpProperty[idx]; prp; prp=prp->next) if (!strncmp(prp->prpName,prop,EQP_PROP_SIZE)) break;
  if (prp != NULL) 
  {
    prpId0 = prp->prpId;
    if (dout != NULL)
    {
      if (dout->dArrayLength != prp->prpSizeOut || dout->dFormat != prp->prpFormatOut) isNew = -1;
    }
    else
    {
      if (prp->prpSizeOut > 0) isNew = -1;
    }
    if (din != NULL)
    {
      if (din->dArrayLength != prp->prpSizeIn || din->dFormat != prp->prpFormatIn) isNew = -1;
    }
    else
    {
      if (prp->prpSizeIn > 0) isNew = -1;
    }
    if (acc != prp->prpAccessMode) isNew = -1;
  }
  else
  {
    isNew = -1;
  }

  if (!isNew) ccerr(0);

  if ((prp=(ExportPropertyListStruct *)calloc(1,sizeof(ExportPropertyListStruct))) == NULL) ccerr(out_of_server_memory);
  strncpy(prp->prpName,prop,EQP_PROP_SIZE);
  if ((i=getAliasTableIndexFromName(prop)) >= 0)
    strncpy(prp->prpAlias,gAliasTable[i].alias,EQP_PROP_SIZE);
  prp->next = el->EqpProperty[idx];
  el->EqpProperty[idx] = prp;
  if (prpId < 0) prpId = el->EqpNumProperties;
  prp->prpId = prpId0 >= 0 ? prpId0 : MAKE_PROPERTY_ID(prpId,idx);
  el->EqpNumProperties++;

  if (dout != NULL)
  {
    prp->prpSizeOut = dout->dArrayLength;
    prp->prpFormatOut = dout->dFormat; 
    prp->prpArrayType = atype;
    if (rowlength > 0)
    {
      prp->rowSize = rowlength;
      prp->numRows = (UINT16)(prp->prpSizeOut/rowlength);
    }
    else
    {
      prp->rowSize = (UINT16)prp->prpSizeOut;
      prp->numRows = 1;
    }
    if (!(prp->prpArrayType & AT_DOUBLE)) prp->prpArrayType |= prp->prpSizeOut > 1 ? AT_SINGLE : AT_SCALAR;
    if (strlen(dout->dTag)) 
    {
      strncpy(prp->strTagOut,dout->dTag,8);
      if (dout->dFormat == CF_STRUCT)
      {
        if (getStructFromTag(dout->dTag) == NULL) ccerr(invalid_structure_tag);
        if ((ssize=GETSRVSTRUCTSIZE(dout->dTag)) == 0) ccerr(un_allocated);
        prp->prpSizeOut *= ssize;
        prp->structSizeOut = ssize;
      }
    }
  }
  if (din != NULL)
  {
    prp->prpSizeIn = din->dArrayLength;
    prp->prpFormatIn = din->dFormat;
    if (strlen(din->dTag)) 
    {
      strncpy(prp->strTagIn,din->dTag,8);
      if (din->dFormat == CF_STRUCT)
      {
        if (getStructFromTag(din->dTag) == NULL) ccerr(invalid_structure_tag);
        if ((ssize=GETSRVSTRUCTSIZE(din->dTag)) == 0) ccerr(un_allocated);
        prp->prpSizeIn *= ssize;
        prp->structSizeIn = ssize;
      }
    }
  }
  prp->prpAccessMode = acc;
  strncpy(strbuf,dsc,127); strbuf[127] = 0;
  ParsePropertyDescription(strbuf,prp);
  if (rdr != NULL && strlen(rdr) != 0)
  {
    if ((prd=(PrpRedirBlk *)SystemCalloc(1,sizeof(PrpRedirBlk))) == NULL) ccerr(out_of_server_memory);
    prd->next = prp->prd;
    prp->prd = prd;
    strncpy(strbuf,rdr,127);
    ParseRedirector(strbuf,prp->prd);
  }
  if ((prp->devNames=GetEqpModuleNames(NULL,prp->prpName,prp->prpSizeOut,NULL,NULL)) != NULL)
  {
    prp->numDevices = prp->prpSizeOut;
    el->hasPropertyDeviceNames = TRUE;
    prp->hasQueriedForDeviceNames = TRUE;
  }
err:
  return cc ? -cc : (int)PROPERTY_NUM_INDEX(prp->prpId);
}
__declspec(dllexport) short RegisterPropertyEx(char *eqm,char *prop,long sizIn,short fmtIn,char *tagIn,
    									   long sizOut,short fmtOut,char *tagOut,short acc,char *dsc)
{
  DTYPE dout,din,*dinPtr = NULL;
  dout.dArrayLength = sizOut;
  dout.dFormat = fmtOut;
  strncpy(dout.dTag,tagOut,8);
//  if (acc & CA_WRITE)
//  {
    din.dArrayLength = sizIn;
    din.dFormat = fmtIn;
    strncpy(din.dTag,tagIn,8);
    dinPtr = &din;
//  }
  return (short)RegisterPropertyInformation(eqm,prop,&dout,dinPtr,acc,AT_UNKNOWN,0,dsc,-1,NULL);
}
__declspec(dllexport) short RegisterProperty(char *eqm,char *prop,short siz,short fmt,short acc,char *dsc) /* Appends property list for eqm */
{
  return RegisterPropertyEx(eqm,prop,siz,fmt,"",siz,fmt,"",acc,dsc);
}

__declspec(dllexport) short PASCAL GetRegisteredPropertyList(char *eqm,NAME32 *prpNames,int *nprps)
{
  ExportListStruct *el;
  ExportPropertyListStruct *prp;
  int n,l,idx;

  if (eqm == NULL ||
      prpNames == NULL ||
      nprps == NULL) return argument_list_error;
  n = *nprps;
  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strncmp(eqm,el->EqpName,EQP_NAME_SIZE)) continue;
    for (l=0,idx=0; idx<PROPERTY_HASH_SIZE+1 && l<n; idx++)
      for (prp=el->EqpProperty[idx]; prp && l<n; prp=prp->next)
      {
        if (l > 0 && !strncmp(prpNames[l-1].name,prp->prpName,32)) continue;
        strncpy(prpNames[l].name,prp->prpName,32);
        l++;
      }
  }
  *nprps = l;
  return 0;
}
__declspec(dllexport) short PASCAL GetRegisteredPropertyListStruct(char *eqm,char *prpName,ExportPropertyListStruct *prpLstStruct)
{
  if (eqm == NULL ||
      prpName == NULL ||
      prpLstStruct == NULL) return argument_list_error;
  memcpy(prpLstStruct,GetProperyListStruct(eqm,prpName),sizeof(ExportPropertyListStruct));
  return 0;
}
ExportPropertyListStruct *GetProperyListStruct(char *eqm,char *prpName)
{
  ExportListStruct *el = findTagID(eqm);
  ExportPropertyListStruct *prp = NULL;
  int idx;
  if (el == NULL || prpName == NULL) return NULL;
  idx = (int)(ElfHash((unsigned char *)prpName) % PROPERTY_HASH_SIZE);
  for (prp=el->EqpProperty[idx]; prp && strncmp(prp->prpName,prpName,32); prp=prp->next);
  return prp;
}
int isChannelArray(char *eqm,char *prpName)
{
  ExportPropertyListStruct *prp = GetProperyListStruct(eqm,prpName);
  if (prp == NULL || !(prp->prpArrayType & AT_CHANNEL)) return FALSE;
  return TRUE;
}
int isSpectrumArray(char *eqm,char *prpName)
{
  ExportPropertyListStruct *prp = GetProperyListStruct(eqm,prpName);
  if (prp == NULL || !(prp->prpArrayType & AT_SPECTRUM)) return FALSE;
  if (prp->prpSizeOut <= 1) return FALSE;
  return TRUE;
}
void getWildcardTokens(char *wildcardString,char *beginToken,char *endToken,char *midToken)
{
  char *s,*c,lclStr[128];

  if (wildcardString == NULL || beginToken == NULL || endToken == NULL || midToken == NULL) return;
  strncpy(lclStr,wildcardString,32); lclStr[32] = 0;
  s = lclStr;
  beginToken[0] = endToken[0] = midToken[0] = 0; /* start from scratch */
  if (*s == '*') 
  { /* no begin token */
    while (*s == '*') s++;
    if ((c=strchr(s,'*')) != NULL) 
    { /* there's a midToken ("*abc*") */
      *c = 0;
      strncpy(midToken,s,32);
    }
    else
    { /* there's an end token */
      strncpy(endToken,s,32);
    }
  }
  else
  {
    strncpy(beginToken,s,32);
    if ((c=strchr(beginToken,'*')) != NULL) 
    { /* found begin Token end */
      *c++ = 0;
      s = c; /* re-position (could still be an end token */
      while (*s == '*') s++; /* is it "abc*xyz" ? */
      if (strlen(s) > 0) strncpy(endToken,s,32);
    }
  }
}
int wildcardMatches(char *devName,char *beginToken,int beginLen,char *endToken,int endLen,char *midToken,int midLen)
{
  int dlen;
  if (devName == NULL || beginToken == NULL || endToken == NULL || midToken == NULL) return FALSE;
  if ((dlen=strlen(devName)) == 0) return FALSE;
  if (beginLen > 0)
  { /* check against begin token */
    if (strncmp(devName,beginToken,beginLen)) return FALSE;
  }
  if (endLen > 0)
  { /* check against end token */
    if (strncmp(&devName[dlen-endLen],endToken,endLen)) return FALSE;
  }
  if (midLen > 0)
  { /* check against mid token */
    if (strstr(devName,midToken) == NULL) return FALSE;
  }
  return TRUE;
}
int isContiguous(char *eqm,char *devName,NAME16 *devlist,int *start,int *len)
{ /* called in the case of local multi-channel arrays */
  int i, n, devlistsize = 0, inSequence = FALSE, finSequence = FALSE;
  char beginToken[32], endToken[32], midToken[32];
  int blen = 0, elen = 0, mlen = 0;
  ExportListStruct *el=findTagID(eqm);

  if (el == NULL || devlist == NULL || start == NULL || len == NULL) return FALSE;

  if (!strcmp(devName,"*"))  
  { /* channel array with '*' is continguous */
    *start = 0;
    return TRUE;
  }
  getWildcardTokens(devName,beginToken,endToken,midToken);

  /* calculate these once outside the loop ... */
  blen = strlen(beginToken);
  mlen = strlen(midToken);
  elen = strlen(endToken);

  for (i=0,n=0; i<*len; i++)
  {
    if (!wildcardMatches(devlist[i].name,beginToken,blen,endToken,elen,midToken,mlen))
    {
      if (inSequence) finSequence = TRUE;
      continue;
    }
    if (finSequence) return FALSE; /* found another match -> not contiguous ! */
    if (!inSequence) *start = i;
    inSequence = TRUE;
    n++;
  }
  *len = n;
  return inSequence;
}
int isLocal(char *eqm,char *devName,char *prpName)
{ /* not a registered multi-channel array */
  ExportListStruct *el=findTagID(eqm);
  ExportPropertyListStruct *prp = GetProperyListStruct(eqm,prpName);
  PrpRedirBlk *prd = NULL;
  char beginToken[32], endToken[32], midToken[32];
  int i, blen = 0, elen = 0, mlen = 0;

  if (el == NULL || prp == NULL) return FALSE;
  getWildcardTokens(devName,beginToken,endToken,midToken);
  /* calculate these once outside the loop ... */
  blen = strlen(beginToken);
  mlen = strlen(midToken);
  elen = strlen(endToken);
  if (el->EqpModuleRdr != NULL) 
  { /* has device redirection list, check it */
    for (i=0; i<el->EqpNumModules; i++)
    {
      if (strlen(el->EqpModuleRdr[i].name) > 0 &&
          wildcardMatches(el->EqpModuleName[i].name,beginToken,blen,endToken,elen,midToken,mlen))
        return FALSE;
    }
  }
  if (prp->prd != NULL)
  { /* has property redirector, check it */
    for (prd=prp->prd; prd != NULL; prd=prd->next)
    {
      if (wildcardMatches(prd->prpRedirection[2].name,beginToken,blen,endToken,elen,midToken,mlen))
        return FALSE;
    }
  }
  return TRUE; /* appears to be local so-far */
}
__declspec(dllexport) short GetPropertyId(char *eqm,char *prpName)
{
  ExportListStruct *el = findTagID(eqm);
  ExportPropertyListStruct *prp = NULL;
  int idx;

  if (el == NULL || prpName == NULL) return -1;
  idx = (int)(ElfHash((unsigned char *)prpName) % PROPERTY_HASH_SIZE);

  for (prp=el->EqpProperty[idx]; prp; prp=prp->next)
  {
    if (!strncmp(prp->prpName,prpName,32)) return PROPERTY_NUM_INDEX(prp->prpId);
  }
  return -1;
}
__declspec(dllexport) short PASCAL GetNumberRegisteredDevices(char *eqm)
{
  ExportListStruct *el = findTagID(eqm);
  if (el == NULL) return -argument_list_error;
  return el->EqpNumModules;
}
short GetPropertyOutputFormat(char *eqm,char *prpName)
{
  ExportListStruct *el = findTagID(eqm);
  ExportPropertyListStruct *prp = NULL;
  int idx;

  if (el == NULL || prpName == NULL) return -1;
  idx = (int)(ElfHash((unsigned char *)prpName) % PROPERTY_HASH_SIZE);

  for (prp=el->EqpProperty[idx]; prp; prp=prp->next)
  {
    if (!strncmp(prp->prpName,prpName,32)) return prp->prpFormatOut;
  }
  return -1;
}
