#include "project.def"
#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <malloc.h>
#include <sys\timeb.h>
#include <winsock.h>
#include "errors.h"
#include "database.h"
#include "winsrv.h"
#include "prolog.h"
#include "history.h"

char *getCaller(void)
{
  static char name[48];
  strncpy(name,clientName,16);
  return name;
}
int getNumContracts(char *eqname)
{
  int i,n=0;
  if (eqname == NULL || strlen(eqname) == 0) return ncontract;
  for (i=0; i<ncontract; i++)
   if (!strncmp(ContractList[i]->contract.EqpName,eqname,EQP_NAME_SIZE)) n++;
  return n;
}
int getNumConsumers(char *eqname)
{
  int i,n=0;
  if (eqname == NULL || strlen(eqname) == 0) return nconsumer;
  for (i=0; i<ncontract; i++)
   if (!strncmp(ContractList[i]->contract.EqpName,eqname,EQP_NAME_SIZE) &&
       ContractList[i]->nconsumer > n) n = ContractList[i]->nconsumer;
  return n;
}
int getNumCallers(char *eqname)
{
  return getNumConsumers(eqname);
}
int historyRecordDepth(char *prpName,UINT16 *shortDepth,UINT16 *longDepth)
{
  int i;
  for (i=0; i<nHistoryRecords && strncmp(prpName,historyRecords[i].c.EqpProperty,32); i++);
  if (i < nHistoryRecords) 
  {
    if (shortDepth != NULL) *shortDepth = historyRecords[i].depthShort;
    if (longDepth != NULL) *longDepth = historyRecords[i].depthLong;
    return historyRecords[i].depthShort;
  }
  return 0;
}
int getLastAccessList(CONTRACT *con,char *dataIn,char *data)
{
  int slen = 0;
  switch (LFMT(con->EqpFormatOut))
  {
    case CF_NAME16:slen = 16; break;
    case CF_NAME32:slen = 32; break;
    default: return illegal_format;
  }
  switch (con->EqpSizeOut)
  {
    case 5: strncpy(data + slen*4,getDataTimeString(gLastWriteAccessTime,-1),slen);
    case 4: strncpy(data + slen*3,gLastWriteAccessDev,slen);
    case 3: strncpy(data + slen*2,gLastWriteAccessPrp,slen);
    case 2: strncpy(data + slen*1,gLastWriteAccessAddr,16);
    case 1: strncpy(data,gLastWriteAccessUsr,16);
      break;
    default:
      return dimension_error;
  }
  return 0;
}
int propsQuery(CONTRACT *con,char *dataIn,char *data)
{
  int n,idx,jdx,j,k,devnr=-1,cc=0;
  UINT32 l,len;
  short fmt;
  ExportPropertyListStruct *prp;
  ExportListStruct *el;
  PropertyQueryStructEx *pqsx = NULL;
  PropertyQueryStruct *pqs = NULL;
  XPropertyQueryStruct *xpqs = NULL;
  int tgtprplen = 0,useAliasName=FALSE,uselist=FALSE,hasdevgroup=FALSE;
  char tgtprp[33], *bp=NULL, *mp=NULL;
  char btok[32], etok[32], mtok[32];
  int blen = 0, elen = 0, mlen = 0, checkWildCard = FALSE;

  if (con->EqpSizeOut == 0) return illegal_read_write;

  if (strcmp(con->EqpProperty,"*") && strchr(con->EqpProperty,'*') != NULL)
  { /* property names query with non-trivial wildcard string */
    getWildcardTokens(con->EqpProperty,btok,etok,mtok);
    blen = strlen(btok); mlen = strlen(mtok); elen = strlen(etok);
    checkWildCard = TRUE;
  }

  if ((len=con->EqpSizeIn) > 0)
  {
    memset(tgtprp,0,33);
    switch (LFMT(con->EqpFormatIn))
    {
      case CF_TEXT:
        if (len < 1 || len > 32) return illegal_read_write;
        strncpy(tgtprp,dataIn,len);
        break;
      case CF_NAME16:
        if (len != 1) return illegal_read_write;
        strncpy(tgtprp,((NAME16 *)dataIn)->name,16);
        break;
      case CF_NAME32:
        if (len != 1) return illegal_read_write;
        strncpy(tgtprp,((NAME32 *)dataIn)->name,32);
        break;
      default:  return illegal_format;
    }
    if (nAliasTableEntries > 0) /* is the incoming name an alias ? */
    {
      if ((j=getAliasTableIndex(tgtprp)) >= 0) 
      {
        if (!IsMetaProperty(gAliasTable[j].name,&bp,&mp)) bp = gAliasTable[j].name;
        strncpy(tgtprp,bp,EQP_PROP_SIZE);
        useAliasName = TRUE;
        jdx = j;
      }
    }
    tgtprplen = strlen(tgtprp);
  }
  switch (fmt=LFMT(con->EqpFormatOut))
  {
    case CF_STRUCT:
      if (con->EqpSizeOut%XPROPERTYQUERYSTRUCT_SIZE == 0 && !strncmp(con->strTagOut,"XPQS",8))
      {
        n = con->EqpSizeOut/XPROPERTYQUERYSTRUCT_SIZE;
        xpqs = (XPropertyQueryStruct *)data;
        for (el=ExportList; el != NULL; el=el->next)
        {
          if (strncmp(con->EqpName,el->EqpName,EQP_NAME_SIZE)) continue;
          if (strlen(el->group) > 0) hasdevgroup = TRUE;
          if (el->EqpModulePrpList != NULL) uselist = TRUE;
          devnr = GetDeviceNumber(el->EqpName,con->EqpDeviceName);
          for (l=0,idx=0; idx<PROPERTY_HASH_SIZE+1 && (int)l<n; idx++)
	          for (prp=el->EqpProperty[idx]; prp && (int)l<n; prp=prp->next)
          {
            if (el->EqpModulePrpList != NULL && devnr >=0 && el->EqpModulePrpList[devnr].prpListSize > 0) 
            { /* there's a device-specific property list */
              for (k=0; k<el->EqpModulePrpList[devnr].prpListSize; k++) 
                if (!strncmp(el->EqpModulePrpList[devnr].prpList[k].name,prp->prpName,EQP_PROP_SIZE)) break;
              if (k == el->EqpModulePrpList[devnr].prpListSize) continue; /* not found in list */
            }
            if (checkWildCard && !wildcardMatches(prp->prpName,btok,blen,etok,elen,mtok,mlen)) continue;
            if (tgtprplen && strncmp(prp->prpName,tgtprp,32)) continue;
            strncpy(xpqs[l].prpName,prp->prpName,32);
            if (useAliasName) strncpy(xpqs[l].prpName,gAliasTable[jdx].alias,32);
            strncpy(xpqs[l].prpDescription,prp->prpDescription,64);
            if (mp != NULL && strlen(xpqs[l].prpDescription) < 48) 
            {
              strcat(xpqs[l].prpDescription," ");
              strncat(xpqs[l].prpDescription,mp,15);
            }
            xpqs[l].prpSize = prp->prpSizeOut;
            xpqs[l].prpSizeIn = prp->prpSizeIn;
            xpqs[l].prpFormat = (BYTE)BFMT(prp->prpFormatOut);
            xpqs[l].prpFormatIn = (BYTE)BFMT(prp->prpFormatIn);
            strncpy(xpqs[l].prpTag,prp->strTagOut,8);
            strncpy(xpqs[l].prpTagIn,prp->strTagIn,8);
            if (prp->structSizeOut > 0 && prp->strTagOut[0] != 0) xpqs[l].prpSize /= prp->structSizeOut;
            if (prp->structSizeIn > 0 && prp->strTagIn[0] != 0) xpqs[l].prpSizeIn /= prp->structSizeIn;
            xpqs[l].prpGraphType = prp->egu.graph;
            xpqs[l].prpMaxValue = (float)prp->egu.max;
            xpqs[l].prpMinValue = (float)prp->egu.min;
            if (prp->prd != NULL)
              memcpy(xpqs[l].prpRedirection,prp->prd->prpRedirection->name,32);
            strncpy(xpqs[l].prpUnits,prp->egu.units,16);
            xpqs[l].prpNumOverloads++;
            xpqs[l].prpAccess = (BYTE)prp->prpAccessMode;
            historyRecordDepth(prp->prpName,&xpqs[l].prpHistoryDepthShort,&xpqs[l].prpHistoryDepthLong);
            xpqs[l].prpArrayType = prp->prpArrayType;
            xpqs[l].numRows = prp->numRows;
            xpqs[l].rowSize = prp->rowSize;
            xpqs[l].rngMaxValue = (float)prp->xegu.max;
            xpqs[l].rngMinValue = (float)prp->xegu.min;
            strncpy(xpqs[l].rngUnits,prp->xegu.units,16);
            l++;
          }
          if (tgtprplen) for (j=0; j<(int)l; j++) xpqs[j].prpNumOverloads = l;
          if ((int)l < n) con->EqpSizeOut = l*XPROPERTYQUERYSTRUCT_SIZE;
          return l == 0 ? non_existent : 0;
        }
      }
      if (con->EqpSizeOut%PROPERTYQUERYSTRUCTEX_SIZE == 0)
      {
        n = con->EqpSizeOut/PROPERTYQUERYSTRUCTEX_SIZE;
        pqsx = (PropertyQueryStructEx *)data;
        for (el=ExportList; el != NULL; el=el->next)
        {
          if (strncmp(con->EqpName,el->EqpName,EQP_NAME_SIZE)) continue;
          if (strlen(el->group) > 0) hasdevgroup = TRUE;
          if (el->EqpModulePrpList != NULL) uselist = TRUE;
          devnr = GetDeviceNumber(el->EqpName,con->EqpDeviceName);
          for (l=0,idx=0; idx<PROPERTY_HASH_SIZE+1 && (int)l<n; idx++)
	          for (prp=el->EqpProperty[idx]; prp && (int)l<n; prp=prp->next)
          {
            if (el->EqpModulePrpList != NULL && devnr >=0 && el->EqpModulePrpList[devnr].prpListSize > 0) 
            { /* there's a device-specific property list */
              for (k=0; k<el->EqpModulePrpList[devnr].prpListSize; k++) 
                if (!strncmp(el->EqpModulePrpList[devnr].prpList[k].name,prp->prpName,EQP_PROP_SIZE)) break;
              if (k == el->EqpModulePrpList[devnr].prpListSize) continue; /* not found in list */
            }
            if (checkWildCard && !wildcardMatches(prp->prpName,btok,blen,etok,elen,mtok,mlen)) continue;
            if (tgtprplen && strncmp(prp->prpName,tgtprp,32)) continue;
            strncpy(pqsx[l].prpName,prp->prpName,32);
            if (useAliasName) strncpy(pqsx[l].prpName,gAliasTable[jdx].alias,32);
            strncpy(pqsx[l].prpDescription,prp->prpDescription,64);
            pqsx[l].prpSize = prp->prpSizeOut;
            pqsx[l].prpSizeIn = prp->prpSizeIn;
            pqsx[l].prpFormat = (BYTE)BFMT(prp->prpFormatOut);
            pqsx[l].prpFormatIn = (BYTE)BFMT(prp->prpFormatIn);
            strncpy(pqsx[l].prpTag,prp->strTagOut,8);
            strncpy(pqsx[l].prpTagIn,prp->strTagIn,8);
            if (prp->structSizeOut > 0 && prp->strTagOut[0] != 0) pqsx[l].prpSize /= prp->structSizeOut;
            if (prp->structSizeIn > 0 && prp->strTagIn[0] != 0) pqsx[l].prpSizeIn /= prp->structSizeIn;
            pqsx[l].prpGraphType = prp->egu.graph;
            pqsx[l].prpMaxValue = (float)prp->egu.max;
            pqsx[l].prpMinValue = (float)prp->egu.min;
            if (prp->prd != NULL)
              memcpy(pqsx[l].prpRedirection,prp->prd->prpRedirection->name,32);
            strncpy(pqsx[l].prpUnits,prp->egu.units,16);
            pqsx[l].prpNumOverloads++;
            pqsx[l].prpAccess = (BYTE)prp->prpAccessMode;
            historyRecordDepth(prp->prpName,&pqsx[l].prpHistoryDepthShort,&pqsx[l].prpHistoryDepthLong);
            l++;
          }
          if (tgtprplen) for (j=0; j<(int)l; j++) pqsx[j].prpNumOverloads = l;
          if ((int)l < n) con->EqpSizeOut = l*PROPERTYQUERYSTRUCTEX_SIZE;
          return l == 0 ? non_existent : 0;
        }
      }
      if (con->EqpSizeOut%PROPERTYQUERYSTRUCT_SIZE == 0)
      {
        n = con->EqpSizeOut/PROPERTYQUERYSTRUCT_SIZE;
        pqs = (PropertyQueryStruct *)data;
        for (el=ExportList; el != NULL; el=el->next)
        {
          if (strlen(el->group) > 0) hasdevgroup = TRUE;
          if (el->EqpModulePrpList != NULL) uselist = TRUE;
          devnr = GetDeviceNumber(el->EqpName,con->EqpDeviceName);
          if (strncmp(con->EqpName,el->EqpName,EQP_NAME_SIZE)) continue;
          for (l=0,idx=0; idx<PROPERTY_HASH_SIZE+1 && (int)l<n; idx++)
	        for (prp=el->EqpProperty[idx]; prp && (int)l<n; prp=prp->next)
          {
            if (el->EqpModulePrpList != NULL && devnr >=0 && el->EqpModulePrpList[devnr].prpListSize > 0) 
            { /* there's a device-specific property list */
              for (k=0; k<el->EqpModulePrpList[devnr].prpListSize; k++) 
                if (!strncmp(el->EqpModulePrpList[devnr].prpList[k].name,prp->prpName,EQP_PROP_SIZE)) break;
              if (k == el->EqpModulePrpList[devnr].prpListSize) continue; /* not found in list */
            }
            if (checkWildCard && !wildcardMatches(prp->prpName,btok,blen,etok,elen,mtok,mlen)) continue;
            if (tgtprplen && strncmp(prp->prpName,tgtprp,32)) continue;
            strncpy(pqs[l].prpName,prp->prpName,32);
            if (useAliasName) strncpy(pqs[l].prpName,prp->prpAlias,32);
            if (useAliasName) strncpy(pqs[l].prpName,gAliasTable[jdx].alias,32);
            strncpy(pqs[l].prpDescription,prp->prpDescription,32);
            pqs[l].prpSize = (UINT16)prp->prpSizeOut;
            pqs[l].prpFormat = (BYTE)BFMT(prp->prpFormatOut);
            pqs[l].prpAccess = (BYTE)prp->prpAccessMode;
            l++;
          }
          if ((int)l < n) con->EqpSizeOut = l*PROPERTYQUERYSTRUCT_SIZE;
          return l == 0 ? non_existent : 0;
        }
      }
      return non_existent_elem;
    case CF_NAME32:
      n = con->EqpSizeOut;
      for (el=ExportList; el != NULL; el=el->next)
      {
        if (strncmp(con->EqpName,el->EqpName,EQP_NAME_SIZE)) continue;
        if (strlen(el->group) > 0) hasdevgroup = TRUE;
        if (el->EqpModulePrpList != NULL) uselist = TRUE;
        devnr = GetDeviceNumber(el->EqpName,con->EqpDeviceName);
        for (l=0,idx=0; idx<PROPERTY_HASH_SIZE+1 && (int)l<n; idx++)
        {
          for (prp=el->EqpProperty[idx]; prp && (int)l<n; prp=prp->next)
          {
            if (el->EqpModulePrpList != NULL && devnr >=0 && el->EqpModulePrpList[devnr].prpListSize > 0) 
            { /* there's a device-specific property list */
              for (k=0; k<el->EqpModulePrpList[devnr].prpListSize; k++) 
                if (!strncmp(el->EqpModulePrpList[devnr].prpList[k].name,prp->prpName,EQP_PROP_SIZE)) break;
              if (k == el->EqpModulePrpList[devnr].prpListSize) continue; /* not found in list */
            }
            if (checkWildCard && !wildcardMatches(prp->prpName,btok,blen,etok,elen,mtok,mlen)) continue;
            if (tgtprplen && strncmp(prp->prpName,tgtprp,32)) continue;
            if (l > 0 && !strncmp(((NAME32 *)data)[l-1].name,prp->prpName,32)) continue;
            if (l > 1 && !strncmp(((NAME32 *)data)[l-2].name,prp->prpName,32)) continue;
            strncpy(((NAME32 *)data)[l].name,prp->prpName,32);
            l++;
            if (prp->prpAlias[0] != 0)
            {
              strncpy(((NAME32 *)data)[l].name,prp->prpAlias,32);
              l++;
            }
          }
        }
        if (uselist || hasdevgroup) cc = has_query_function | CE_SENDDATA;
        if ((int)l < n) con->EqpSizeOut = l;
        return l == 0 ? non_existent : cc;
      }
      return non_existent_elem;
    case CF_USTRING:
      n = con->EqpSizeOut;
      for (el=ExportList; el != NULL; el=el->next)
      {
        if (strncmp(con->EqpName,el->EqpName,EQP_NAME_SIZE)) continue;
        if (strlen(el->group) > 0) hasdevgroup = TRUE;
        if (el->EqpModulePrpList != NULL) uselist = TRUE;
        devnr = GetDeviceNumber(el->EqpName,con->EqpDeviceName);
        for (l=0,idx=0; idx<PROPERTY_HASH_SIZE+1 && (int)l<n; idx++)
        {
          for (prp=el->EqpProperty[idx]; prp && (int)l<n; prp=prp->next)
          {
            if (el->EqpModulePrpList != NULL && devnr >=0 && el->EqpModulePrpList[devnr].prpListSize > 0) 
            { /* there's a device-specific property list */
              for (k=0; k<el->EqpModulePrpList[devnr].prpListSize; k++) 
                if (!strncmp(el->EqpModulePrpList[devnr].prpList[k].name,prp->prpName,EQP_PROP_SIZE)) break;
              if (k == el->EqpModulePrpList[devnr].prpListSize) continue; /* not found in list */
            }
            if (checkWildCard && !wildcardMatches(prp->prpName,btok,blen,etok,elen,mtok,mlen)) continue;
            if (tgtprplen && strncmp(prp->prpName,tgtprp,32)) continue;
            if (l > 0 && !strncmp(((DUSTRING *)data)[l-1].str,prp->prpName,32)) continue;
            if (l > 1 && !strncmp(((DUSTRING *)data)[l-2].str,prp->prpName,32)) continue;
            strncpy(((DUSTRING *)data)[l].str,prp->prpName,32);
            strncpy(&((DUSTRING *)data)[l].str[32],prp->prpDescription,48);
            ((DUSTRING *)data)[l].ival = (UINT32)prp->prpSizeOut;
            ((DUSTRING *)data)[l].tm = (UINT32)prp->prpFormatOut;
            l++;
            if (prp->prpAlias[0] != 0)
            {
              strncpy(((DUSTRING *)data)[l].str,prp->prpAlias,32);
              strncpy(&((DUSTRING *)data)[l].str[32],prp->prpDescription,48);
              ((DUSTRING *)data)[l].ival = (UINT32)prp->prpSizeOut;
              ((DUSTRING *)data)[l].tm = (UINT32)prp->prpFormatOut;
              l++;
            }
          }
        }
        if (uselist || hasdevgroup) cc = has_query_function | CE_SENDDATA;
        if ((int)l < n) con->EqpSizeOut = l;
        return l == 0 ? non_existent : cc;
      }
      return non_existent_elem;
    default:
      return non_existent_elem;
  }
  return illegal_format;
}
int ndeviceQuery(CONTRACT *con,char *dataIn,char *data)
{
  ExportListStruct *el;
  if (con->EqpSizeOut != 1) return illegal_read_write;
  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strncmp(con->EqpName,el->EqpName,EQP_NAME_SIZE)) continue;
    switch (LFMT(con->EqpFormatOut))
    {
      case CF_SHORT:
        *(short *)data = el->EqpNumModules; break;
      case CF_LONG:
        *(int *)data = (int)el->EqpNumModules; break;
      case CF_FLOAT:
        *(float *)data = (float)el->EqpNumModules; break;
      case CF_DOUBLE:
        *(double *)data = (double)el->EqpNumModules; break;
      default:
        return illegal_format;
    }
    return 0;
  }
  return non_existent_elem;
}
int deviceQuery(CONTRACT *con,char *dataIn,char *data)
{
  int k,i,slen,start,cc = 0;
  ExportListStruct *el;
  char btok[32], etok[32], mtok[32];
  int blen = 0, elen = 0, mlen = 0, checkWildCard = FALSE;

  if (con->EqpSizeOut == 0) return illegal_read_write;
  start = con->EqpDeviceName[0] == '#' ? atoi(&con->EqpDeviceName[1])-1 : 0;
  if (start < 0) start = 0;

  if (strcmp(con->EqpDeviceName,"*") && strchr(con->EqpDeviceName,'*') != NULL)
  { /* device query with non-trivial wildcard string */
    getWildcardTokens(con->EqpDeviceName,btok,etok,mtok);
    blen = strlen(btok); mlen = strlen(mtok); elen = strlen(etok);
    checkWildCard = TRUE;
  }

  switch (LFMT(con->EqpFormatOut))
  {
    case CF_NAME8: 
    case CF_NAME16:
    case CF_NAME32:slen = fmtsizeof(con->EqpFormatOut); break;
    default: return illegal_format;
  }
  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(con->EqpName,el->EqpName,EQP_NAME_SIZE)) break;
  if (el == NULL) return non_existent_elem;
  for (i=0,k=0; i<(int)con->EqpSizeOut && i+start<el->EqpNumModules; i++)
  {
    if (el->EqpModuleName && strlen(el->EqpModuleName[i+start].name))
    {
      if (checkWildCard && 
          !wildcardMatches(el->EqpModuleName[i+start].name,btok,blen,etok,elen,mtok,mlen))
        continue;
      strncpy(&data[k*slen],el->EqpModuleName[i+start].name,slen);
    }
    else 
    {
      if (checkWildCard) continue;
      sprintf(&data[k*slen],"#%d",i);
    }
    k++;
  }
  if (el->hasPropertyDeviceNames) 
  {
    cc = has_query_function | CE_SENDDATA;
  }
  if (k < (int)con->EqpSizeOut) con->EqpSizeOut = k;
  return cc;
}
int npropsQuery(CONTRACT *con,char *dataIn,char *data)
{
  ExportListStruct *el;

  if (con->EqpSizeOut != 1) return illegal_read_write;
  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strncmp(con->EqpName,el->EqpName,EQP_NAME_SIZE)) continue;
    switch (LFMT(con->EqpFormatOut))
    {
      case CF_SHORT:
        *(short *)data = el->EqpNumProperties; break;
      case CF_LONG:
        *(int *)data = (int)el->EqpNumProperties; break;
      case CF_FLOAT:
        *(float *)data = (float)el->EqpNumProperties; break;
      case CF_DOUBLE:
        *(double *)data = (double)el->EqpNumProperties; break;
      default:
        return illegal_format;
    }
    return 0;
  }
  return non_existent_elem;
}
int alarmsQuery(CONTRACT *con,char *dataIn,char *data)
{
  AMS *ams = (AMS *)data;
  UINT16 n;
  int start;
  if (LFMT(con->EqpFormatOut) != CF_STRUCT) return illegal_format;
  if (con->EqpSizeOut%AMS_SIZE) return illegal_read_write;
  start = (short)(con->EqpDeviceName[0] == '#' ? atoi(&con->EqpDeviceName[1]) : 0);
  n = (UINT16)con->EqpSizeOut/AMS_SIZE;
  getAlarms(con->EqpName,ams,n,(short)start);
  return 0;
}
int alarmsQueryEx(CONTRACT *con,char *dataIn,char *data)
{
  AMS *ams = (AMS *)data;
  AMRB *amrb = (AMRB *)dataIn;
  int n,start;
  int size = (int)con->EqpSizeOut;
  int sizeIn = (int)con->EqpSizeIn;

  if (LFMT(con->EqpFormatIn) != CF_STRUCT || LFMT(con->EqpFormatOut) != CF_STRUCT)
    return illegal_format;
  if (sizeIn != AMRB_SIZE) return dimension_error;
  if (size%AMS_SIZE) return illegal_read_write;
  if ((n=size/AMS_SIZE) == 0) return buffer_too_small;
  start = con->EqpDeviceName[0] == '#' ? atoi(&con->EqpDeviceName[1]) : 0;
  getAlarmsEx(con->EqpName,(AMS *)ams,(short)n,(short)amrb->almMinSeverity,
              (time_t)amrb->almStartTime,(time_t)amrb->almStopTime,(short)start);
  return 0;
}
int nalarmsQuery(CONTRACT *con,char *dataIn,char *data)
{
  short sc,ntim,nsev;

  if (LFMT(con->EqpFormatOut) != CF_LONG) return illegal_format;
  switch (con->EqpSizeOut)
  {
    case 3:
      *(UINT32 *)data = numAlarmsEx(con->EqpName,&((time_t *)data)[1],&sc,&ntim,&nsev);
      ((UINT32 *)data)[2] = (UINT32)sc;
      break;
    case 5:
      *(UINT32 *)data = numAlarmsEx(con->EqpName,&((time_t *)data)[1],&sc,&ntim,&nsev);
      ((UINT32 *)data)[2] = (UINT32)sc;
      ((UINT32 *)data)[3] = (UINT32)ntim;
      ((UINT32 *)data)[4] = (UINT32)nsev;
      break;
    default: return illegal_read_write;
  }
  if (!gRequireReadFromCAS) 
  {
    if ((gRequireReadFromCAS=isCASInUserList()) != FALSE)
    {
      feclog("ALM: Central Alarm Server attached");
    }
  }
  return 0;
}
int alarmDefsQuery(CONTRACT *con,char *dataIn,char *data)
{
  ADS *ads = (ADS *)data;
  UINT16 n;
  if (LFMT(con->EqpFormatOut) != CF_STRUCT) return illegal_format;
  if (con->EqpSizeOut%ADS_SIZE) return illegal_read_write;
  n = (UINT16)con->EqpSizeOut/ADS_SIZE;
  getAlarmDefs(con->EqpName,ads,n);
  return 0;
}
int nalarmDefsQuery(CONTRACT *con,char *dataIn,char *data)
{
  if (con->EqpSizeOut != 1) return illegal_read_write;
  if (LFMT(con->EqpFormatOut) != CF_SHORT) return illegal_format;
  *(short *)data = numAlarmDefs(con->EqpName);
  return 0;
}
int msgDisplay(CONTRACT *con,char *data,char *dataOut)
{
  time_t t = time(NULL);
  int len = (int)con->EqpSizeIn;
  if (len == 0) return illegal_read_write;
  if (LFMT(con->EqpFormatIn) != CF_TEXT) return illegal_format;
  feclog(data);
  strncpy(MSGLastMessageString,data,256);
  isNewMessage = TRUE;
  return 0;
}
int activityQuery(CONTRACT *con,char *dataIn,char *data)
{
  ActivityQueryStruct aqs;

  if (LFMT(con->EqpFormatOut) != CF_STRUCT) return illegal_format;
  if (con->EqpSizeOut != ACTIVITYQUERYSTRUCT_SIZE) return illegal_read_write;

  memset(aqs.reserved,0,4);
  strncpy(aqs.FecName,gFecName,FEC_NAME_SIZE);
  aqs.localtime = time(NULL);
  aqs.starttime = startuptime;
  aqs.systemPollingRate = SysPoll;
  aqs.numBkgTsks = 0;
  aqs.numTotalContracts = ncontract - nExpiredContracts;
  aqs.numTotalClients = nconsumer;
  aqs.numTargetContracts = getNumContracts(con->EqpName);
  aqs.numTargetClients = getNumConsumers(con->EqpName);
  aqs.numConnections = 0;
  aqs.numConnectionTimeouts = 0;
  aqs.numConnectionArrivals = 0;
  aqs.numUDPpkts = UDPcount;
  aqs.numTCPpkts = TCPcount;
  aqs.numIPXpkts = IPXcount;
  aqs.numSPXpkts = SPXcount;
  memcpy(data,&aqs,sizeof(ActivityQueryStruct));
  return 0;
}
int contractsQuery(CONTRACT *con,char *dataIn,char *data)
{
  ContractQueryStruct *cqs = (ContractQueryStruct *)data;
  ContractListStruct *cl;
  int i, ncon;

  if (LFMT(con->EqpFormatOut) != CF_STRUCT) return illegal_format;
  if (con->EqpSizeOut%CONTRACT_QUERY_SIZE) return illegal_read_write;
  if ((ncon=con->EqpSizeOut/CONTRACT_QUERY_SIZE) == 0) return dimension_error;

  for (i=0; i<ncon && i<ncontract; i++)
  {
    cl = ContractList[i];
    memcpy(&cqs[i].contract,&cl->contract,sizeof(CONTRACT));
    cqs[i].pollrate = cl->PollingRate;
    cqs[i].nconsumers = cl->nconsumer;
  }
  return 0;
}
int clientsQuery(CONTRACT *con,char *dataIn,char *data)
{
  ClientQueryStruct *cqs = (ClientQueryStruct *)data;
  CLN *cln;
  int i, ncln;

  if (LFMT(con->EqpFormatOut) != CF_STRUCT) return illegal_format;
  if (con->EqpSizeOut%CLIENT_QUERY_SIZE) return illegal_read_write;
  if ((ncln=con->EqpSizeOut/CLIENT_QUERY_SIZE) == 0) return dimension_error;

  for (i=0; i<ncln && i<nconsumer; i++)
  {
    if (Consumers[i] == NULL) continue;
    cln = Consumers[i];
    memcpy(&cqs[i].cln,cln,sizeof(CLNQS));
  }
  return 0;
}
int getFecLogFile(CONTRACT *con,char *dataIn,char *data)
{
  char fn[128];
  int slen, cc;

  if (con->EqpDeviceName[0] != '#') 
  {
    strncpy(fn,con->EqpDeviceName,16); fn[16] = 0;
  }
  if (con->EqpSizeIn > 0)
  {
    if (con->EqpAccess & CA_WRITE)
    { /* write the text file submitted to disk */
      if (con->EqpSizeIn == 0) return illegal_read_write;
      if ((cc=writeFecLogFile(fn,dataIn,(int)con->EqpSizeIn)) != 0) return cc;
    }
    else
    { /* alternate way of supplying a (long) file name */
      switch (LFMT(con->EqpFormatIn))
      {
        case CF_NAME8:
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          slen = fmtsizeof(con->EqpFormatIn);
          break;
        case CF_TEXT:
          slen = con->EqpSizeIn;
          break;
        default: return illegal_format;
      }
      strncpy(fn,dataIn,slen); fn[slen] = 0;
    }
  }
  if (con->EqpSizeOut > 0)
  {
    if (LFMT(con->EqpFormatOut) != CF_TEXT) return illegal_format;
    if ((cc=readFecLogFile(fn,data,(int)con->EqpSizeOut)) != 0) return cc;
    con->EqpSizeOut = strlen(data);
  }
  return 0;
}
int getFecLogFileList(CONTRACT *con,char *dataIn,char *data)
{
  int cc, n, slen = 0;
  char path[256];
  if (con->EqpSizeOut > 0)
  {
    switch (LFMT(con->EqpFormatIn))
    {
      case CF_NAME8:
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME48:
      case CF_NAME64:
        slen = fmtsizeof(con->EqpFormatIn);
        break;
      case CF_TEXT:
        slen = con->EqpSizeIn;
        break;
      default: 
        break;
    }
    if (slen > 0) strncpy(path,dataIn,slen); 
    path[slen] = 0; 
    applyFecLogPath(path,128);
    if (LFMT(con->EqpFormatOut) != CF_NAME64) return illegal_format;
    n = (int)con->EqpSizeOut;
    if ((cc=scanFecLogFiles(path,NULL,(NAME64 *)data,&n,0)) != 0) return cc;
    con->EqpSizeOut = (UINT32)n;
  }
  return 0;
}
int getFecLog(CONTRACT *con,char *dataIn,char *data)
{
  if (con->EqpSizeOut > 0)
  {
    if (LFMT(con->EqpFormatOut) != CF_TEXT) return illegal_format;
    readFecLog(data,con->EqpSizeOut);
  }
  return 0;
}
int getSelfTestFile(CONTRACT *con,char *dataIn,char *data)
{
  int bytesread;
  if (con->EqpSizeOut > 0)
  {
    if (LFMT(con->EqpFormatOut) != CF_TEXT) return illegal_format;
    bytesread = readSelfTest(con->EqpName,data,(int)con->EqpSizeOut);
    if (bytesread == 0) return no_such_file;
    con->EqpSizeOut = bytesread;
  }
  return 0;
}
int setLogDepth(CONTRACT *con,char *dataIn,char *data)
{
  if (con->EqpSizeIn > 0)
  {
    if (!(con->EqpAccess&CA_WRITE)) return illegal_read_write;
    if (LFMT(con->EqpFormatIn) != CF_SHORT) return illegal_format;
    FeclogDepth = *(short *)dataIn;
  }
  if (con->EqpSizeOut > 0)
  {
    if (LFMT(con->EqpFormatOut) != CF_SHORT) return illegal_format;
    *(short *)data = (short)FeclogDepth;
  }
  return 0;
}
int setLogCommands(CONTRACT *con,char *dataIn,char *data)
{
  if (con->EqpSizeIn > 0)
  {
    if (!(con->EqpAccess&CA_WRITE)) return illegal_read_write;
    if (LFMT(con->EqpFormatIn) != CF_SHORT) return illegal_format;
    putCommandsInFeclog = *(short *)dataIn;
  }
  if (con->EqpSizeOut > 0)
  {
    if (LFMT(con->EqpFormatOut) != CF_SHORT) return illegal_format;
    *(short *)data = (short)putCommandsInFeclog;
  }
  return 0;
}
int setDebugLevel(CONTRACT *con,char *dataIn,char *data)
{
  short level;
  
  if (con->EqpSizeIn > 0)
  {
    if (!con->EqpAccess&CA_WRITE) return illegal_read_write;
    if (con->EqpSizeIn != 1) return dimension_error;
    if (LFMT(con->EqpFormatIn) != CF_SHORT) return illegal_format;
    level = *(short *)dataIn;
    DebugServer(level);
  }
  if (con->EqpSizeOut > 0)
  {
    if (LFMT(con->EqpFormatOut) != CF_SHORT) return illegal_format;
    *(short *)data = (short)debug;
  }
  return 0;
}
int getDebugBuffer(CONTRACT *con,char *dataIn,char *data)
{
  if (con->EqpSizeOut != BUFLEN+26) return illegal_read_write;
  if (LFMT(con->EqpFormatOut) != CF_TEXT) return illegal_format;
  memcpy(data,txtbuffer,BUFLEN+26);
  return 0;
}
int lockAccess(CONTRACT *con,char *dataIn,char *data)
{
  char s[256];
  short sval[2];
  int actionTaken = FALSE;
  int passed = FALSE;
  CLN *cln = EQPconsumerlist->client; 

  if (cln == NULL) return code_failure;
  if (con->EqpAccess&CA_WRITE)
  {
    if (con->EqpSizeIn != 2) return illegal_read_write;
    if (LFMT(con->EqpFormatIn) != CF_SHORT) return illegal_format;
    sval[0] = ((short *)dataIn)[0];
    sval[1] = ((short *)dataIn)[1];
    if (!putCommandsInFeclog)
    {
        sprintf(s,"ACCESSLOCK %d called from %s",sval[0],getCaller());
        feclog(s);
    }
    if (sval[0] < 0 || sval[0] > LOCK_ABORT) return out_of_range;
    if (IsMemberLockSet(cln)) passed = TRUE;
    else if (currentAccessLock.lockType != LOCK_PERSISTENT) passed = TRUE;
    else if (sval[0] == LOCK_ABORT) passed = TRUE;
    if (!passed) return access_denied;
    if (sval[0] == LOCK_ABORT || sval[0] == LOCK_UNLOCKED)
    {
      memset(&currentAccessLock,0,sizeof(currentAccessLock));
    }
    else
    {
      memcpy(&currentAccessLock.cln,cln,sizeof(CLN));
      currentAccessLock.lockType = sval[0];
      currentAccessLock.lockDuration = sval[1];
      currentAccessLock.lockStart = time(NULL);
    }
    actionTaken = TRUE;
  } 
  if (con->EqpSizeOut == 2)
  { 
    if (LFMT(con->EqpFormatOut) != CF_NAME32) return illegal_format;
    strncpy(data,currentAccessLock.cln.userName,16);
	strncpy(&data[32],inet_ntoa(currentAccessLock.cln.IPaddress.sin_addr),32);
    actionTaken = TRUE;
  }
    
  return actionTaken ? 0 : illegal_read_write;
}
__declspec(dllexport) short PASCAL GetLocalHistoryData(
      char *eqp,char *prp, char *dev,
      DTYPE *dout,DTYPE *din)
{
  CONTRACT cn;
  strncpy(cn.EqpName,eqp,EQP_NAME_SIZE);
  strncpy(cn.EqpProperty,prp,EQP_PROP_SIZE);
  strncpy(cn.EqpDeviceName,dev,EQP_DEVICE_SIZE);
  cn.EqpSizeOut = dout->dArrayLength;
  cn.EqpFormatOut = BFMT(dout->dFormat);
  cn.EqpSizeIn = din->dArrayLength;
  cn.EqpFormatIn = BFMT(din->dFormat);
  return HISTQueryFunction(&cn,din->data.cptr,dout->data.cptr);
}
int HISTQueryFunction(CONTRACT *con,char *dataIn,char *data)
{
  long numread, nr;
  char *c, Property[32];
  time_t stop=time(NULL)-1; /* default values */
  double dstop = (double)stop, dstart = 0;
  float lwr=0,upr=0;
  int n,fmt,cc=0,index=0,sample=0,recsreadFlag=0,isSnapshot=0;
  long *lptr;
  double *dptr;
  BYTE *dataptr = (BYTE *)data;

  strncpy(Property,con->EqpProperty,32);
  if ((c=strstr(Property,".AR")) != NULL)  
  {
    *c = 0;
    return getRemoteHistoryServer(con->EqpName,Property);
  }
  if (strstr(Property,".HIST@") != NULL || strstr(Property,".HST@") != NULL)
    isSnapshot = TRUE;
  if ((c=strstr(Property,".HIST")) == NULL &&
      (c=strstr(Property,".HST")) == NULL) return illegal_property;
  *c = 0;
  if (con->EqpSizeOut == 0) return illegal_read_write;
  n = (int)con->EqpSizeOut;
  switch (fmt = LFMT(con->EqpFormatOut))
  {
    case CF_DOUBLE: case CF_FLOAT: case CF_LONG: case CF_INTFLTINT:
      recsreadFlag = 0;
      break;
    default:
      break;
  }
  if (con->EqpSizeIn > 0)
  {
    switch (LFMT(con->EqpFormatIn))
    {
      case CF_LONG:
        lptr = (long *)dataIn;
        switch (con->EqpSizeIn)
        {
          case 5: recsreadFlag = (int)lptr[4];
          case 4: sample = (int)lptr[3];
          case 3: index = (int)lptr[2];
          case 2: if (lptr[1] < stop) stop = lptr[1];
                  dstop = (double)stop;
          case 1: dstart = (double)lptr[0];
          default: break;
        }
        if (sample < 0) sample = 0;
        break;
      case CF_DOUBLE:
        dptr = (double *)dataIn;
        switch (con->EqpSizeIn)
        {
          case 5: recsreadFlag = (int)dptr[4];
          case 4: sample = (int)dptr[3];
          case 3: index = (int)dptr[2];
          case 2: if (dptr[1] < stop) stop = (time_t)dptr[1];
                  dstop = (double)stop;
          case 1: dstart = dptr[0]; 
          default: break;
        }
        if (sample < 0) sample = 0;
        break;
      case CF_FWINDOW:
        dstart = (double)((FWINDOW *)dataIn)[0].i1val;
        dstop = (double)((FWINDOW *)dataIn)[0].i2val;
        lwr = ((FWINDOW *)dataIn)[0].f1val;
        upr = ((FWINDOW *)dataIn)[0].f2val;
        if (lwr < upr) sample = -1;
        break;
      case CF_WINDOW:
        dstart = (double)((DWINDOW *)dataIn)[0].t1;
        dstop = (double)((DWINDOW *)dataIn)[0].t2;
        sample = ((DWINDOW *)dataIn)[0].i2val;
        break;
      default:
        return illegal_format;
    }
  }
  else
  {
    dstart = (double)getStartTimeFromDepth(con->EqpName,Property,n);
    if (dstart < 0) ccerr((int)(-dstart));
  }

  if (dstart < 0) dstart = 0;
  if (dstop < dstart) dstop = (double)time(NULL);
  if (isSnapshot) 
  { 
    if (dstart < 1.0) return argument_list_error;
    if (fmt > CF_FLOAT) return illegal_format;
    dstop = dstart;
    /* snapshot read doesn't use sample, index or recsread ... */
    sample = 0;
    index = 0;
    recsreadFlag = 0;
    lwr = upr = 0.0;
  }

  numread = getLTSData(con->EqpName,Property,con->EqpDeviceName,index,dstart,dstop,lwr,upr,
                dataptr,LFMT(con->EqpFormatOut),n,sample,recsreadFlag);

  if (numread < 0) ccerr((int)(-numread));
  if (numread > 0) 
  {
    if (!isSnapshot && numread < (long)con->EqpSizeOut - 1) 
    {
      dataptr += numread * fmtsizeof(fmt);
      dstart = getDataTimeStamp() + 1.0;
      n -= numread;
    }
    else 
    {
      ccerr(0);
    }
  }

  nr = getSTSData(con->EqpName,Property,con->EqpDeviceName,index,dstart,dstop,lwr,upr,
                dataptr,LFMT(con->EqpFormatOut),n,sample,recsreadFlag);

  if (nr < 0) ccerr((int)(-nr));
  numread += nr;
  if (numread < (int)con->EqpSizeOut - 1) con->EqpSizeOut = numread;
err:
  return  cc;
}
int gateFunction(CONTRACT *con,char *dataIn,char *data)
{ /* keep up apperances ... */
  return 0;
}
int EGUQueryFunction(CONTRACT *con,char *dataIn,char *data)
{
  int idx,len;
  
  ExportPropertyListStruct *prp;
  ExportListStruct *el;
  char *c, Property[32];
  DUSTRING *ustr;

  if ((len=con->EqpSizeOut) == 0) return illegal_read_write;
  strncpy(Property,con->EqpProperty,32);
  if ((c=strstr(Property,".EGU")) == NULL) return illegal_property;
  *c = 0;
  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strncmp(con->EqpName,el->EqpName,EQP_NAME_SIZE)) continue;
    for (idx=0; idx<PROPERTY_HASH_SIZE+1; idx++)
      for (prp=el->EqpProperty[idx]; prp; prp=prp->next)
      {
	    if (strncmp(Property,prp->prpName,32)) continue;
        goto checkformat;
      }
    return illegal_property;
  }
  return non_existent_elem;
checkformat:
  switch (LFMT(con->EqpFormatOut))
  {
    case CF_USTRING:
        ustr = (DUSTRING *)data;
        ustr->ival  = (UINT32)prp->egu.graph;
        ustr->f1val = (float)prp->egu.min;
        ustr->f2val = (float)prp->egu.max;
        ustr->tm = startuptime;
        strncpy(ustr->str,prp->egu.units,16);
     	return 0;
    case CF_NAME16:
        if (len != 1) return illegal_read_write;
        strncpy((char *)data,prp->egu.units,16);
     	return 0;
    case CF_NAME32:
        if (len != 1) return illegal_read_write;
        strncpy((char *)data,prp->egu.units,32);
     	return 0;
    case CF_TEXT:
        if (len < 1 || len > 16) return illegal_read_write;
        strncpy((char *)data,prp->egu.units,len);
     	return 0;
    case CF_FLOAT:
        if (len != 2) return illegal_read_write;
        ((float *)data)[0] = (float)prp->egu.min;
        ((float *)data)[1] = (float)prp->egu.max;
     	return 0;
    default:
        return illegal_format;
  }
  return illegal_format;
}
int NAMQueryFunction(CONTRACT *con,char *dataIn,char *data)
{
  int i,slen;
  ExportPropertyListStruct *prp;
  char *c,tgtprp[64];
  strncpy(tgtprp,con->EqpProperty,32);
  if ((c=strstr(tgtprp,".NAM")) != NULL) *c = 0;
  if ((prp=AssignDeviceListFromProperty(con->EqpName,tgtprp)) != NULL)
  {
    if (con->EqpSizeOut == 0) return illegal_read_write;
    switch (LFMT(con->EqpFormatOut))
    {
      case CF_NAME16:
      case CF_NAME32:slen = fmtsizeof(con->EqpFormatOut); break;
      default: return illegal_format;
    }
    for (i=0; i<(int)con->EqpSizeOut && i<prp->numDevices; i++)
    {
      strncpy(&data[i*slen],prp->devNames[i].name,16);
    }
    return 0;
  }
  return deviceQuery(con,dataIn,data);
}
int DESCQueryFunction(CONTRACT *con,char *dataIn,char *data)
{
  int cc=0;
  int idx,j,len;
  ExportPropertyListStruct *prp;
  ExportListStruct *el;
  PropertyQueryStructEx *pqsx = NULL;
  PropertyQueryStruct *pqs = NULL;
  char *c, Property[32];

  strncpy(Property,con->EqpProperty,32);
  if ((c=strstr(Property,".DESC")) == NULL &&
      (c=strstr(Property,".DESC")) == NULL) return illegal_property;
  *c = 0;
  len = con->EqpSizeOut;
  switch (LFMT(con->EqpFormatOut))
  {
    case CF_TEXT:
      if (len < 1 || len > 64) return illegal_read_write;
      break;
    case CF_NAME16:
      if (len != 1) return illegal_read_write;
      len = 16;
      break;
    case CF_NAME32:
      if (len != 1) return illegal_read_write;
      len = 32;
      break;
    case CF_NAME48:
      if (len != 1) return illegal_read_write;
      len = 48;
      break;
    case CF_NAME64:
      if (len != 1) return illegal_read_write;
      len = 64;
      break;
    default:  return illegal_format;
  }

  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strncmp(con->EqpName,el->EqpName,EQP_NAME_SIZE)) continue;
    for (j=0,idx=0; idx<PROPERTY_HASH_SIZE+1; idx++)
      for (prp=el->EqpProperty[idx]; prp; prp=prp->next)
    {
      if (strncmp(prp->prpName,Property,32)) continue;
      strncpy(data,prp->prpDescription,len);
      return 0;
    }
    return illegal_property;
  }
  return non_existent_elem;
}
int rmvRegisteredUser(CONTRACT *con,char *dataIn,char *data)
{
  NAME16 usr, *usrlst = NULL;
  int i, len = (int)con->EqpSizeIn, lstsiz = 0;
  if (!(con->EqpAccess&CA_WRITE)) return illegal_read_write;
  if (!(con->EqpAccess&CA_LAST)) return illegal_mode; /* this must be a synchronous call */
  switch (LFMT(con->EqpFormatIn))
  {
    case CF_NAME16:
      usrlst = (NAME16 *)dataIn;
      lstsiz = len;
      break;
    case CF_NAME32:
      usrlst = (NAME16 *)dataIn;
      lstsiz = len;
      for (i=0; i<len; i++) strncpy(usrlst[i].name,((NAME32 *)dataIn)[i].name,16);
      break;
    case CF_TEXT:
      memset(usr.name,0,16);
      if (len < 1) return illegal_read_write;
      if (len > 16) len = 16;
      strncpy(usr.name,dataIn,len);
      usrlst = &usr;
      lstsiz = 1;
      break;
    default: return illegal_format;
  }
  return removeRegisteredUser(usrlst,lstsiz);
}
int addRegisteredUser(CONTRACT *con,char *dataIn,char *data)
{
  NAME16 usr, *usrlst = NULL;
  int i, len = (int)con->EqpSizeIn, lstsiz = 0;
  if (NrOfUsers == 0) return not_allowed; 
  if (!(con->EqpAccess&CA_WRITE)) return illegal_read_write;
  if (!(con->EqpAccess&CA_LAST)) return illegal_mode; /* this must be a synchronous call */
  switch (LFMT(con->EqpFormatIn))
  {
    case CF_NAME16:
      usrlst = (NAME16 *)dataIn;
      lstsiz = len;
      break;
    case CF_NAME32:
      usrlst = (NAME16 *)dataIn;
      lstsiz = len;
      for (i=0; i<len; i++) strncpy(usrlst[i].name,((NAME32 *)dataIn)[i].name,16);
      break;
    case CF_TEXT:
      memset(usr.name,0,16);
      if (len < 1) return illegal_read_write;
      if (len > 16) len = 16;
      strncpy(usr.name,dataIn,len);
      usrlst = &usr;
      lstsiz = 1;
      break;
    default: return illegal_format;
  }
  return appendRegisteredUserList(usrlst,lstsiz);
}
int getDeniedUserList(CONTRACT *con,char *dataIn,char *data)
{
  int i;
  if (con->EqpSizeOut == 0) return illegal_read_write;
  if (LFMT(con->EqpFormatOut) != CF_NAME16) return illegal_format;
  for (i=0; i<(int)con->EqpSizeOut && i<NrOfDeniedUsers; i++)
  {
    strncpy(((NAME16 *)data)[i].name,DeniedUserList[i].UserName,16);
  }
  return 0;
}
int getRegisteredUserList(CONTRACT *con,char *dataIn,char *data)
{
  int i;
  if (con->EqpSizeOut == 0) return illegal_read_write;
  if (LFMT(con->EqpFormatOut) != CF_NAME16) return illegal_format;
  for (i=0; i<(int)con->EqpSizeOut && i<NrOfUsers; i++)
  {
    strncpy(((NAME16 *)data)[i].name,UserList[i].UserName,16);
  }
  return 0;
}
int getAppVersion(CONTRACT *con,char *dataIn,char *data)
{
  int len;
  if (LFMT(con->EqpFormatOut) != CF_TEXT) return illegal_format;
  if ((len=con->EqpSizeOut) < 1) return illegal_read_write;
  strncpy(data,appmkversion,len);
  return 0;
}
int getTINEVersion(CONTRACT *con,char *dataIn,char *data)
{
  char s[32];
  int len;

  if (LFMT(con->EqpFormatOut) != CF_TEXT) return illegal_format;
  if ((len=con->EqpSizeOut) < 1) return illegal_read_write;
  sprintf(s,"%d.%02d.%04d",MAJOR_VERSION,MINOR_VERSION,VERSION_REVISION);
  strncpy(data,s,len);
  return 0;
}
int getServerTimeStr(CONTRACT *con,char *data,time_t t)
{
  int len;
  char s[32];
  if ((len=(int)con->EqpSizeOut) < 1) return illegal_read_write;

  switch (LFMT(con->EqpFormatOut))
  {
    case CF_TEXT:
      strncpy(s,ctime(&t),24); s[24] = 0;
      switch (len)
      {
        case 32: case 26: case 24: strncpy(data,s,24); break;
        case 16: strncpy(data,&s[4],15); data[15] = 0; break;
        default: return illegal_read_write;
      }
      break;
    case CF_LONG:
      *(UINT32 *)data = (UINT32)t;
      break;
    default:
      return illegal_format;
  }
  return 0;
}
int getServerStarttime(CONTRACT *con,char *dataIn,char *data)
{
  return getServerTimeStr(con,data,startuptime);
}
int getAppDate(CONTRACT *con,char *dataIn,char *data)
{
  return getServerTimeStr(con,data,appmktime);
}
int getServerOS(CONTRACT *con,char *dataIn,char *data)
{
  int len;
  if (LFMT(con->EqpFormatOut) != CF_TEXT) return illegal_format;
  if ((len=con->EqpSizeOut) < 1) return illegal_read_write;
  strncpy(data,"WIN32",len);
  return 0;
}
int getNotPostedCount(CONTRACT *con,char *dataIn,char *data)
{
  if (con->EqpSizeOut != 1) return illegal_read_write;
  switch (LFMT(con->EqpFormatOut))
  {
    case CF_SHORT:
      *(short *)data = (short)gCntNotPosted; break;
    case CF_LONG:
      *(int *)data = gCntNotPosted; break;
    case CF_FLOAT:
      *(float *)data = (float)gCntNotPosted; break;
    case CF_DOUBLE:
      *(double *)data = (double)gCntNotPosted; break;
    default: return illegal_format;
  }
  return 0;

}
int T_numQuery(CONTRACT *con,char *data,short num)
{
  if (con->EqpSizeOut != 1) return illegal_read_write;
  switch (LFMT(con->EqpFormatOut))
  {
    case CF_SHORT:
      *(short *)data = num; break;
    case CF_LONG:
      *(int *)data = (int)num; break;
    case CF_FLOAT:
      *(float *)data = (float)num; break;
    case CF_DOUBLE:
      *(double *)data = (double)num; break;
    default: return illegal_format;
  }
  return 0;
}
int nhistoryRecordsQuery(CONTRACT *con,char *dataIn,char *data)
{
  int i,n=0;
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strcmp(con->EqpName,historyRecords[i].c.EqpName)) continue;
    n++;
  }
  return T_numQuery(con,data,(short)n);
}
int nstockpropsQuery(CONTRACT *con,char *dataIn,char *data)
{
  return T_numQuery(con,data,(short)NrStockProperties);
}
int getNumRegisteredIPNets(CONTRACT *con,char *dataIn,char *data)
{
  return T_numQuery(con,data,(short)NrOfIPNets);
}
int getNumRegisteredUsers(CONTRACT *con,char *dataIn,char *data)
{
  return T_numQuery(con,data,(short)NrOfUsers);
}
int historyRecordsQuery(CONTRACT *con,char *dataIn,char *data)
{
  int i,n=0;

  if (LFMT(con->EqpFormatOut) != CF_NAME32) return illegal_format;
  for (i=0; i<nHistoryRecords && i<(int)con->EqpSizeOut; i++)
  {
    if (strcmp(con->EqpName,historyRecords[i].c.EqpName)) continue;
    strncpy(((NAME32 *)data)[n++].name,historyRecords[i].c.EqpProperty,32);
  }
  return 0;
}
int getRegisteredIPNetList(CONTRACT *con,char *dataIn,char *data)
{
  int i;
  struct sockaddr_in *lst = IPNetList;
  if (con->EqpSizeOut == 0) return illegal_read_write;
  if (LFMT(con->EqpFormatOut) != CF_NAME16) return illegal_format;
  for (i=0; i<(int)con->EqpSizeOut && i<NrOfIPNets; i++)
  {
    strncpy(((NAME16 *)data)[i].name,inet_ntoa(lst[i].sin_addr),16);
  }
  return 0;
}
int stockpropsQuery(CONTRACT *con,char *dataIn,char *data)
{
  int i,k,len,n,fmt;
  PropertyQueryStructEx *pqsx = NULL;
  PropertyQueryStruct *pqs = NULL;
  XPropertyQueryStruct *xpqs = NULL;
  int tgtprplen = 0;
  char *c,tgtprp[33];

  if ((len=con->EqpSizeIn) > 0)
  {
    memset(tgtprp,0,33);
    switch (LFMT(con->EqpFormatIn))
    {
      case CF_TEXT:
        if (len < 1 || len > 32) return illegal_read_write;
        strncpy(tgtprp,dataIn,len);
        break;
      case CF_NAME16:
        if (len != 1) return illegal_read_write;
        strncpy(tgtprp,((NAME16 *)dataIn)->name,16);
        break;
      case CF_NAME32:
        if (len != 1) return illegal_read_write;
        strncpy(tgtprp,((NAME32 *)dataIn)->name,32);
        break;
      default:  return illegal_format;
    }
    if ((c=strchr(tgtprp,'*')) != NULL)
    {
      *c = 0;
      tgtprplen=strlen(tgtprp);
    }
    else tgtprplen = 32;
  }

  switch (fmt=LFMT(con->EqpFormatOut))
  {
    case CF_NAME32:
      if (con->EqpSizeOut == 0) return illegal_read_write;
      for (i=0,k=0; i<(int)con->EqpSizeOut && i<NrStockProperties; i++)
      {
        if (tgtprplen && strncmp(StockProperty[i].prpName,tgtprp,tgtprplen)) continue;
        if (k > 0 && !strncmp(((NAME32 *)data)[k-1].name,StockProperty[i].prpName,32)) continue;
        strncpy(((NAME32 *)data)[k].name,StockProperty[i].prpName,32);
	      k++;
      }
      if (k < (int)con->EqpSizeOut) con->EqpSizeOut = k;
      return 0;
    case CF_USTRING:
      if (con->EqpSizeOut == 0) return illegal_read_write;
      for (i=0,k=0; i<(int)con->EqpSizeOut && i<NrStockProperties; i++)
      {
        if (tgtprplen && strncmp(StockProperty[i].prpName,tgtprp,tgtprplen)) continue;
        if (k > 0 && !strncmp(((DUSTRING *)data)[k-1].str,StockProperty[i].prpName,32)) continue;
        strncpy(((DUSTRING *)data)[k].str,StockProperty[i].prpName,32);
        strncpy(&((DUSTRING *)data)[k].str[32],StockProperty[i].prpDescription,48);
        ((DUSTRING *)data)[k].ival = (UINT32)StockProperty[i].prpSize;
        ((DUSTRING *)data)[k].tm = (UINT32)StockProperty[i].prpFormat;
        k++;
      }
      if (k < (int)con->EqpSizeOut) con->EqpSizeOut = k;
      return 0;
    case CF_STRUCT:
      if (con->EqpSizeOut%XPROPERTYQUERYSTRUCT_SIZE == 0 && !strncmp(con->strTagOut,"XPQS",8))
      {
        n = con->EqpSizeOut/XPROPERTYQUERYSTRUCT_SIZE;
        xpqs = (XPropertyQueryStruct *)data;
        for (i=0,k=0; k<n && i<NrStockProperties; i++)
        {
          if (tgtprplen && strncmp(StockProperty[i].prpName,tgtprp,tgtprplen)) continue;
          strncpy(xpqs[k].prpName,StockProperty[i].prpName,32);
          strncpy(xpqs[k].prpDescription,StockProperty[i].prpDescription,64);
          xpqs[k].prpSize = StockProperty[i].prpSize;
          xpqs[k].prpSizeIn = StockProperty[i].prpSizeIn;
          xpqs[k].prpFormat = BFMT(StockProperty[i].prpFormat);
          xpqs[k].prpFormatIn = BFMT(StockProperty[i].prpFormatIn);
          xpqs[k].prpAccess = (BYTE)StockProperty[i].prpAccessMode;
          xpqs[k].prpArrayType = StockProperty[i].prpArrayType;
          xpqs[k].prpNumOverloads = 1;
          strncpy(xpqs[k].prpTag,StockProperty[i].prpStructTag,8);
          k++;
        }
        if (tgtprplen) for (i=0; i<k; i++) xpqs[i].prpNumOverloads = k;
        if (k < n) con->EqpSizeOut = k*XPROPERTYQUERYSTRUCT_SIZE;
        return 0;
      }
      if (con->EqpSizeOut%PROPERTYQUERYSTRUCTEX_SIZE == 0)
      {
        n = con->EqpSizeOut/PROPERTYQUERYSTRUCTEX_SIZE;
        pqsx = (PropertyQueryStructEx *)data;
        for (i=0,k=0; k<n && i<NrStockProperties; i++)
        {
          if (tgtprplen && strncmp(StockProperty[i].prpName,tgtprp,tgtprplen)) continue;
          strncpy(pqsx[k].prpName,StockProperty[i].prpName,32);
          strncpy(pqsx[k].prpDescription,StockProperty[i].prpDescription,64);
          pqsx[k].prpSize = StockProperty[i].prpSize;
          pqsx[k].prpSizeIn = StockProperty[i].prpSizeIn;
          pqsx[k].prpFormat = BFMT(StockProperty[i].prpFormat);
          pqsx[k].prpFormatIn = BFMT(StockProperty[i].prpFormatIn);
          pqsx[k].prpAccess = (BYTE)StockProperty[i].prpAccessMode;
          strncpy(pqsx[k].prpTag,StockProperty[i].prpStructTag,8);
          pqsx[k].prpNumOverloads = 1;
          k++;
        }
        if (tgtprplen) for (i=0; i<k; i++) pqsx[i].prpNumOverloads = k;
        if (k < n) con->EqpSizeOut = k*PROPERTYQUERYSTRUCTEX_SIZE;
        return 0;
      }
      if (con->EqpSizeOut%PROPERTYQUERYSTRUCT_SIZE == 0)
      {
        n = con->EqpSizeOut/PROPERTYQUERYSTRUCT_SIZE;
        pqs = (PropertyQueryStruct *)data;
        for (i=0,k=0; i<n && i<NrStockProperties; i++)
        {
          if (tgtprplen && strncmp(StockProperty[i].prpName,tgtprp,tgtprplen)) continue;
          strncpy(pqs[k].prpName,StockProperty[i].prpName,32);
          strncpy(pqs[k].prpDescription,StockProperty[i].prpDescription,32);
          pqs[k].prpSize = (UINT16)StockProperty[i].prpSize;
          pqs[k].prpFormat = BFMT(StockProperty[i].prpFormat);
          pqs[k].prpAccess = (BYTE)StockProperty[i].prpAccessMode;
          k++;
        }
        if (k < n) con->EqpSizeOut = k*PROPERTYQUERYSTRUCT_SIZE;
        return 0;
      }
    default:
      return illegal_format;
  }
  return illegal_format;
}
int getServerDescription(CONTRACT *con,char *dataIn,char *data)
{
  int len;
  if (LFMT(con->EqpFormatOut) != CF_TEXT) return illegal_format;
  if ((len=con->EqpSizeOut) < 1) return illegal_read_write;
  strncpy(data,RPCFecEx.desc,len);
  return 0;
}
int getServerLocation(CONTRACT *con,char *dataIn,char *data)
{
  int len;
  if (LFMT(con->EqpFormatOut) != CF_TEXT) return illegal_format;
  if ((len=con->EqpSizeOut) < 1) return illegal_read_write;
  strncpy(data,RPCFecEx.loc,len);
  return 0;
}
int getServerStats(CONTRACT *con,char *dataIn,char *data)
{
  UINT32 statsBuffer[16];
  int i;
  if (LFMT(con->EqpFormatOut) != CF_LONG) return illegal_format;
  statsBuffer[0] = AveBusyTime;
  statsBuffer[1] = gCycleCounts;
  statsBuffer[2] = gMaxCycleCounts;
  statsBuffer[3] = gSingleLinkCount;
  statsBuffer[4] = gClientMisses;
  statsBuffer[5] = gClientReconnects;
  statsBuffer[6] = gClientRetries;
  statsBuffer[7] = gContractMisses;
  statsBuffer[8] = gContractDelays;
  statsBuffer[9] = gBurstLimitReachedCount;
  statsBuffer[10] = (UINT32)(gDataTimeStampOffset * 1000);
  for (i=0; i<(int)con->EqpSizeOut && i<11; i++) ((UINT32 *)data)[i] = statsBuffer[i];
  return 0;
}
int taggedStructQuery(CONTRACT *con,char *dataIn,char *data)
{
  int cc,k,len;
  char tag[9];

  if (con->EqpSizeOut == 0) return illegal_read_write;
  if (LFMT(con->EqpFormatOut) != CF_INTINT) return illegal_format;
  if ((len=con->EqpSizeIn) == 0) return illegal_read_write;
  memset(tag,0,9);
  switch (LFMT(con->EqpFormatIn))
  {
    case CF_TEXT:
      if (len < 1 || len > 8) return illegal_read_write;
      strncpy(tag,dataIn,len);
      break;
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME64:
      if (len != 1) return illegal_read_write;
      strncpy(tag,dataIn,8);
      break;
    default:  return illegal_format;
  }
  k = (int)con->EqpSizeOut;
  cc = queryStructFormat(tag,(INTINT *)data,&k);
  if (k < (int)con->EqpSizeOut) con->EqpSizeOut = k;
  return cc;
}
int getServerAliasList(CONTRACT *con,char *dataIn,char *data)
{
  int len,i,j;
  if (con->EqpSizeOut == 0) return illegal_read_write;
  switch (LFMT(con->EqpFormatOut))
  {
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME64:
      len = MIN(fmtsizeof(con->EqpFormatOut),32);
      break;
    default:  return illegal_format;
  }
  for (i=0,j=0; i<(int)con->EqpSizeOut && j<nAliasTableEntries; i += 2,j++)
  {
     strncpy(((NAME32 *)data)[i].name,gAliasTable[j].name,len);
     strncpy(((NAME32 *)data)[i+1].name,gAliasTable[j].alias,len);
  }
  if (i < (int)con->EqpSizeOut) con->EqpSizeOut = i;
  return 0;
}
StockPropertyStructEx StockProperty[] =
{
     {"PROPS",512,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE|AT_CHANNEL,"","Exported Properties",propsQuery},
     {"PROPS",100,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"PQS","Exported Properties",propsQuery},
     {"PROPS",20,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"PQSX","Exported Properties",propsQuery},
     {"PROPS",20,32,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"PQSX","Exported Properties",propsQuery},
     {"PROPS",10,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"XPQS","Exported Properties",propsQuery},
     {"PROPS",10,32,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"XPQS","Exported Properties",propsQuery},
     {"NPROPS",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of Exported Properties",npropsQuery},
     {"PROPERTIES",512,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE,"","Exported Properties",propsQuery},
     {"PROPERTIES",100,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"PQS","Exported Properties",propsQuery},
     {"PROPERTIES",20,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"PQSX","Exported Properties",propsQuery},
     {"PROPERTIES",20,32,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"PQSX","Exported Properties",propsQuery},
     {"PROPERTIES",10,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"XPQS","Exported Properties",propsQuery},
     {"PROPERTIES",10,32,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"XPQS","Exported Properties",propsQuery},
     {"NPROPERTIES",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of Exported Properties",npropsQuery},
     {"DEVICES",512,0,CF_NAME16,CF_NULL,CA_READ,AT_SINGLE,"","Device Names",deviceQuery},
     {"DEVICES",512,1,CF_NAME16,CF_NAME16,CA_READ,AT_SINGLE,"","Device Names",deviceQuery},
     {"NDEVICES",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of Devices",ndeviceQuery},
     {"ALARMS",512,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"AMS","Total Current Alarm List",alarmsQuery},
     {"NALARMS",5,0,CF_LONG,CF_NULL,CA_READ,AT_SINGLE|AT_STRUCTURED,"","Current Alarm Information",nalarmsQuery},
     {"ALMDEFS",32,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"ADS","Defined Alarm Descriptions",alarmDefsQuery},
     {"NALMDEFS",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of Alarm Defined Alarms",nalarmDefsQuery},
     {"ALARMSEXT",512,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"AMS","Specified Current Alarm List",alarmsQueryEx}, 
     {"MESSAGE",0,256,CF_NULL,CF_TEXT,CA_WRITE,AT_SINGLE,"","Accept Message",msgDisplay},
     {"ACTIVITY",68,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"","Current Server Activity",activityQuery},  
     {"ACTIVITY",1,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"AQS","Current Server Activity",activityQuery},  
     {"CONTRACTS",920,0,CF_STRUCT,CF_NULL,CA_READ,AT_UNKNOWN,"","Current Contract List",contractsQuery},
     {"CONTRACTS",10,0,CF_STRUCT,CF_NULL,CA_READ,AT_UNKNOWN,"CONQS","Current Contract List",contractsQuery},
     {"CLIENTS",540,0,CF_STRUCT,CF_NULL,CA_READ,AT_UNKNOWN,"","Current Client List",clientsQuery},
     {"CLIENTS",10,0,CF_STRUCT,CF_NULL,CA_READ,AT_UNKNOWN,"CLNQS","Current Client List",clientsQuery},
     {"ACCESSLOCK",0,2,CF_NULL,CF_SHORT,CA_WRITE,AT_UNKNOWN,"","lock type/lock duration",lockAccess},
     {"ACCESSLOCK",2,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE,"","lock owner/address",lockAccess},
     {"ADDUSER",0,16,CF_NULL,CF_TEXT,CA_WRITE,AT_UNKNOWN,"","Add User to Access List",addRegisteredUser},
     {"DELUSER",0,16,CF_NULL,CF_TEXT,CA_WRITE,AT_UNKNOWN,"","Remove User from Access List",rmvRegisteredUser},
     {"USERS",256,0,CF_NAME16,CF_NULL,CA_READ,AT_SINGLE,"","Read User Access List",getRegisteredUserList},
     {"DENIEDUSERS",256,0,CF_NAME16,CF_NULL,CA_READ,AT_SINGLE,"","Read Denied User List",getDeniedUserList},
     {"NUSERS",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of Users with Access",getNumRegisteredUsers},
     {"STOCKPROPS",512,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE,"","Stock Properties",stockpropsQuery},
     {"STOCKPROPS",68*10,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"","Stock Properties",stockpropsQuery},
     {"STOCKPROPS",188*10,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"","Stock Properties",stockpropsQuery},
     {"STOCKPROPS",188*10,32,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"","Stock Properties",stockpropsQuery},
     {"STOCKPROPS",10,32,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"XPQS","Stock Properties",stockpropsQuery},
     {"NSTOCKPROPS",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of Stock Properties",nstockpropsQuery},
     {"IPNETS",256,0,CF_NAME16,CF_NULL,CA_READ,AT_SINGLE,"","Read IP Access List",getRegisteredIPNetList},
     {"NIPNETS",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of IP Nets with Access",getNumRegisteredIPNets},
     {"DEBUGLEVEL",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","get debug level",setDebugLevel},
     {"DEBUGLEVEL",1,1,CF_SHORT,CF_SHORT,CA_READ|CA_WRITE,AT_SCALAR,"","get/set debug level",setDebugLevel},
     {"DEBUGBUFFER",BUFLEN+26,0,CF_TEXT,CF_NULL,CA_READ,AT_SINGLE,"","Debug display window",getDebugBuffer},
     {"LOGCOMMANDS",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","read en/dis-able command logging flag",setLogCommands},
     {"LOGCOMMANDS",1,1,CF_SHORT,CF_SHORT,CA_READ|CA_WRITE,AT_SCALAR,"","en/dis-able command logging",setLogCommands},
     {"LOGDEPTH",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","get log depth in lines",setLogDepth},
     {"LOGDEPTH",1,1,CF_SHORT,CF_SHORT,CA_READ|CA_WRITE,AT_SCALAR,"","get/set log depth in lines",setLogDepth},
     {"LOGFILE",8000,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","get log file",getFecLog},
     {"SRVVERSION",16,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","TINE Server Version Number",getTINEVersion},
     {"SRVOS",16,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","General Server OS",getServerOS},
     {"SRVDESC",64,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","Server Description",getServerDescription},
     {"SRVLOCATION",32,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","Server Location",getServerLocation},
     {"SRVSTATS",16,0,CF_LONG,CF_NULL,CA_READ,AT_SINGLE|AT_STRUCTURED,"","Server Statistics counters",getServerStats},
     {"SRVALIASLIST",512,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE,"","Alias List (Name/Alias pairs)",getServerAliasList},
     {"NHISTORIES",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of history properties",nhistoryRecordsQuery},
     {"HISTORIES",32,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE,"","history properties",historyRecordsQuery},
     {"STRUCTFORMAT",512,1,CF_INTINT,CF_NAME16,CA_READ,AT_SINGLE,"","Registered structure information",taggedStructQuery},
     {"SRVSTARTTIME",32,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","Server Startup Time",getServerStarttime},
     {"SRVSTARTTIME",1,0,CF_LONG,CF_NULL,CA_READ,AT_SCALAR,"","Server Startup Time",getServerStarttime},
     {"APPDATE",32,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","Application Compile Time",getAppDate},
     {"APPDATE",1,0,CF_LONG,CF_NULL,CA_READ,AT_SCALAR,"","Application Compile  Time",getAppDate},
     {"APPVERSION",32,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","Application Version Number",getAppVersion},
     {"SRVSELFTEST",32000,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","get self-test file",getSelfTestFile},
     {"SRVLOGFILES",32,0,CF_NAME64,CF_NULL,CA_READ,AT_SCALAR,"","get log file list",getFecLogFileList},
     {"SRVLOGFILES",32,64,CF_NAME64,CF_TEXT,CA_READ,AT_SCALAR,"","get log file list",getFecLogFileList},
     {"SRVLOGFILE",32000,64,CF_TEXT,CF_TEXT,CA_READ,AT_SCALAR,"","get specific log file ",getFecLogFile},
     {"SRVLASTACCESS",5,0,CF_NAME32,CF_NULL,CA_READ,AT_SCALAR,"","get most recent access",getLastAccessList},
     {"SRVNOTPOSTED",1,0,CF_LONG,CF_NULL,CA_READ,AT_SCALAR,"","Number Contracts not posted",getNotPostedCount},
};
int NrStockProperties = (sizeof(StockProperty)/sizeof(StockPropertyStructEx));

HashXRefStruct **StockPropXRefTable = NULL;

int makeStockPropXRefTable(void)
{
  int k,idx;
  HashXRefStruct *xref;

  if ((StockPropXRefTable=(HashXRefStruct **)calloc(STOCKPROP_HASH_SIZE,sizeof(HashXRefStruct *))) == NULL)
    return out_of_server_memory;
  for (k=0; k<NrStockProperties; k++)
  {
    idx = (int)(ElfHash((unsigned char *)StockProperty[k].prpName) % STOCKPROP_HASH_SIZE);
    for (xref=StockPropXRefTable[idx]; xref && xref->index != k; xref = xref->next)
    {
      if (!strncmp(StockProperty[k].prpName,StockProperty[xref->index].prpName,32)) break;
    }
    if (xref == NULL) 
    {
      if ((xref=(HashXRefStruct *)calloc(1,sizeof(HashXRefStruct))) == NULL) return out_of_server_memory;
      xref->index = k;
      xref->next = StockPropXRefTable[idx];
      StockPropXRefTable[idx] = xref;
    }
  }
  return 0;
}

StockModifierStruct StockModifier[] =
{
  {".EGU",EGUQueryFunction},
  {".HIST",HISTQueryFunction},
  {".HST",HISTQueryFunction},
  {".HIST@",HISTQueryFunction},
  {".HST@",HISTQueryFunction},
  {".ARCH",HISTQueryFunction},
  {".ARC",HISTQueryFunction},
  {".AR",HISTQueryFunction},
  {".NAM",NAMQueryFunction},
  {".DESC",DESCQueryFunction},
  {".DSC",DESCQueryFunction},
  {".BIT.",gateFunction},
  {".MASK.",gateFunction},
  {".GATE.",gateFunction},
};
int NrStockModifiers = (sizeof(StockModifier)/sizeof(StockModifierStruct));

short IsStockProperty(char *Property)
{
  if (strchr(Property,'*') != NULL) return STOCK_PROP; /* maps to "PROPERTIES" */
  return GetStockPropertyId(Property) < 0 ? EMPTY_PROP : STOCK_PROP;
}

short IsMetaProperty(char *Property,char **baseProperty,char **metaProperty)
{
  int i;
  char *c;
  static char metaPropertyString[64];

  strncpy(metaPropertyString,Property,EQP_PROP_SIZE);
  for (i=0; i<NrStockModifiers; i++)
  {
    if ((c=strstr(metaPropertyString,StockModifier[i].modName)) != NULL)
    {
      *c++ = 0;
      if (baseProperty != NULL) *baseProperty = metaPropertyString;
      if (metaProperty != NULL) *metaProperty = c;
      return TRUE;
    }
  }
  return FALSE;
}

short IsGatedModifier(CONTRACT *con,char *baseProperty,UINT32 *gate)
{
  char *c;
  int bit;

  if (baseProperty == NULL || gate == NULL) return FALSE;
  if (LFMT(con->EqpFormatOut) != CF_BYTE &&
      LFMT(con->EqpFormatOut) != CF_SHORT &&
      LFMT(con->EqpFormatOut) != CF_LONG) return FALSE;
  strncpy(baseProperty,con->EqpProperty,EQP_PROP_SIZE);
  if ((c=strstr(baseProperty,".BIT.")) != NULL)
  {
    *c = 0;
    c += 5;
    bit = atoi(c);
    if (bit < 0 || bit >= 8*fmtsizeof(con->EqpFormatOut)) return FALSE;
    *gate = 0x01 << bit;
    return GATE_BOOLEAN;
  }
  if ((c=strstr(baseProperty,".MASK.")) != NULL)
  {
    *c = 0;
    c += 6;
    *gate = strtol(c,NULL,0);
    return GATE_BOOLEAN;
  }
  if ((c=strstr(baseProperty,".GATE.")) != NULL)
  {
    *c = 0;
    c += 6;
    *gate = strtol(c,NULL,0);
    return GATE_FILTER;
  }
  return FALSE;
}
